! T3X Runtime Class: String Functions
! Copyright (C) 1999-2003 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.

module string();

public class string(t3x, char)

 object	t[t3x], chr[char];

 public const	MAXLEN = 32767;

 public length(s) return t.memscan(s, 0, MAXLEN);

 public copy(a, b) t.memcopy(a, b, t.memscan(b, 0, MAXLEN)+1);

 public comp(a, b) do var k1, k2;
	k1 := t.memscan(a, 0, MAXLEN)+1;
	k2 := t.memscan(b, 0, MAXLEN)+1;
	return t.memcomp(a, b, k1>k2-> k1: k2);
 end

 public icomp(a, b) do var i, c1, c2;
	i := 0;
	while (1) do
		c1 := chr.lcase(a::i);
		c2 := chr.lcase(b::i);
		if (c1 \= c2) return c1-c2;
		if (\c1) return 0;
		i := i+1;
	end
 end

 public scan(s, c) return t.memscan(s, c, t.memscan(s, 0, MAXLEN));

 public rscan(s, c) do var k, i;
	k := t.memscan(s, 0, MAXLEN);
	for (i=k-1, -1, -1) if (s::i = c) return i;
	return %1;
 end

 public find(a, b) do var ka, kb, i;
	ka := t.memscan(a, 0, MAXLEN);
	kb := t.memscan(b, 0, MAXLEN);
	for (i=0, ka-kb+1)
		if (\t.memcomp(@a::i, b, kb))
			return i;
	return %1;
 end

 public rfind(a, b) do var ka, kb, i;
	ka := t.memscan(a, 0, MAXLEN);
	kb := t.memscan(b, 0, MAXLEN);
	for (i=ka-kb, -1, -1)
		if (\t.memcomp(@a::i, b, kb))
			return i;
	return %1;
 end

 public xlate(s, c1, c2) do var i;
	i := 0;
	while (s::i) do
		if (s::i = c1) s::i := c2;
		i := i+1;
	end
	return s;
 end

 _digits() return "0123456789ABCDEF";

 public numtostr(s, n, r) do
	var	digits;
 	var	i;
	var	ds::66, dp;

	digits := _digits();
	i := 0;
	! handle sign
	if (r < 0) do
		r := -r;
		if (n < 0) do
			n := -n;
			s::i := '-';
			i := i+1;
		end
	end
	dp := 0;
	! generate reverse numeric string
	while (1) do
		ds::dp := digits::(n mod r);
		dp := dp+1;
		n := n ./ r;
		if (\n) leave;
	end
	! reverse the string
	while (dp) do
		dp := dp-1;
		s::i := ds::dp;
		i := i+1;
	end
	s::i := 0;
	return s;
 end

! %[max][:F][U][LR][BCDSX%]
! Fillchar
! Unsigned
! Left, Right
! Character, Decimal, String, heXadecimal, %
 public format(s, tmpl, args) do
	const	LEFT=-1, NONE=0, RIGHT=1;	! adjust modes
	const	MAXPAD = 256;	! max. padding characters
	var	adj;		! adjustment type, see above
	var	op, ap;
	var	i, j, k;
	var	max;		! max. field size
	var	p, x;
	var	unsigned;	! unsigned flag for %D (1=unsigned)
	var	nbuf::MAXPAD+1;
	var	fc;		! fill character (for padding)

	i := 0;		! template index
	j := 0;		! buffer index
	ap := 0;	! argument index
	while (tmpl::i) do
		ie (tmpl::i = '%') do
			i := i+1;
			! set defaults
			adj := NONE;
			max := -1;	! automatic length
			unsigned := 0;
			fc := '\s';
			! accept max field length
			while (chr.digit(tmpl::i)) do
				if (max = -1) max := 0;
				max := max*10 + tmpl::i - '0';
				i := i+1;
			end
			! accept fill character
			if (tmpl::i = ':') do
				fc := tmpl::(i+1);
				i := i+2;
			end
			! accept unsigned flag
			if (tmpl::i = 'u' \/ tmpl::i = 'U') do
				unsigned := 1;
				i := i+1;
			end
			! accept adjustment type
			ie (tmpl::i = 'l' \/ tmpl::i = 'L') do
				adj := LEFT;
				i := i+1;
			end
			else if (tmpl::i = 'r' \/ tmpl::i = 'R') do
				adj := RIGHT;
				i := i+1;
			end
			x := chr.ucase(tmpl::i);	! get type
			op := ap;
			ie (x = 'B') do
				p := self.numtostr(nbuf, args[ap],
					unsigned-> 2: -2);
				if (adj = NONE) adj := RIGHT;
				ap := ap+1;
			end
			else ie (x = 'C') do
				nbuf::0 := args[ap];
				nbuf::1 := 0;
				p := nbuf;
				ap := ap+1;
			end
			else ie (x = 'D') do
				p := self.numtostr(nbuf, args[ap],
					unsigned-> 10: -10);
				if (adj = NONE) adj := RIGHT;
				ap := ap+1;
			end
			else ie (x = 'S') do
				p := args[ap];
				ap := ap+1;
				if (adj = NONE) adj := LEFT;
			end
			else ie (x = 'X') do
				p := self.numtostr(nbuf, args[ap],
					unsigned-> 16: -16);
				ap := ap+1;
				if (adj = NONE) adj := RIGHT;
			end
			! default: copy character after %
			else do
				s::j := tmpl::i;
				j := j+1;
			end
			if (op \= ap) do	! op=ap means invalid type
				k := self.length(p);
				if (max > 0) max := max mod MAXPAD;
				ie (max <= k) do
					! no padding required
					t.memcopy(@s::j, p, k);
					j := j+k;
				end
				else do
					t.memfill(@s::j, fc, max);
					ie (adj = LEFT)
						t.memcopy(@s::j, p, k);
					else
						t.memcopy(
							@s::(j+max-k), p, k);
					j := j+max;
				end
			end
		end
		else do
			! no %-sequence: copy one char
			s::j := tmpl::i;
			j := j+1;
		end
		if (tmpl::i) i := i+1;
	end
	s::j := 0;
	return s;
 end

 public strtonum(s, r, lp) do
	var	digits;
	var	g, i, j;
	var	v;
	var	lim;

	digits := _digits();
	i := 0;
	v := 0;
	g := 1;
	! skip leading white space
	while (s::i = '\s' \/ s::i = '\t') i := i+1;
	! ignore leading +
	ie (s::i = '+') do
		i := i+1;
	end
	! accept leading - or %
	else if (s::i = '-' \/ s::i = '%') do
		g := -1;
		i := i+1;
	end
	! if a limit is supplied, use it
	lim := lp /\ lp[0]-> lp[0]: MAXLEN;
	for (i=i, lim) do
		j := t.memscan(digits, chr.ucase(s::i), r);
		if (j < 0) leave;
		v := v*r + j;
	end
	! if a limit was supplied, replace it with
	! the number of characters parsed
	if (lp) lp[0] := i;
	return v*g;
 end

! %[len][:delimiter][BCDSWX%]
! Unsigned
! Character, Decimal, String, White, heXadecimal, %
! %[group] (literal []s)
! %W = %[\s\t]
 public parse(s, tmpl, args) do
	var	i, j, k, x;
	var	n, m;
	var	ap;
	var	len,		! max. length to parse
		dlm;		! delimiter
	var	cbuf::256;

	i := 0;		! template index
	j := 0;		! string (source) index
	k := self.length(s);
	ap := 0;	! argument index
	while (tmpl::i /\ j < k) do
		ie (tmpl::i = '%') do
			i := i+1;
			len := -1;	! no limit on length
			dlm := -1;	! no delimiter
			! accept length
			while (chr.digit(tmpl::i)) do
				if (len = -1) len := 0;
				len := len*10 + tmpl::i - '0';
				i := i+1;
			end
			! accept delimiter
			if (tmpl::i = ':') do
				dlm := tmpl::(i+1);
				i := i+2;
			end
			x := chr.ucase(tmpl::i);	! get type
			ie (x = 'B') do
				n := len=-1-> 0: len;
				args[ap][0] := self.strtonum(@s::j, 2, @n);
				ap := ap+1;
				j := j+n;
			end
			else ie (x = 'C') do
				args[ap][0] := s::j;
				ap := ap+1;
				j := j+1;
			end
			else ie (x = 'D') do
				n := len=-1-> 0: len;
				args[ap][0] := self.strtonum(@s::j, 10, @n);
				ap := ap+1;
				j := j+n;
			end
			else ie (x = 'S') do
				if (len = -1) do
					! no limit: copy rest of source
					if (dlm = -1) dlm := tmpl::(i+1);
					len := t.memscan(@s::j, dlm, k-j+1);
					If (len < 1) len := k-j+1;
				end
				if (len > k-j+1) len := k-j+1;
				t.memcopy(args[ap], @s::j, len);
				args[ap]::len := 0;
				ap := ap+1;
				j := j+len;
			end
			else ie (x = 'X') do
				n := len=-1-> 0: len;
				args[ap][0] := self.strtonum(@s::j, 16, @n);
				ap := ap+1;
				j := j+n;
			end
			else ie (x = 'W' \/ x = '[') do
				ie (x = 'W') do
					cbuf::0 := '\t';
					cbuf::1 := '\f';
					cbuf::2 := '\s';
					m := 3;
				end
				else do
					m := 0;
					i := i+1;
					while (tmpl::i /\ tmpl::i \= ']') do
						cbuf::m := tmpl::i;
						i := i+1;
						m := m+1;
					end
				end
				while (	j < k /\
					t.memscan(cbuf, s::j, m) >= 0
				)
					j := j+1;
			end
			else do
				! default: match character after %
				if (tmpl::i \= s::j) leave;
				j := j+1;
			end
		end
		else do
			! default: match literal character
			if (tmpl::i \= s::j) leave;
			j := j+1;
		end
		if (tmpl::i) i := i+1;
	end
	return ap;
 end

end

do end
