! TXDPF -- Tcode8 Dead Procedure Filter
! Copyright (C) 1999,2000,2002,2003 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: TXDPF <INFILE >OUTFILE

module txdpf(t3x, tcode);

object	t[t3x];

const	BUFLEN		= 1024;		! I/O buffer size
const	MAXLABEL	= 4096;		! max. label ID
const	MAXREF		= 8191;		! max. number of references / module
const	MAXIPROC	= 2048;		! max. number of interface procs

! procedure reference flags
const	UNUSED	= -1,	! label unused
	LEAF	= -2;	! label is a leaf (used but has no callees)

var	Inbuf::BUFLEN, Ilen, Iptr;	! input buffer, length, ptr
var	Outbuf::BUFLEN, Optr;		! output buffer, ptr
var	Infile, Outfile;		! input file, output file
var	Procs[MAXLABEL],		! procedure call trees
	Uflag::MAXLABEL;		! procedure 'called' flags
var	Iprocs::MAXIPROC;		! interface procedure usage map
var	CRefs;				! global data referenes to code
var	Entrypoint;			! label of initial entry point
var	Skip;				! flag, used to suspend output
var	Tmpname::256;			! temporary file name
var	Addr;				! Input position
var	Pass;				! Pass number
var	Bpw;				! t.bpw() cache
var	XFFFFFFFF;			! Constant 0xFFFFFFFF
var	Na, L1, L2, Sf;			! Tcode flags

! reference (call) structure
struct	REF =	Rdest,	! label of called procedure
		Rnext;	! next REF

var	Refs[REF*MAXREF], Rtop;		! references, top ptr

length(s) return t.memscan(s, 0, 32767);

init() do var i;
	if (t.getenv("TMP", Tmpname, 256) < 1 /\
		t.getenv("tmp", Tmpname, 256) < 1
	) do
		Tmpname::0 := '.';
		Tmpname::1 := 0;
	end
	i := length(Tmpname);
	t.memcopy(@Tmpname::i, "/__txdpf.tmp", 13);
	Bpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Ilen := 0;
	Iptr := 0;
	Optr := 0;
	Addr := 0;
	Infile := 0;
	Outfile := 1;
	for (i=0, MAXLABEL) Procs[i] := UNUSED;
	Crefs := UNUSED;
	Entrypoint := UNUSED;
	Rtop := 0;
	Skip := 0;
	t.memfill(Iprocs, 0, MAXIPROC);
end

writep(f, s) t.write(f, s, length(s));

nl(f) do var b::3;
	writep(f, t.newline(b));
end

! convert number to padded string
var	ntoa_buf::64;
ntoa(v, w) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
	ntoa_buf::63:= 0;
	i := 62;
	while (v \/ i = 62) do
		ntoa_buf::i := v mod 10 + '0'; i := i-1;
		v := v/10;
		w := w-1;
	end
	if (g) do
		ntoa_buf::i := '-'; i := i-1;
		w := w-1;
	end
	while (w>0) do
		ntoa_buf::i := '\s'; i := i-1;
		w := w-1;
	end
	return @ntoa_buf::(i+1);
end

error(m) do
	writep(T3X.SYSERR, "TXDPF(");
	writep(T3X.SYSERR, ntoa(Pass, 0));
	writep(T3X.SYSERR, "): ");
	writep(T3X.SYSERR, ntoa(Addr, 0));
	writep(T3X.SYSERR, ": ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	halt 1;
end

rdch() do
	if (Iptr >= Ilen) do
		if (Ilen < 0) error("unexpected EOF");
		Ilen := t.read(Infile, Inbuf, BUFLEN);
		if (Ilen < 1) do
			Ilen := -1;
			return -1;
		end
		Iptr := 0;
	end
	Iptr := Iptr+1;
	Addr := Addr+1;
	return Inbuf::(Iptr-1);
end

flush() do
	if (t.write(Outfile, Outbuf, Optr) \= Optr)
		error("file write error");
	Optr := 0;
end

wrch(c) do
	if (Skip) return 0;
	if (Optr >= BUFLEN) flush();
	Outbuf::Optr := c;
	Optr := Optr+1;
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Bpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Bpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdop(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (Bpw < 4 /\ m /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		error("value out of range");
	return sext32(n);
end

wrop(n, k) do
	ie (k = 0) do
		wrch(n);
	end
	else ie (k = 1) do
		wrch(n >> 8);
		wrch(n);
	end
	else do
		wrch(n >> 24);
		wrch(n >> 16);
		wrch(n >> 8);
		wrch(n);
	end
end

rdcode() do var fl, op;
	op := rdch();
	if (op < 0) return -1;
	ie (op & tcode.IARFLAG) do
		fl := op;
		op := op << 8 | rdch();
		Na := fl & tcode.IARMASK >> tcode.IAMOFF;
		L1 := fl & tcode.IL1MASK >> tcode.IL1OFF;
		L2 := fl & tcode.IL2MASK >> tcode.IL2OFF;
		Sf := fl & tcode.ISTFLAG;
	end
	else do
		Na := 0;
		Sf := 0;
	end
	return op;
end

wrcode(op) do
	if (op >> 8 & tcode.IARFLAG) wrch(op >> 8);
	wrch(op);
end

gen1(i, o) do var k;
	ie (-127 <= o /\ o <= 127) do
		wrch(i>>8);
		k := 0;
	end
	else ie (-32767 <= o /\ o <= 32767) do
		wrch(i >> 8 | (1 << tcode.IL1OFF));
		k := 1;
	end
	else do
		wrch(i >> 8 | (3 << tcode.IL1OFF));
		k := 3;
	end
	wrch(i);
	wrop(o, k);
end

! advance to next Tcode instruction
next(op) do var a, i;
	wrcode(op);
	if (Na) do
		a := rdop(L1);
		wrop(a, L1);
		if (Na > 1) do
			a := rdop(L2);
			wrop(a, L2);
		end
	end
	if (Sf) for (i=0, a) wrch(rdch());
	return rdcode();
end

! Add new reference (a) and link into list pointed to by p
newref(p, a) do var n;
	if (Rtop+REF >= REF*MAXREF) error("out of space");
	n := @Refs[Rtop];
	Rtop := Rtop+REF;
	n[Rdest] := a;
	n[Rnext] := p[0];
	p[0] := n;
	return n;
end

! Add reference, if not already contained in p[0]
addref(p, a) do var q;
	q := p[0];
	while (q \= UNUSED /\ q \= LEAF) do
		if (q[Rdest] = a) return 0;
		q := q[Rnext];
	end
	newref(p, a);
end

! Phase 1: build call tree
phase1() do var op, o2, a, plab, prev;
	Pass := 1;
	Addr := 0;
	Infile := T3X.SYSIN;
	Outfile := t.open(Tmpname, T3X.OWRITE);
	if (Outfile < 0) error("could not create tmp file");
	plab := -1;	! label of current procedure
	op := rdcode();
	prev := -1;	! label created in previous round
	while (op \= -1) do
		o2 := op & tcode.IOPMASK;
		if (	prev \= -1 /\
			(o2 = tcode.IHDR & tcode.IOPMASK \/
			 o2 = tcode.IMHDR & tcode.IOPMASK)
		) do
			plab := prev;
			if (plab >= MAXLABEL)
				error("procedure label too big");
			Procs[plab] := LEAF;
		end
		prev := -1;
		ie (	o2 = tcode.ICALL & tcode.IOPMASK \/
			o2 = tcode.ILDLAB & tcode.IOPMASK \/
			o2 = tcode.ICREF & tcode.IOPMASK \/
			o2 = tcode.ICLAB & tcode.IOPMASK
		) do
			wrcode(op);
			a := rdop(L1);
			if (a >= MAXLABEL) error("label too big");
			wrop(a, L1);
			ie (o2 = tcode.ICLAB & tcode.IOPMASK) do
				ie (a = Entrypoint)	! IEP has no HDR
					plab := a;
				else
					prev := a;
			end
			else ie (o2 = tcode.ICREF & tcode.IOPMASK /\ plab = -1)
			do
				! CREF not in procedure
				addref(@CRefs, a);
			end
			else do
				if (plab >= MAXLABEL)
					error("procedure label too big");
				! CALL and LDLAB make no sense when not
				!  in a procedure
				if (plab = -1)
					error("illegal reference");
				addref(@Procs[plab], a);
			end
			op := rdcode();
		end
		else ie (o2 = tcode.IICALL & tcode.IOPMASK) do
			wrcode(op);
			a := rdop(L1);
			wrop(a, L1);
			if (a >= MAXIPROC) error("interface label too big");
			Iprocs::a := 1;
			op := rdcode();
		end
		else ie (o2 = tcode.IINIT & tcode.IOPMASK) do
			wrcode(op);
			a := rdop(L1);
			wrop(a, L1);
			if (a \= 8) error("unsupported Tcode version");
			a := rdop(L2);
			wrop(a, L2);
			if (Entrypoint \= UNUSED)
				error("multiple entrypoints");
			Entrypoint := a;
			op := rdcode();
		end
		else do
			! end of procedure: clear plab
			if (	o2 = tcode.IEND & tcode.IOPMASK \/
				o2 = tcode.IENDM & tcode.IOPMASK
			)
				plab := -1;
			op := next(op);
		end
	end
	flush();
	t.close(Outfile);
end

! mark all procedures which may be called by p
mark(p) do var q, r;
	if (Uflag::p) return 0;
	Uflag::p := 1;
	q := Procs[p];
	while (q \= UNUSED /\ q \= LEAF) do
		mark(q[Rdest]);
		q := q[Rnext];
	end
end

! Phase 2: mark reachable procedures
phase2() do var i, p;
	Pass := 2;
	Addr := 0;
	for (i=0, MAXLABEL) Uflag::i := 0;
	if (Entrypoint \= UNUSED) mark(Entrypoint);
	p := CRefs;
	while (p \= UNUSED /\ p \= LEAF) do
		mark(p[Rdest]);
		p := p[Rnext];
	end
end

! Phase 3: copy (saved) input to output, skipping unused procedures
phase3() do var op, o2, onext, a, i, s, fl;
	Pass := 3;
	Addr := 0;
	Ilen := 0;
	Iptr := 0;
	Infile := t.open(Tmpname, T3X.OREAD);
	if (Infile < 0) error("could not reopen tmp file");
	Outfile := T3X.SYSOUT;
	op := rdcode();
	while (op \= -1) do
		fl := op >> 8 & tcode.IARFLAG-> op>>8: 0;
		o2 := op & tcode.IOPMASK;
		ie (o2 = tcode.ICLAB & tcode.IOPMASK) do
			a := rdop(fl & tcode.IL1MASK >> tcode.IL1OFF);
			op := rdcode();
			o2 := op & tcode.IOPMASK;
			! Found a procedure (CLAB n HDR|MHDR).
			! If it is unused, suspend output.
			if (	o2 = tcode.IHDR & tcode.IOPMASK \/
				o2 = tcode.IMHDR & tcode.IOPMASK
			)
				Skip := \Uflag::a;
			gen1(tcode.ICLAB, a);
			onext := next(op);
			! Found the end of a procedure (CLAB n END|ENDM)
			! Resume output.
			if (	o2 = tcode.IEND & tcode.IOPMASK \/
				o2 = tcode.IENDM & tcode.IOPMASK
			)
				Skip := 0;
			op := onext;
		end
		else ie (o2 = tcode.IIPROC & tcode.IOPMASK) do
			a := rdop(L1);
			s := Skip;
			Skip := \Iprocs::a;
			wrcode(op);
			wrop(a, L1);
			a := rdop(L2);
			wrop(a, L2);
			for (i=0, a) wrch(rdch());
			op := rdcode();
			o2 := op & tcode.IOPMASK;
			if (o2 = tcode.ICMAP & tcode.IOPMASK) do
				op := next(op);
				o2 := op & tcode.IOPMASK;
			end
			Skip := s;
		end
		else do
			op := next(op);
		end
	end
	t.close(Infile);
	flush();
	t.remove(Tmpname);
end

dumprefs() do var i, p;
	for (i=0, MAXLABEL) do
		if (Procs[i] \= UNUSED) do
			writep(T3X.SYSOUT, Uflag::i-> "*": "\s");
			writep(T3X.SYSOUT, ntoa(i,5));
			writep(T3X.SYSOUT, ": ");
			p := Procs[i];
			while (p \= UNUSED /\ p \= LEAF) do
				writep(T3X.SYSOUT, ntoa(p[Rdest],0));
				if (p[Rnext] \= UNUSED /\ p[Rnext] \= LEAF)
					writep(T3X.SYSOUT, ",");
				p := p[Rnext];
			end
			nl(T3X.SYSOUT);
		end
	end
	p := CRefs;
	writep(T3X.SYSOUT, " CREFs: ");
	while (p \= -1) do
		writep(T3X.SYSOUT, ntoa(p[Rdest],0));
		if (p[Rnext] \= UNUSED /\ p[Rnext] \= LEAF)
			writep(T3X.SYSOUT, ",");
		p := p[Rnext];
	end
	nl(T3X.SYSOUT);
	writep(T3X.SYSOUT, "  INIT: ");
	writep(T3X.SYSOUT, ntoa(Entrypoint,0));
	nl(T3X.SYSOUT);
end

do
	init();
	phase1();
	phase2();
	phase3();
	!dumprefs();
end
