! TXLD -- Load Tcode8 Programs
! Copyright (C) 1999,2000,2002,2003 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: CAT MODULE ... | TXLD >OUTFILE

module txld(t3x, tcode);

object	t[t3x];

const	BUFSIZE	=	1024;	! I/O buffer size
const	TEXTLEN	=	256;	! max. symbol size
const	PUBSIZE	=	6144;	! public symbol table size (words)
const	EXTSIZE	=	2048;	! external symbol map size (words)
const	INFSIZE	=	2048;	! interface symbol table size (words)
const	REFSIZE =	1024;	! interface map size (words)
const	MAXMODS	=	256;	! max. number of input modules
const	LIBMAX =	256;	! space for external module names

! public symbol structure
struct	PSYM = P_left, P_right, P_lab;
const	P_name = PSYM;	! name is appended to above struct

! external/public mappig structure
struct	XMAP = X_id, X_lab;

var	Buffer::BUFSIZE, Cp, Ep;	! input buffer, char ptr, end ptr
var	Obuffer::BUFSIZE, Obp;		! output buffer, free ptr
var	Symbols;			! ptr to symbol table (Pubtbl)
var	Isymbols;			! ptr to Interface sym. table (Inftbl)
var	Pubtbl[PUBSIZE], Ptop;		! public symbol table, top ptr
var	Exttbl[EXTSIZE], Etop;		! ext. symbol table, top ptr
var	Inftbl[EXTSIZE], Itop;		! interface symbol table, top ptr
var	Reftbl[REFSIZE], Rtop;		! ref. symbol table, top ptr
var	Modlist[MAXMODS], Mtop;		! Module list, top ptr
var	Liblist[LIBMAX], Ltop;		! Library list, top ptr
var	Addr;				! curr. input address
var	Base, Bot;			! base address, max address+1
var	Ibase, Ibot;			! interface base, interface max
var	Errflag;			! Error flag
var	Bpw;
var	XFFFFFFFF;			! Constant 0xFFFFFFFF

init() do
	Bpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Ep := 1;
	Cp := 1;
	Obp := 0;
	Addr := 0;
	Base := 0;
	Bot := 0;
	Ibase := 0;
	Ibot := 0;
	Ptop := 0;
	Etop := 0;
	Mtop := 0;
	Ltop := 0;
	Itop := 0;
	Rtop := 0;
	Symbols := 0;
	Isymbols := 0;
	Errflag := 0;
end

length(s) return t.memscan(s, 0, 32767);

writep(f, s) t.write(f, s, length(s));

nl(f) do var b::3;
	writep(f, t.newline(b));
end

! convert number to string
var	ntoa_buf::32;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
	ntoa_buf::31:= 0;
	i := 30;
	while (v \/ i = 30) do
		ntoa_buf::i := v mod 10 + '0'; i := i-1;
		v := v/10;
	end
	if (g) do
		ntoa_buf::i := '-'; i := i-1;
	end
	return @ntoa_buf::(i+1);
end

report(m, n) do var o;
	writep(T3X.SYSERR, "TXLD: ");
	writep(T3X.SYSERR, ntoa(Addr));
	writep(T3X.SYSERR, ": ");
	writep(T3X.SYSERR, n);
	writep(T3X.SYSERR, ": ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	Errflag := 1;
end

error(m) do
	writep(T3X.SYSERR, "TXLD: ");
	writep(T3X.SYSERR, ntoa(Addr));
	writep(T3X.SYSERR, ": ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	halt 1;
end

rdch() do
	if (Cp >= Ep) do
		Ep := t.read(T3X.SYSIN, Buffer, BUFSIZE);
		Cp := 0;
		if (Ep < 0) Ep := 0;
		if (\Ep) return %1;
	end
	Cp := Cp+1;
	Addr := Addr+1;
	return Buffer::(Cp-1);
end

flush() do
	t.write(T3X.SYSOUT, Obuffer, Obp);
	Obp := 0;
end

wrch(n) do
	if (Obp >= BUFSIZE) flush();
	Obuffer::Obp := n;
	Obp := Obp+1;
end

! find symbol and parent (pp)
findsym(name, pp) do var k, y, r, parent;
	y := Symbols;
	parent := 0;
	k := length(name);
	while (y) do
		r := t.memcomp(name, @y[P_name], k+1);
		if (\r) do
			if (pp) pp[0] := parent->
				@parent[r<0-> P_left: P_right]: 0;
			return y;
		end
		parent := y;
		y := r<0-> y[P_left] : y[P_right];
	end
	if (pp) pp[0] := parent-> @parent[r<0-> P_left: P_right]: 0;
	return 0;
end

addsym(name, lab) do var k, y, new, pp;
	y := findsym(name, @pp);
	if (y) do
		report("duplicate symbol", name);
		return y;
	end
	k := length(name);
	if ((Ptop + k + 2) / 2 + PSYM >= PUBSIZE)
		error("symbol table overflow");
	new := @Pubtbl[Ptop];
	Ptop := Ptop + (k+2)/2 + PSYM;
	new[P_lab] := lab;
	new[P_left] := 0;
	new[P_right] := 0;
	t.memcopy(@new[P_name], name, k+1);
	ie (pp) do
		pp[0] := new;
	end
	else do
		Symbols := new;
	end
	return new;
end

! associate an external label with a real (public) label
addref(name, id) do var y, a, new;
	y := findsym(name, 0);
	ie (\y) do
		report("unresolved external", name);
		a := -1;
	end
	else do
		a := y[P_lab];
	end
	if (Etop + XMAP >= EXTSIZE)
		error("link map overflow");
	new := @Exttbl[Etop];
	Etop := Etop+XMAP;
	new[X_id] := id;
	new[X_lab] := a;
	return a;
end

lookup(id) do var i;
	for (i=0, Etop, XMAP) do
		if (Exttbl[i+X_id] = id)
			return Exttbl[i+X_lab];
	end
	report("bad external reference", ntoa(id));
end

! find interface symbol and parent (pp)
findisym(name, pp) do var k, y, r, parent;
	y := Isymbols;
	parent := 0;
	k := length(name);
	while (y) do
		r := t.memcomp(name, @y[P_name], k+1);
		if (\r) do
			if (pp) pp[0] := parent->
				@parent[r<0-> P_left: P_right]: 0;
			return y;
		end
		parent := y;
		y := r<0-> y[P_left] : y[P_right];
	end
	if (pp) pp[0] := parent-> @parent[r<0-> P_left: P_right]: 0;
	return 0;
end

addisym(name, lab) do var k, y, new, pp;
	y := findisym(name, @pp);
	if (y) do
		report("duplicate interface symbol", name);
		return y;
	end
	k := length(name);
	if ((Itop + k + 2) / 2 + PSYM >= INFSIZE)
		error("interface symbol table overflow");
	new := @Inftbl[Itop];
	Itop := Itop + (k+2)/2 + PSYM;
	new[P_lab] := lab;
	new[P_left] := 0;
	new[P_right] := 0;
	t.memcopy(@new[P_name], name, k+1);
	ie (pp) do
		pp[0] := new;
	end
	else do
		Isymbols := new;
	end
	return new;
end

! associate an external label with a real (public) label
addiref(name, id) do var y, a, new;
	y := findisym(name, 0);
	ie (\y) do
		report("unresolved interface", name);
		a := -1;
	end
	else do
		a := y[P_lab];
	end
	if (Rtop + XMAP >= REFSIZE)
		error("reference map overflow");
	new := @Reftbl[Rtop];
	Rtop := Rtop+XMAP;
	new[X_id] := id;
	new[X_lab] := a;
	return a;
end

ilookup(id) do var i;
	for (i=0, Rtop, XMAP) do
		if (Reftbl[i+X_id] = id)
			return Reftbl[i+X_lab];
	end
	report("bad interface reference", ntoa(id));
end

! memorize init address of module
addmodule(id) do
	if (Mtop >= MAXMODS) error("too many modules");
	Modlist[Mtop] := id;
	Mtop := Mtop+1;
end

addlib(name, len) do
	if (Ltop + len > LIBMAX) error("too many libraries");
	Liblist::Ltop := len;
	t.memcopy(@Liblist::(Ltop+1), name, len);
	Ltop := Ltop + len+1;
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Bpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Bpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdop(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (Bpw < 4 /\ m /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		error("value out of range");
	return sext32(n);
end

wrop(n) do
	ie (-127 <= n /\ n <= 127) do
		wrch(n);
	end
	else ie (-32767 <= n /\ n <= 32767) do
		wrch(n >> 8);
		wrch(n);
	end
	else do
		wrch(n >> 24);
		wrch(n >> 16);
		wrch(n >> 8);
		wrch(n);
	end
end

gen1(i, o) do
	ie (-127 <= o /\ o <= 127) wrch(i>>8);
	else ie (-32767 <= o /\ o <= 32767) wrch(i >> 8 | (1 << tcode.IL1OFF));
	else wrch(i >> 8 | (3 << tcode.IL1OFF));
	wrch(i);
	wrop(o);
end

opsize(n) return -127 <= n /\ n <= 127 -> 0:
		-32767 <= n /\ n <= 32767 -> 1: 3;
	

mkflags(na, sf, a1, a2) do var fl;
	if (\na) return 0;
	fl := na << tcode.IAMOFF | tcode.IARFLAG;
	fl := sf | fl;
	fl := opsize(a1) << tcode.IL1OFF | fl;
	if (na > 1) fl := opsize(a2) << tcode.IL2OFF | fl;
	return fl;
end

load() do
	var	op, fl, na, sf, a1, a2, i, n, k, skip;
	var	scan, sbuf::TEXTLEN;

	addlib("t3x", 4);
	skip := 0;
	while (1) do
		scan := 0;	! 0 = pass thru, 1 = collect
		op := rdch();
		if (\Ep) leave;
		ie (op & tcode.IARFLAG) do
			fl := op;
			op := rdch();
			na := fl & tcode.IARMASK >> tcode.IAMOFF;
			sf := fl & tcode.ISTFLAG;
		end
		else do
			fl := 0;
			na := 0;
			sf := 0;
		end
		if (op .>= tcode.IENDOFSET) error("illegal opcode");
		if (na) do
			a1 := rdop(fl & tcode.IL1MASK >> tcode.IL1OFF);
			if (na > 1)
				a2 := rdop(fl & tcode.IL2MASK >> tcode.IL2OFF);
		end
		if (skip) do
			skip := skip-1;
			loop;
		end
		if (op = tcode.IINIT & tcode.IOPMASK) do
			if (a1 \= 8) error("unsupported Tcode version");
			Base := Bot;
			Ibase := Ibot;
			Etop := 0;
			Rtop := 0;
			skip := 2; ! swallow CALL,HALT
		end
		ie (	op = tcode.ICLAB & tcode.IOPMASK \/
			op = tcode.IDLAB & tcode.IOPMASK \/
			op = tcode.ICREF & tcode.IOPMASK \/
			op = tcode.IDREF & tcode.IOPMASK \/
			op = tcode.IINCG & tcode.IOPMASK \/
			op = tcode.ILDG & tcode.IOPMASK \/
			op = tcode.ILDGV & tcode.IOPMASK \/
			op = tcode.ILDLAB & tcode.IOPMASK \/
			op = tcode.ISAVG & tcode.IOPMASK \/
			op = tcode.ICALL & tcode.IOPMASK \/
			op = tcode.IBRF & tcode.IOPMASK \/
			op = tcode.IBRT & tcode.IOPMASK \/
			op = tcode.INBRF & tcode.IOPMASK \/
			op = tcode.INBRT & tcode.IOPMASK \/
			op = tcode.IJUMP & tcode.IOPMASK \/
			op = tcode.IUNEXT & tcode.IOPMASK \/
			op = tcode.IDNEXT & tcode.IOPMASK \/
			op = tcode.IPUB & tcode.IOPMASK \/
			op = tcode.IINIT & tcode.IOPMASK
		) do
			! adjust address (relocate)
			ie (op = tcode.IINIT & tcode.IOPMASK) do
				a2 := a2 + Base;
				if (a2 >= Bot) Bot := a2+1;
			end
			else do
				a1 := a1 + Base;
				if (a1 >= Bot) Bot := a1+1;
			end
		end
		else ie (op = tcode.IIPROC & tcode.IOPMASK) do
			! adjust address (relocate)
			a1 := a1 + Ibase;
			if (a1 >= Ibot) Ibot := a1;
		end
		else ie (op = tcode.ICALX & tcode.IOPMASK) do
			i := lookup(a1);
			if (i \= -1) do
				op := tcode.ICALL & tcode.IOPMASK;
				a1 := i;
			end
		end
		else if (op = tcode.IICALX & tcode.IOPMASK) do
			i := ilookup(a1);
			if (i \= -1) do
				op := tcode.IICALL & tcode.IOPMASK;
				a1 := i;
			end
		end
		if (	op = tcode.IEXT & tcode.IOPMASK \/
			op = tcode.IPUB & tcode.IOPMASK \/
			op = tcode.IIREF & tcode.IOPMASK \/
			op = tcode.ILIB & tcode.IOPMASK
		)
			scan := 1;
		if (op = tcode.IINIT & tcode.IOPMASK) do
			addmodule(a2);
			scan := 1;		! swallow all but first INIT
			if (\Base) do;		! redirect first INIT to CLAB 0
				wrch(fl & ~(3 << tcode.IL2OFF));
				wrch(op);
				wrop(a1);
				wrop(0);
			end
		end
		if (\scan) do
			if (fl) wrch(mkflags(na, sf, a1, a2));
			wrch(op);
			if (na) do
				wrop(a1);
				if (na > 1) wrop(a2);
			end
		end
		if (sf) do
			k := (op = tcode.ISTR & tcode.IOPMASK \/
				op = tcode.ILIB & tcode.IOPMASK)-> a1: a2;
			for (i=0, k) do
				n := rdch();
				sbuf::i := n;
				if (\scan) wrch(n);
			end
			sbuf::i := 0;
			if (	op = tcode.ILIB & tcode.IOPMASK /\
				t.memcomp(sbuf, "t3x", 4)
			)
				addlib(sbuf, a1+1);
		end
		if (Base = 0 /\ op = tcode.IINIT & tcode.IOPMASK)
			gen1(tcode.IJUMP, 0);	! jump to main procedure
		if (op = tcode.IPUB & tcode.IOPMASK)
			addsym(sbuf, a1);
		if (op = tcode.IIPROC & tcode.IOPMASK)
			addisym(sbuf, a1);
		i := 0;
		if (op = tcode.IEXT & tcode.IOPMASK)
			i := addref(sbuf, a1) = -1;
		if (op = tcode.IIREF & tcode.IOPMASK)
			i := addiref(sbuf, a1) = -1;
		if (i = -1) do
			wrch(fl);
			wrch(op);
			wrop(a1);
			wrop(a2);
			for (i=0, a2) wrch(sbuf::i);
		end
	end
	! the new main procedure begins here
	gen1(tcode.ICLAB, 0);
	! emit an init call for each module
	for (i=0, Mtop) if (Modlist[i]) do
		gen1(tcode.ICALL, Modlist[i]);
		gen1(tcode.ICLEAN, 0);
		wrch(tcode.IPOP);
	end
	! in case the main program returns
	gen1(tcode.IHALT, 0);
end

wrlibdir() do var ld, i, n::3, ln;
	if (Ltop = 0) return;
	ld := t.open("LIBDIR", T3X.OWRITE);
	if (ld = -1) error("cannot create LIBDIR file");
	i := 0;
	t.newline(n);
	ln := n::1-> 2: 1;
	while (i < Ltop) do
		t.write(ld, @Liblist::(i+1), Liblist::i - 1);
		t.write(ld, n, ln);
		i := i + Liblist::i + 1;
	end
	t.close(ld);
end

do
	init();
	load();
	wrlibdir();
	flush();
	if (Errflag) halt 1;
end
