! TXOPT -- A Tcode8 Optimizer.
! Copyright (C) 1999,2000,2002,2003 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: TXOPT <INFILE >OUTFILE

module txopt(t3x, tcode);

object	t[t3x];

const	POOLSIZE	= 16000;	! size of memory pool
const	STACKLEN	= 256;		! expression stack size
const	BUFLEN		= 1024; 	! I/O buffer size
const	STRLEN		= 256;		! string length
const	MAXLABEL	= 4096;		! max. labels per procedure

! Tcode instruction node
struct	TCI =
	TC_OP,
	TC_A1,
	TC_A2,
	TC_STR,
	TC_LEFT,
	TC_RIGHT;

struct	COMMTBL =
	CM_PREC,
	CM_INSTR,
	CM_REPL;

struct	TRUTH_VALUES = TRUE, FALSE, UNKNOWN;

var	Pool[POOLSIZE], Ptop;		! memory pool, top ptr
var	Inbuf::BUFLEN, Cp, Ep;		! input buffer, char ptr, end ptr
var	Outbuf::BUFLEN, Obp;		! output buffer, end ptr
var	Addr;				! current input address
var	Checkmagic;			! magic check still to do?
var	Xstack[STACKLEN], Xsp;		! expression stack
var	Comtab;				! table of commutative ops
var	Baselbl;			! base label
var	Labels[MAXLABEL];		! pointers to labelled code
var	Bpw;				! t.bpw() cache
var	XFFFFFFFF;			! Constant 0xFFFFFFFF

init() do
	Bpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Ptop := 0;
	Cp := 0;
	Ep := 0;
	Obp := 0;
	Addr := 0;
	Checkmagic := 1;
	Comtab := [
	! P  instruction	replacement	(P = precedence)
	[ 1, tcode.INORMB,	tcode.INORMB	],
	[ 2, tcode.IMUL,	tcode.IMUL	],
	[ 2, tcode.IUMUL,	tcode.IUMUL	],
	[ 3, tcode.IADD,	tcode.IADD	],
	[ 4, tcode.IBAND,	tcode.IBAND	],
	[ 4, tcode.IBOR,	tcode.IBOR	],
	[ 4, tcode.IBXOR,	tcode.IBXOR	],
	[ 5, tcode.IEQU,	tcode.IEQU	],
	[ 5, tcode.INEQU,	tcode.INEQU	],
	[ 6, tcode.ILESS,	tcode.IGRTR	],
	[ 6, tcode.IGRTR,	tcode.ILESS	],
	[ 6, tcode.ILTEQ,	tcode.IGTEQ	],
	[ 6, tcode.IGTEQ,	tcode.ILTEQ	],
	[ 6, tcode.IULESS,	tcode.IUGRTR	],
	[ 6, tcode.IUGRTR,	tcode.IULESS	],
	[ 6, tcode.IULTEQ,	tcode.IULTEQ	],
	[ 6, tcode.IUGTEQ,	tcode.IUGTEQ	],
	[ 9, tcode.ILDG,	0		],
	[ 9, tcode.ILDL,	0		],
	[ 9, tcode.ILDI,	0		],
	[ 0, 0,			0		],
	];
end

length(s) return t.memscan(s, 0, 32767);

writep(f, s) t.write(f, s, length(s));

nl(f) do var b::3;
	writep(f, t.newline(b));
end

! convert number to string
var	ntoa_buf::64;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
	ntoa_buf::63:= 0;
	i := 62;
	while (v \/ i = 62) do
		ntoa_buf::i := v mod 10 + '0'; i := i-1;
		v := v/10;
	end
	if (g) do
		ntoa_buf::i := '-'; i := i-1;
	end
	return @ntoa_buf::(i+1);
end

fatal(m, n) do
	writep(T3X.SYSERR, "TXOPT: ");
	writep(T3X.SYSERR, ntoa(Addr));
	writep(T3X.SYSERR, ": ");
	writep(T3X.SYSERR, m);
	if (n) do
		writep(T3X.SYSERR, ": ");
		writep(T3X.SYSERR, n);
	end
	nl(T3X.SYSERR);
	writep(T3X.SYSERR, "terminating."); nl(T3X.SYSERR);
	halt 1;
end

rdch() do
	if (Cp >= Ep) do
		if (Ep = -1) return -1;
		Ep := t.read(T3X.SYSIN, Inbuf, BUFLEN);
		if (Ep < 1) do
			Ep := -1;
			return -1;
		end
		Cp := 0;
	end
	Cp := Cp+1;
	Addr := Addr+1;
	return Inbuf::(Cp-1);
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Bpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Bpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdop(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (Bpw < 4 /\ m /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		fatal("value out of range", 0);
	return sext32(n);
end

rdcode() do var op;
	op := rdch();
	if (op = -1) return -1;
	if (op & tcode.IARFLAG) op := op << 8 | rdch();
	return op;
end

alloc(n) do var p;
	n := (n+Bpw-1)/Bpw;
	if (n+Ptop .>= POOLSIZE) fatal("out of memory", 0);
	p := @Pool[Ptop];
	Ptop := Ptop+n;
	return p;
end

savestr(s, k) do var p;
	p := alloc(k);
	t.memcopy(p, s, k);
	return p;
end

newnode(op, a1, a2, l, r) do var ti;
	ti := alloc(TCI*t.bpw());
	ti[TC_OP] := op;
	ti[TC_A1] := a1;
	ti[TC_A2] := a2;
	ti[TC_STR] := 0;
	ti[TC_LEFT] := l;
	ti[TC_RIGHT] := r;
	return ti;
end

flush() do
	if (t.write(T3X.SYSOUT, Outbuf, Obp) \= Obp)
		fatal("file write error", 0);
	Obp := 0;
end

wrch(c) do
	if (Obp >= BUFLEN) flush();
	Outbuf::Obp := c;
	Obp := Obp + 1;
end

wrop(a, k) do
	if (k = 0) do
		wrch(a);
		return;
	end
	if (k = 1) do
		wrch(a >> 8);
		wrch(a);
		return;
	end
	wrch(a >> 24);
	wrch(a >> 16);
	wrch(a >> 8);
	wrch(a);
end

emitnode(node) do var op, fl, na, sf, i, k, s;
	op := node[TC_OP];
	if (op & tcode.IOPMASK = tcode.IGLUE & tcode.IOPMASK) return;
	ie (op >> 8 & tcode.IARFLAG) do
		fl := op >> 8;
		na := fl & tcode.IARMASK >> tcode.IAMOFF;
		sf := fl & tcode.ISTFLAG;
		wrch(fl);
	end
	else do
		fl := 0;
		na := 0;
		sf := 0;
	end
	wrch(op);
	if (na) do
		k := node[TC_A1];
		wrop(k, fl & tcode.IL1MASK >> tcode.IL1OFF);
		if (na > 1) do
			k := node[TC_A2];
			wrop(k, fl & tcode.IL2MASK >> tcode.IL2OFF);
		end
	end
	if (sf) do
		s := node[TC_STR];
		for (i=0, k) wrch(s::i);
	end
end

writeproc(tree) do var parent, tmp;
	parent := 0;
	while (tree) do
		ie (tree[TC_LEFT]) do
			tmp := tree[TC_LEFT];
			tree[TC_LEFT] := parent;
			parent := tree;
			tree := tmp;
		end
		else ie (tree[TC_RIGHT]) do
			tmp := tree[TC_RIGHT];
			tree[TC_RIGHT] := parent;
			parent := tree;
			tree := tmp;
		end
		else do
			emitnode(tree);
			tree := parent;
			if (\tree) leave;
			ie (tree[TC_LEFT]) do
				parent := tree[TC_LEFT];
				tree[TC_LEFT] := 0;
			end
			else do
				parent := tree[TC_RIGHT];
				tree[TC_RIGHT] := 0;
			end
		end
	end
end

! copy an instruction into memory
! op = opcode of instruction to copy
copyi(op) do
	var	k, j;
	var	o2, fl, na, sf, a1, a2;
	var	sbuf::STRLEN;
	var	ti;

	fl := op >> 8;
	o2 := op & tcode.IOPMASK;
	ie (op >> 8 & tcode.IARFLAG) do
		na := fl & tcode.IARMASK >> tcode.IAMOFF;
		sf := fl & tcode.ISTFLAG;
	end
	else do
		fl := 0;
		na := 0;
		sf := 0;
	end
	if (o2 .>= tcode.IENDOFSET) fatal("illegal opcode", 0);
	if (na) do
		k := rdop(fl & tcode.IL1MASK >> tcode.IL1OFF);
		a1 := k;
		if (na > 1) do
			k := rdop(fl & tcode.IL2MASK >> tcode.IL2OFF);
			a2 := k;
		end
	end
	if (o2 = tcode.ICLAB & tcode.IOPMASK /\ a1 < Baselbl)
		Baselbl := a1;
	if (o2 = tcode.IINIT & tcode.IOPMASK /\ a1 \= 8)
		fatal("unsupported Tcode version", 0);
	if (sf) do
		for (j=0, k) sbuf::j := rdch();
		sbuf::j := 0;
	end
	ti := newnode(op, a1, a2, 0, 0);
	if (sf) ti[TC_STR] := savestr(sbuf, k+1);
	return ti;
end

isload(x) return
	x & tcode.IOPMASK >= tcode.ILDG & tcode.IOPMASK /\
	x & tcode.IOPMASK <= tcode.ISELF & tcode.IOPMASK;

isbinop(x) return
	x & tcode.IOPMASK >= tcode.IMUL & tcode.IOPMASK /\
	x & tcode.IOPMASK <= tcode.IUGTEQ & tcode.IOPMASK \/
	x & tcode.IOPMASK >= tcode.IDEREF & tcode.IOPMASK /\
	x & tcode.IOPMASK <= tcode.INORMB &tcode.IOPMASK;

isexpr(x) return
	x = tcode.INEG \/
	x = tcode.ILNOT \/
	x = tcode.IBNOT \/
	isbinop(x) \/
 	isload(x);

link(tree, ti) do
	if (\tree) return ti;
	ti[TC_LEFT] := tree;
	return ti;
end

! dumptree(tree, indent) do var i;
!	while (tree) do
!		for (i=0, indent) writep(T3X.SYSOUT, ".");
!		writep(T3X.SYSOUT, rt.lookup(tree[TC_OP] & tcode.IOPMASK));
!		writep(T3X.SYSOUT, "\n");
!		dumptree(tree[TC_LEFT], indent+1);
!		tree := tree[TC_RIGHT];
!		indent := indent + 1;
!	end
! end

copyx(op, pnop) do var ti, tree, i, failed;
	Xsp := 0;
	failed := 0;
	while (isexpr(op)) do
		ti := copyi(op);
		ie (isload(op)) do
			if (Xsp >= STACKLEN)
				fatal("expression stack overflow", 0);
			Xstack[Xsp] := ti;
			Xsp := Xsp+1;
		end
		else ie (isbinop(op)) do
			if (Xsp < 2) do
				failed := 1;
				leave;
			end
			ti[TC_LEFT] := Xstack[Xsp-2];
			ti[TC_RIGHT] := Xstack[Xsp-1];
			Xstack[Xsp-2] := ti;
			Xsp := Xsp-1;
		end
		else do	! unary operation
			if (Xsp < 1) do
				failed := 1;
				leave;
			end
			ti[TC_LEFT] := Xstack[Xsp-1];
			Xstack[Xsp-1] := ti;
		end
		op := rdcode();
	end
	if (failed) do
		if (Xsp >= STACKLEN) fatal("expression stack overflow", 0);
		Xstack[Xsp] := ti;
		Xsp := Xsp + 1;
		op := rdcode();
	end
	pnop[0] := op;
	tree := 0;
	for (i=0, Xsp) tree := link(tree,
		newnode(tcode.IGLUE, 0, 0, 0, Xstack[i]));
	return newnode(tcode.IGLUE, 0, 0, 0, tree);
end

! read a procedure and its prefix (data declarations) into memory
readproc() do var op, o2, k, ti, tree;
	Ptop := 0;
	tree := 0;
	op := rdcode();
	o2 := op & tcode.IOPMASK;
	if (Checkmagic /\ o2 \= tcode.IINIT & tcode.IOPMASK)
		fatal("magic match failed", 0);
	Checkmagic := 0;
	! read everything up to beginning of next procedure
	while ( op \= -1 /\
		o2 \= tcode.IHDR & tcode.IOPMASK /\
		o2 \= tcode.IMHDR & tcode.IOPMASK
	) do
		ti := copyi(op);
		tree := link(tree, ti);
		op := rdcode();
		o2 := op & tcode.IOPMASK;
	end
	writeproc(tree);
	Ptop := 0;
	tree := 0;
	if (op = -1) return -1;
	Baselbl := 32767;
	! read everything up to end of procedure
	while ( op \= -1 /\
		o2 \= tcode.IEND & tcode.IOPMASK /\
		o2 \= tcode.IENDM & tcode.IOPMASK
	) do
		ie (isexpr(op)) do
			ti := copyx(op, @op);
		end
		else do
			ti := copyi(op);
			op := rdcode();
		end
		tree := link(tree, ti);
		o2 := op & tcode.IOPMASK;
	end
	if (op = -1) fatal("unexpected end of procedure", 0);
	return link(tree, newnode(op, 0, 0, 0, 0));
end

isop(x, op) return x /\ x[TC_OP] & tcode.IOPMASK = op & tcode.IOPMASK;

isconst(x) return isop(x, tcode.INUM);

isval(x, v) return isconst(x) /\ x[TC_A1] = v;

optype(x) do
	if (-127 <= x /\ x <= 127) return 0;
	if (-32767 <= x /\ x <= 32767) return 1;
	return 3;
end

mkop(ti, op, a1, a2, l, r) do var fl, na;
	fl := op >> 8;
	na := fl & tcode.IARMASK >> tcode.IAMOFF;
	if (na) do
		fl := optype(a1) << tcode.IL1OFF | fl;
		if (na > 1) fl := optype(a2) << tcode.IL2OFF | fl;
	end
	ti[TC_OP] := fl << 8 | op;
	ti[TC_A1] := a1;
	ti[TC_A2] := a2;
	ti[TC_STR] := 0;
	ti[TC_LEFT] := l;
	ti[TC_RIGHT] := r;
end

mkconst(ti, n) mkop(ti, tcode.INUM, n, 0, 0, 0);

zapl(tree, op) do
	tree[TC_OP] := op;
	tree[TC_LEFT] := 0;
end

zapr(tree, op) do
	tree[TC_OP] := op;
	tree[TC_RIGHT] := 0;
end

fold_mul(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0) \/ isval(r, 0)) return mkconst(tree, 0);
	if (isval(l, 1)) return zapl(tree, tcode.IGLUE);
	if (isval(r, 1)) return zapr(tree, tcode.IGLUE);
	if (isval(l, -1)) return zapl(tree, tcode.INEG);
	if (isval(r, -1)) return zapr(tree, tcode.INEG);
end

fold_div(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0)) return mkconst(tree, 0);
	if (isval(r, 0)) fatal("divide by zero", 0);
	if (isval(r, 1)) zapr(tree, tcode.IGLUE);
	if (isval(r, -1)) zapr(tree, tcode.INEG);
end

fold_umul(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0) \/ isval(r, 0)) return mkconst(tree, 0);
	if (isval(l, 1)) return zapl(tree, tcode.IGLUE);
	if (isval(r, 1)) return zapr(tree, tcode.IGLUE);
end

fold_udiv(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0)) return mkconst(tree, 0);
	if (isval(r, 0)) fatal("divide by zero", 0);
	if (isval(r, 1)) zapr(tree, tcode.IGLUE);
end

fold_mod(tree) return fold_udiv(tree);

fold_add(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0)) return zapl(tree, tcode.IGLUE);
	if (isval(r, 0)) return zapr(tree, tcode.IGLUE);
end

fold_sub(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0)) return zapl(tree, tcode.INEG);
	if (isval(r, 0)) return zapr(tree, tcode.IGLUE);
end

fold_band(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0) \/ isval(r, 0)) return mkconst(tree, 0);
end

fold_bor(tree) return fold_add(tree);

fold_bshl(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0) \/ isval(r, 0)) return zapr(tree, tcode.IGLUE);
end

fold_bshr(tree) return fold_bshl(tree);

fold_norm(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(l, 0)) return zapl(tree, tcode.IGLUE);
end

fold_normb(tree) do var l, r;
	l := tree[TC_LEFT];
	r := tree[TC_RIGHT];
	if (\l \/ \r) return;
	if (isval(r, 0)) return zapr(tree, tcode.IGLUE);
	if (isval(l, 0)) return zapl(tree, tcode.IGLUE);
end

foldx(tree) do var op, al, ar;
	if (\tree) return;
	foldx(tree[TC_LEFT]);
	foldx(tree[TC_RIGHT]);
	op := tree[TC_OP] & tcode.IOPMASK;
	if (	(op = tcode.ILNOT & tcode.IOPMASK \/
		 op = tcode.IBNOT & tcode.IOPMASK \/
		 op = tcode.INEG & tcode.IOPMASK) /\
		isconst(tree[TC_LEFT])
	) do
		al := tree[TC_LEFT][TC_A1];
		ie (op = tcode.ILNOT & tcode.IOPMASK) al := \al;
		else ie (op = tcode.IBNOT & tcode.IOPMASK) al := ~al;
		else al := -al;
		mkconst(tree, al);
		return;
	end
	if (	isbinop(op) /\
		isconst(tree[TC_LEFT]) /\
		isconst(tree[TC_RIGHT])
	) do
		al := tree[TC_LEFT][TC_A1];
		ar := tree[TC_RIGHT][TC_A1];
		ie (op = tcode.IMUL & tcode.IOPMASK) al := al * ar;
		else ie (op = tcode.IDIV & tcode.IOPMASK) do
			if (\ar) fatal("divide by zero", 0);
			al := al / ar;
		end
		else ie (op = tcode.IUMUL & tcode.IOPMASK) al := al .* ar;
		else ie (op = tcode.IUDIV & tcode.IOPMASK) do
			if (\ar) fatal("divide by zero", 0);
			al := al ./ ar;
		end
		else ie (op = tcode.IMOD & tcode.IOPMASK) do
			if (\ar) fatal("divide by zero", 0);
			al := al mod ar;
		end
		else ie (op = tcode.IADD & tcode.IOPMASK) al := al + ar;
		else ie (op = tcode.ISUB & tcode.IOPMASK) al := al - ar;
		else ie (op = tcode.IBAND & tcode.IOPMASK) al := al & ar;
		else ie (op = tcode.IBOR & tcode.IOPMASK) al := al | ar;
		else ie (op = tcode.IBXOR & tcode.IOPMASK) al := al ^ ar;
		else ie (op = tcode.IBSHL & tcode.IOPMASK) al := al << ar;
		else ie (op = tcode.IBSHR & tcode.IOPMASK) al := al >> ar;
		else ie (op = tcode.IEQU & tcode.IOPMASK) al := al = ar;
		else ie (op = tcode.INEQU & tcode.IOPMASK) al := al \= ar;
		else ie (op = tcode.ILESS & tcode.IOPMASK) al := al < ar;
		else ie (op = tcode.IGRTR & tcode.IOPMASK) al := al > ar;
		else ie (op = tcode.ILTEQ & tcode.IOPMASK) al := al <= ar;
		else ie (op = tcode.IGTEQ & tcode.IOPMASK) al := al >= ar;
		else ie (op = tcode.IULESS & tcode.IOPMASK) al := al .< ar;
		else ie (op = tcode.IUGRTR & tcode.IOPMASK) al := al .> ar;
		else ie (op = tcode.IULTEQ & tcode.IOPMASK) al := al .<= ar;
		else ie (op = tcode.IUGTEQ & tcode.IOPMASK) al := al .>= ar;
		else fatal("internal: foldx: bad op", 0);
		mkconst(tree, al);
		return;
	end
	if (op = tcode.IMUL & tcode.IOPMASK) return fold_mul(tree);
	if (op = tcode.IDIV & tcode.IOPMASK) return fold_div(tree);
	if (op = tcode.IMUL & tcode.IOPMASK) return fold_umul(tree);
	if (op = tcode.IUMUL & tcode.IOPMASK) return fold_udiv(tree);
	if (op = tcode.IUDIV & tcode.IOPMASK) return fold_mod(tree);
	if (op = tcode.IMOD & tcode.IOPMASK) return fold_mod(tree);
	if (op = tcode.IADD & tcode.IOPMASK) return fold_add(tree);
	if (op = tcode.ISUB & tcode.IOPMASK) return fold_sub(tree);
	if (op = tcode.IBAND & tcode.IOPMASK) return fold_band(tree);
	if (op = tcode.IBOR & tcode.IOPMASK) return fold_bor(tree);
	if (op = tcode.IBSHL & tcode.IOPMASK) return fold_bshl(tree);
	if (op = tcode.IBSHR & tcode.IOPMASK) return fold_bshr(tree);
	if (op = tcode.INORM & tcode.IOPMASK) return fold_norm(tree);
	if (op = tcode.INORMB & tcode.IOPMASK) return fold_normb(tree);
end

fold_incr(tree) do var op, xop, newop, incr, expr;
	op := tree[TC_OP] & tcode.IOPMASK;
	ie (op = tcode.ISAVG & tcode.IOPMASK) xop := tcode.ILDG;
	else ie (op = tcode.ISAVL & tcode.IOPMASK) xop := tcode.ILDL;
	else ie (op = tcode.ISAVI & tcode.IOPMASK) xop := tcode.ILDI;
	else return 0;
	if (\tree[TC_LEFT] \/ tree[TC_LEFT][TC_OP] \= tcode.IGLUE)
		return 0;
	expr := tree[TC_LEFT];
	if (\expr[TC_RIGHT] \/ expr[TC_RIGHT][TC_OP] \= tcode.IGLUE)
		return 0;
	expr := expr[TC_RIGHT];
	if (\expr[TC_RIGHT]) return 0;
	expr := expr[TC_RIGHT];
	if (\isconst(expr[TC_RIGHT])) return 0;
	incr := expr[TC_RIGHT][TC_A1];
	op := expr[TC_OP] & tcode.IOPMASK;
	ie (op = tcode.IADD & tcode.IOPMASK) ;
	else ie (op = tcode.ISUB & tcode.IOPMASK) incr := -incr;
	else return 0;
	op := expr[TC_LEFT][TC_OP] & tcode.IOPMASK;
	if (op \= xop & tcode.IOPMASK) return 0;
	if (tree[TC_A1] \= expr[TC_LEFT][TC_A1]) return 0;
	ie (op = tcode.ILDG & tcode. IOPMASK) newop := tcode.IINCG;
	else ie (op = tcode.ILDL & tcode. IOPMASK) newop := tcode.IINCL;
	else ie (op = tcode.ILDI & tcode. IOPMASK) newop := tcode.IINCI;
	else return 0;
	tree[TC_OP] := tcode.IGLUE;
	mkop(expr, newop, expr[TC_LEFT][TC_A1], incr, 0, 0);
	return 1;
end

foldexpr(tree) do
	while (tree) do
		if (	\fold_incr(tree) /\
			tree[TC_RIGHT] /\
			tree[TC_RIGHT][TC_OP] = tcode.IGLUE & tcode.IOPMASK
		) do
			foldx(tree[TC_RIGHT]);
		end
		tree := tree[TC_LEFT];
	end
end

truthval(tree, pti) do
	if (	isop(tree, tcode.IGLUE) /\
		isop(tree[TC_RIGHT], tcode.IGLUE) /\
		isconst(tree[TC_RIGHT][TC_RIGHT])
	) do
		pti[0] := tree[TC_RIGHT][TC_RIGHT];
		return tree[TC_RIGHT][TC_RIGHT][TC_A1]-> TRUE: FALSE;
	end
	return UNKNOWN;
end

islognot(tree, pti) do
	if (	isop(tree, tcode.IGLUE) /\
		isop(tree[TC_RIGHT], tcode.IGLUE) /\
		isop(tree[TC_RIGHT][TC_RIGHT], tcode.ILNOT)
	) do
		pti[0] := tree[TC_RIGHT][TC_RIGHT];
		return 1;
	end
	return 0;
end

repl(ti, op) ti[TC_OP] := ti[TC_OP] & ~tcode.IOPMASK | (op & tcode.IOPMASK);

foldcond(tree) do var op, tv, ti, bf;
	while (tree) do
		op := tree[TC_OP] & tcode.IOPMASK;
		if (	op = tcode.IBRT & tcode.IOPMASK \/
			op = tcode.IBRF & tcode.IOPMASK
		) do
			bf := op = tcode.IBRF & tcode.IOPMASK;
			tv := truthval(tree[TC_LEFT], @ti);
			ie (tv = TRUE) do
				ti[TC_OP] := tcode.IGLUE;
				ie (bf)
					tree[TC_OP] := tcode.IGLUE;
				else
					repl(tree, tcode.IJUMP);
			end
			else ie (tv = FALSE) do
				ti[TC_OP] := tcode.IGLUE;
				ie (bf)
					repl(tree, tcode.IJUMP);
				else
					tree[TC_OP] := tcode.IGLUE;
			end
			else if (islognot(tree[TC_LEFT], @ti)) do
				ti[TC_OP] := tcode.IGLUE;
				repl(tree, bf-> tcode.IBRT: tcode.IBRF);
			end
		end
		tree := tree[TC_LEFT];
	end
end

common(x1, x2) do var na;
	if (x1 = 0) return x2 = 0;
	if (x2 = 0) return x1 = 0;
	na := x1[TC_OP] >> 8 & tcode.IARMASK >> tcode.IAMOFF;
	if (	x1[TC_OP] \= x2[TC_OP] \/
		na > 0 /\ x1[TC_A1] \= x2[TC_A1] \/
		na > 1 /\ x1[TC_A2] \= x2[TC_A2]
	)
		return 0;
	if (\x1) return 1;
	return common(x1[TC_LEFT], x2[TC_LEFT]) /\
		common(x1[TC_RIGHT], x2[TC_RIGHT]);
end

eliminate(left, right) do
	if (	common(left[TC_LEFT], right[TC_LEFT]) /\
		common(left[TC_RIGHT], right[TC_RIGHT])
	) do
		mkop(right[TC_LEFT], tcode.IDUP, 0, 0, 0, 0);
		mkconst(right[TC_RIGHT], 0);
	end
end

cxelim(tree) do var expr;
	while (tree) do
		if (isop(tree, tcode.IGLUE)) do
			expr := tree[TC_RIGHT];
			ie (	isop(expr, tcode.IGLUE) /\
				isop(expr[TC_LEFT], tcode.IGLUE) /\
				isop(expr[TC_LEFT][TC_RIGHT], tcode.INORM) /\
				expr[TC_RIGHT] /\
				isop(expr[TC_RIGHT][TC_LEFT], tcode.IDEREF)
			) do
				eliminate(expr[TC_LEFT][TC_RIGHT],
					expr[TC_RIGHT][TC_LEFT]);
			end
			else if (isop(expr, tcode.IGLUE) /\
				isop(expr[TC_LEFT], tcode.IGLUE) /\
				isop(expr[TC_LEFT][TC_RIGHT], tcode.INORMB) /\
				expr[TC_RIGHT] /\
				isop(expr[TC_RIGHT][TC_LEFT], tcode.IDREFB)
			) do
				eliminate(expr[TC_LEFT][TC_RIGHT],
					expr[TC_RIGHT][TC_LEFT]);
			end
		end
		tree := tree[TC_LEFT];
	end
end

comop(ti) do var i, op;
	if (ti = 0) return 0;
	op := ti[TC_OP] & tcode.IOPMASK;
	i := 0;
	while (Comtab[i][0]) do
		if (op = Comtab[i][CM_INSTR] & tcode.IOPMASK)
			return Comtab[i];
		i := i+1;
	end
end

xreorder(expr) do var base, left, right;
	if (\expr) return;
	xreorder(expr[TC_LEFT]);
	xreorder(expr[TC_RIGHT]);
	base := comop(expr);
	if (\base) return;
	left := comop(expr[TC_LEFT]);
	if (\left) return;
	right := comop(expr[TC_RIGHT]);
	if (\right) return;
	if (base[CM_REPL] /\ left[CM_PREC] > right[CM_PREC]) do var tmp;
		tmp := expr[TC_RIGHT];
		expr[TC_RIGHT] := expr[TC_LEFT];
		expr[TC_LEFT] := tmp;
		expr[TC_OP] := base[CM_REPL];
	end
end

reorder(tree) do
	while (tree) do
		if (	tree[TC_RIGHT] /\
			tree[TC_RIGHT][TC_OP] = tcode.IGLUE & tcode.IOPMASK
		) do
			xreorder(tree[TC_RIGHT]);
		end
		tree := tree[TC_LEFT];
	end
end

findlabels(tree) do var op, last, lbl;
	last := 0;
	t.memfill(Labels, 0, MAXLABEL*t.bpw());
	while (tree) do
		op := tree[TC_OP] & tcode.IOPMASK;
		if (op = tcode.ICLAB & tcode.IOPMASK /\ last) do
			lbl := tree[TC_A1] - Baselbl;
			if (lbl < 0 \/ lbl >= MAXLABEL)
				fatal("label out of range", 0);
			if (Labels[lbl]) fatal("duplicate label", 0);
			Labels[lbl] := last;
		end
		if (tree[TC_OP] \= tcode.IGLUE \/ tree[TC_RIGHT])
			last := tree;
		tree := tree[TC_LEFT];
	end
end

dest(ti) do var lbl;
	lbl := ti[TC_A1] - Baselbl;
	ie (lbl >= 0 /\ lbl < MAXLABEL)
		return Labels[lbl];
	else
		return 0;
end

redirect(dl) do var lbl;
	while (1) do
		lbl := dl-Baselbl;
		if (lbl < 0 \/ lbl > MAXLABEL \/ Labels[lbl] = 0) leave;
		if (	Labels[lbl][TC_OP] & tcode.IOPMASK \=
			tcode.IJUMP & tcode.IOPMASK
		)
			leave;
		dl := Labels[lbl][TC_A1];
	end
	return dl;
end

jumpredir(tree) do var op, d;
	findlabels(tree);
	while (tree) do
		op := tree[TC_OP] & tcode.IOPMASK;
		if (op = tcode.IJUMP & tcode.IOPMASK) do
			d := dest(tree);
			if (d /\ d[TC_LEFT] /\ d[TC_LEFT][TC_LEFT] = tree)
				tree[TC_OP] := tcode.IGLUE;
			tree[TC_A1] := redirect(tree[TC_A1]);
		end
		tree := tree[TC_LEFT];
	end
end

do var tree;
	init();
	while (1) do
		tree := readproc();
		if (tree = -1) leave;
		foldexpr(tree);
		foldcond(tree);
		cxelim(tree);
		reorder(tree);
		jumpredir(tree);
		writeproc(tree);
	end
	flush();
end
