! TXTRN -- a T3X->Tcode translator, Version 8
! Copyright (C) 1997-2004 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: TXTRN <T3X-PROGRAM >TCODE-PROGRAM
!
! TXTRN is a compiler for a simple, portable, recursive,
! block-structured, typeless, object-oriented language
! called T3X.
! The translator reads a source program from sysin and
! writes an equivalent Tcode8 program program to sysout.
! Compiler messages are sent to syserr.

module txtrn(t3x, tcode);

object	t[t3x];

const	BUFSIZE=	1026,	! input buffer size
	OBUFL=		1024,	! output buffer size
	SYMTBLSIZE=	8192,	! symbol table size
	NAMEPOOLSIZE=	8192,	! name pool size
	TEXTLEN=	129,	! max. token length
	MAXTBL=		128;	! max. member per table

const	META=	256;

! symbol table flags
const	CNST=	1,	! Symbol Flags
	GLOBL=	2,
	PROC=	4,
	PROTO=	8,
	IPROC=	32,
	PUBLC=	64,
	EXTRN=	128,
	CLSS=	256,
	OBJCT=	512,
	IVAR=	1024,
	REDIR=	2048,
	LINKD=	4096;

struct SYMREC = SNAME,
		SVAL,	! CNST: value		other: address
		SSIZE,	! CLSS: elements	OBJCT: ptr to class
		SFLAGS;

! byte-order-independent SYMREC
struct XSYMREC =
		XHHVAL,	XHLVAL, XLHVAL, XLLVAL,
		XHHSIZE, XHLSIZE, XLHSIZE, XLLSIZE,
		XHFLGS,	XLFLGS;

var	Classpaths,	! system class paths [see initvars()]
	Syscp::TEXTLEN,	! system CP
	Usrcp::TEXTLEN;	! user CP

var	Symbols[SYMTBLSIZE], St, Dt;	! symbol table, top ptr, dir ptr
var	Sym_equ, Sym_mod, Sym_sub,	! ptrs to frequently used
	Sym_add, Sym_mul, Sym_bor,	!  operators
	Sym_bnot;
var	Names::NAMEPOOLSIZE, Nt;	! name pool, top ptr
var	Line;				! input line no
var	File::TEXTLEN;			! input file name
var	Modname::TEXTLEN, Modctx;	! module name, context
var	Errcount;			! error counter
var	Token;				! curr. token
var	Text::TEXTLEN;			! token text
var	Value;				! token value
var	Op;				! curr. operator
var	Initdone;			! init sequence emitted
var	Label, Xlabel;			! next label / ext. label
var	Ilabel, Infid;			! next inf. label / inf. proc id
var	Startlab,			! loop context: start /
	Exitlab;			!  end label
var	Endlab;				! proc context: end label
var	Looplevl;			! Locladdr of loop context
var	Locladdr;			! next free local addr
var	Buffer::BUFSIZE;		! input buffer
var	Cp, Ep;				!  char ptr, end ptr
var	Lowp, Nomore;			!  low water mark, EOF flag
var	Obuf::OBUFL, Obp;		! output buffer, free ptr
var	Last::10, LL;			! delay buffer, length
var	Ops;				! operator table
var	Bcslot;				! INTERFACE slot
var	Debug;				! #debug flag
var	Xbpw;				! #bpw parameter
var	Classctx;			! curr. class context
var	Classoff;			! next free ivar addr
var	Infctx;				! Interface context
var	Selfobj;			! self
var	Method_adjust;			! addr adjustment flag (1=in method)
var	B4F, B8F,			! constant 0xFFFF and 0xFFFFFFFF
	B16F;				! and 0xFFFFFFFFFFFFFFFF
var	B15, B31;			! constant 0x8000 and 0x80000000
var	Bpw;				! T.BPW() cache

! Token definitions

const	ENDFILE =	-1;

struct	Tokenlist =

	! General
	SYMBOL, NUMBER, STRING,

	! Interpunctuation
	ADDROP, ASSIGN, BINOP, BYTEOP, COLON, COMMA, COND, CONOP, DISOP,
	DOT, LBRACK, LPAREN, METACO, METAOP, RBRACK, RPAREN, SEMI, UNOP,

	! Keywords
	KCALL, KCLASS, KCONST, KDECL, KDO, KELSE, KEND, KFOR, KHALT,
	KICLASS, KIDECL, KIE, KIF, KLEAVE, KLOOP, KMODULE, KOBJECT,
	KPACKED, KPUBLIC, KRETURN, KSEND, KSTRUCT, KVAR, KWHILE;

! Operator structure
struct	OPER =
	OPREC,
	OLEN,
	ONAME,
	OTOK,
	OCODE;

decl	fatal(2), factor(0), expr(0), stmt(0), compound(0), declaration(1);

length(a) return t.memscan(a, 0, 32767);

strequ(a, b) return \t.memcomp(a, b, length(a)+1);

strcpy(a, b) do var k;
	k := length(b);
	t.memcopy(a, b, k+1);
	return k;
end

writep(f, s) t.write(f, s, length(s));

nl(f) do var b::3;
	t.write(f, t.newline(b), length(b));
end

! convert number to string
var	ntoa_buf::32;
ntoa(v, w) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
	ntoa_buf::31:= 0;
	i := 30;
	while (v \/ i = 30) do
		ntoa_buf::i := v mod 10 + '0';
		i := i-1;
		v := v/10;
		w := w-1;
	end
	if (g) do
		ntoa_buf::i := '-';
		i := i-1;
		w := w-1;
	end
	while (w>0) do
		ntoa_buf::i := '\s';
		i := i-1;
		w := w-1;
	end
	return @ntoa_buf::(i+1);
end

error(m, s) do var o, col;
	col := ": ";
	writep(T3X.SYSERR, "TXTRN: ");
	ie (File::0) do
		writep(T3X.SYSERR, File);
		writep(T3X.SYSERR, col);
	end
	else if (Modname::0) do
		writep(T3X.SYSERR, Modname);
		writep(T3X.SYSERR, col);
	end
	writep(T3X.SYSERR, ntoa(Line, 0));
	writep(T3X.SYSERR, col);
	writep(T3X.SYSERR, m);
	if (s) do
		writep(T3X.SYSERR, col);
		writep(T3X.SYSERR, s);
	end
	nl(T3X.SYSERR);
	Errcount := Errcount+1;
	if (Errcount >= 20) do
		Errcount := 0;
		fatal("too many errors", 0);
	end
end

fatal(m, s) do
	error(m, s);
	writep(T3X.SYSERR, "terminating.");
	nl(T3X.SYSERR);
	halt 1;
end

fillbuf() do var i;
	if (Nomore) return 0;
	! move remaining input to start of buffer
	for (i=Cp, Ep) Buffer::(i-Cp) := Buffer::i;
	i := Ep-Cp;
	Cp := 0;
	Ep := t.read(T3X.SYSIN, @Buffer::i, BUFSIZE/2-1);
	if (Ep < 1) Nomore := 1;
	Ep := Ep + i;
	Lowp := Ep-TEXTLEN;
end

eof() return Nomore /\ Cp >= Ep;

getce() do var c;
	c := Buffer::Cp;
	Cp := Cp+1;
	if (c \= '\\') return c;
	c := Buffer::Cp;
	Cp := Cp+1;
	if (c = 'a' \/ c = 'A') return '\a';
	if (c = 'b' \/ c = 'B') return '\b';
	if (c = 'e' \/ c = 'E') return '\e';
	if (c = 'f' \/ c = 'F') return '\f';
	if (c = 'n' \/ c = 'N') return '\n';
	if (c = 'q' \/ c = 'Q' \/ c = '"') return '"' | META;
	if (c = 'r' \/ c = 'R') return '\r';
	if (c = 's' \/ c = 'S') return '\s';
	if (c = 't' \/ c = 'T') return '\t';
	if (c = 'v' \/ c = 'V') return '\v';
	return c;
end

findkw(s) ie (s::0 = 'c') do
		if (strequ(s, "call")) return KCALL;
		if (strequ(s, "class")) return KCLASS;
		if (strequ(s, "const")) return KCONST;
		return 0;
	end
	else ie (s::0 = 'd') do
		if (strequ(s, "decl")) return KDECL;
		if (strequ(s, "do")) return KDO;
		return 0;
	end
	else ie (s::0 = 'e') do
		if (strequ(s, "else")) return KELSE;
		if (strequ(s, "end")) return KEND;
		return 0;
	end
	else ie (s::0 = 'f') do
		if (strequ(s, "for")) return KFOR;
		return 0;
	end
	else ie (s::0 = 'h') do
		if (strequ(s, "halt")) return KHALT;
		return 0;
	end
	else ie (s::0 = 'i') do
		if (strequ(s, "iclass")) return KICLASS;
		if (strequ(s, "idecl")) return KIDECL;
		if (strequ(s, "ie")) return KIE;
		if (strequ(s, "if")) return KIF;
		return 0;
	end
	else ie (s::0 = 'l') do
		if (strequ(s, "leave")) return KLEAVE;
		if (strequ(s, "loop")) return KLOOP;
		return 0;
	end
	else ie (s::0 = 'm') do
		if (strequ(s, "mod")) return BINOP;
		if (strequ(s, "module")) return KMODULE;
		return 0;
	end
	else ie (s::0 = 'o') do
		if (strequ(s, "object")) return KOBJECT;
		return 0;
	end
	else ie (s::0 = 'p') do
		if (strequ(s, "packed")) return KPACKED;
		if (strequ(s, "public")) return KPUBLIC;
		return 0;
	end
	else ie (s::0 = 'r') do
		if (strequ(s, "return")) return KRETURN;
		return 0;
	end
	else ie (s::0 = 's') do
		if (strequ(s, "send")) return KSEND;
		if (strequ(s, "struct")) return KSTRUCT;
		return 0;
	end
	else ie (s::0 = 'v') do
		if (strequ(s, "var")) return KVAR;
		return 0;
	end
	else ie (s::0 = 'w') do
		if (strequ(s, "while")) return KWHILE;
		return 0;
	end
	else do
		return 0;
	end

findop(c) do var nmatch, i, pos;
	nmatch := 0;	! number of matched chars
	i := 0;
	pos := -1;	! position of matched operator, -1=none
	while (nmatch < Ops[i][OLEN]) do
		if (c = Ops[i][ONAME]::nmatch) do
			c := Buffer::Cp;
			Cp := Cp+1;
			pos := i;
			nmatch := nmatch+1;
			if (Ops[i][OLEN] > nmatch) do
				! more chars to match in this slot
				pos := -1;
				i := i-1;
			end
		end
		i := i+1;
	end
	Cp := Cp-1;	! reject non-operator char
	Op := pos;
	return pos = -1-> 0: Ops[pos][OTOK];
end

findop2(s) do var i;
	i := 0;
	while (Ops[i][OLEN]) do
		if (strequ(Ops[i][ONAME], s)) return i;
		i := i+1;
	end
	fatal("bad name in findop2()", s);
end

scan() do
	var	c, i, bc::5;
	var	radix, lastv, dv, ovfl;

	if (Cp >= Lowp) fillbuf();
	c := Buffer::Cp;
	Cp := Cp+1;
	! skip over white space and comments
	while (1) do
		while (	c = '\s' \/ c = '\t' \/ c = '\n' \/
			c = '\r' \/ c = '\f'
		) do
			if (c = '\n') Line := Line+1;
			c := Buffer::Cp;
			Cp := Cp+1;
			if (Cp >= Lowp) do
				if (eof()) return ENDFILE;
				fillbuf();
			end
		end
		if (c \= '!') leave;	! no comment
		while (c \= '\n') do
			if (Cp >= Lowp) do
				if (eof()) return ENDFILE;
				fillbuf();
			end
			c := Buffer::Cp;
			Cp := Cp+1;
		end
	end
	if (eof()) return ENDFILE;
	if (	'a' <= c /\ c <= 'z' \/
		'A' <= c /\ c <= 'Z' \/
		c = '_'
	) do
		i := 0;
		while (1) do
			ie ('A' <= c /\ c <= 'Z')
				c := c-('A'-'a');
			else if (\('a' <= c /\ c <= 'z' \/
				'0' <= c /\ c <= '9' \/
				c = '_')
			)
				leave;
			if (i >= TEXTLEN-1) fatal("symbol too long", 0);
			Text::i := c;
			i := i+1;
			c := Buffer::Cp;
			Cp := Cp+1;
		end
		Text::i := 0;
		Cp := Cp-1;
		c := findkw(Text);
		if (c) do
			if (c = BINOP) Op := Sym_mod;
			return c;
		end
		return SYMBOL;
	end
	if ('0' <= c /\ c <= '9' \/ c = '%') do
		Value := 0;
		radix := 10;
		ovfl := 0;
		i := 0;			! sign flag
		if (c = '%') do
			i := 1;
			c := Buffer::Cp;
			Cp := Cp+1;
		end
		if (c = '0') do		! may be '0x'
			c := Buffer::Cp;
			Cp := Cp+1;
			ie (c = 'x' \/ c = 'X') do
				c := Buffer::Cp;
				Cp := Cp+1;
				radix := 16;
			end
			else if (c = 'b' \/ c = 'B') do
				c := Buffer::Cp;
				Cp := Cp+1;
				radix := 2;
			end
		end
		while (1) do
			lastv := Value;
			ie (radix = 16 /\ 'A' <= c /\ c <= 'F')
				dv := c-('A'-10);
			else ie (radix = 16 /\ 'a' <= c /\ c <= 'f')
				dv := c-('a'-10);
			else ie ('0' <= c /\ c <= '9')
				dv := c-'0';
			else
				leave;
			Value := Value .* radix + dv;
			if (\ovfl /\ (Value - dv) ./ radix \= lastv) do
				error("number out of range", 0);
				ovfl := 1;
			end
			c := Buffer::Cp;
			Cp := Cp+1;
		end
		Cp := Cp-1;
		if (i) Value := -Value;
		return NUMBER;
	end
	if (c = '\'') do
		Value := getce() & ~META;
		c := Buffer::Cp;
		Cp := Cp+1;
		if (c \= '\'') error("missing '\\''", 0);
		return NUMBER;
	end
	if (c = '"') do
		i := 0;
		c := getce();
		while (c \= '"') do
			if (i >= TEXTLEN-1) fatal("string too long", 0);
			Text::i := c & ~META;
			i := i+1;
			c := getce();
			if (eof()) fatal("EOF in string", 0);
		end
		Text::i := 0;
		return STRING;
	end
	i := findop(c);
	if (i) return i;
	! convert input char to hex string
	bc::0 := '0';
	bc::1 := 'x';
	bc::2 := c/16 + (c/16 > 9 -> 'A'-10 : '0');
	bc::3 := c mod 16 + (c mod 16 > 9 -> 'A'-10 : '0');
	bc::4 := 0;
	fatal("bad input character", bc);
end

synch(tok) do
	ie (tok) do	! find sync token
		while (Token \= tok) do
			Token := scan();
			if (Token = ENDFILE)
				fatal("EOF found in error recovery", 0);
		end
	end
	else do		! swallow one token
		Token := scan();
	end
end

extend(x, cbpw) do
	if (cbpw < Bpw) do
		if (cbpw = 2 /\ x & B15) x := B8F & ~B4F | x;
		if (Bpw > 4 /\ x & B31) x := B16F & ~B8F | x;
	end
	return x;
end

adjustcache(cache, k, cbpw) do var i, y, x;
	if (Bpw < 4) for (i=0, k) do
		x := @cache::(i*XSYMREC);
		if (	\(x::XHHSIZE = 0x00 /\ x::XHLSIZE = 0x00 \/
			  x::XHHSIZE = 0xFF /\ x::XHLSIZE = 0xFF /\
			  x::XLHSIZE & 0x80)
			\/
			\(x::XHHVAL = 0x00 /\ x::XHLVAL = 0x00 \/
			  x::XHHVAL = 0xFF /\ x::XHLVAL = 0xFF /\
			  x::XLHVAL & 0x80)
		)
			fatal("CLASSDIR has long values", 0);
	end
	ie (Bpw >= 4) do
		! Reverse order is important. Otherwise, source
		! fields may be overwritten.
		for (i=k-1, -1, -1) do
			x := @cache::(i*XSYMREC);
			y := @cache[i*SYMREC];
			! reverse order, see above
			y[SFLAGS] := x::XHFLGS << 8 | x::XLFLGS;
			y[SSIZE] := extend((x::XHHSIZE << 24)
				| (x::XHLSIZE << 16)
				| (x::XLHSIZE << 8)
				| x::XLLSIZE, cbpw);
			y[SVAL] := extend((x::XHHVAL << 24)
				| (x::XHLVAL << 16)
				| (x::XLHVAL << 8)
				| x::XLLVAL, cbpw);
			y[SNAME] := 0;
		end
	end
	else do
		for (i=0, k) do
			x := @cache::(i*XSYMREC);
			y := @cache[i*SYMREC];
			y[SNAME] := 0;
			y[SVAL] := extend((x::XHHVAL << 24)
				| (x::XHLVAL << 16)
				| (x::XLHVAL << 8)
				| x::XLLVAL, cbpw);
			y[SSIZE] := extend((x::XHHSIZE << 24)
				| (x::XHLSIZE << 16)
				| (x::XLHSIZE << 8)
				| x::XLLSIZE, cbpw);
			y[SFLAGS] := x::XHFLGS << 8 | x::XLFLGS;
		end
	end
end

dumptbl(hd, tbl, lim) do var i, y, err;
	err := T3X.SYSERR;
	writep(err, hd);
	nl(err);
	for (i=0, lim, SYMREC) do
		y := @tbl[i];
		writep(err, ntoa(y[SVAL], 6));
		writep(err, ntoa(y[SSIZE], 6));
		writep(err, "\s");
		writep(err, y[SNAME]);
		writep(err, " (");
		writep(err, y[SFLAGS] & CNST-> " CNST": "");
		writep(err, y[SFLAGS] & GLOBL-> " GLOBL": "");
		writep(err, y[SFLAGS] & PROC-> " PROC": "");
		writep(err, y[SFLAGS] & IPROC-> " IPROC": "");
		writep(err, y[SFLAGS] & PROTO-> " PROTO": "");
		writep(err, y[SFLAGS] & PUBLC-> " PUBLC": "");
		writep(err, y[SFLAGS] & EXTRN-> " EXTRN": "");
		writep(err, y[SFLAGS] & CLSS-> " CLSS": "");
		writep(err, y[SFLAGS] & OBJCT-> " OBJCT": "");
		writep(err, y[SFLAGS] & IVAR-> " IVAR": "");
		writep(err, y[SFLAGS] & REDIR-> " REDIR": "");
		writep(err, " )");
		nl(err);
	end
end

rdclsdir(dirname, reqd) do
	var	i, j, k, lib;
	var	magic::4;
	var	rf;
	var	n, nbase;
	var	ndt, nlen;
	var	nb::2;
	var	cbpw;

	rf := "corrupt CLASSDIR";
	lib := t.open(dirname, T3X.OREAD);
	if (lib = -1) do
		if (\reqd) return 0;
		fatal("class not available", dirname);
	end
	if (	t.read(lib, magic, 4) \= 4 \/
		t.memcomp(magic, "CDR", 3)
	)
		fatal(rf, "#1");
	cbpw := magic::3;
	! get number of entries
	if (t.read(lib, nb, 2) \= 2) fatal(rf, "#2");
	n := nb::0 << 8 | nb::1;
	if (Dt + n*XSYMREC >= St)
		fatal("class directory cache overflow", 0);
	nlen := n*XSYMREC;
	ndt := Dt + n*SYMREC;	! new length
	! append entries
	if (t.read(lib, @Symbols[Dt], nlen) \= nlen)
		fatal(rf, "#3");
	adjustcache(@Symbols[Dt], n, cbpw);
	nbase := Nt;		! beginning of new name pool
	! append names
	Nt := Nt + t.read(lib, @Names::Nt, NAMEPOOLSIZE-Nt);
	if (Nt >= NAMEPOOLSIZE \/ Nt < nbase)
		fatal("name space overflow (rdclsdir)", 0);
	j := Dt;
	k := nbase;
	! fix SNAME references in appended symbols
	for (i=nbase, Nt) do
		if (Names::i = 0) do
			Symbols[j+SNAME] := @Names::k;
			k := i+1;
			j := j+SYMREC;
		end
		if (j > ndt) leave;
	end
	! consistency check
	if (j \= ndt) fatal(rf, "#4");
	Dt := ndt;
	t.close(lib);
end

wrclsdir() do
	const	WBUFLEN = 256;
	var	k, i, j, f, lib;
	var	buf::WBUFLEN, wp;
	var	wf, pub, magic;
	var	nb::2, p, q;

	wf := "error writing CLASSDIR";
	k := 0;
	pub := 0;
	! count items to export and fix flags
	for (i=0, Dt, SYMREC) do
		f := Symbols[i+SFLAGS];
		! count only symbols in public classes
		if (f & CLSS) pub := f & PUBLC;
		! clear LINKD and set EXTRN in PROC entries
		!  except for IPROCS
		ie ((f & (PROC|IPROC)) = PROC)
			Symbols[i+SFLAGS] := f & ~LINKD | (f->EXTRN: 0);
		!  clear LINKD on IPROCS
		else if (f & IPROC)
			Symbols[i+SFLAGS] := f & ~LINKD;
		if (pub) k := k+1;
	end
	if (k = 0) return 0;
	lib := t.open("CLASSDIR", T3X.OWRITE);
	if (lib = -1) fatal("could not update CLASSDIR", 0);
	nb::0 := k >> 8;
	nb::1 := k;
	magic := "CDR#";
	magic::3 := Bpw > 4-> 4: Bpw;
	if (	t.write(lib, magic, 4) \= 4 \/
		t.write(lib, nb, 2) \= 2
	)
		fatal(wf, "#1");
	pub := 0;
	wp := 0;	! output buffer ptr
	! write the symbol table entries
	for (i=0, Dt, SYMREC) do
		! export only public classes
		if (Symbols[i+SFLAGS] & CLSS)
			pub := Symbols[i+SFLAGS] & PUBLC;
		if (pub) do
			! flush buffer, if required
			if (wp + XSYMREC >= WBUFLEN) do
				if (t.write(lib, buf, wp) \= wp)
					fatal(wf, "#2");
				wp := 0;
			end
			! append one entry
			p := @buf::wp;
			q := @Symbols[i];
			p::XHHVAL  := q[SVAL] >> 24;
			p::XHLVAL  := q[SVAL] >> 16;
			p::XLHVAL  := q[SVAL] >> 8;
			p::XLLVAL  := q[SVAL];
			p::XHHSIZE := q[SSIZE] >> 24;
			p::XHLSIZE := q[SSIZE] >> 16;
			p::XLHSIZE := q[SSIZE] >> 8;
			p::XLLSIZE := q[SSIZE];
			p::XHFLGS := q[SFLAGS] >> 8;
			p::XLFLGS := q[SFLAGS];
			wp := wp+XSYMREC;
		end
	end
	pub := 0;
	! write the symbol names
	for (i=0, Dt, SYMREC) do
		if (Symbols[i+SFLAGS] & CLSS)
			pub := Symbols[i+SFLAGS] & PUBLC;
		if (\pub) loop;
		j := length(Symbols[i+SNAME]);
		if (j + wp >= WBUFLEN) do
			if (t.write(lib, buf, wp) \= wp)
				fatal(wf, "#3");
			wp := 0;
		end
		t.memcopy(@buf::wp, Symbols[i+SNAME], j+1);
		wp := wp+j+1;
	end
	! flush output buffer
	if (wp /\ t.write(lib, buf, wp) \= wp)
		fatal(wf, "#4");
	t.close(lib);
end

findivar(base, lim, name, rpt) do var i, s;
	i := base+SYMREC;
	while (i < lim /\ Symbols[i+SFLAGS] & IVAR) do
		s := Symbols[i+SNAME];
		if (s::0 = name::0 /\ strequ(name, s)) return i;
		i := i+SYMREC;
	end
	if (rpt) error("undefined symbol", name);
	return -1;
end

findintclass(name) do var i, j, s;
	for (i=0, Dt, SYMREC) do
		s := Symbols[i+SNAME];
		if (	Symbols[i+SFLAGS] & CLSS /\
			s::0 = name::0 /\ strequ(name, s)
		)
			return i;
	end
	return -1;
end

mkclasspath(dest, base, dir) do var i, j, sfx;
	sfx := ".dir";
	j := 0;
	i := 0;
	while (base::i) do
		dest::j := base::i;
		i := i+1;
		j := j+1;
	end
	dest::j := '/';
	j := j+1;
	i := 0;
	while (dir::i) do
		dest::j := dir::i;
		i := i+1;
		j := j+1;
	end
	i := 0;
	while (sfx::i) do
		dest::j := sfx::i;
		i := i+1;
		j := j+1;
	end
	dest::j := 0;
end

findclass(name) do
	var	cl, i;
	var	cldir::TEXTLEN;

	cl := findintclass(name);
	if (cl \= -1) return cl;
	i := 0;
	while (cl = -1) do
		if (\Classpaths[i]) leave;
		if (Classpaths[i]::0) do
			mkclasspath(cldir, Classpaths[i], name);
			rdclsdir(cldir, 0);
			cl := findintclass(name);
		end
		i := i+1;
	end
	if (cl = -1) fatal("class not available", name);
	return cl;
end

findredclass(ctx, name) do var i, cl;
	i := ctx+SYMREC;	! the REDIR list is before the context
	while (i < SYMTBLSIZE) do
		if (\(Symbols[i+SFLAGS] & REDIR)) leave;
		cl := Symbols[i+SVAL];
		if (strequ(Symbols[cl+SNAME], name)) return cl;
		i := i + SYMREC;
	end
	return -1;
end

findclassref(name, rpt) do var k;
	k := Classctx = -1-> -1: findredclass(Classctx, name);
	if (k = -1 /\ Modctx \= -1) k := findredclass(Modctx, name);
	if (rpt /\ k = -1) error("class not available", name);
	return k;
end

findsym(name, rpt) do var i, s;
	i := St;
	while (i < SYMTBLSIZE) do
		s := Symbols[i+SNAME];
		if (s::0 = name::0 /\ strequ(name, s)) return i;
		i := i+SYMREC;
	end
	if (rpt) error("undefined symbol", name);
	return -1;
end

newsym(name, size, val, flags) do var k, sp;
	k := length(name);
	if (findintclass(name) \= -1)
		error("duplicate symbol (class conflict)", name);
	sp := findsym(name, 0);
	if (name::0 \= '*' /\ name::0 \= '+' /\ sp \= -1) do
		if (	flags & (PROC|PROTO) \= PROC \/ 
			Symbols[sp+SFLAGS] & (PROC|PROTO) \= (PROC|PROTO)
		) do
			error("duplicate symbol", name);
			return sp;
		end
	end
	St := St-SYMREC;
	if (St < Dt) fatal("too many symbols", name);
	if (Nt + k >= NAMEPOOLSIZE) fatal("out of name space", name);
	strcpy(@Names::Nt, name);
	Symbols[St+SNAME] := @Names::Nt;
	Nt := Nt+k+1;
	Symbols[St+SSIZE] := size;
	Symbols[St+SVAL] := val;
	Symbols[St+SFLAGS] := flags | ((Classctx \= -1)-> IVAR: 0);
	return St;
end

rptdecls(from, to) do var i, s;
	i := from;
	while (i < to) do
		s := @Symbols[i];
		if (s[SFLAGS] & PROTO)
			error("missing procedure definition",
				s[SNAME]);
		i := i+SYMREC;
	end
end

newlab() do
	if (\Label) fatal("out of labels", 0);
	Label := Label+1;
	return Label-1;
end

flush() do
	t.write(T3X.SYSOUT, Obuf, Obp);
	Obp := 0;
end

decl	initseq(0);

commit() do var i;
	if (\Initdone) initseq();
	for (i=0, LL) do
		if (Obp >= OBUFL) flush();
		Obuf::Obp := Last::i;
		Obp := Obp + 1;
	end
end

gen0(instr) do
	commit();
	Last::0 := instr;
	LL := 1;
end

gen1(instr, arg) do
	commit();
	Last::0 := instr >> 8;
	Last::1 := instr;
	ie (-127 <= arg /\ arg <= 127) do
		Last::2 := arg;
		LL := 3;
	end
	else ie (-32767 <= arg /\ arg <= 32767) do
		Last::0 := (1 << tcode.IL1OFF) | Last::0;
		Last::2 := arg >> 8;
		Last::3 := arg;
		LL := 4;
	end
	else do
		Last::0 := (3 << tcode.IL1OFF) | Last::0;
		Last::2 := arg >> 24;
		Last::3 := arg >> 16;
		Last::4 := arg >> 8;
		Last::5 := arg;
		LL := 6;
	end
end

gen2(instr, a1, a2) do
	commit();
	Last::0 := instr >> 8;
	Last::1 := instr;
	ie (-127 <= a1 /\ a1 <= 127) do
		Last::2 := a1;
		LL := 3;
	end
	else ie (-32767 <= a1 /\ a1 <= 32767) do
		Last::0 := (1 << tcode.IL1OFF) | Last::0;
		Last::2 := a1 >> 8;
		Last::3 := a1;
		LL := 4;
	end
	else do
		Last::0 := (3 << tcode.IL1OFF) | Last::0;
		Last::2 := a1 >> 24;
		Last::3 := a1 >> 16;
		Last::4 := a1 >> 8;
		Last::5 := a1;
		LL := 6;
	end
	ie (-127 <= a2 /\ a2 <= 127) do
		Last::LL := a2;
		LL := LL + 1;
	end
	else ie (-32767 <= a2 /\ a2 <= 32767) do
		Last::0 := (1 << tcode.IL2OFF) | Last::0;
		Last::LL := a2 >> 8;
		Last::(LL+1) := a2;
		LL := LL + 2;
	end
	else do
		Last::0 := (3 << tcode.IL2OFF) | Last::0;
		Last::LL := a2 >> 24;
		Last::(LL+1) := a2 >> 16;
		Last::(LL+2) := a2 >> 8;
		Last::(LL+3) := a2;
		LL := LL + 4;
	end
end

initseq() do
	Initdone := 1;
	gen2(tcode.IINIT, 8, 1);
	gen1(tcode.ICALL, 1);
	gen1(tcode.IHALT, 0);
	if (Xbpw) gen1(tcode.IBPW, Xbpw);
	gen1(tcode.IDATA, 0);
end

stack(n) if (n) gen1(tcode.ISTACK, n);

! adjust addresses of local variables in methods
! (saved caller context uses up an additional word)
mtdadj(n) return n<0-> n: n+Method_adjust;

! generate linkage information for public/external symbols
! type = public (PUB), external (EXT), interface (IPROC,IREF)
! base = index of class
! k    = index of symbol
linkage(type, base, k) do var i, j, lab, s::TEXTLEN*2+1;
	! symbols are imported/exported as: classname_symbolname
	ie (type = tcode.IPUB \/ type = tcode.IIPROC) do
		i := strcpy(s, Symbols[Classctx+SNAME]);
		s::i := '_';
		i := i+1;
		strcpy(@s::i, Symbols[k+SNAME]);
		lab := Symbols[k+SVAL];
	end
	else do
		i := strcpy(s, Symbols[base+SNAME]);
		s::i := '_';
		i := i+1;
		strcpy(@s::i, Symbols[k+SNAME]);
		! assign external label to EXTs and IREFs
		ie (type = tcode.IEXT) do
			if (\Xlabel) fatal("out of external labels", 0);
			lab := Xlabel;
			Xlabel := Xlabel+1;
			Symbols[k+SVAL] := lab;
		end
		else do
			if (\Ilabel) fatal("out of interface labels", 0);
			lab := Ilabel;
			Ilabel := Ilabel+1;
			Symbols[k+SVAL] := lab;
		end
	end
	j := length(s);
	gen2(type, lab, j);
	for (i=0, j) gen0(s::i);
end

! generate code to reference a data object
loadsym(k) do var f, z;
	f := Symbols[k+SFLAGS];
	if (f & PROC) return 0;
	! get size (atomic objects have z=0)
	z := (f & OBJCT)-> Symbols[Symbols[k+SSIZE]+SSIZE]:
		Symbols[k+SSIZE];
	ie (k = Selfobj) do
		gen0(tcode.ISELF);
	end
	else ie (f & GLOBL) do
		ie (f & IVAR)
			gen1(z-> tcode.ILDIV: tcode.ILDI, Symbols[k+SVAL]);
		else
			gen1(z-> tcode.ILDGV: tcode.ILDG, Symbols[k+SVAL]);
	end
	else do
		gen1(z-> tcode.ILDLV: tcode.ILDL, mtdadj(Symbols[k+SVAL]));
	end
end

! generate code to change an atomic data object
savesym(k) do var f;
	f := Symbols[k+SFLAGS];
	ie (f & GLOBL)
		gen1(f & IVAR-> tcode.ISAVI: tcode.ISAVG, Symbols[k+SVAL]);
	else
		gen1(tcode.ISAVL, mtdadj(Symbols[k+SVAL]));
end

! start true branch (branch to false part) of a conditional
! n     = dest label
! flags = 0=unconditional branch, 1=branch on true, 2=branch on false
tcond(n, flags) do
	ie (flags) do
		gen1(flags=1-> tcode.INBRT: tcode.INBRF, n);
		! swallow value, if branch failed
		gen0(tcode.IPOP);
	end
	else do
		gen1(tcode.IBRF, n);
	end
end

! start false branch (branch to end) of conditional
fcond(y, n) do
	gen1(tcode.IJUMP, y);	! jump to end (from true branch)
	gen1(tcode.ICLAB, n);	! set label of false branch
end

! end conditional (false branch)
econd(y) gen1(tcode.ICLAB, y);	! place end label

match(tok, m) do
	ie (tok = Token)
		Token := scan();
	else
		error(m, 0);
end

xsemi() match(SEMI, "missing ';'");

xlparen() match(LPAREN, "missing '('");

xrparen() match(RPAREN, "missing ')'");

xrbrack() match(RBRACK, "missing ']'");

xcomma(swallow) ie (Token = COMMA)
		if (swallow) Token := scan();
	else
		error("missing ','", 0);

equsign() ie (Token \= BINOP \/ Op \= Sym_equ)
		error("missing '='", 0);
	else
		Token := scan();

forcesym(msg) ie (Token = SYMBOL) do
		return 0;
	end
	else do
		error(msg-> msg: "symbol expected", 0);
		synch(SEMI);
		Token := scan();
		return -1;
	end

classconst() do var k, v;
	k := findclassref(Text, 0);
	if (k = -1) error("undefined symbol", Text);
	Token := scan();
	if (k = -1) return -1;
	ie (Token = DOT) do
		Token := scan();
		if (forcesym(0)) return -1;
		k := findivar(k, Dt, Text, 0);
		ie (k \= -1) do
			if (\(Symbols[k+SFLAGS] & CNST))
				error("not a class constant", Text);
			v := Symbols[k+SVAL];
		end
		else do
			error("no such class constant", Text);
		end
		Token := scan();
	end
	else do
		error("missing '.' after class name", 0);
	end
	return v;
end

constfac() do var k, v, osub, onot;
	v := 1;
	osub := 0;
	onot := 0;
	ie (Token = BINOP /\ Op = Sym_sub) do
		osub := 1;
		Token := scan();
	end
	else if (Token = UNOP /\ Op = Sym_bnot) do
		onot := 1;
		Token := scan();
	end
	ie (Token = NUMBER) do
		v := Value;
		Token := scan();
	end
	else ie (Token = SYMBOL) do
		k := findsym(Text, 0);
		ie (k \= -1) do
			ie (Symbols[k+SFLAGS] & CNST)
				v := Symbols[k+SVAL];
			else ie (Symbols[k+SFLAGS] & CLSS) do
				v := Symbols[k+SSIZE];
			end
			else do
				error("not a constant", Text);
			end
			Token := scan();
		end
		else do
			v := classconst();
		end
	end
	else do
		error("constant value expected", 0);
	end
	if (osub) v := -v;
	if (onot) v := ~v;
	return v;
end

constexpr() do var v, o;
	v := constfac();
	while (Token = BINOP) do
		o := Op;
		Token := scan();
		ie (o = Sym_add)
			v := v + constfac();
		else ie (o = Sym_bor)
			v := v | constfac();
		else ie (o = Sym_mul)
			v := v * constfac();
		else ie (o = Sym_sub)
			v := v - constfac();
		else
			leave;
	end
	return v;
end

! generate procedure call
! k   = index of proc/var symbol
! ind = 1=indirect call
pcall(k, ind) do var n, dd::9, dl, i;
	n := 0;
	ie (k = -1) do
		error("call of non-procedure", 0);
	end
	else do
		ie (ind) do
			! ignore CALL before real procedure names
			ie (Symbols[k+SFLAGS] & PROC /\
				\(Symbols[k+SFLAGS] & PUBLC)
			)
				ind := 0;
			! allow only CALLs to atomic variables
			else if ((Symbols[k+SFLAGS] &
				 (CNST|CLSS|OBJCT|PUBLC)) \/
				Symbols[k+SSIZE]
			)
				error("bad indirect call",
					Symbols[k+SNAME]);
		end
		else do
			if (	\(Symbols[k+SFLAGS] & PROC) \/
				Symbols[k+SFLAGS] & PUBLC
			)
				error("call of non-procedure",
					Symbols[k+SNAME]);
		end
	end
	xlparen();
	if (ind) do
		! save reference to procedure pointer
		t.memcopy(dd, Last, 10);
		dl := LL;
		LL := 0;
	end
	! compile argument list
	if (Token \= RPAREN) while (1) do
		expr();
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (\ind /\ k >= 0 /\ Symbols[k+SSIZE] \= n)
		error("wrong number of arguments", Symbols[k+SNAME]);
	if (k >= 0) do
		ie (ind) do
			! re-emit reference to proc ptr
			commit();
			t.memcopy(Last, dd, 10);
			LL := dl;
			gen0(tcode.ICALR);
		end
		else ie (Symbols[k+SFLAGS] & EXTRN) do
			gen1(tcode.ICALX, Symbols[k+SVAL]);
		end
		else do
			gen1(tcode.ICALL, Symbols[k+SVAL]);
		end
		gen1(tcode.ICLEAN, n);
	end
	xrparen();
end

strlit(s) do var i, l, k;
	k := length(s);
	l := newlab();
	gen1(tcode.IDLAB, l);
	gen1(tcode.ISTR, k);
	for (i=0, k) gen0(s::i);
	Token := scan();
	return l;
end

! compile byte vector
bytevec() do var tbl[MAXTBL], v, i, p;
	Token := scan();
	if (Token \= LBRACK) do
		error("packed vector expected", 0);
		synch(0);
		return 0;
	end
	p := 0;
	Token := scan();
	while (Token = SYMBOL \/ Token = NUMBER) do
		tbl[p] := constexpr();
		if (tbl[p] > 255 \/ tbl[p] < -128)
			error("packed table member too big", 0);
		if (p >= MAXTBL) fatal("table too big", 0);
		p := p+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (\p) error("empty table", 0);
	v := newlab();
	gen1(tcode.IDLAB, v);
	gen1(tcode.ISTR, p);
	for (i=0, p) gen0(tbl[i]);
	xrbrack();
	return v;
end

table() do var tv[MAXTBL], tf[MAXTBL], p, v, i, k, dynamic;
	p := 0;
	dynamic := 0;
	Token := scan();
	while (	Token = SYMBOL \/ Token = NUMBER \/
		Token = LBRACK \/ Token = STRING \/
		Token = ADDROP \/ Token = KPACKED \/
		Token = LPAREN \/
		Token = BINOP /\ Op = Sym_sub \/
		Token = UNOP /\ Op = Sym_bnot
	) do
		ie (dynamic \/ Token = LPAREN) do
			tv[p] := newlab();
			tf[p] := tcode.IDLAB;
			if (\dynamic) Token := scan();
			expr();
			gen1(tcode.ISAVG, tv[p]);
			ie (Token = RPAREN) do
				Token := scan();
				dynamic := 0;
			end
			else do
				dynamic := 1;
			end
		end
		else ie (Token = SYMBOL \/ Token = NUMBER \/
			Token = BINOP /\ Op = Sym_sub \/
			Token = UNOP /\ Op = Sym_bnot
		) do
			tv[p] := constexpr();
			tf[p] := tcode.IDATA;
		end
		else ie (Token = STRING) do
			tv[p] := strlit(Text);
			tf[p] := tcode.IDREF;
		end
		else ie (Token = KPACKED) do
			tv[p] := bytevec();
			tf[p] := tcode.IDREF;
		end
		else ie (Token = ADDROP) do
			Token := scan();
			ie (Token = SYMBOL) do
				k := findsym(Text, 1);
			end
			else do
				error("symbol expected", 0);
				k := -1;
			end
			if (k >= 0) do
				ie (Symbols[k+SFLAGS] & (CNST)) error(
					"cannot take @const", Text);
				else ie (Symbols[k+SFLAGS] & PROC)
					tf[p] := tcode.ICREF;
				else ie (Symbols[k+SFLAGS] & GLOBL = GLOBL)
					tf[p] := tcode.IDREF;
				else
					error("unknown address", Text);
			end
			tv[p] := Symbols[k+SVAL];
			Token := scan();
		end
		else do
			tv[p] := table();
			tf[p] := tcode.IDREF;
		end
		if (p >= MAXTBL) fatal("table too big", 0);
		p := p+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	v := newlab();
	gen1(tcode.IDLAB, v);
	if (\p) error("empty table", 0);
	for (i=0, p) do
		ie (tf[i] = tcode.IDLAB) do
			gen1(tcode.IDLAB, tv[i]);
			gen1(tcode.IDATA, 0);
		end
		else do
			gen1(tf[i], tv[i]);
		end
	end
	xrbrack();
	return v;
end

linkmethod(base, m) do
	! If LINKD is not set, this method has not been referenced before.
	! In this case, generate linkage information and set LINKD.
	if (m = -1) return;
	ie (Symbols[m+SFLAGS] & (EXTRN|LINKD) = EXTRN) do
		linkage(tcode.IEXT, base, m);
		Symbols[m+SFLAGS] := Symbols[m+SFLAGS] | LINKD;
	end
	else if (Symbols[m+SFLAGS] & (IPROC|LINKD) = IPROC) do
		linkage(tcode.IIREF, base, m);
		Symbols[m+SFLAGS] := Symbols[m+SFLAGS] | LINKD;
	end
end

! search method 'name' in class 'base' in class dir cache
searchmethod(base, name) do var m;
	m := findivar(base, Dt, name, 0);
	linkmethod(base, m);
	return m;
end

! find an accessible method
! k = index of object
findmethod(k) do var m;
	m := -1;
	! if k is the current class context, search the symbol table
	ie (Symbols[k+SSIZE] = Classctx) do
		m := findivar(St-SYMREC, Classctx, Text, 0);
		linkmethod(Classctx, m);
	end
	! else search the class of k in the class dir cache
	else do
		m := searchmethod(Symbols[k+SSIZE], Text);
	end
	if (m \= -1 /\ \(Symbols[m+SFLAGS] & PROC)) m := -1;
	return m;
end

! send message to object
! k  = index of object
! cl = index of class
sendmsg(k, cl) do
	var	m, n;
	var	dd[9], dl, i;

	! save reference to object (receiver)
	t.memcopy(dd, Last, 10);
	dl := LL;
	LL := 0;
	Token := scan();
	if (cl = -1 /\ (k = -1 \/ Symbols[k+SFLAGS] & OBJCT \= OBJCT)) do
		error("invalid object", k=-1-> 0: Symbols[k+SNAME]);
		k := -1;
	end
	if (forcesym("method name expected")) return 0;
	! if an explicit class is given,
	!  search it for the requested message
	!  (used by SEND operator)
	ie (cl \= -1) do
		m := searchmethod(cl, Text);
	end
	! else search the message in the class of k
	else do
		m := k=-1-> -1: findmethod(k);
	end
	if (m = -1 /\ k \= -1) do
		error("cannot answer message", Text);
	end
	if (m \= -1 /\ Symbols[m+SFLAGS] & (PUBLC|PROC) \= (PUBLC|PROC))
		error("cannot answer message", m=-1-> 0: Symbols[m+SNAME]);
	Token := scan();
	if (Token \= LPAREN) do
		error("incomplete message", m=-1-> 0: Symbols[m+SNAME]);
		synch(SEMI);
		return 0;
	end
	xlparen();
	n := 0;
	! compile argument list
	if (Token \= RPAREN) while (1) do
		expr();
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (m \= -1 /\ Symbols[m+SSIZE] \= n)
		error("wrong number of arguments", Symbols[m+SNAME]);
	! re-emit reference to receiving object
	commit();
	t.memcopy(Last, dd, 10);
	LL := dl;
	if (m >= 0) do
		gen1(Symbols[m+SFLAGS] & EXTRN-> tcode.ICALX:
			Symbols[m+SFLAGS] & IPROC -> tcode.IICALX:
			tcode.ICALL, Symbols[m+SVAL]);
		gen1(tcode.ICLEAN, n+1);
	end
	xrparen();
end

! compute address of a symbol (lvalue)
address() do var tok, k;
	if (Token \= SYMBOL) do
		error("symbol expected", 0);
		return -1;
	end
	k := findsym(Text, 0);
	ie (k = -1) do
		! no symbol? try class constants...
		gen1(tcode.INUM, classconst());
		return -1;	! constants are anonymous
	end
	else do
		Token := scan();
		ie (Symbols[k+SFLAGS] & CNST)
			gen1(tcode.INUM, Symbols[k+SVAL]);
		else ie (Symbols[k+SFLAGS] & CLSS)
			gen1(tcode.INUM, Symbols[k+SSIZE]);
		else
			loadsym(k);
	end
	! process subscripts
	while (Token = LBRACK \/ Token = BYTEOP) do
		if (k >= 0 /\ (Symbols[k+SFLAGS] & (CNST|PROC|OBJCT|CLSS)))
			error("bad subscript", Symbols[k+SNAME]);
		tok := Token;
		Token := scan();
		ie (tok = BYTEOP) do
			factor();
			gen0(tcode.IDREFB);
			return -1;
		end
		else do
			expr();
			gen0(tcode.IDEREF);
			xrbrack();
		end
		k := -1;	! vector elements are anonymous
	end
	return k;
end

sendop() do var p, cl;
	Token := scan();
	xlparen();
	if (forcesym("missing variable in SEND")) return -1;
	p := findsym(Text, 1);
	if (p < 0) do
		synch(SEMI);
		return 0;
	end
	if (Symbols[p+SFLAGS] & (CNST|CLSS|OBJCT|PROC)) do
		error("bad pointer in SEND", Symbols[p+SNAME]);
		synch(SEMI);
		return 0;
	end
	Token := scan();
	xcomma(1);
	if (forcesym("missing class in SEND")) return -1;
	cl := findclassref(Text, 1);
	if (cl < 0) do
		synch(SEMI);
		return 0;
	end
	Token := scan();
	xcomma(0);
	loadsym(p);
	sendmsg(0, cl);
	xrparen();
end

factor() do var o, k, ind;
	ie (Token = NUMBER) do
		gen1(tcode.INUM, Value);
		Token := scan();
	end
	else ie (Token = SYMBOL \/ Token = KCALL) do
		ind := 0;
		if (Token = KCALL) do
			Token := scan();
			ind := 1;
		end
		k := address();
		ie (Token = DOT) do
			if (ind) error("bad CALL before message", 0);
			sendmsg(k, -1);
		end
		else ie (Token = LPAREN)
			pcall(k, ind);
		else if (k >= 0 /\ Symbols[k+SFLAGS] & PROC \/ ind)
			error("incomplete procedure call", Symbols[k+SNAME]);
	end
	else ie (Token = KSEND) do
		sendop();
	end
	else ie (Token = STRING) do
		gen1(tcode.ILDLAB, strlit(Text));
	end
	else ie (Token = KPACKED) do
		gen1(tcode.ILDLAB, bytevec());
	end
	else ie (Token = LBRACK) do
		gen1(tcode.ILDLAB, table());
	end
	else ie (Token = ADDROP) do
		Token := scan();
		k := address();
		ie (k >= 0) do
			! clear referencing code (procedures generate none)
			if (\(Symbols[k+SFLAGS] & PROC)) LL := 0;
			ie (Symbols[k+SFLAGS] & (CNST))
				error("cannot take @const", Symbols[k+SNAME]);
			else ie (Symbols[k+SFLAGS] & PROC)
				gen1(tcode.ILDLAB, Symbols[k+SVAL]);
			else ie (Symbols[k+SFLAGS] & GLOBL)
				gen1(Symbols[k+SFLAGS] & IVAR-> tcode.ILDIV:
					tcode.ILDGV, Symbols[k+SVAL]);
			else
				gen1(tcode.ILDLV, mtdadj(Symbols[k+SVAL]));
		end
		else do
			! reduce vector member values to references
			ie (Last::0 = tcode.IDEREF)
				Last::0 := tcode.INORM;
			else ie (Last::0 = tcode.IDREFB)
				Last::0 := tcode.INORMB;
			else
				error("bad lvalue", 0);
		end
	end
	else ie (Token = BINOP) do
		if (Op \= Sym_sub) error("bad unary operator", 0);
		Token := scan();
		factor();
		gen0(tcode.INEG);
	end
	else ie (Token = UNOP) do
		o := Op;
		Token := scan();
		factor();
		gen0(Ops[o][OCODE]);
	end
	else ie (Token = LPAREN) do
		Token := scan();
		expr();
		xrparen();
	end
	else do
		error("bad expression", 0);
		synch(0);
	end
end

emitop(stk, sp) do
	sp := sp-1;
	gen0(Ops[stk[sp]][OCODE]);
	return sp;
end

! falling precedence parser for binary expressions
binary() do var stk[5], sp;
	sp := 0;
	factor();
	while (Token = BINOP) do
		while (sp /\ Ops[Op][OPREC] <= Ops[stk[sp-1]][OPREC])
			sp := emitop(stk, sp);
		! this should never happen:
		if (sp >= 5) fatal("binary(): stack overflow", 0);
		stk[sp] := Op;
		sp := sp+1;
		Token := scan();
		factor();
	end
	while (sp) sp := emitop(stk, sp);
end

conj() do var e;
	e := 0;
	binary();
	if (Token = CONOP) e := newlab();
	while (Token = CONOP) do
		Token := scan();
		tcond(e, 2);
		binary();
	end
	if (e) econd(e);
end

disj() do var e;
	e := 0;
	conj();
	if (Token = DISOP) e := newlab();
	while (Token = DISOP) do
		Token := scan();
		tcond(e, 1);
		conj();
	end
	if (e) econd(e);
end

! parse expressions (conditional operations, that is)
expr() do var y, n;
	disj();
	if (Token = COND) do
		y := newlab();
		n := newlab();
		Token := scan();
		tcond(n, 0);
		expr();
		fcond(y, n);
		match(COLON, "missing ':'");
		expr();
		econd(y);
	end
end

! parse IF (alt=0) or IE (alt=1) statements
ie_stmt(alt) do var y, n;
	n := newlab();
	if (alt) y := newlab();
	Token := scan();
	xlparen();
	expr();
	xrparen();
	tcond(n, 0);
	stmt();
	if (alt) do
		fcond(y, n);
		match(KELSE, "missing 'ELSE'");
		stmt();
	end
	econd(alt-> y: n);
end

while_stmt() do var s, e, levl;
	! save context of outer loop, if any
	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	! create new loop context
	Startlab := newlab();
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	gen1(tcode.ICLAB, Startlab);
	expr();
	tcond(Endlab, 0);
	xrparen();
	stmt();
	gen1(tcode.IJUMP, Startlab);
	econd(Endlab);
	! restore outer loop context
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end

! see also while_stmt()
for_stmt() do var k, step, r, s, e, levl;
	step := 1;
	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	Startlab := newlab();
	r := newlab();		! re-entrance label (\= Startlab in FOR)
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	! init part
	ie (Token = SYMBOL) do
		k := findsym(Text, 1);
		if (k \= -1 /\ Symbols[k+SFLAGS] & (CNST|PROC|CLSS|OBJCT))
			error("bad lvalue in 'FOR'", Text);
		if (k \= -1 /\ Symbols[k+SSIZE])
			error("index may not be a vector", Text);
	end
	else do
		k := -1;
	end
	match(SYMBOL, "symbol name expected");
	equsign();
	expr();
	if (k \= -1) savesym(k);
	xcomma(1);
	! limit part
	gen1(tcode.ICLAB, r);
	if (k \= -1) loadsym(k);
	expr();
	if (Token = COMMA) do
		Token := scan();
		step := constexpr();
	end
	xrparen();
	gen1(step>0-> tcode.IUNEXT: tcode.IDNEXT, Endlab);
	! body
	stmt();
	! increment part
	gen1(tcode.ICLAB, Startlab);
	ie (Symbols[k+SFLAGS] & GLOBL)
		ie (Symbols[k+SFLAGS] & IVAR)
			gen2(tcode.IINCI, Symbols[k+SVAL], step);
		else
			gen2(tcode.IINCG, Symbols[k+SVAL], step);
	else
		gen2(tcode.IINCL, mtdadj(Symbols[k+SVAL]), step);
	gen1(tcode.IJUMP, r);
	gen1(tcode.ICLAB, Endlab);
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end

asg_or_call() do var k;
	if (forcesym("bad statement")) return 0;
	k := address();
	ie (Token = ASSIGN) do
		Token := scan();
		! assign a value to an atomic data object
		ie (k \= -1) do
			ie (Symbols[k+SFLAGS] & (CNST|PROC|OBJCT|CLSS))
				error("bad lvalue", Symbols[k+SNAME]);
			else if (Symbols[k+SSIZE])
				error("missing subscript",
					Symbols[k+SNAME]);
			LL := 0;
			expr();
			savesym(k);
		end
		! assign a value to a vector member
		else do
			k := Last::0;
			ie (k = tcode.IDEREF)
				Last::0 := tcode.INORM;
			else ie (k = tcode.IDREFB)
				Last::0 := tcode.INORMB;
			else
				error("bad lvalue", 0);
			expr();
			gen0(k=tcode.IDEREF-> tcode.ISTORE: tcode.ISTORB);
		end
		xsemi();
	end
	else ie (Token = LPAREN) do
		pcall(k, 0);
		gen0(tcode.IPOP);	! discard return value
		xsemi();
	end
	else ie (Token = DOT) do
		sendmsg(k, -1);
		gen0(tcode.IPOP);	! discard return value
		xsemi();
	end
	else do
		error("bad statement", 0);
		synch(SEMI);
		xsemi();
	end
end

mcomment() do var c, state, level;
	state := 0;
	level := 1;
	while (level) do
		if (Cp >= Lowp) do
			if (eof()) fatal("EOF in meta comment", 0);
			fillbuf();
		end
		c := Buffer::Cp;
		Cp := Cp+1;
		ie (c = '!') do
			ie (state = 0) do
				state := -1;
			end
			else if (state = 1) do
				level := level + 1;
				state := 0;
			end
		end
		else ie (c = '#') do
			ie (state = 0) do
				state := 1;
			end
			else if (state = -1) do
				level := level - 1;
				state := 0;
			end
		end
		else do
			if (c = '\n') Line := Line+1;
			state := 0;
		end
	end
	Token := scan();
end

metacmd() do var badsyn;
	badsyn := 0;
	Token := scan();
	ie (Token \= SYMBOL) do
		badsyn := 1;
	end
	else ie (strequ(Text, "classpath")) do
		Token := scan();
		ie (Token \= STRING) do
			badsyn := 1;
		end
		else do
			strcpy(Usrcp, Text);
			Classpaths[0] := Usrcp;
			Token := scan();
		end
	end
	else ie (strequ(Text, "debug")) do
		Debug := 1;
		Token := scan();
	end
	else ie (strequ(Text, "bpw")) do
		if (Initdone)
			error("#bpw must precede all other declarations", 0);
		Token := scan();
		ie (Token \= NUMBER)
			badsyn := 1;
		else
			Xbpw := Value;
		if (Xbpw \= 0 /\ Xbpw \= 2 /\ Xbpw \= 4 /\ Xbpw \= 8)
			error("#bpw value must be in {0,2,4,8}", 0);
		Token := scan();
	end
	else do
		Token := scan();
		badsyn := 1;
	end
	if (badsyn) error("invalid # syntax", 0);
	xsemi();
end

stmt() do
	if (Debug) gen1(tcode.ILINE, Line);
	ie (Token = KDO) do
		! this statement has a return value, if the compund
		! statement has one
		return compound();
	end
	else ie (Token = KIE) do
		ie_stmt(1);
	end
	else ie (Token = KIF) do
		ie_stmt(0);
	end
	else ie (Token = KWHILE) do
		while_stmt();
	end
	else ie (Token = KFOR) do
		for_stmt();
	end
	else ie (Token = KLEAVE) do
		Token := scan();
		xsemi();
		if (\Endlab) error("'LEAVE' in wrong context", 0);
		! release local storage
		stack(-(Locladdr - Looplevl));
		gen1(tcode.IJUMP, Endlab);
	end
	else ie (Token = KLOOP) do
		Token := scan();
		xsemi();
		if (\Startlab) error("'LOOP' in wrong context", 0);
		! release local storage
		stack(-(Locladdr - Looplevl));
		gen1(tcode.IJUMP, Startlab);
	end
	else ie (Token = KRETURN) do
		Token := scan();
		ie (Token = SEMI)
			gen1(tcode.INUM, 0);
		else
			expr();
		xsemi();
		if (\Exitlab) error("'RETURN' in wrong context", 0);
		gen0(tcode.IPOP);
		! release local storage
		stack(-Locladdr+1);
		gen1(tcode.IJUMP, Exitlab);
		return 1;	! this statement has a return value
	end
	else ie (Token = KHALT) do
		Token := scan();
		ie (Token = SEMI)
			gen1(tcode.IHALT, 0);
		else
			gen1(tcode.IHALT, constexpr());
		xsemi();
	end
	else ie (Token = SEMI) do
		Token := scan();
	end
	else ie (Token = KELSE) do
		Token := scan();
		error("'ELSE' with no matching 'IE'", 0);
	end
	else ie (Token = KCALL) do
		factor();
		gen0(tcode.IPOP);	! discard return value
		xsemi();
	end
	else ie (Token = KSEND) do
		sendop();
		gen0(tcode.IPOP);	! discard return value
		xsemi();
	end
	else ie (Token = METACO) do
		mcomment();
	end
	else ie (Token = METAOP) do
		metacmd();
	end
	else do
		asg_or_call();
	end
	return 0;	! this statement has no return value
end

cleanup() do var k, la;
	la := Locladdr;
	! find innermost local context
	k := findsym("*", 0);
	if (k \= -1) do
		! restore outer context
		St := k+SYMREC;
		Nt := Symbols[k+SSIZE];
		Locladdr := Symbols[k+SVAL];
	end
	! release local storage
	stack(-(la - Locladdr));
end

compound() do var ot, rv;
	ot := Locladdr;
	! create new block context
	newsym("*", Nt, Locladdr, 0);
	Token := scan();
	while (	Token = KVAR \/ Token = KCONST \/ Token = KSTRUCT \/
		Token = METACO \/ Token = METAOP \/ Token = KOBJECT
	)
		declaration(1);
	! create local storage
	stack(Locladdr - ot);
	rv := 0;
	while (Token \= KEND) do
		if (eof()) fatal("unexpected EOF", 0);
		rv := stmt();
	end
	Token := scan();
	cleanup();
	! if the last statement in this block had a return value,
	! this block also has one
	return rv;
end

! adjust local symbol addresses
! k = first symbol to adjust
! n = number of arguments
adjust(p, k, n) do
	n := n+3;	! compensate for return/context information
	for (p=p, k, SYMREC) Symbols[p+SVAL] := Symbols[p+SVAL]-n;
end

! generate debug information
dumpsym(k, local) do var i, j, s;
	s := Symbols[k+SNAME];
	j := length(s);
	ie (local)
		if (s::0 \= '*')
			gen2(tcode.ILSYM, mtdadj(Symbols[k+SVAL]), j);
	else ie (Symbols[k+SFLAGS] & IVAR)
		gen2(tcode.IISYM, Symbols[k+SVAL], j);
	else
		gen2(tcode.IGSYM, Symbols[k+SVAL], j);
	for (i=0, j) gen0(s::i);
end

! procedure (pub=0) or method (pub=1) definition
procdef(pub) do var p, otbl, i, k, lv, n, addr, rv;
	n := 0;			! number of formal args
	addr := pub->1:2;	! first local addr to use
	Method_adjust := pub->1:0;
	p := findsym(Text, 0);	! look up potential prototype
	Exitlab := newlab();
	k := newsym(Text, 0, 0, PROC | LINKD | (pub->PUBLC:0));
	! create procedure context, lv = local var index
	lv := newsym("*", Nt, 1, 0)+SYMREC;
	Token := scan();
	xlparen();
	! build argument list
	if (Token \= RPAREN) while (1) do
		if (Token = SYMBOL) do
			newsym(Text, 0, addr, 0);
			addr := addr+1;
		end
		match(SYMBOL, "argument name expected");
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	! save procedure type
	Symbols[k+SSIZE] := n;
	! if there is a prototype, make sure the
	!  redefinition is valid
	ie (p \= -1 /\ (Symbols[p+SFLAGS] & PROTO)) do
		if (Symbols[p+SSIZE] \= n)
			error("argument count mismatch",
				Symbols[k+SNAME]);
		! copy prototype label
		Symbols[k+SVAL] := Symbols[p+SVAL];
		! make prototype entry invalid
		Symbols[p+SFLAGS] := Symbols[p+SFLAGS] & IVAR;
		Symbols[p+SNAME] := "?";
	end
	else do
		! no prototype: assign label
		Symbols[k+SVAL] := newlab();
	end
	if (pub /\ Symbols[Classctx+SFLAGS] & PUBLC)
		linkage(tcode.IPUB, 0, k);
	gen1(tcode.ICLAB, Symbols[k+SVAL]);
	gen0(pub-> tcode.IMHDR: tcode.IHDR);
	gen1(tcode.IHINT, pub->n+1: n);
	adjust(St, lv-SYMREC, n);
	if (Debug) do
		for (i=St, lv, SYMREC) dumpsym(i, 1);
	end
	xrparen();
	rv := stmt();
	cleanup();
	if (\rv) do
		! no return value? generate zero rv
		gen1(tcode.INUM, 0);
		gen0(tcode.IPOP);
	end
	gen1(tcode.ICLAB, Exitlab);
	gen0(pub-> tcode.IENDM: tcode.IEND);
	Exitlab := 0;
	Method_adjust := 0;
end

procdecl(inf) do var k, map;
	Token := scan();
	if (inf /\ Classctx = -1) do
		error("'IDECL' in wrong context", 0);
		synch(SEMI);
		return;
	end
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0, inf-> PROC|IPROC|PUBLC: PROC|PROTO);
		Token := scan();
		xlparen();
		Symbols[k+SSIZE] := constexpr();
		ie (inf) do
			xcomma(1);
			map := constexpr();
			if (\Infid) fatal("out of interface IDs", 0);
			Symbols[k+SVAL] := Infid;
			Infid := Infid + 1;
		end
		else do
			Symbols[k+SVAL] := newlab();
		end
		xrparen();
		if (inf) do
			linkage(tcode.IIPROC, Classctx, k);
			gen2(tcode.ICMAP, Symbols[k+SSIZE], map);
		end
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end

! allocate space for a variable
allocvar(k, local, bv) do
	if (bv /\ Xbpw) do
		Symbols[k+SSIZE] := (Symbols[k+SSIZE] + Xbpw - 1) / Xbpw;
	end
	ie (local) do
		Symbols[k+SVAL] := Locladdr;
		! local address space grows down !
		if (Symbols[k+SSIZE])
			Symbols[k+SVAL] :=
				Symbols[k+SVAL] +
				Symbols[k+SSIZE]-1;
		Locladdr := Locladdr +
			(Symbols[k+SSIZE]->
			Symbols[k+SSIZE]: 1);
	end
	! if there's a class ctx, it's an instance var
	else ie (Classctx \= -1) do
		Symbols[k+SVAL] := Classoff;
		Classoff := Classoff +
			(Symbols[k+SSIZE]->
			Symbols[k+SSIZE]: 1);
	end
	! else it's a global
	else do
		Symbols[k+SVAL] := newlab();
		gen1(tcode.IDLAB, Symbols[k+SVAL]);
		gen1(tcode.IVEC, Symbols[k+SSIZE]);
	end
end

vardef(pub, local) do var n, tok, k;
	tok := Token;
	Token := scan();
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0,
			(tok=KCONST-> CNST: local-> 0 : GLOBL) |
			(pub-> PUBLC: 0));
		Token := scan();
		ie (tok = KCONST) do
			equsign();
			Symbols[k+SVAL] := constexpr();
		end
		else do
			if (Token = LBRACK \/ Token = BYTEOP) do
				tok := Token;
				Token := scan();
				n := constexpr();
				if (tok \= BYTEOP) xrbrack();
				if (n < 1) error("bad vector size",
					Symbols[k+SNAME]);
				Symbols[k+SSIZE] := n;
			end
			allocvar(k, local, tok = BYTEOP);
			if (Debug) dumpsym(k, local);
		end
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end

structdecl(pub) do var base, count;
	Token := scan();
	if (forcesym("struct name expected")) return 0;
	base := newsym(Text, 0, 0, CNST | (pub-> PUBLC: 0));
	Token := scan();
	equsign();
	count := 0;
	while (1) do
		if (forcesym("member name expected")) return 0;
		newsym(Text, 0, count, CNST | (pub-> PUBLC: 0));
		Token := scan();
		count := count+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	Symbols[base+SVAL] := count;
	xsemi();
end

objdef(local) do var k, i, cl;
	Token := scan();
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0, OBJCT | (local->0:GLOBL));
		Token := scan();
		match(LBRACK, "missing '['");
		if (forcesym(0)) return 0;
		cl := findclassref(Text, 1);
		Token := scan();
		Symbols[k+SSIZE] := cl=-1-> -1: Symbols[cl+SSIZE];
		allocvar(k, local, 0);
		Symbols[k+SSIZE] := cl=-1-> -1: cl;
		xrbrack();
		if (cl = -1) St := St+SYMREC;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end

redirlist() do var b;
	xlparen();
	while (Token = SYMBOL) do
		b := findclass(Text);
		if (b \= -1) newsym("+", 0, b, REDIR);
		Token := scan();
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xrparen();
end

libname() do var k, i;
	xlparen();
	ie (Token = STRING) do
		k := length(Text);
		gen1(tcode.ILIB, k);
		for (i=0, k) gen0(Text::i);
		Token := scan();
	end
	else do
		error("library name expected in 'ICLASS'", 0);
		synch(RPAREN);
	end
	xrparen();
end

accept()
	while (	Token = KVAR \/ Token = KCONST \/ Token = SYMBOL \/
		Token = KDECL \/ Token = KSTRUCT \/ Token = KPUBLIC \/
		Token = KCLASS \/ Token = KICLASS \/ Token = KIDECL \/
		Token = KOBJECT \/ Token = KMODULE \/ Token = METAOP \/
		Token = METACO
	)
		declaration(0);

classdecl(pub, inf) do
	var	cl;
	var	p, i, k;
	var	endcl, nbase;
	var	thiscl::TEXTLEN;

	if (pub /\ \Modname::0)
		error("public class in anonymous module", 0);
	if (Classctx \= -1) error("nested class", 0);
	Token := scan();
	if (forcesym(0)) return 0;
	strcpy(thiscl, Text);
	Token := scan();
	endcl := St;
	ie (inf)
		libname();
	else
		redirlist();
	nbase := Nt;
	cl := newsym(thiscl, 0, -1, CLSS | (pub->PUBLC:0));
	Classctx := cl;
	Classoff := 0;
	if (inf) Infctx := Classctx;
	Symbols[Selfobj+SSIZE] := Classctx;
	accept();
	Classctx := -1;
	Infctx := -1;
	Symbols[cl+SSIZE] := Classoff = 0-> 1: Classoff;
	match(KEND, "missing END");
	newsym(Modname::0-> Modname: "**", 0, 0, PUBLC);
	rptdecls(St, cl);
	for (p=cl, St-SYMREC, -SYMREC) do
		if (p = cl \/ Symbols[p+SFLAGS] & PUBLC = PUBLC) do
			k := length(Symbols[p+SNAME])+1;
			t.memcopy(@Names::nbase, Symbols[p+SNAME], k);
			for (i=0, SYMREC) Symbols[Dt+i] := Symbols[p+i];
			Symbols[Dt+SNAME] := @Names::nbase;
			nbase := nbase+k;
			Dt := Dt+SYMREC;
		end
	end
	Symbols[Dt-SYMREC+SFLAGS] := 0;
	Nt := nbase;
	St := endcl;
end

moddecl() do var k, i, j;
	Token := scan();
	if (forcesym("module name expected")) return 0;
	if (Modname::0) error("module redefinition", Text);
	k := length(Text);
	Modname::0 := '*';
	strcpy(@Modname::1, Text);
	Modname::(k+1) := '*';
	Modname::(k+2) := 0;
	Token := scan();
	redirlist();
	Modctx := St-SYMREC;
	xsemi();
	! delete any classes defined in this module from the
	!  class directory cache
	for (i=0, Dt, SYMREC) do
		if (Symbols[i+SFLAGS] & CLSS) k := i;
		if (	\Symbols[i+SFLAGS] /\
			strequ(Symbols[i+SNAME], Modname)
		) do
			if (Dt-i-SYMREC)
				for (j=0, (Dt-i-SYMREC))
					Symbols[k+j] := Symbols[i+SYMREC+j];
			Dt := Dt-(i-k+SYMREC);
			i := k;
		end
	end
end

declaration(local) do var pub;
	pub:= 0;
	if (Token = KPUBLIC) do
		pub := 1;
		Token := scan();
		if (	(Token \= SYMBOL /\ Token \= KSTRUCT /\
			 Token \= KCONST \/ Classctx = -1) /\
			Token \= KCLASS /\ Token \= KICLASS
		)
			error("'PUBLIC' in wrong context", 0);
	end
	ie (Token = KVAR \/ Token = KCONST) do
		vardef(pub, local);
	end
	else ie (Token = KDECL /\ \local) do
		procdecl(0);
	end
	else ie (Token = KIDECL /\ \local) do
		if (Infctx = -1)
			error("'IDECL' in wrong context", 0);
		procdecl(1);
	end
	else ie (Token = KSTRUCT) do
		structdecl(pub);
	end
	else ie (Token = SYMBOL /\ \local) do
		procdef(pub);
	end
	else ie (Token = KOBJECT) do
		objdef(local);
	end
	else ie (Token = KCLASS /\ \local) do
		classdecl(pub, 0);
	end
	else ie (Token = KICLASS /\ \local) do
		classdecl(pub, 1);
	end
	else ie (Token = KMODULE /\ \local) do
		if (Classctx \= -1)
			error("'MODULE' in wrong context", 0);
		moddecl();
	end
	else ie (Token = METACO) do
		mcomment();
	end
	else ie (Token = METAOP) do
		metacmd();
	end
	else do
		ie (local)
			error("bad local declaration", 0);
		else
			error("bad declaration", 0);
		synch(0);
	end
end

init() do var k;
	Classpaths := [
		"",	! empty slot for #classpath
		".",
		"",	! reserved for $T3XDIR/classes
		"/usr/local/t3x/r8/classes",
		"C:/v/t3xr8/classes",
		0
	];
	k := t.getenv("T3XDIR", Syscp, TEXTLEN-9);
	if (k < 1) k := t.getenv("t3xdir", Syscp, TEXTLEN-9);
	if (k > 0) do
		t.memcopy(@Syscp::k, "/classes", 9);
		Classpaths[2] := Syscp;
	end
	Bpw := t.bpw();
	St := SYMTBLSIZE;
	Dt := 0;
	Nt := 0;
	Line := 1;
	File::0 := 0;
	Modname::0 := 0;
	Modctx := -1;
	Errcount := 0;
	Initdone := 0;
	Label := 2;
	Xlabel := 1;
	Ilabel := 1;
	Infid := 1;
	Startlab := 0;
	Endlab := 0;
	Exitlab := 0;
	Looplevl := 0;
	Locladdr := 1;
	Cp := 0;
	Ep := 0;
	Lowp := 0;
	Nomore := 0;
	Obp := 0;
	LL := 0;
	Bcslot := 0;
	Debug := 0;
	Xbpw := Bpw;
	Classctx := -1;
	Infctx := -1;
	Method_adjust := 0;
	B4F := 0xFF;
	B4F := B4F << 8 | 0xFF;
	B8F := B4F << 16 | B4F;
	B16F := B8F << 32 | B8F;
	B15 := 0x80;
	B15 := B15 << 8;
	B31 := B15 << 16;
	! Do not change the order of this table - see findop()
	Ops := [
		[ 6, 1, "+",	BINOP,	tcode.IADD	],
		[ 7, 1, "*",	BINOP,	tcode.IMUL	],
		[ 0, 1, ";",	SEMI,	0		],
		[ 0, 1, ",",	COMMA,	0		],
		[ 0, 1, "(",	LPAREN,	0		],
		[ 0, 1, ")",	RPAREN,	0		],
		[ 0, 1, "[",	LBRACK,	0		],
		[ 0, 1, "]",	RBRACK,	0		],
		[ 3, 1, "=",	BINOP,	tcode.IEQU	],
		[ 5, 1, "&",	BINOP,	tcode.IBAND	],
		[ 5, 1, "|",	BINOP,	tcode.IBOR	],
		[ 5, 1, "^",	BINOP,	tcode.IBXOR	],
		[ 0, 1, "@",	ADDROP,	0		],
		[ 0, 1, "~",	UNOP,	tcode.IBNOT	],
		[ 0, 1, ":",	COLON,	0		],
		[ 0, 2, "::",	BYTEOP,	0		],
		[ 0, 2, ":=",	ASSIGN,	0		],
		[ 0, 1, "\\",	UNOP,	tcode.ILNOT	],
		[ 1, 2, "\\/",	DISOP,	0		],
		[ 3, 2, "\\=",	BINOP,	tcode.INEQU	],
		[ 4, 1, "<",	BINOP,	tcode.ILESS	],
		[ 5, 2, "<<",	BINOP,	tcode.IBSHL	],
		[ 4, 2, "<=",	BINOP,	tcode.ILTEQ	],
		[ 4, 1, ">",	BINOP,	tcode.IGRTR	],
		[ 5, 2, ">>",	BINOP,	tcode.IBSHR	],
		[ 4, 2, ">=",	BINOP,	tcode.IGTEQ	],
		[ 6, 1, "-",	BINOP,	tcode.ISUB	],
		[ 0, 2, "->",	COND,	0		],
		[ 7, 1, "/",	BINOP,	tcode.IDIV	],
		[ 2, 2, "/\\",	CONOP,	0		],
		[ 0, 1, ".",	DOT,	0		],
		[ 7, 2, "./",	BINOP,	tcode.IUDIV	],
		[ 7, 2, ".*",	BINOP,	tcode.IUMUL	],
		[ 4, 2, ".<",	BINOP,	tcode.IULESS	],
		[ 4, 3, ".<=",	BINOP,	tcode.IULTEQ	],
		[ 4, 2, ".>",	BINOP,	tcode.IUGRTR	],
		[ 4, 3, ".>=",	BINOP,	tcode.IUGTEQ	],
		[ 0, 1, "#",	METAOP,	0		],
		[ 0, 2, "#!",	METACO,	0		],
		[ 7, 3, "mod",	BINOP,	tcode.IMOD	],
		[ 0, 0, 0,	0,	0			]
	];
	Selfobj := newsym("self", 0, 0, OBJCT|GLOBL);
	Sym_equ := findop2("=");
	Sym_sub := findop2("-");
	Sym_mod := findop2("mod");
	Sym_add := findop2("+");
	Sym_mul := findop2("*");
	Sym_bor := findop2("|");
	Sym_bnot := findop2("~");
end

do
	init();
	rdclsdir("CLASSDIR", 0);
	Token := scan();
	accept();
	if (Token \= KDO) fatal("'DO' or declaration expected", 0);
	gen1(tcode.ICLAB, 1);
	gen0(tcode.IHDR);
	gen1(tcode.IHINT, 0);
	compound();
	if (Token \= ENDFILE) error("trailing declarations", 0);
	rptdecls(St, SYMTBLSIZE);
	gen0(tcode.IEND);
	commit();
	flush();
	if (\Errcount) wrclsdir();
	if (Errcount) halt 1;
end
