! Explode the T3X Manual into small pieces for easier online browsing.
! Copyright (C) 2000-2004 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! Usage:
! tx explode-manual.t
! mkdir t3x-manual
! txx explode-manual

module explode_manual(t3x, string, iostream);

object	t[t3x], str[string];

const	BUFLEN	= 1024;

object	In[iostream], Out[iostream], Index[iostream],
	Index2[iostream], Toc[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN, Ndxbuf::BUFLEN,
	Ndx2buf::BUFLEN, Tocbuf::BUFLEN;
var	Outfile::128;
var	Outpath::128+20;
var	Line::128, New::160;
var	Previous::128, This::128, Next::128;
var	Indent, Indent2;
var	Verbose;


Head() return [
	"<HTML>",
	"<HEAD>",
	"<TITLE>",
	"",
	"</TITLE>",
	"<META name=\"description\" content=\"T3X Manual, Online Edition\">",
	"<META name=\"keywords\" content=\"T3X, reference, manual\">",
	"",
	"<STYLE>",
	" A { text-decoration: none }",
	" A:hover { background: #00FFFF }",
	"</STYLE>",
	"</HEAD>",
	"<BODY bgcolor=\"#FFFFFF\" text=\"#000000\" link=\"#0000FF\"",
	" vlink=\"#0000FF\">",
	"",
	"<TABLE width=\"100%\" bgcolor=\"#40C0C0\" cellspacing=5",
	" cellpadding=5 border=0>",
	" <TR>",
	"  <TD bgcolor=\"#FFFFE0\">",
	"   <TABLE width=\"100%\" cellspacing=0 cellpadding=0 border=0>",
	"    <TR>",
	"     <TD bgcolor=\"#FFFFE0\">",
	"      <FONT face=\"lucida, sans-serif\" size=5>",
	"       T3X <TT>-</TT> A Minimum Procedural Language<BR>",
	"      </FONT>",
	"      <B>Version 8.1.5, Online Edition</B>",
	"     </TD>",
	"     <TD align=\"right\">",
	"      Copyright(C) 1996-2004<BR>",
	"<A href=\"/nmh/index.html\">Nils M Holm</A>",
	"     </TD>",
	"    </TR>",
	"   </TABLE>",
	"  </TD>",
	" </TR>",
	"</TABLE>",
	"<TABLE cellpadding=0 cellspacing=0 border=0>",
	" <TR><TD><SMALL>&nbsp;</SMALL></TD></TR>",
	"</TABLE>",
	"",
0 ];

Tail() return [
	"",
	"</BODY>",
	"</HTML>",
0 ];

var	Ndx1::64, Ndx2::64;
Menu() do
	str.format(Ndx1, " <A href=\"index.html#%S\">TOC</A> |",
		[(Outfile)]);
	str.format(Ndx2, " <A href=\"index2.html#%S\">Index</A> |",
		[(Outfile)]);
	return [
		"",
		"<TABLE width=\"100%\" cellspacing=0 cellpadding=0 border=0>",
		"<TR><TD align=\"left\" width=\"40%\"><SMALL>Previous:<BR>",
		"",
		"</SMALL></TD><TD align=\"center\">",
		(Ndx1),
		(Ndx2),
		" <A href=\"../index.html\">Back</A>",
		"</TD><TD align=\"right\" width=\"40%\"><SMALL>Next:<BR>",
		"",
		"</SMALL></TD></TR>",
		"</TABLE>",
		"",
	0 ];
end

Indexhead() return [
	"<TABLE width=\"100%\" cellspacing=0 cellpadding=0 border=0>",
	"<TR><TD align=\"center\">",
	"<SMALL><A href=\"../index.html\">[back to documents]</A></SMALL>",
	"</TD></TR></TABLE>",
	"<H1>Table of Contents (<A href=\"index2.html\">compact</A>)</H1>",
0 ];

Index2head() return [
	"<TABLE width=\"100%\" cellspacing=0 cellpadding=0 border=0>",
	"<TR><TD align=\"center\">",
	"<SMALL><A href=\"../index.html\">[back to documents]</A></SMALL>",
	"</TD></TR></TABLE>",
	"<H1>Table of Contents (<A href=\"index.html\">full</A>)</H1>",
0 ];


m(s) t.write(T3X.SYSERR, s, str.length(s));


error(msg, n) do var s::256;
	m("EXPLODE-MANUAL: ");
	m(str.format(s, msg, [(n)]));
	m(t.newline(s));
	halt 1;
end


iwrites(s) do
	Index.writes(s);
	Index.writes(t.newline([0,0,0]));
end


i2writes(s) do
	Index2.writes(s);
	Index2.writes(t.newline([0,0,0]));
end


append(ios, list) do var i, b::3;
	i := 0; while(list[i]) do
		send(ios, iostream, writes(list[i]));
		send(ios, iostream, writes(t.newline(b)));
		i := i+1;
	end
end


appendmenu(ios) do var mnu;
	mnu := Menu();
	mnu[3] := Previous;
	mnu[9] := Next;
	append(ios, mnu);
end


init() do var h;
	if (In.open("T3X.html", Inbuf, BUFLEN, iostream.FREAD) = %1)
		error("failed to open manual: T3X.html", 0);
	if (Index.open("t3x-manual/index.html", Ndxbuf, BUFLEN,
		iostream.FWRITE) = %1)
			error("failed to create index: t3x-manual/index.html",
				0);
	if (Index2.open("t3x-manual/index2.html", Ndx2buf, BUFLEN,
		iostream.FWRITE) = %1)
			error("failed to create index: t3x-manual/index2.html",
				0);
	if (Toc.open("t3x-manual/TOC", Tocbuf, BUFLEN, iostream.FWRITE) = %1)
		error("failed to create t3x-manual/TOC", 0);
	Outfile::0 := 0;
	Indent2 := 0;
	Indent := 0;
	h := Head();
	h[3] := "T3X -- A Minimum Procedural Language, Online Edition";
	append(Index, h);
	append(Index2, h);
	append(Index, Indexhead());
	append(Index2, Index2head());
	str.copy(Previous, "-/-");
	Next::0 := 0;
	Verbose := 0;
end


makeindex(id, name) do var nbuf::128, s::256, p, k;
	k := str.length(id);
	str.copy(nbuf, name);
	nbuf::(str.length(nbuf)-1) := 0;
	p := str.scan(nbuf, '<');
	if (p >= 0) nbuf::p := 0;
	while (Indent2 < k) do
		i2writes("<DL>");
		Indent2 := Indent2 + 1;
	end
	while (Indent2 > k) do
		i2writes("</DL>");
		Indent2 := Indent2 - 1;
	end
	i2writes(str.format(s,
		"<DD><A name=\"%S\">&nbsp;</A><A href=\q%S.html\q>%S</A></DD>",
			[ (id), (id), (nbuf)]));
	Indent2 := k;
	if (k > 2) return;
	while (Indent < k) do
		iwrites("<DL>");
		Indent := Indent + 1;
	end
	while (Indent > k) do
		iwrites("</DL>");
		Indent := Indent - 1;
	end
	iwrites(str.format(s,
		"<DD><A name=\"%S\">&nbsp;</A><A href=\q%S.html\q>%S</A></DD>",
			[ (id), (id), (nbuf)]));
end


maketoc(id, name) do var nbuf::128, s::256, p, k;
	k := str.length(id);
	str.copy(nbuf, name);
	nbuf::(str.length(nbuf)-1) := 0;
	p := str.scan(nbuf, '<');
	if (p >= 0) nbuf::p := 0;
	Toc.writes(str.format(s, "<A href=\q%S.html\q>%S</A>",
		[ (id), (nbuf)]));
	Toc.writes(t.newline(s));
end


preprocess() do var p, q, s::256, nl::3, n;
	if (Verbose) t.write(T3X.SYSOUT, "Generating TOC, chapter: ", 25);
	In.reads(Line, 128);
	n := 0;
	while (\In.eof()) do
		p := str.find(Line, "<A name=\q");
		if (p \= -1) do
			str.copy(Outfile, @Line::(p+9));
			p := str.scan(Outfile, '"');
			if (p = -1) error("bad anchor", 0);
			Outfile::p := 0;
			maketoc(Outfile, @Outfile::(p+2));
		end
		if (Verbose /\ n \= Outfile::0) do
			t.write(T3X.SYSOUT, @Outfile::0, 1);
			n := Outfile::0;
		end
		In.reads(Line, 128);
	end
	Toc.close();
	In.reset();
	In.move(0, iostream.SEEK_SET);
	if (Verbose) t.write(T3X.SYSOUT, t.newline(s), str.length(s));
end


explode() do var p, q, s::256, nl::3, h, meta::256, title::256, descr::256;
	if (Toc.open("t3x-manual/TOC", Tocbuf, BUFLEN, iostream.FREAD) = %1)
		error("cannot reopen TOC: t3x-manual/TOC", 0);
	In.reads(Line, 128);
	Outfile::0 := 0;
	Toc.reads(This, 128); Next::0 := 0;
	while (\In.eof()) do
		p := str.find(Line, "<A name=\q");
		if (p \= -1) do
			if (Outfile::0) do
				appendmenu(Out);
				append(Out, Tail());
				Out.close();
			end
			if (Next::0) do
				str.copy(Previous, This);
				str.copy(This, Next);
			end
			Toc.reads(Next, 128);
			if (Toc.eof()) str.copy(Next, "-/-");
			str.copy(Outfile, @Line::(p+9));
			p := str.scan(Outfile, '"');
			if (p = -1) error("bad anchor", 0);
			Outfile::p := 0;
			makeindex(Outfile, @Outfile::(p+2));
			str.format(Outpath, "t3x-manual/%S.html",
				[(Outfile)]);
			ie (Out.open(Outpath, Outbuf, BUFLEN,
				iostream.FWRITE) = -1
			)
				error("failed to create: %S", Outpath);
			else if (Verbose)
				t.write(T3X.SYSOUT,
					str.format(s, "+ %S%S",
						[(Outpath), (t.newline(nl))]),
					str.length(s));
			h := Head();
			str.copy(s, This);
			p := str.scan(s, '>');
			if (p >= 0) str.copy(s, @This::p);
			p := str.scan(s, ' ');
			if (p >= 0) str.copy(s, @s::(p+1));
			p := str.scan(s, '<');
			if (p >= 0) s::p := 0;
			str.format(meta,
				"<META name=\qkeywords\q content=\q%s\q>",
					[(s)]);
			h[7] := meta;
			str.format(title, "T3X Manual: %s",
				[(s)]);
			h[3] := title;
			str.format(descr,
		"<META name=\"description\" content=\"T3X Manual: %s\">",
				[(s)]);
			h[5] := descr;
			append(Out, h);
			appendmenu(Out);
		end
		p := str.find(Line, "<A href=\q#");
		ie (p \= -1) do
			p := p+9;
			q := str.scan(@Line::p, '"')-1;
			if (q < 0) error("bad hyperlink", 0);
			t.memcopy(New, Line, p);
			t.memcopy(@New::p, @Line::(p+1), q);
			p := p+q;
			t.memcopy(@New::p, ".html", 5);
			str.copy(@New::(p+5), @Line::(p+1));
			if (Outfile::0) Out.writes(New);
		end
		else do
			if (Outfile::0) Out.writes(Line);
		end
		In.reads(Line, 128);
	end
	appendmenu(Out);
	append(Out, Tail());
	Out.close();
	while (Indent) do
		iwrites("</DL>");
		Indent := Indent - 1;
	end
	while (Indent2) do
		i2writes("</DL>");
		Indent2 := Indent2 - 1;
	end
	append(Index, Tail());
	append(Index2, Tail());
	Index.close();
	Index2.close();
	Toc.close();
	t.remove("t3x-manual/TOC");
end


do var a::3;
	init();
	if (t.getarg(1, a, 3) /\ \str.comp(a, "-v")) Verbose := 1;
	preprocess();
	explode();
end
