! TXCGEXE -- Tcode->EXE translator
! Copyright (C) 2000-2003 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcgexe(t3x, tcode, string, iostream);

object	t[t3x], str[string];

const	BPW =		2;

const	BUFSIZE	=	1024;	! I/O buffer size
const	MAXLABEL=	4096;	! max. label ID
const	MAXBSSREF=	512;	! max. data->BSS references
const	LIBMAX=		256;	! max space for library names
const	LIBPATHMAX=	1024;	! max space for library paths
const	DIRBUFSIZE=	1024;	! library directory buffer size
const	MAXINFPROC=	256;	! max. number of interface procedures
const	NAMEPOOLSIZE=	1024;	! interface procedure name pool size

struct INF =
	INF_ENTRY,
	INF_OFFSET,
	INF_NAME;

object	In[iostream], Out[iostream],
	Data[iostream];		! data segment stream
var	R_op, R_fl;		! Rejection queue
var	Inbuf::BUFSIZE;
var	Outbuf::BUFSIZE;
var	Databuf::BUFSIZE;
var	Addr;
var	Labels[MAXLABEL],
	Ltypes::MAXLABEL;	! label types: Text, Data, BSS
var	Fl, Na, Sf, L1, L2;	! Tcode flags
var	BSSrefs[MAXBSSREF], Bt;	! data->BSS references, top ptr
var	Ctop, Dtop, Btop, Bss;	! free address pointers, BSS start address
var	Firstpass;
var	Iqueue, Aqueue,		! 1-instruction queue (instr, arg)
	Queued,			! queue active indicator (1=active),
	Active,			! accumulator active flag (1=active),
	Depth;			! RT stack depth
var	Dataname;
var	Tmpfile::256,		! tmp file name buffers
	Tmpdata::256;
var	Infprocs[INF*MAXINFPROC],	! interface procedures,
	Ninf;				! top ptr
var	Names::NAMEPOOLSIZE, Nt;	! interface name pool, top ptr
var	Infmap[MAXINFPROC];		! interface map
var	Verbose;
var	KFFFF;			! constant 65535
var	Lbpw;
var	XFFFFFFFF;

! variables to bind code sequences
var	LOAD_GLOB,
	LOAD_LOC,
	LOAD_IVAR,
	LOAD_LOC_VEC,
	LOAD_IVAR_VEC,
	LOAD_CONST,
	LOAD_SELF,
	PUSH_SEQ,
	POP_SEQ,
	DEREF_SEQ,
	DEREF_BYTE_SEQ,
	STORE_GLOB,
	STORE_LOC,
	STORE_IVAR,
	STORE_SEQ,
	STORE_BYTE,
	NORM_SEQ,
	INC_GLOB,
	INC_LOC,
	INC_IVAR,
	NEG_SEQ,
	NOT_SEQ,
	LNOT_SEQ,
	CMP_ZERO_SEQ,
	MUL_SEQ,
	DIV_SEQ,
	MOD_SEQ,
	UMUL_SEQ,
	UDIV_SEQ,
	ADD_SEQ,
	SUB_SEQ,
	AND_SEQ,
	OR_SEQ,
	XOR_SEQ,
	SHL_SEQ,
	SHR_SEQ,
	CMP_SEQ,
	SWAP_SEQ,
	BE_SEQ,
	BNE_SEQ,
	BL_SEQ,
	BG_SEQ,
	BLE_SEQ,
	BGE_SEQ,
	BUL_SEQ,
	BUG_SEQ,
	BULE_SEQ,
	BUGE_SEQ,
	JMP_SEQ,
	JMPS_SEQ,
	TRUE_SEQ,
	FALSE_SEQ,
	PROLOGUE_SEQ,
	EPILOGUE_SEQ,
	MPROLOGUE_SEQ,
	MEPILOGUE_SEQ,
	CALL_REG,
	CALL_SEQ,
	ADJUST_SEQ,
	CLEANUP_SEQ,
	HALT_SEQ;

! variables to bind addressing modes
struct	MODES =
	MODE_GLOB,	! acc/memory
	MODE_LOC,	! acc/[sp+offset] (local)
	MODE_IVAR,	! acc/[bx+offset] (instance)
	MODE_CONST,	! acc/immediate
	MODE_SELF,	! acc/bx (self)
	MODE_STACK;	! acc/[sp] (popped value)

! Bind machine code fragments to variables.
! When a variable is bound to an array of sequences,
! then there is one entry for each of the above
! addressing modes. Otherwise, the instruction
! has an implied mode.
! The format of each entry is as follows:
!
!	member#		description
!	1		length of sequence
!	2		number of arguments	(0,1,2)
!	3		offset of argument 1	(optional)
!	4		offset of argument 2	(optional)
!	5		beginning of sequence
!
! Arguments are filled with values like 0xDDDD, 0xEEEE, 0xFFFF.
! When an argument is at the end of a sequence, it is omitted.
!
mkopcodes() do
	LOAD_GLOB :=
	packed[ 2, 1, 2, 0x8B, 0x06 ];	! MOV AX,[EEEE]

	LOAD_LOC :=
	packed[ 2, 1, 2, 0x8B, 0x86 ];	! MOV AX,[BP+DDDD]

	LOAD_IVAR :=
	packed[ 2, 1, 2, 0x8B, 0x87 ];	! MOV AX,[BX+DDDD]

	LOAD_LOC_VEC :=
	packed[ 5, 1, 1, 0xB8, 0xDD, 0xDD,  	! MOV AX,DDDD
		0x01, 0xE8 ];			! ADD AX,BP

	LOAD_IVAR_VEC :=
	packed[ 5, 1, 1, 0xB8, 0xDD, 0xDD,  	! MOV AX,DDDD
		0x01, 0xD8 ];			! ADD AX,BX

	LOAD_CONST :=
	packed[ 1, 1, 1, 0xB8 ];	! MOV AX,EEEE

	LOAD_SELF :=
	packed[ 2, 0, 0x89, 0xD8 ];	! MOV AX,BX

	PUSH_SEQ :=
	packed[ 1, 0, 0x50 ];	! PUSH AX

	POP_SEQ :=
	packed[ 1, 0, 0x58 ];	! POP AX

	DEREF_SEQ := [
	packed[ 10, 1, 2, 0x8B, 0x36, 0xEE, 0xEE,	! MOV SI,[EEEE]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 10, 1, 2, 0x8B, 0xB6, 0xDD, 0xDD,	! MOV SI,[BP+DDDD]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 10, 1, 2, 0x8B, 0xB7, 0xDD, 0xDD,	! MOV SI,[BX+DDDD]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 9, 1, 1, 0xBE, 0xFF, 0xFF,	! MOV DX,FFFF
		0xD1, 0xE6,			! SHL SI,1
		0x03, 0xF0,			! ADD SI,AX
		0x8B, 0x04 ],			! MOV AX,[SI]

	packed[ 8, 0, 0x8B, 0xF3,	! MOV SI,BX
		0xD1, 0xE6,		! SHL SI,1
		0x03, 0xF0,		! ADD SI,AX
		0x8B, 0x04 ],		! MOV AX,[SI]

	packed[ 7, 0, 0xD1, 0xE0,	! SHL AX,1
		0x5E,			! POP SI
		0x03, 0xF0,		! ADD SI,AX
		0x8B, 0x04 ],		! MOV AX,[SI]
	];

	DEREF_BYTE_SEQ := [
	packed[ 10, 1, 2,
		0x03, 0x06, 0xEE, 0xEE,	! ADD AX,[EEEE]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x03, 0x86, 0xDD, 0xDD,	! ADD AX,[BP+DDDD]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x03, 0x87, 0xDD, 0xDD,	! ADD AX,[BX+DDDD]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x81, 0xC0, 0xFF, 0xFF,	! ADD AX,FFFF
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 8, 0, 0x01, 0xD8,	! ADD AX,BX
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 7, 0, 0x5E,	! POP SI
		0x03, 0xF0,	! ADD SI,AX
		0x31, 0xC0,	! XOR AX,AX
		0x8A, 0x04 ]	! MOV AL,[SI]
	];

	STORE_GLOB :=
	packed[ 2, 1, 2, 0x89, 0x06 ];	! MOV [EEEE],AX

	STORE_LOC :=
	packed[ 2, 1, 2, 0x89, 0x86 ];	! MOV [BP+DDDD],AX

	STORE_IVAR :=
	packed[ 2, 1, 2, 0x89, 0x87 ];	! MOV [BX+DDDD],AX

	STORE_SEQ :=
	packed[ 3, 0, 0x5E,	! POP SI
		0x89, 0x04 ];	! MOV [SI],AX

	STORE_BYTE :=
	packed[ 3, 0, 0x5E,	! POP SI
		0x88, 0x04 ];	! MOV [SI],AL

	NORM_SEQ := [
	packed[ 8, 1, 2, 0x8B, 0x16, 0xEE, 0xEE,	! MOV DX,[EEEE]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 8, 1, 2, 0x8B, 0x96, 0xDD, 0xDD,	! MOV DX,[BP+DDDD]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 8, 1, 2, 0x8B, 0x97, 0xDD, 0xDD,	! MOV DX,[BX+DDDD]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 7, 1, 1, 0xBA, 0xFF, 0xFF,	! MOV DX,FFFF
		0xD1, 0xE2,			! SHL DX,1
		0x01, 0xD0 ],			! ADD AX,DX

	packed[ 6, 0, 0x89, 0xDA,	! MOV DX,BX
		0xD1, 0xE2,		! SHL DX,1
		0x01, 0xD0 ],		! ADD AX,DX

	packed[ 5, 0, 0xD1, 0xE0,	! SHL AX,1
		0x5A,			! POP DX
		0x01, 0xD0 ]		! ADD AX,DX
	];

	INC_GLOB :=
	packed[ 6, 2, 2, 4, 0x81, 0x06 ];	! ADD WORD PTR [EEEE],FFFF

	INC_LOC :=
	packed[ 6, 2, 2, 4, 0x81, 0x86 ];	! ADD WORD PTR [BP+DDDD],FFFF

	INC_IVAR :=
	packed[ 6, 2, 2, 4, 0x81, 0x87 ];	! ADD WORD PTR [BX+DDDD],FFFF

	NEG_SEQ :=
	packed[ 2, 0, 0xF7, 0xD8 ];	! NEG AX

	NOT_SEQ :=
	packed[ 2, 0, 0xF7, 0xD0 ];	! NOT AX

	LNOT_SEQ :=
	packed[ 6, 0, 0xF7, 0xD8,	! NEG AX
		0x19, 0xC0,		! SBB AX,AX
		0xF7, 0xD0 ];		! NOT AX

	CMP_ZERO_SEQ :=
	packed[ 2, 0, 0x09, 0xC0 ];	! OR AX,AX

	MUL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xF7, 0xE9 ],			! IMUL CX

	packed[ 2, 0, 0xF7, 0xEB ],	! IMUL BX

	packed[ 3, 0, 0x59,	! POP CX
		0xF7, 0xE9 ]	! IMUL CX
	];

	DIV_SEQ := [
	packed[ 7, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 7, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 7, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 6, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x99,				! CWD 
		0xF7, 0xF9 ],			! IDIV CX

	packed[ 2, 0, 0xF7, 0xFB ],	! IDIV BX

	packed[ 6, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x99,		! CWD 
		0xF7, 0xF9 ]	! IDIV CX
	];

	MOD_SEQ := [
	packed[ 10, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 10, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 10, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 9, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x31, 0xD2,			! XOR DX,DX
		0xF7, 0xF1,			! DIV CX
		0x89, 0xD0 ],			! MOV AX,DX

	packed[ 6, 0, 0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF3,		! DIV BX
		0x89, 0xD0 ],		! MOV AX,DX

	packed[ 9, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF1,	! DIV CX
		0x89, 0xD0 ]	! MOV AX,DX
	];

	UMUL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xF7, 0xE1 ],			! MUL CX

	packed[ 2, 0, 0xF7, 0xE3 ],	! MUL BX

	packed[ 3, 0, 0x59,	! POP CX
		0xF7, 0xE1 ]	! MUL CX
	];

	UDIV_SEQ := [
	packed[ 8, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 8, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 8, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 7, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x31, 0xD2,			! XOR DX,DX
		0xF7, 0xF1 ],			! DIV CX

	packed[ 4, 0, 0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF3 ],		! DIV BX

	packed[ 7, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF1 ]	! DIV CX
	];

	ADD_SEQ := [
	packed[ 2, 1, 2, 0x03, 0x06 ],	! ADD AX,[EEEE]

	packed[ 2, 1, 2, 0x03, 0x86 ],	! ADD AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x03, 0x87 ],	! ADD AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xC0 ],	! ADD AX,FFFF

	packed[ 2, 0, 0x01, 0xD8 ],	! ADD AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x01, 0xD0 ]	! ADD AX,DX
	];

	SUB_SEQ := [
	packed[ 2, 1, 2, 0x2B, 0x06 ],	! SUB AX,[EEEE]

	packed[ 2, 1, 2, 0x2B, 0x86 ],	! SUB AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x2B, 0x87 ],	! SUB AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xE8 ],	! SUB AX,FFFF

	packed[ 2, 0, 0x29, 0xD8 ],	! SUB AX,BX

	packed[ 5, 0, 0x5A,	! POP DX
		0x87, 0xD0,	! XCHG DX,AX
		0x29, 0xD0 ]	! SUB AX,DX
	];

	AND_SEQ := [
	packed[ 2, 1, 2, 0x23, 0x06 ],	! AND AX,[EEEE]

	packed[ 2, 1, 2, 0x23, 0x86 ],	! AND AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x23, 0x87 ],	! AND AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xE0 ],	! AND AX,FFFF

	packed[ 2, 0, 0x21, 0xD8 ],	! AND AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x21, 0xD0 ]	! AND AX,DX
	];

	OR_SEQ := [
	packed[ 2, 1, 2, 0x0B, 0x06 ],	! OR AX,[EEEE]

	packed[ 2, 1, 2, 0x0B, 0x86 ],	! OR AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x0B, 0x87 ],	! OR AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xC8 ],	! OR AX,FFFF

	packed[ 2, 0, 0x09, 0xD8 ],	! OR AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x09, 0xD0 ]	! OR AX,DX
	];

	XOR_SEQ := [
	packed[ 2, 1, 2, 0x33, 0x06 ],	! XOR AX,[EEEE]

	packed[ 2, 1, 2, 0x33, 0x86 ],	! XOR AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x33, 0x87 ],	! XOR AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xF0 ],	! XOR AX,FFFF

	packed[ 2, 0, 0x31, 0xD8 ],	! XOR AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x31, 0xD0 ]	! XOR AX,DX
	];

	SHL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE, 	! MOV CX,[EEEE]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xD3, 0xE0 ],			! SHL AX,CL

	packed[ 4, 0, 0x89, 0xD9,	! MOV CX,BX
		0xD3, 0xE0 ],		! SHL AX,CL

	packed[ 5, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0xD3, 0xE0 ]	! SHL AX,CL
	];

	SHR_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xD3, 0xE8 ],			! SHR AX,CL

	packed[ 4, 0,  0x89, 0xD9,	! MOV CX,BX
		0xD3, 0xE8 ],		! SHR AX,CL

	packed[ 5, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0xD3, 0xE8 ]	! SHR AX,CL
	];

	CMP_SEQ := [
	packed[ 2, 1, 2, 0x3B, 0x06 ],	! CMP AX,[EEEE]

	packed[ 2, 1, 2, 0x3B, 0x86 ],	! CMP AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x3B, 0x87 ],	! CMP AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xF8 ],	! CMP AX,FFFF

	packed[ 2, 0, 0x39, 0xD8 ],	! CMP AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x39, 0xC2 ]	! CMP DX,AX
	];

	SWAP_SEQ := packed[ 4, 0, 0x89, 0xE6,	! MOV SI,SP
		0x87, 0x04 ];			! XCHG AX,[SI]

	BE_SEQ := packed[ 1, 1, 1, 0x74 ];	! JZ 02D4

	BNE_SEQ := packed[ 1, 1, 1, 0x75 ];	! JNZ 02D4

	BL_SEQ := packed[ 1, 1, 1, 0x7C ];	! JL 02D4

	BG_SEQ := packed[ 1, 1, 1, 0x7F ];	! JG 02D4

	BLE_SEQ := packed[ 1, 1, 1, 0x7E ];	! JLE 02D4

	BGE_SEQ := packed[ 1, 1, 1, 0x7D ];	! JGE 02D4

	BUL_SEQ := packed[ 1, 1, 1, 0x72 ];	! JB 02D4

	BUG_SEQ := packed[ 1, 1, 1, 0x77 ];	! JA 02D4

	BULE_SEQ := packed[ 1, 1, 1, 0x76 ];	! JBE 02D4

	BUGE_SEQ := packed[ 1, 1, 1, 0x73 ];	! JNB 02D4

	JMP_SEQ := packed[ 1, 1, 1, 0xE9 ];	! JMP CDCC

	JMPS_SEQ := packed[ 1, 1, 1, 0xEB ];	! JMPS 02D4

	TRUE_SEQ := packed[ 3, 0, 0xB8, 0xFF, 0xFF ];	! MOV AX,FFFF

	FALSE_SEQ := packed[ 2, 0, 0x31, 0xC0 ];	! XOR AX,AX

	PROLOGUE_SEQ := packed[ 3, 0, 0x55,	! PUSH BP
		0x89, 0xE5 ];			! MOV BP,SP

	EPILOGUE_SEQ := packed[ 2, 0, 0x5D,	! POP BP
		0xC3 ];				! RET 

	MPROLOGUE_SEQ := packed[ 7, 0, 0x55,	! PUSH BP
		0x89, 0xE5,			! MOV BP,SP
		0x53,				! PUSH BX
		0x8B, 0x5E, 0x04 ];		! MOV BX,[BP+04]

	MEPILOGUE_SEQ := packed[ 3, 0, 0x5B,	! POP BX
		0x5D,				! POP BP
		0xC3 ];				! RET 

	CALL_REG := packed[ 2, 0, 0xFF, 0xD0 ];	! CALL AX

	CALL_SEQ := packed[ 1, 1, 1, 0xE8 ];	! CALL CDCC

	ADJUST_SEQ := packed[ 2, 1, 2, 0x81, 0xEC ];	! SUB SP,FFFF

	CLEANUP_SEQ := packed[ 2, 1, 2, 0x81, 0xC4 ];	! ADD SP,FFFF

	HALT_SEQ := packed[ 7, 1, 1, 0xB8, 0xFF, 0xFF,	! MOV AX,FFFF
		0xB4, 0x4C,				! MOV AH,4C
		0xCD, 0x21 ];				! INT 21
end

init() do var i, tmpname::240;
	Lbpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Ninf := 0;
	Nt := 0;
	Bt := 0;
	R_op := %1;
	Addr := 0;
	Active := 0;
	Queued := 0;
	Depth := 0;
	Dataname := 0;
	Verbose := 0;
	KFFFF := 0x3FFF;
	KFFFF := (KFFFF << 2) + 3;
	t.memfill(Infmap, 0, MAXINFPROC*t.bpw());
	t.memfill(Labels, -1, MAXLABEL*t.bpw());
	if (t.getenv("TMP", tmpname, 240) < 1 /\
		t.getenv("tmp", tmpname, 240) < 1
	) do
		tmpname::0 := '.';
		tmpname::1 := 0;
	end
	i := str.length(tmpname);
	t.memcopy(Tmpfile, tmpname, i+1);
	t.memcopy(@Tmpfile::i, "/__txcgexe.tmp", 15);
	t.memcopy(Tmpdata, tmpname, i+1);
	t.memcopy(@Tmpdata::i, "/__txcgexe.data", 16);
	mkopcodes();
end

error(msg, name) do var buf::128;
	t.write(T3X.SYSERR, "TXCGEXE(", 8);
	t.write(T3X.SYSERR, Firstpass-> "1): ": "2): ", 4);
	t.write(T3X.SYSERR, str.numtostr(buf, Addr, 10), str.length(buf));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	if (name) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, name, str.length(name));
	end
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	t.remove(Tmpfile);
	t.remove(Tmpdata);
	halt 1;
end

internal(msg) do
	t.write(T3X.SYSERR, "internal error: ", 16);
	error(msg, 0);
end

wrdata(n) do
	if (Dtop + Btop .>= KFFFF-256) error("data segment overflow", 0);
	Data.wrch(n & 255);
	Data.wrch(n >> 8);
	Dtop := Dtop+BPW;
end

rdch() do var ch;
	ch := In.rdch();
	if (Firstpass /\ \In.eof()) Out.wrch(ch);
	Addr := Addr+1;
	return ch;
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Lbpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Lbpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdarg(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (m /\ (m \= 0xFFFF))	! \/ n & 0x8000 = 0))
		error("value out of range", 0);
	return sext32(n);
end

rdcode() do var op;
	ie (R_op >= 0) do
		op := R_op;
		Fl := R_fl;
		R_op := %1;
	end
	else do
		op := rdch();
		ie (op & tcode.IARFLAG) do
			Fl := op;
			op := rdch();
		end
		else do
			Fl := 0;
		end
	end
	Na := Fl & tcode.IARMASK >> tcode.IAMOFF;
	Sf := Fl & tcode.ISTFLAG;
	if (Na) do
		L1 := Fl & tcode.IL1MASK >> tcode.IL1OFF;
		if  (Na > 1)
			L2 := Fl & tcode.IL2MASK >> tcode.IL2OFF;
	end
	return op;
end

reject(op) do
	R_op := op;
	R_fl := Fl;
end

wrch(ch) do
	if (Ctop .>= KFFFF) error("text segment overflow", 0);
	Ctop := Ctop+1;
	if (Firstpass) return;
	Out.wrch(ch);
end

! emit a code sequence with merged in arguments
emit(code, args) do var k, n, i, j, p0, p1;
	p0 := -1;
	p1 := -1;
	k := code::0;
	n := code::1;
	i := 2;
	if (n) do
		p0 := code::2;
		i := 3;
		if (n > 1) do
			p1 := code::3;
			i := 4;
		end
	end
	if (p0 >= k) k := k + BPW;
	if (p1 >= k) k := k + BPW;
	for (j=0, k) do
		ie (j = p0) do
			wrch(args[0] & 255);
			wrch(args[0] >> 8);
			j := j+1;
		end
		else ie (j = p1) do
			wrch(args[1] & 255);
			wrch(args[1] >> 8);
			j := j+1;
		end
		else do
			wrch(code::(i+j));
		end
	end
end

! emit a PC-relative jump/call instruction
emitj(code, arg) do var k, n, i, j, p0, a;
	p0 := -1;
	k := code::0;
	n := code::1;
	i := 2;
	if (n) do
		p0 := code::2;
		i := 3;
	end
	if (p0 >= k) k := k + BPW;
	a := arg - Ctop - k;
	for (j=0, k) do
		ie (j = p0) do
			wrch(a & 255);
			wrch(a >> 8);
			j := j+1;
		end
		else do
			wrch(code::(i+j));
		end
	end
end

! emit a conditional jump
emit8(code, arg) do
	wrch(code::3);	! assume 1,1,1 prefix
	wrch(arg);
end

mklabel(id, lt, a) do
	if (id >= MAXLABEL) error("mklabel: label too big", 0);
	if (Labels[id] \= -1) error("label redefinition", 0);
	Labels[id] := a;
	Ltypes::id := lt;
end

find(id) do
	if (id >= MAXLABEL) error("find: label too big", 0);
	if (Labels[id] = -1 /\ \Firstpass) error("undefined label", 0);
	return Labels[id] + (Ltypes::id = 'B'-> Bss: 0);
end

push() do
	if (\Active) internal("accumulator not active in PUSH()");
	emit(PUSH_SEQ, 0);
	Depth := Depth+1;
	Active := 0;
end

pop() do
	if (\Depth) do
		if (Active) do
			Active := 0;
			return;
		end
		internal("stack underflow in POP()");
	end
	emit(POP_SEQ, 0);
	Depth := Depth-1;
	Active := 1;
end

! load queued value into accumulator
spill(m, a) do
	if (Active) push();
	ie (m = tcode.ILDG & tcode.IOPMASK)
		emit(LOAD_GLOB, [(find(a))]);
	else ie (m = tcode.ILDGV & tcode.IOPMASK)
		emit(LOAD_CONST, [(find(a))]);
	else ie (m = tcode.ILDL & tcode.IOPMASK)
		emit(LOAD_LOC, [(-a*BPW)]);
	else ie (m = tcode.ILDLV & tcode.IOPMASK)
		emit(LOAD_LOC_VEC, [(-a*BPW)]);
	else ie (m = tcode.ILDI & tcode.IOPMASK)
		emit(LOAD_IVAR, [(a*BPW)]);
	else ie (m = tcode.ILDIV & tcode.IOPMASK)
		emit(LOAD_IVAR_VEC, [(a*BPW)]);
	else ie (m = tcode.ILDLAB & tcode.IOPMASK)
		emit(LOAD_CONST, [(find(a))]);
	else ie (m = tcode.INUM & tcode.IOPMASK)
		emit(LOAD_CONST, [(a)]);
	else ie (m = tcode.ISELF & tcode.IOPMASK)
		emit(LOAD_SELF, 0);
	else
		internal("unknown instruction in LOAD()");
	Active := 1;
	Queued := 0;
end

! generate load instruction
load(m, a) do
	ie (\Queued) do
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
		return;
	end
	else do
		spill(Iqueue, Aqueue);
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
	end
end

unary(op) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("accumulator not active in UNARY()");
		pop();
	end
	ie (op = tcode.INEG & tcode.IOPMASK) emit(NEG_SEQ, 0);
	else ie (op = tcode.IBNOT & tcode.IOPMASK) emit(NOT_SEQ, 0);
	else ie (op = tcode.ILNOT & tcode.IOPMASK) emit(LNOT_SEQ, 0);
	else internal("unknown instruction in UNARY()");
end

emitb(seq, mode) emit(seq[mode], [(Aqueue)]);

wm(s) do var nl::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
end

undefd_addr() do
	wm("TXNCG: Addresses of local variables and anonymous vectors");
	wm("       are not supported in binary operations.");
	wm("       This may be considered a bug of TXNCG.");
	wm("       See the documentation for details.");
	internal("unsupported addressing mode in binary()");
end

binary(op) do var mode, op2, flg, lab, rel;
	if (\Active /\ Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in BINARY()");
		pop();
	end
	ie (Queued) do
		ie (Iqueue = tcode.ILDG & tcode.IOPMASK)
			mode := MODE_GLOB;
		else ie (Iqueue = tcode.ILDGV & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ILDL & tcode.IOPMASK)
			mode := MODE_LOC;
		else ie (Iqueue = tcode.ILDLV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDI & tcode.IOPMASK)
			mode := MODE_IVAR;
		else ie (Iqueue = tcode.ILDIV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDLAB & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.INUM & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ISELF & tcode.IOPMASK)
			mode := MODE_SELF;
		else
			internal("bad mode in BINARY()");
		! these instructions reference labels
		if (	Iqueue = tcode.ILDG & tcode.IOPMASK \/
			Iqueue = tcode.ILDGV & tcode.IOPMASK \/
			Iqueue = tcode.ILDLAB & tcode.IOPMASK
		) 
			Aqueue := find(Aqueue);
	end
	else do
		if (\Depth) internal("stack underflow in BINARY()");
		mode := MODE_STACK;
		Depth := Depth-1;
	end
	Queued := 0;
	Active := 1;
	! convert Tcode-style local/ivar addresses into real addresses
	ie (mode = MODE_LOC) Aqueue := -Aqueue*BPW;
	else if (mode = MODE_IVAR) Aqueue := Aqueue*BPW;
	ie (op = tcode.INORMB & tcode.IOPMASK) emitb(ADD_SEQ, mode);
	else ie (op = tcode.INORM & tcode.IOPMASK) emitb(NORM_SEQ, mode);
	else ie (op = tcode.IDEREF & tcode.IOPMASK) emitb(DEREF_SEQ, mode);
	else ie (op = tcode.IDREFB & tcode.IOPMASK) emitb(DEREF_BYTE_SEQ,mode);
	else ie (op = tcode.IMUL & tcode.IOPMASK) emitb(MUL_SEQ, mode);
	else ie (op = tcode.IDIV & tcode.IOPMASK) emitb(DIV_SEQ, mode);
	else ie (op = tcode.IMOD & tcode.IOPMASK) emitb(MOD_SEQ, mode);
	else ie (op = tcode.IUMUL & tcode.IOPMASK) emitb(UMUL_SEQ, mode);
	else ie (op = tcode.IUDIV & tcode.IOPMASK) emitb(UDIV_SEQ, mode);
	else ie (op = tcode.IADD & tcode.IOPMASK) emitb(ADD_SEQ, mode);
	else ie (op = tcode.ISUB & tcode.IOPMASK) emitb(SUB_SEQ, mode);
	else ie (op = tcode.IBAND & tcode.IOPMASK) emitb(AND_SEQ, mode);
	else ie (op = tcode.IBOR & tcode.IOPMASK) emitb(OR_SEQ, mode);
	else ie (op = tcode.IBXOR & tcode.IOPMASK) emitb(XOR_SEQ, mode);
	else ie (op = tcode.IBSHL & tcode.IOPMASK) emitb(SHL_SEQ, mode);
	else ie (op = tcode.IBSHR & tcode.IOPMASK) emitb(SHR_SEQ, mode);
	else do
		emitb(CMP_SEQ, mode);
		op2 := rdcode();
		ie (	op2 = tcode.IBRT & tcode.IOPMASK \/
			op2 = tcode.IBRF & tcode.IOPMASK
		) do
			pop();
			flg := op2=tcode.IBRF & tcode.IOPMASK;
			lab := rdarg(L1);
			ie (op = tcode.IEQU & tcode.IOPMASK)
				rel := flg-> BE_SEQ: BNE_SEQ;
			else ie (op = tcode.INEQU & tcode.IOPMASK)
				rel := flg-> BNE_SEQ: BE_SEQ;
			else ie (op = tcode.ILESS & tcode.IOPMASK)
				rel := flg-> BL_SEQ: BGE_SEQ;
			else ie (op = tcode.IGRTR & tcode.IOPMASK)
				rel := flg-> BG_SEQ: BLE_SEQ;
			else ie (op = tcode.ILTEQ & tcode.IOPMASK)
				rel := flg-> BLE_SEQ: BG_SEQ;
			else ie (op = tcode.IGTEQ & tcode.IOPMASK)
				rel := flg-> BGE_SEQ: BL_SEQ;
			else ie (op = tcode.IULESS & tcode.IOPMASK)
				rel := flg-> BUL_SEQ: BUGE_SEQ;
			else ie (op = tcode.IUGRTR & tcode.IOPMASK)
				rel := flg-> BUG_SEQ: BULE_SEQ;
			else ie (op = tcode.IULTEQ & tcode.IOPMASK)
				rel := flg-> BULE_SEQ: BUG_SEQ;
			else ie (op = tcode.IUGTEQ & tcode.IOPMASK)
				rel := flg-> BUGE_SEQ: BUL_SEQ;
			else internal("unknown relation in BINARY()");
			emit8(rel, 3);
			emitj(JMP_SEQ, find(lab));
		end
		else do
			reject(op2);
			ie (op = tcode.IEQU & tcode.IOPMASK) rel := BNE_SEQ;
			else ie (op = tcode.INEQU & tcode.IOPMASK)
				rel := BE_SEQ;
			else ie (op = tcode.ILESS & tcode.IOPMASK)
				rel := BGE_SEQ;
			else ie (op = tcode.IGRTR & tcode.IOPMASK)
				rel := BLE_SEQ;
			else ie (op = tcode.ILTEQ & tcode.IOPMASK)
				rel := BG_SEQ;
			else ie (op = tcode.IGTEQ & tcode.IOPMASK)
				rel := BL_SEQ;
			else ie (op = tcode.IULESS & tcode.IOPMASK)
				rel := BUGE_SEQ;
			else ie (op = tcode.IUGRTR & tcode.IOPMASK)
				rel := BULE_SEQ;
			else ie (op = tcode.IULTEQ & tcode.IOPMASK)
				rel := BUG_SEQ;
			else ie (op = tcode.IUGTEQ & tcode.IOPMASK)
				rel := BUL_SEQ;
			else
				internal("unknown relation in BINARY()");
			emit8(rel, 5);
			emit(TRUE_SEQ, 0);
			emit8(JMPS_SEQ, 2);
			emit(FALSE_SEQ, 0);
		end
	end
end

store(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in STORE()");
		pop();
	end
	ie (op = tcode.ISAVG & tcode.IOPMASK)
		emit(STORE_GLOB, [(find(arg))]);
	else ie (op = tcode.ISAVL & tcode.IOPMASK)
		emit(STORE_LOC, [(-arg*BPW)]);
	else ie (op = tcode.ISAVI & tcode.IOPMASK)
		emit(STORE_IVAR, [(arg*BPW)]);
	else ie (op = tcode.ISTORB & tcode.IOPMASK)
		emit(STORE_BYTE, 0);
	else ie (op = tcode.ISTORE & tcode.IOPMASK)
		emit(STORE_SEQ, 0);
	else
		internal("bad mode in STORE()");
	if (	op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	)
		Depth := Depth-1;
	pop();
	! Leftover values are OK when computing dynamic tables
	! if (Depth) internal("leftover values in STORE()");
end

pcall(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.ICALR & tcode.IOPMASK) do
		if (\Active) do
			if (\Depth)
				internal("no active reg in PCALL(CALR)");
			pop();
		end
		emit(CALL_REG, 0);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK) do
		if (Active) push();
		emitj(CALL_SEQ, find(arg));
	end
	else do
		if (Active) push();
		ie (Firstpass) do
			emit(LOAD_CONST, [0]);
			emitj(CALL_SEQ, 0);
		end
		else do
			if (arg .>= MAXINFPROC \/ \Infmap[arg])
				error("bad interface mapping", 0);
			emit(LOAD_CONST, [(Infmap[arg][INF_OFFSET])]);
			emitj(CALL_SEQ, Infmap[arg][INF_ENTRY]);
		end
	end
	Active := 1;
end

pclean(arg) do
	Depth := Depth-arg;
	if (arg) emit(CLEANUP_SEQ, [(arg*BPW)]);
end

branch(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.IJUMP & tcode.IOPMASK) do
		emitj(JMP_SEQ, find(arg));
		if (Depth \/ Active) pop();
	end
	else do
		if (\Active) do
			if (\Depth) internal("too few operands in BRANCH()");
			pop();
		end
		emit(CMP_ZERO_SEQ, 0);
		if (	op = tcode.IBRT & tcode.IOPMASK \/
			op = tcode.IBRF & tcode.IOPMASK
		)
			pop();
		emit8(op=tcode.IBRT & tcode.IOPMASK \/
			op=tcode.INBRT & tcode.IOPMASK-> BE_SEQ: BNE_SEQ, 3);
		emitj(JMP_SEQ, find(arg));
	end
end

genstr(s, k) do var i;
	if (Dtop + Btop + k .>= KFFFF-256)
		error("data segment overflow", 0);
	for (i=0, k) Data.wrch(s::i);
	Data.wrch(0);
	Dtop := Dtop + k+1;
	if (Dtop & 1) do
		Data.wrch(0);
		Dtop := Dtop+1;
	end
end

mapinf(id, name, len) do var i, ip;
	for (i=0, Ninf) do
		ip := @Infprocs[i*INF];
		if (	ip[INF_NAME]::0 = name::0 /\
			\t.memcomp(ip[INF_NAME], name, len+1)
		)
			leave;
	end
	if (i >= Ninf) error("unresolved interface procedure", name);
	if (id >= MAXINFPROC) error("interface ID out of range", name);
	Infmap[id] := ip;
end

w(s) do var b::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
end

generate(op, a1, a2, sbuf) do
	ie (	op = tcode.ILDG & tcode.IOPMASK \/
		op = tcode.ILDGV & tcode.IOPMASK \/
		op = tcode.ILDL & tcode.IOPMASK \/
		op = tcode.ILDLV & tcode.IOPMASK \/
		op = tcode.ILDI & tcode.IOPMASK \/
		op = tcode.ILDIV & tcode.IOPMASK \/
		op = tcode.ILDLAB & tcode.IOPMASK \/
		op = tcode.INUM & tcode.IOPMASK \/
		op = tcode.ISELF & tcode.IOPMASK
	) do
		load(op, a1);
	end
	else ie (op = tcode.INEG & tcode.IOPMASK \/
		op = tcode.ILNOT & tcode.IOPMASK \/
		op = tcode.IBNOT & tcode.IOPMASK
	) do
		unary(op);
	end
	else ie (op = tcode.INORMB & tcode.IOPMASK \/
		op = tcode.INORM & tcode.IOPMASK \/
		op = tcode.IMUL & tcode.IOPMASK \/
		op = tcode.IDIV & tcode.IOPMASK \/
		op = tcode.IMOD & tcode.IOPMASK \/
		op = tcode.IADD & tcode.IOPMASK \/
		op = tcode.ISUB & tcode.IOPMASK \/
		op = tcode.IBAND & tcode.IOPMASK \/
		op = tcode.IBOR & tcode.IOPMASK \/
		op = tcode.IBXOR & tcode.IOPMASK \/
		op = tcode.IBSHL & tcode.IOPMASK \/
		op = tcode.IBSHR & tcode.IOPMASK \/
		op = tcode.IEQU & tcode.IOPMASK \/
		op = tcode.INEQU & tcode.IOPMASK \/
		op = tcode.ILESS & tcode.IOPMASK \/
		op = tcode.IGRTR & tcode.IOPMASK \/
		op = tcode.ILTEQ & tcode.IOPMASK \/
		op = tcode.IGTEQ & tcode.IOPMASK \/
		op = tcode.IUMUL & tcode.IOPMASK \/
		op = tcode.IUDIV & tcode.IOPMASK \/
		op = tcode.IULESS & tcode.IOPMASK \/
		op = tcode.IUGRTR & tcode.IOPMASK \/
		op = tcode.IULTEQ & tcode.IOPMASK \/
		op = tcode.IUGTEQ & tcode.IOPMASK \/
		op = tcode.IDREFB & tcode.IOPMASK \/
		op = tcode.IDEREF & tcode.IOPMASK
	) do
		binary(op);
	end
	else ie (op = tcode.ISAVG & tcode.IOPMASK \/
		op = tcode.ISAVL & tcode.IOPMASK \/
		op = tcode.ISAVI & tcode.IOPMASK \/
		op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	) do
		store(op, a1);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK \/
		op = tcode.ICALR & tcode.IOPMASK \/
		op = tcode.IICALL & tcode.IOPMASK
	) do
		pcall(op, a1);
	end
	else ie (op = tcode.IPOP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		pop();
	end
	else ie (op = tcode.ICLEAN & tcode.IOPMASK) do
		pclean(a1);
	end
	else ie (op = tcode.ISTACK & tcode.IOPMASK) do
		emit(ADJUST_SEQ, [(a1*BPW)]);
	end
	else ie (op = tcode.IINCG & tcode.IOPMASK) do
		emit(INC_GLOB, [(find(a1)),(a2)]);
	end
	else ie (op = tcode.IINCI & tcode.IOPMASK) do
		emit(INC_IVAR, [(a1*BPW),(a2)]);
	end
	else ie (op = tcode.IINCL & tcode.IOPMASK) do
		emit(INC_LOC, [(-a1*BPW),(a2)]);
	end
	else ie (op = tcode.IUNEXT & tcode.IOPMASK \/
		op = tcode.IDNEXT & tcode.IOPMASK
	) do
		if (Queued) spill(Iqueue, Aqueue);
		if (\Active \/ \Depth)
			internal("too few operands in GENERATE(?NEXT)");
		emit(CMP_SEQ[MODE_STACK], 0);
		Depth := Depth-1;
		emit8(op=tcode.IUNEXT & tcode.IOPMASK-> BL_SEQ: BG_SEQ, 3);
		emitj(JMP_SEQ, find(a1));
		pop();
	end
	else ie (op = tcode.IBRF & tcode.IOPMASK \/
		op = tcode.IBRT & tcode.IOPMASK \/
		op = tcode.INBRF & tcode.IOPMASK \/
		op = tcode.INBRT & tcode.IOPMASK \/
		op = tcode.IJUMP & tcode.IOPMASK
	) do
		branch(op, a1);
	end
	else ie (op = tcode.IHDR & tcode.IOPMASK) do
		emit(PROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IEND & tcode.IOPMASK) do
		emit(EPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END", 0);
		Active := 0;
	end
	else ie (op = tcode.IMHDR & tcode.IOPMASK) do
		emit(MPROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IENDM & tcode.IOPMASK) do
		emit(MEPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END", 0);
		Active := 0;
	end
	else ie (op = tcode.IINIT & tcode.IOPMASK) do
		if (a1 \= 8) error("unsupported Tcode version", 0);
	end
	else ie (op = tcode.IBPW & tcode.IOPMASK) do
		if (a1 \= 2) error("unsupported machine word size", 0);
	end
	else ie (op = tcode.ICLAB & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Firstpass) mklabel(a1, 'C', Ctop);
	end
	else ie (op = tcode.IDLAB & tcode.IOPMASK) do
		if (Dataname) do
			if (Firstpass) mklabel(Dataname, 'D', Dtop);
		end
		Dataname := a1;
	end
	else ie (op = tcode.IVEC & tcode.IOPMASK) do
		if (Firstpass) do
			mklabel(Dataname, 'B', Btop);
			if (Dtop + Btop + a1 .>= KFFFF-256)
				error("data segment overflow", 0);
			Btop := Btop + (a1=0-> BPW: a1*BPW);
		end
		Dataname := 0;
	end
	else ie (op = tcode.IDATA & tcode.IOPMASK \/
		op = tcode.IDREF & tcode.IOPMASK \/
		op = tcode.ICREF & tcode.IOPMASK
	) do
		if (Firstpass) do
			if (Dataname) mklabel(Dataname, 'D', Dtop);
			ie (op = tcode.IDATA & tcode.IOPMASK) do
				wrdata(a1);
			end
			else do
				! memorize data->BSS references
				if (Ltypes::a1 = 'B') do
					if (Bt >= MAXBSSREF)
						error(
						"too many BSS references", 0);
					BSSrefs[Bt] := Dtop;
					Bt := Bt+1;
				end
				wrdata(find(a1));
			end
		end
		Dataname := 0;
	end
	else ie (op = tcode.ISTR & tcode.IOPMASK) do
		if (Firstpass) do
			mklabel(Dataname, 'D', Dtop);
			genstr(sbuf, a1);
		end
		Dataname := 0;
	end
	else ie (op = tcode.IHALT & tcode.IOPMASK) do
		emit(HALT_SEQ, [(a1)]);
	end
	else ie (op = tcode.IDUP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		push();
		Active := 1;
	end
	else ie (op = tcode.ISWAP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Depth < 1) internal("too few operands in GENERATE(SWAP)");
		emit(SWAP_SEQ, 0);
	end
	else ie (op = tcode.IIPROC & tcode.IOPMASK) do
		if (\Firstpass) mapinf(a1, sbuf, a2);
	end
	else ie (op = tcode.IHINT & tcode.IOPMASK \/
		op = tcode.ILIB & tcode.IOPMASK \/
		op = tcode.IGLUE & tcode.IOPMASK \/
		op = tcode.ICMAP & tcode.IOPMASK \/
		op = tcode.IPUB & tcode.IOPMASK \/
		op = tcode.IEXT & tcode.IOPMASK \/
		op = tcode.IIREF & tcode.IOPMASK
	) do
		! ignore
	end
	else ie (op = tcode.ICALX & tcode.IOPMASK) do
		error("unresolved CALX (external reference)", 0);
	end
	else ie (op = tcode.IICALX & tcode.IOPMASK) do
		error("unresolved ICALX (external interface)", 0);
	end
	else do
		internal("unknown instruction in GENERATE()");
	end
end

xlate() do
	var	op, a1, a2, i, j, k, ch;
	var	sbuf[256];

	op := rdcode();
	if (op \= tcode.IINIT & tcode.IOPMASK) error("magic match failed", 0);
	while (\In.eof()) do
		if (op >= tcode.IENDOFSET) error("invalid opcode", 0);
		if (Na) do
			a1 := rdarg(L1);
			k := a1;
			if (Na > 1) do
				a2 := rdarg(L2);
				k := a2;
			end
		end
		if (Sf) do
			for (i=0, k) sbuf::i := rdch();
			sbuf::i := 0;
		end
		generate(op, a1, a2, sbuf);
		op := rdcode();
	end
end

addinfproc(ep, off, name, len) do var n;
	if (len + 1 + Nt >= NAMEPOOLSIZE) error("too many library names", 0);
	if (Ninf >= MAXINFPROC) error("too many interface procedures", 0);
	n := @Infprocs[Ninf*INF];
	Ninf := Ninf + 1;
	n[INF_ENTRY] := ep;
	n[INF_OFFSET] := off;
	n[INF_NAME] := @Names::Nt;
	t.memcopy(@Names::Nt, name, len+1);
	Nt := Nt + len + 1;
end

finddir(buf, len) do var i;
	for (i=len-6, -1, -1) do
		if (buf::i = 'T' /\ \t.memcomp(@buf::i, "T3XLIB", 6))
			return i+6;
	end
	return -1;
end

relocate(buf, n) do var i, r, p;
	for (i=0, n) do
		r := @buf::(i*2+6);
		p := r::1 << 8 | r::0;
		p := p+Ctop;
		r::0 := p;
		r::1 := p>>8;
	end
end

loadlib(path) do var buf::DIRBUFSIZE, ld, k, i, msg, n, nproc;
	msg := "corrupt library file";
	ld := t.open(path, T3X.OREAD);
	if (ld < 0) return -1;
	k := t.read(ld, buf, DIRBUFSIZE);
	if (k = DIRBUFSIZE) do
		t.seek(ld, 1024, T3X.SEEK_END);
		k := t.read(ld, buf, DIRBUFSIZE);
		if (k \= DIRBUFSIZE) error("file read error", path);
	end
	i := finddir(buf, k);
	if (i < 0) error(msg, path);
	n := 0;
	if (\Firstpass) do
		nproc := 0;
		while (buf::i) do
			if (i >= k) error(msg, path);
			addinfproc(Ctop+3, n, @buf::(i+1), buf::i-2);
			i := i + buf::i;
			n := n+BPW;
			nproc := nproc + 1;
		end
	end
	t.seek(ld, 0, T3X.SEEK_SET);
	k := t.read(ld, buf, DIRBUFSIZE);
	if (\Firstpass) relocate(buf, nproc);
	while (1) do
		i := finddir(buf, k);
		ie (i < 0) do
			ie (k < DIRBUFSIZE) do
				internal("LOADLIB: missing signature");
			end
			else do
				if (\Firstpass) Out.write(buf, DIRBUFSIZE/2);
				Ctop := Ctop + DIRBUFSIZE/2;
				t.memcopy(buf, @buf::(DIRBUFSIZE/2),
					DIRBUFSIZE/2);
				k := t.read(ld, @buf::(DIRBUFSIZE/2),
					DIRBUFSIZE/2);
				if (k >= 0) k := k + DIRBUFSIZE/2;
			end
		end
		else do
			if (\Firstpass) Out.write(buf, i-6);
			Ctop := Ctop+i-6;
			leave;
		end
	end
	t.close(ld);
end

loadlibs() do
	var	path::256, buf::LIBMAX+1;
	var	ld, k, i, kp, ks, kt, n, msg;

	msg := "too many libraries";
	ld := t.open("LIBDIR", T3X.OREAD);
	if (ld = -1) return;
	ie (t.getenv("T3XDIR", buf, LIBMAX) < 0 /\
		t.getenv("t3xdir", buf, LIBMAX) < 0
	)
		path::0 := 0;
	else
		str.format(path, "%S/lib/", [(buf)]);
	n := t.read(ld, buf, LIBMAX+1);
	t.close(ld);
	if (n > LIBMAX) error(msg, 0);
	kp := path::0-> str.length(path): 0;
	ks := 2;
	for (i=0, n) if (buf::i = '\n' \/ buf::i = '\r') buf::i := 0;
	i := 0;
	while (i < n) do
		k := str.length(@buf::i);
		if (kp + ks + k + 1 >= 256) error(msg, 0);
		t.memcopy(@path::kp, @buf::i, k);
		i := i + k+1;
		while (buf::i = 0 /\ i < n) i := i+1;
		kt := kp + k;
		t.memcopy(@path::kt, ".x", ks);
		kt := kt + ks;
		path::kt := 0;
		if (loadlib(@path::kp) < 0) do
			if (loadlib(path) < 0)
				error("no such extension library",
					@path::kp);
		end
	end
end

pass1() do
	Firstpass := 1;
	Bss := 0;
	Ctop := 16;	! 16 bytes reserved for segment setup
	Dtop := 0;
	Btop := 0;
	In.create(T3X.SYSIN, Inbuf, BUFSIZE, iostream.FREAD);
	if (Out.open(Tmpfile, Outbuf, BUFSIZE, iostream.FWRITE) = %1)
		error("cannot create temporary file", 0);
	if (Data.open(Tmpdata, Databuf, BUFSIZE, iostream.FWRITE) = %1)
		error("cannot create data file", 0);
	loadlibs();
	xlate();
	Out.close();
	Data.close();
end

wrwrd(x) do
	Out.wrch(x & 255);
	Out.wrch(x >> 8);
end

wrbs(k, v) do var i;
	for (i=0, k) Out.wrch(v::i);
end

! emit segment setup code
! tlen = text segment size in paragraphs (16-byte blocks)
segsetup(tlen) do
	wrbs(2, packed[ 0x8C, 0xC8 ]);		! MOV AX,CS
	wrbs(1, packed[ 0x05 ]); wrwrd(tlen);	! ADD AX,tlen
	wrbs(2, packed[ 0x8E, 0xD0 ]);		! MOV SS,AX
	wrbs(3, packed[ 0xBC, 0xFE, 0xFF ]);	! MOV SP,$FFFE
	wrbs(2, packed[ 0x8E, 0xD8 ]);		! MOV DS,AX
	wrbs(2, packed[ 0x8C, 0xC2 ]);		! MOV DX,ES
	wrbs(2, packed[ 0x8E, 0xC0 ]);		! MOV ES,AX
end

! generate EXE header
writeexehd() do
	const   delta = 32;     ! size of non-{TEXT,DATA} area (header)
	var     k, m, n;
	var     ltext, ldata;

	ltext := Ctop + 15 & ~15;	! paragraphs, rounded up
	ldata := Bss + 15 & ~15;
	m := ltext mod 512;
	n := (ldata+delta) mod 512;
	k := ltext ./ 512 + (ldata+delta) ./ 512;	! total size / 512
	ie (m+n >= 512) do				! fix up total size
		if (m+n \= 512) k := k+1;
		m := m+n-512;
	end
	else do
		m := m+n;
	end
	! write EXE header
	wrbs(2, packed[ 0x4D, 0x5A ]);	! Magic
	wrwrd(m);			! Size MOD 512
	wrwrd(k+1);			! Size / 512 + 1
	wrwrd(0);			! Num. of RELOC entries
	wrwrd(2);			! (header size + 15) / 16
	wrwrd(0);			! MINALLOC
	wrwrd(~0);			! MAXALLOC
	wrwrd(ltext./16+2);		! Initial SS
	wrwrd(~2);			! Initial SP
	wrwrd(0);			! Checksum (0=ignore)
	wrwrd(0);			! IEP, IP
	wrwrd(0);			! IEP, CS
	wrwrd(28);			! Offset of RELOC
	wrwrd(0);			! Number of overlays
	wrbs(4, packed[ 0xFF, 0xFF, 0xFF, 0xFF ]);
	return ltext;
end

! fill text segment up to next paragraph boundary
fill(len) do var z::16;
	t.memfill(z, 0, 16);
	if (len .> 0) do
		if (len mod 16) Out.write(z, 16 - len mod 16);
	end
end

! append data segment
appenddata(file) do var ch, k, n, bp, a;
	if (In.open(file, Inbuf, BUFSIZE, iostream.FREAD) = %1)
		error("cannot open data segment file", 0);
	a := 0;
	bp := 1;
	n := Bt-> BSSrefs[0]: -1;
	while (1) do
		ch := In.rdch();
		if (In.eof()) leave;
		if (a = n) do
			! address references BSS:
			! fix up by adding BSS start address
			n := bp >= Bt-> -1: BSSrefs[bp];
			bp := bp+1;
			k := (In.rdch() << 8 | ch) + Bss;
			if (Out.wrch(k & 255) < 0)
				error("file write error", 0);
			ch := k>>8;
			a := a+1;
		end
		if (Out.wrch(ch) < 0) error("file write error", 0);
		a := a+1;
	end
	In.close();
end

pass2() do var ltext, i, buf::4;
	Bss := Dtop;
	Out.create(T3X.SYSOUT, Outbuf, BUFSIZE, iostream.FWRITE);
	ltext := writeexehd();
	segsetup(ltext >> 4);
	if (In.open(Tmpfile, Inbuf, BUFSIZE, iostream.FREAD) = %1)
		error("cannot re-open temporary file", 0);
	Firstpass := 0;
	Ctop := 16;
	loadlibs();
	Addr := 0;
	Active := 0;
	xlate();
	In.close();
	t.remove(Tmpfile);
	fill(Ctop);
	appenddata(Tmpdata);
	t.remove(Tmpdata);
	Out.flush();
	if (Verbose) do var b::100;
		t.write(2, str.format(b, "text=%UD data=%UD bss=%UD%s",
			[(Ctop), (Dtop), (Btop), (t.newline(buf))]),
			str.length(b));
	end
end

do var abuf::4;
	init();
	if (t.getarg(1, abuf, 4) = 2 /\ \t.memcomp(abuf, "-v", 3))
		Verbose := 1;
	pass1();	! build data segment
	pass2();	! build text segment
end
