! TXCGUC -- Tcode->Ugly C translator
! Copyright (C) 2000,2001,2002,2003,2004 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcguc(t3x, tcode, string, iostream);

object	t[t3x], str[string];

const	BUFSIZE =		1024;		! I/O buffer size
const	MAXINF =		1024;		! max. interface procedures
const	MAXLABEL =		8192;		! max. number of labels
const	NAMEPOOLSIZE =	8192;		! interface name pool size
const	MAXPROTO =		1024;		! max. functions

const	STACKSIZE =		32768;

struct	INFPROC =
		INFLABEL,
		INFNARG,
		INFCMAP,
		INFNAME;

var		Phase;
object	In[iostream], Out[iostream], Tmp[iostream];
var		R_op, R_fl;					! Rejection queue
var		Inbuf::BUFSIZE;
var		Outbuf::BUFSIZE;
var		Tmpbuf::BUFSIZE;
var		Tmpfile::256;
var		Segment;
var		Addr;
var		Tlabel;
var		Iqueue, Aqueue,				! 1-instruction queue (instr, arg)
		Queued,						! queue active indicator (1=active),
		Active,						! accumulator active flag (1=active),
		Depth;						! RT stack depth
var		Tprefix, Aprefix;
var		Infprocs[MAXINF*INFPROC],	! interface procedure table
		Infmax;						! # of interfaces
var		Names::NAMEPOOLSIZE, Ntop;	! Name pool, top of pool
var		Lname[MAXLABEL],			! Label/address pool
		Laddr[MAXLABEL];
var		Prototbl[MAXPROTO],			! Prototype table
		Ptop;						! top of Prototbl
var		Ltop;						! top of L/A pool
var		Dtop;						! top of data area
var		Dsize;						! size of data area
var		Fl, Na, Sf, L1, L2;			! Tcode flags
var		Lbpw;						! t.bpw() cache
var		XFFFFFFFF;					! Constant 0xFFFFFFFF
var		Verbose;

var	BANNER,
	HEAD,
	TAIL,
	DECL_WORD,
	DECL_CREF,
	DECL_DREF,
	DECL_CODE_LABEL,
	DECL_TMP_LABEL,
	LOAD_GLOB,
	LOAD_LOC,
	LOAD_IVAR,
	LOAD_GLOB_VEC,
	LOAD_LOC_VEC,
	LOAD_IVAR_VEC,
	LOAD_DLAB,
	LOAD_CLAB,
	LOAD_CONST,
	LOAD_SELF,
	PUSH_SEQ,
	POP_SEQ,
	DEREF_SEQ,
	DEREF_BYTE_SEQ,
	STORE_GLOB,
	STORE_LOC,
	STORE_IVAR,
	STORE_SEQ,
	STORE_BYTE,
	NORM_SEQ,
	NORM_BYTE_SEQ,
	INC_GLOB,
	INC_LOC,
	INC_IVAR,
	NEG_SEQ,
	NOT_SEQ,
	LNOT_SEQ,
	EXT_BOOL_SEQ,
	SAVE_CMP_RESULT,
	MUL_SEQ,
	DIV_SEQ,
	MOD_SEQ,
	UMUL_SEQ,
	UDIV_SEQ,
	ADD_SEQ,
	SUB_SEQ,
	AND_SEQ,
	OR_SEQ,
	XOR_SEQ,
	SHL_SEQ,
	SHR_SEQ,
	CMP_EQ_SEQ,
	CMP_NE_SEQ,
	CMP_LT_SEQ,
	CMP_GT_SEQ,
	CMP_LE_SEQ,
	CMP_GE_SEQ,
	CMP_ULT_SEQ,
	CMP_UGT_SEQ,
	CMP_ULE_SEQ,
	CMP_UGE_SEQ,
	SWAP_SEQ,
	BE_SEQ,
	BNE_SEQ,
	BL_SEQ,
	BG_SEQ,
	BLE_SEQ,
	BGE_SEQ,
	BUL_SEQ,
	BUG_SEQ,
	BULE_SEQ,
	BUGE_SEQ,
	JMP_SEQ,
	TRUE_SEQ,
	FALSE_SEQ,
	EPILOGUE_SEQ,
	MPROLOGUE_SEQ,
	MEPILOGUE_SEQ,
	CALL_REG,
	CALL_SEQ,
	ADJUST_SEQ,
	CLEANUP_SEQ,
	HALT_SEQ;

! variables to bind addressing modes
struct	MODES =
		MODE_GLOB,		! acc/memory
		MODE_LOC,		! acc/(local)
		MODE_IVAR,		! acc/(instance)
		MODE_CONST,		! acc/immediate
		MODE_SELF,		! acc/(self)
		MODE_STACK;		! acc/(popped value)

mkops() do
	BANNER := [
		"/*",
		" * File generated by TXCGUC, do not edit",
		" */",
	0 ];
	HEAD := [
		"#ifdef plan9",
		" #include <u.h>",
		" #include <libc.h>",
		"#else",
		" #include <stdlib.h>",
		" #include <string.h>",
		"#endif",
		"#define BPW	4",
		"#define XBPW	2",
		"#define CELLS	%1",
		"#define B	unsigned char",
		"#define C	int",
		"#define U	unsigned int",
		"C	A;",
		"B	*BP;",
		"C	*P;",
		"C	T;",
		"C	X;",
		"U	S = CELLS;",
		"U	F = CELLS;",
		"C	*I = 0;",
		"C	M[CELLS];",
		"U	*UM = (U*)M;",
		"#define push(X)	M[--S] = X",
		"#define pop()	M[S++]",
		"#ifdef plan9",
		" void exit(int n) {",
		"	char b[4];",
		"	sprint(b, \"%%d\", n & 255);",
		"	exits(n? b: nil);",
		" }",
		"#endif",
		"void init(void) {",
	0 ];
	TAIL := [ "}", 0 ];
	DECL_WORD := [ "M[%1] = %2;", 0 ];
	DECL_CREF := [ "{ void P%2(void); M[%1] = (C)&P%2; }", 0 ];
	DECL_DREF := [ "M[%1] = (C)&M[%2];", 0 ];
	DECL_CODE_LABEL := [ "A%1:", 0 ];
	DECL_TMP_LABEL := [ "T%1:", 0 ];
	LOAD_GLOB := [ "A = M[%1];", 0 ];
	LOAD_LOC := [ "A = M[F-(%1)];", 0];
	LOAD_IVAR := [ "A = I[%1];", 0 ];
	LOAD_GLOB_VEC := [ "A = (C)&M[%1];", 0 ];
	LOAD_LOC_VEC := [ "A = (C)&M[F-(%1)];", 0];
	LOAD_IVAR_VEC := [ "A = (C)&I[%1];", 0 ];
	LOAD_CONST := [ "A = %1;", 0 ];
	LOAD_CLAB := [ "A = (C)&P%1;", 0 ];
	LOAD_DLAB := [ "A = (C)&M[%1];", 0 ];
	LOAD_SELF := [ "A = (C)I;", 0 ];
	PUSH_SEQ := [ "push(A);", 0 ];
	POP_SEQ := [ "A = pop();" ,0 ];
	DEREF_SEQ := [
		[ "P = (C*)A + M[%1]; A = *P;", 0 ],
		[ "P = (C*)A + M[F-(%1)]; A = *P;", 0 ],
		[ "P = (C*)A + I[%1]; A = *P;", 0 ],
		[ "P = (C*)A + %1; A = *P;", 0 ],
		[ "P = (C*)A + I; A = *P;", 0 ],
		[ "P = (C*)pop() + A; A = *P;", 0 ],
	0 ];
	DEREF_BYTE_SEQ := [ 
		[ "BP = (B*)A + M[%1]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + M[F-(%1)]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + I[%1]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + %1; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + I; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + pop(); A = (C)*BP;", 0 ],
	0 ];
	STORE_GLOB := [ "M[%1] = A;", 0 ];
	STORE_LOC := [ "M[F-(%1)] = A;", 0 ];
	STORE_IVAR := [ "I[%1] = A;", 0 ];
	STORE_SEQ := [ "P = (C*)pop(); *P = A;", 0 ];
	STORE_BYTE := [ "BP = (B*)pop(); *BP = (B)A;", 0 ];
	NORM_SEQ := [
		[ "A += M[%1] << XBPW;", 0 ],
		[ "A += M[F-(%1)] << XBPW;", 0 ],
		[ "A += I[%1] << XBPW;", 0 ],
		[ "A += %1 << XBPW;", 0 ],
		[ "A += (C)I << XBPW;", 0 ],
		[ "A = (A << XBPW) + pop();", 0 ],
	0 ];
	NORM_BYTE_SEQ := [
		[ "A += M[%1];", 0 ],
		[ "A += M[F-(%1)];", 0 ],
		[ "A += I[%1];", 0 ],
		[ "A += %1;", 0 ],
		[ "A += (C)I;", 0 ],
		[ "A += pop();", 0 ],
	0 ];
	INC_GLOB := [ "M[%1] += %2;", 0 ];
	INC_LOC := [ "M[F-(%1)] += %2;", 0 ];
	INC_IVAR := [ "I[%1] += %2;", 0 ];
	NEG_SEQ := [ "A = -A;", 0 ];
	NOT_SEQ := [ "A = ~A;", 0 ];
	LNOT_SEQ := [ "A = A?0:-1;", 0 ];
	EXT_BOOL_SEQ := [ "A = T?-1:0;", 0 ];
	SAVE_CMP_RESULT := [ "T = A;", 0 ];
	MUL_SEQ := [
		[ "A *= M[%1];", 0 ],
		[ "A *= M[F-(%1)];", 0 ],
		[ "A *= I[%1];", 0 ],
		[ "A *= %1;", 0 ],
		[ "A *= (C)I;", 0 ],
		[ "A *= pop();", 0 ],
	0 ];
	DIV_SEQ := [
		[ "A /= M[%1];", 0 ],
		[ "A /= M[F-(%1)];", 0 ],
		[ "A /= I[%1];", 0 ],
		[ "A /= %1;", 0 ],
		[ "A /= (C)I;", 0 ],
		[ "A = pop() / A;", 0 ],
	0 ];
	MOD_SEQ := [
		[ "A %%= M[%1];", 0 ],
		[ "A %%= M[F-(%1)];", 0 ],
		[ "A %%= I[%1];", 0 ],
		[ "A %%= %1;", 0 ],
		[ "A %%= (C)I;", 0 ],
		[ "A = pop() %% A;", 0 ],
	0 ];
	UMUL_SEQ := [
		[ "A = (U)A * UM[%1];", 0 ],
		[ "A = (U)A * UM[F-(%1)];", 0 ],
		[ "A = (U)A * (U)I[%1];", 0 ],
		[ "A = (U)A * (U)%1;", 0 ],
		[ "A = (U)A * (U)I;", 0 ],
		[ "A = (U)A * (U)pop();", 0 ],
	0 ];
	UDIV_SEQ := [
		[ "A = (U)A / UM[%1];", 0 ],
		[ "A = (U)A / UM[F-(%1)];", 0 ],
		[ "A = (U)A / (U)I[%1];", 0 ],
		[ "A = (U)A / (U)%1;", 0 ],
		[ "A = (U)A / (U)I;", 0 ],
		[ "A = (U)pop() / (U)A;", 0 ],
	0 ];
	ADD_SEQ := NORM_BYTE_SEQ;
	SUB_SEQ := [
		[ "A -= M[%1];", 0 ],
		[ "A -= M[F-(%1)];", 0 ],
		[ "A -= I[%1];", 0 ],
		[ "A -= %1;", 0 ],
		[ "A -= (C)I;", 0 ],
		[ "A = pop() - A;", 0 ],
	0 ];
	AND_SEQ := [
		[ "A &= M[%1];", 0 ],
		[ "A &= M[F-(%1)];", 0 ],
		[ "A &= I[%1];", 0 ],
		[ "A &= %1;", 0 ],
		[ "A &= (C)I;", 0 ],
		[ "A &= pop();", 0 ],
	0 ];
	OR_SEQ := [
		[ "A |= M[%1];", 0 ],
		[ "A |= M[F-(%1)];", 0 ],
		[ "A |= I[%1];", 0 ],
		[ "A |= %1;", 0 ],
		[ "A |= (C)I;", 0 ],
		[ "A |= pop();", 0 ],
	0 ];
	XOR_SEQ := [
		[ "A ^= M[%1];", 0 ],
		[ "A ^= M[F-(%1)];", 0 ],
		[ "A ^= I[%1];", 0 ],
		[ "A ^= %1;", 0 ],
		[ "A ^= (C)I;", 0 ],
		[ "A ^= pop();", 0 ],
	0 ];
	SHL_SEQ := [
		[ "A = (U)A << UM[%1];", 0 ],
		[ "A = (U)A << UM[F-(%1)];", 0 ],
		[ "A = (U)A << (U)I[%1];", 0 ],
		[ "A = (U)A << (U)%1;", 0 ],
		[ "A = (U)A << (U)I;", 0 ],
		[ "A = (U)pop() << (U)A;", 0 ],
	0 ];
	SHR_SEQ := [
		[ "A = (U)A >> UM[%1];", 0 ],
		[ "A = (U)A >> UM[F-(%1)];", 0 ],
		[ "A = (U)A >> (U)I[%1];", 0 ],
		[ "A = (U)A >> (U)%1;", 0 ],
		[ "A = (U)A >> (U)I;", 0 ],
		[ "A = (U)pop() >> (U)A;", 0 ],
	0 ];
	CMP_EQ_SEQ := [
		[ "T = A == M[%1];", 0 ],
		[ "T = A == M[F-(%1)];", 0 ],
		[ "T = A == I[%1];", 0 ],
		[ "T = A == %1;", 0 ],
		[ "T = A == (C)I;", 0 ],
		[ "T = pop() == A;", 0 ],
	0 ];
	CMP_NE_SEQ := [
		[ "T = A != M[%1];", 0 ],
		[ "T = A != M[F-(%1)];", 0 ],
		[ "T = A != I[%1];", 0 ],
		[ "T = A != %1;", 0 ],
		[ "T = A != (C)I;", 0 ],
		[ "T = pop() != A;", 0 ],
	0 ];
	CMP_LT_SEQ := [
		[ "T = A < M[%1];", 0 ],
		[ "T = A < M[F-(%1)];", 0 ],
		[ "T = A < I[%1];", 0 ],
		[ "T = A < %1;", 0 ],
		[ "T = A < (C)I;", 0 ],
		[ "T = pop() < A;", 0 ],
	0 ];
	CMP_GT_SEQ := [
		[ "T = A > M[%1];", 0 ],
		[ "T = A > M[F-(%1)];", 0 ],
		[ "T = A > I[%1];", 0 ],
		[ "T = A > %1;", 0 ],
		[ "T = A > (C)I;", 0 ],
		[ "T = pop() > A;", 0 ],
	0 ];
	CMP_LE_SEQ := [
		[ "T = A <= M[%1];", 0 ],
		[ "T = A <= M[F-(%1)];", 0 ],
		[ "T = A <= I[%1];", 0 ],
		[ "T = A <= %1;", 0 ],
		[ "T = A <= (C)I;", 0 ],
		[ "T = pop() <= A;", 0 ],
	0 ];
	CMP_GE_SEQ := [
		[ "T = A >= M[%1];", 0 ],
		[ "T = A >= M[F-(%1)];", 0 ],
		[ "T = A >= I[%1];", 0 ],
		[ "T = A >= %1;", 0 ],
		[ "T = A >= (C)I;", 0 ],
		[ "T = pop() >= A;", 0 ],
	0 ];
	CMP_ULT_SEQ := [
		[ "T = (U)A < UM[%1];", 0 ],
		[ "T = (U)A < UM[F-(%1)];", 0 ],
		[ "T = (U)A < (U)I[%1];", 0 ],
		[ "T = (U)A < (U)%1;", 0 ],
		[ "T = (U)A < (U)I;", 0 ],
		[ "T = (U)pop() < (U)A;", 0 ],
	0 ];
	CMP_UGT_SEQ := [
		[ "T = (U)A > UM[%1];", 0 ],
		[ "T = (U)A > UM[F-(%1)];", 0 ],
		[ "T = (U)A > (U)I[%1];", 0 ],
		[ "T = (U)A > (U)%1;", 0 ],
		[ "T = (U)A > (U)I;", 0 ],
		[ "T = (U)pop() > (U)A;", 0 ],
	0 ];
	CMP_ULE_SEQ := [
		[ "T = (U)A <= UM[%1];", 0 ],
		[ "T = (U)A <= UM[F-(%1)];", 0 ],
		[ "T = (U)A <= (U)I[%1];", 0 ],
		[ "T = (U)A <= (U)%1;", 0 ],
		[ "T = (U)A <= (U)I;", 0 ],
		[ "T = (U)pop() <= (U)A;", 0 ],
	0 ];
	CMP_UGE_SEQ := [
		[ "T = (U)A >= UM[%1];", 0 ],
		[ "T = (U)A >= UM[F-(%1)];", 0 ],
		[ "T = (U)A >= (U)I[%1];", 0 ],
		[ "T = (U)A >= (U)%1;", 0 ],
		[ "T = (U)A >= (U)I;", 0 ],
		[ "T = (U)pop() >= (U)A;", 0 ],
	0 ];
	SWAP_SEQ := [ "X = A; A = M[S]; M[S] = X; ", 0 ];
	BE_SEQ := [ "if (!T) goto %A%2;", 0 ];
	BNE_SEQ := [ "if (T) goto %A%2;", 0 ];
	JMP_SEQ := [ "goto %A%2;", 0 ];
	TRUE_SEQ := [ "A = -1;", 0 ];
	FALSE_SEQ := [ "A = 0;", 0 ];
	EPILOGUE_SEQ := [
		"F = pop();",
		"++S;",
		"}",
	0 ];
	MEPILOGUE_SEQ := [
		"I = (C*)pop();",
		"F = pop();",
		"++S;",
		"}",
	0 ];
	CALL_REG := [ "(*(void(*)(void))A)()", 0 ];
	CALL_SEQ := [ "P%1();", 0 ];
	ADJUST_SEQ := [ "S -= %1;", 0 ];
	CLEANUP_SEQ := [ "S += %1;", 0 ];
	HALT_SEQ := [ "exit(%1);", 0 ];
end

decl	lookup(1);

reset() do
	R_op := %1;
	Addr := 0;
	Tlabel := 0;
	Active := 0;
	Queued := 0;
	Depth := 0;
	Dtop := 0;
end

init() do var i;
	Lbpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Aprefix := "A";
	Tprefix := "T";
	Infmax := 0;
	Ltop := 0;
	Dtop := 0;
	Ntop := 0;
	Verbose := 0;
	if (t.getenv("TMP", Tmpfile, 240) < 1 /\
		t.getenv("tmp", Tmpfile, 240) < 1
	) do
		Tmpfile::0 := '.';
		Tmpfile::1 := 0;
	end
	i := str.length(Tmpfile);
	t.memcopy(@Tmpfile::i, "/_txcguc.tmp", 12);
	reset();
	mkops();
end

error(msg) do var buf::128;
	t.write(T3X.SYSERR, "TXCGUC: ", 8);
	t.write(T3X.SYSERR, str.numtostr(buf, Addr, 10), str.length(buf));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	halt 1;
end

internal(msg) do
	t.write(T3X.SYSERR, "internal error: ", 16);
	error(msg);
end

rdch() do var ch;
	ch := In.rdch();
	Addr := Addr+1;
	if (Phase = 0 /\ ch \= -1) Tmp.wrch(ch);
	return ch;
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Lbpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Lbpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdarg(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (Lbpw < 4 /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		error("value out of range");
	return sext32(n);
end

rdcode() do var op;
	ie (R_op >= 0) do
		op := R_op;
		Fl := R_fl;
		R_op := %1;
	end
	else do
		op := rdch();
		ie (op & tcode.IARFLAG) do
			Fl := op;
			op := rdch();
		end
		else do
			Fl := 0;
		end
	end
	Na := Fl & tcode.IARMASK >> tcode.IAMOFF;
	Sf := Fl & tcode.ISTFLAG;
	if (Na) do
		L1 := Fl & tcode.IL1MASK >> tcode.IL1OFF;
		if  (Na > 1)
			L2 := Fl & tcode.IL2MASK >> tcode.IL2OFF;
	end
	return op;
end

reject(op) do
	R_op := op;
	R_fl := Fl;
end

wrch(ch) Out.wrch(ch);

! emit a code sequence with arguments merged in
_emit(vcode, args) do var i, j, c, b::33, code, radix;
	j := 0;
	while (vcode[j]) do
		code := vcode[j]; j := j+1;
		i := 0;
		while (1) do
			c := code::i; i := i+1;
			if (\c) leave;
			ie (c = '%') do
				radix := -10;
				c := code::i; i := i+1;
				if (c = '+') do
					radix := 10;
					c := code::i; i := i+1;
				end
				ie ('1' <= c /\ c <= '9')
					Out.writes(str.numtostr(b,
						args[c-'1'], radix));
				else ie ('A' <= c /\ c <= 'Z')
					Out.writes(args[c-'A']);
				else ie (\c)
					i := i-1;
				else
					Out.wrch(c);
			end
			else do
				Out.wrch(c);
			end
		end
		Out.wrch('\n');
	end
end

emit(code, args) if (Phase = 2) _emit(code, args);

emitd(code, args) if (Phase = 1) _emit(code, args);

dgrow(n) do
	Dtop := Dtop + n;	! XXX size check?
end

emitl(seq, arg) if (Phase = 2) emit(seq, [(-arg)]);

emiti(seq, arg) if (Phase = 2) emit(seq, [(arg)]);

emitl2(seq, arg, a2) if (Phase = 2) emit(seq, [(-arg), (a2)]);

emiti2(seq, arg, a2) if (Phase = 2) emit(seq, [(arg), (a2)]);

newlabel() do
	Tlabel := Tlabel +1;
	return Tlabel;
end

push() do
	if (\Active) internal("accumulator not active in PUSH()");
	emit(PUSH_SEQ, 0);
	Depth := Depth+1;
	Active := 0;
end

pop() do
	if (\Depth) do
		if (Active) do
			Active := 0;
			return;
		end
		internal("stack underflow in POP()");
	end
	emit(POP_SEQ, 0);
	Depth := Depth-1;
	Active := 1;
end

loadlabel(l) do var a;
	a := lookup(l);
	ie (a < 0)
		emit(LOAD_CLAB, [(l)]);
	else
		emit(LOAD_DLAB, [(a)]);
end

! load queued value into accumulator
spill(m, a) do
	if (Active) push();
	ie (m = tcode.ILDG & tcode.IOPMASK) emit(LOAD_GLOB, [(lookup(a))]);
	else ie (m = tcode.ILDGV & tcode.IOPMASK)
		emit(LOAD_GLOB_VEC, [((lookup(a))]);
	else ie (m = tcode.ILDL & tcode.IOPMASK) emit(LOAD_LOC, [(a)]);
	else ie (m = tcode.ILDLV & tcode.IOPMASK)
		emit(LOAD_LOC_VEC, [(a)]);
	else ie (m = tcode.ILDI & tcode.IOPMASK) emiti(LOAD_IVAR, a);
	else ie (m = tcode.ILDIV & tcode.IOPMASK) emiti(LOAD_IVAR_VEC, a);
	else ie (m = tcode.ILDLAB & tcode.IOPMASK) loadlabel(a);
	else ie (m = tcode.INUM & tcode.IOPMASK) emit(LOAD_CONST, [(a)]);
	else ie (m = tcode.ISELF & tcode.IOPMASK) emit(LOAD_SELF, 0);
	else internal("unknown instruction in LOAD()");
	Active := 1;
	Queued := 0;
end

! generate load instruction
load(m, a) do
	ie (\Queued) do
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
		return;
	end
	else do
		spill(Iqueue, Aqueue);
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
	end
end

unary(op) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("accumulator not active in UNARY()");
		pop();
	end
	ie (op = tcode.INEG & tcode.IOPMASK) emit(NEG_SEQ, 0);
	else ie (op = tcode.IBNOT & tcode.IOPMASK) emit(NOT_SEQ, 0);
	else ie (op = tcode.ILNOT & tcode.IOPMASK) emit(LNOT_SEQ, 0);
	else internal("unknown instruction in UNARY()");
end

emitb(seq, mode)
	if (Phase = 2)
		_emit(seq[mode], [(mode = MODE_GLOB-> lookup(Aqueue): Aqueue)]);

cmp_regs(op, mode) do var op2, rel, flg, lab, t1;
	ie (op = tcode.IEQU) rel := CMP_EQ_SEQ;
	else ie (op = tcode.INEQU & tcode.IOPMASK) rel := CMP_NE_SEQ;
	else ie (op = tcode.ILESS & tcode.IOPMASK) rel := CMP_LT_SEQ;
	else ie (op = tcode.IGRTR & tcode.IOPMASK) rel := CMP_GT_SEQ;
	else ie (op = tcode.ILTEQ & tcode.IOPMASK) rel := CMP_LE_SEQ;
	else ie (op = tcode.IGTEQ & tcode.IOPMASK) rel := CMP_GE_SEQ;
	else ie (op = tcode.IULESS & tcode.IOPMASK) rel := CMP_ULT_SEQ;
	else ie (op = tcode.IUGRTR & tcode.IOPMASK) rel := CMP_UGT_SEQ;
	else ie (op = tcode.IULTEQ & tcode.IOPMASK) rel := CMP_ULE_SEQ;
	else ie (op = tcode.IUGTEQ & tcode.IOPMASK) rel := CMP_UGE_SEQ;
	else internal("unknown relation in BINARY()");
	emitb(rel, mode);
	op2 := rdcode();
	ie (op2 = tcode.IBRT & tcode.IOPMASK \/
		op2 = tcode.IBRF & tcode.IOPMASK
	) do
		pop();
		lab := rdarg(L1);
		emit(op2 = tcode.IBRF & tcode.IOPMASK-> BE_SEQ: BNE_SEQ,
			[(Aprefix), (lab)]);
	end
	else do
		reject(op2);
		emit(EXT_BOOL_SEQ, 0);
	end
end

wm(s) do var nl::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
end

undefd_addr() do
	wm("TXNCG: Addresses of local variables and anonymous vectors");
	wm("	   are not supported in binary operations.");
	wm("	   This may be considered a bug of TXCGUC.");
	wm("	   See the documentation for details.");
	internal("unsupported addressing mode in binary()");
end

binary(op) do var mode;
	if (\Active /\ Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in BINARY()");
		pop();
	end
	ie (Queued) do
		ie (Iqueue = tcode.ILDG & tcode.IOPMASK)
			mode := MODE_GLOB;
		else ie (Iqueue = tcode.ILDGV & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ILDL & tcode.IOPMASK)
			mode := MODE_LOC;
		else ie (Iqueue = tcode.ILDLV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDI & tcode.IOPMASK)
			mode := MODE_IVAR;
		else ie (Iqueue = tcode.ILDIV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDLAB & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.INUM & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ISELF & tcode.IOPMASK)
			mode := MODE_SELF;
		else
			internal("bad mode in BINARY()");
	end
	else do
		if (\Depth) internal("stack underflow in BINARY()");
		mode := MODE_STACK;
		Depth := Depth-1;
	end
	Queued := 0;
	Active := 1;
	! convert Tcode-style local/ivar addresses into real addresses
	ie (op = tcode.INORMB & tcode.IOPMASK) emitb(NORM_BYTE_SEQ, mode);
	else ie (op = tcode.INORM & tcode.IOPMASK) emitb(NORM_SEQ, mode);
	else ie (op = tcode.IDEREF & tcode.IOPMASK) emitb(DEREF_SEQ, mode);
	else ie (op = tcode.IDREFB & tcode.IOPMASK) emitb(DEREF_BYTE_SEQ,mode);
	else ie (op = tcode.IMUL & tcode.IOPMASK) emitb(MUL_SEQ, mode);
	else ie (op = tcode.IDIV & tcode.IOPMASK) emitb(DIV_SEQ, mode);
	else ie (op = tcode.IMOD & tcode.IOPMASK) emitb(MOD_SEQ, mode);
	else ie (op = tcode.IUMUL & tcode.IOPMASK) emitb(UMUL_SEQ, mode);
	else ie (op = tcode.IUDIV & tcode.IOPMASK) emitb(UDIV_SEQ, mode);
	else ie (op = tcode.IADD & tcode.IOPMASK) emitb(ADD_SEQ, mode);
	else ie (op = tcode.ISUB & tcode.IOPMASK) emitb(SUB_SEQ, mode);
	else ie (op = tcode.IBAND & tcode.IOPMASK) emitb(AND_SEQ, mode);
	else ie (op = tcode.IBOR & tcode.IOPMASK) emitb(OR_SEQ, mode);
	else ie (op = tcode.IBXOR & tcode.IOPMASK) emitb(XOR_SEQ, mode);
	else ie (op = tcode.IBSHL & tcode.IOPMASK) emitb(SHL_SEQ, mode);
	else ie (op = tcode.IBSHR & tcode.IOPMASK) emitb(SHR_SEQ, mode);
	else cmp_regs(op, mode);
end

store(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in STORE()");
		pop();
	end
	ie (op = tcode.ISAVG & tcode.IOPMASK) emit(STORE_GLOB, [(lookup(arg))]);
	else ie (op = tcode.ISAVL & tcode.IOPMASK) emitl(STORE_LOC, -arg);
	else ie (op = tcode.ISAVI & tcode.IOPMASK) emiti(STORE_IVAR, arg);
	else ie (op = tcode.ISTORB & tcode.IOPMASK) emit(STORE_BYTE, 0);
	else ie (op = tcode.ISTORE & tcode.IOPMASK) emit(STORE_SEQ, 0);
	else internal("bad mode in STORE()");
	if (op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	)
		Depth := Depth-1;
	pop();
	! Leftover values are OK when computing dynamic tables
	! if (Depth) internal("leftover values in STORE()");
end

findinfproc(id) do var y, i;
	for (i=0, Infmax, INFPROC) do
		y := @Infprocs[i];
		if (y[INFLABEL] = id) return y;
	end
	error("undefined interface procedure");
end

addinfproc(lab, len, narg, cmap, name) do var inf;
	if (len + Ntop + 1 >= NAMEPOOLSIZE \/
		Infmax + INFPROC >= MAXINF
	)
		error("too many interface procedures");
	inf := @Infprocs[Infmax];
	Infmax := Infmax + INFPROC;
	inf[INFLABEL] := lab;
	inf[INFNARG] := narg;
	inf[INFCMAP] := cmap;
	inf[INFNAME] := @Names::Ntop;
	Ntop := Ntop + len+1;
	t.memcopy(inf[INFNAME], name, len+1);
end

lookup(n) do var i;
	for (i=0, Ltop)
		if (Lname[i] = n)
			return Laddr[i];
	return -1;
end

addsym(n) do
	if (Phase \= 1) return;
	if (lookup(n) \= -1) error("re-definition of label");
	if (Ltop >= MAXLABEL) error("too many labels");
	Lname[Ltop] := n;
	Laddr[Ltop] := Dtop;
	Ltop := Ltop+1;
end

pclean(arg) do
	Depth := Depth-arg;
	if (arg) emiti(CLEANUP_SEQ, arg);
end

fixargs() do var op, narg, i, nra;
	op := rdcode();
	if (op \= tcode.ICLEAN & tcode.IOPMASK)
		internal("CLEAN required in fixargs()");
	narg := rdarg(L1);
	if (\narg) return 0;
	pclean(nra);
	return narg;
end

declinfprocs() do var y, i, j, b::100, cmbit;
	for (i=0, Infmax, INFPROC) do
		y := @Infprocs[i];
		Out.writes("C ");
		Out.writes(y[INFNAME]);
		Out.writes("(C *SELF");
		cmbit := 1;
		for (j=1, y[INFNARG]) cmbit := cmbit << 1;
		for (j=0, y[INFNARG]) do
			Out.writes(str.format(b, ", C%sa%d", [(
				y[INFCMAP] & cmbit-> " *": " ", j)]));
			cmbit := cmbit >> 1;
		end
		Out.writes(");\n");
	end
end

callinfproc(id) do var ip, i, cmbit, buf::100;
	if (Phase \= 2) return;
	ip := findinfproc(id);
	Out.writes("A = ");
	Out.writes(ip[INFNAME]);
	Out.writes("((C*)M[S]");
	cmbit := 1;
	for (i=1, ip[INFNARG]) cmbit := cmbit << 1;
	for (i=0, ip[INFNARG]) do
		Out.writes(str.format(buf, ", %sM[S+%d]", [(
			ip[INFCMAP] & cmbit-> "(C*)": "",
			i+1)]));
		cmbit := cmbit >> 1;
	end
	Out.writes(");\n");
end

pcall(op, arg) do var narg;
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.ICALR & tcode.IOPMASK) do
		if (\Active) do
			if (\Depth)
				internal("no active reg in PCALL(CALR)");
			pop();
		end
		emit(CALL_REG, 0);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK) do
		if (Active) push();
		emit(CALL_SEQ, [(arg)]);
	end
	else do ! op = tcode.IICALL & tcode.IOPMASK
		if (Active) push();
		callinfproc(arg);
	end
	Active := 1;
end

branch(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.IJUMP & tcode.IOPMASK) do
		if (arg \= 0) emit(JMP_SEQ, [(Aprefix), (arg)]);
		if (Depth \/ Active) pop();
	end
	else do
		if (\Active) do
			if (\Depth) internal("too few operands in BRANCH()");
			pop();
		end
		emit(SAVE_CMP_RESULT, 0);
		if (op = tcode.IBRT & tcode.IOPMASK \/
			op = tcode.IBRF & tcode.IOPMASK
		)
			pop();
		emit(op=tcode.IBRT & tcode.IOPMASK \/
			op=tcode.INBRT & tcode.IOPMASK-> BNE_SEQ:
			BE_SEQ, [(Aprefix), (arg)]);
	end
end

loop_regs(op, a1) do
	if (Phase = 2) do
		_emit(op=tcode.IUNEXT & tcode.IOPMASK-> CMP_GE_SEQ[MODE_STACK]:
			CMP_LE_SEQ[MODE_STACK], 0);
		emit(BNE_SEQ, [(Aprefix), (a1)]);
	end
	Depth := Depth-1;
	pop();
end

genstr(s, k) do var i, j, b::100;
	if (Phase \= 1) do
		dgrow((k+Lbpw)/Lbpw);
		return;
	end
	Out.writes(str.format(b, "memcpy(&M[%d], \"", [(Dtop)]));
	for (i=0, k) do
		ie (s::i < 32 \/ s::i > 126) do
			Out.wrch('\\');
			str.numtostr(b, s::i, 8);
			for (j=str.length(b), 3) Out.wrch('0');
			Out.writes(b);
		end
		else ie (s::i = '"') do
			Out.writes("\\\"");
		end
		else ie (s::i = '\\') do
			Out.writes("\\\\");
		end
		else do
			Out.wrch(s::i);
		end
	end
	Out.writes("\\0");
	k := k+1;
	while (k & Lbpw-1) do
		Out.writes("\\0");
		k := k+1;
	end
	Out.writes(str.format(b, "\", %d);\n", [(k)]));
	dgrow(k/Lbpw);
end

addproto(id) do var p;
	if (Ptop >= MAXPROTO) error("too many functions");
	Prototbl[Ptop] := id;
	Ptop := Ptop + 1;
end

beginproc(label, pub) do var b::100, i, op, narg;
	op := rdcode();
	if (op = tcode.IHINT & tcode.IOPMASK) narg := rdarg(L1);
	if (Phase = 1) do
		addproto(label);
		return;
	end
	if (Phase = 0) return;
	Out.writes(str.format(b, "void P%d(void) {\n", [(label)]));
	Out.writes("--S;\n");
	Out.writes("push(F);\n");
	Out.writes("F = S;\n");
	if (pub) do
		Out.writes("push((C)I);\n");
		Out.writes("I = (C*)M[F+2];\n");
	end
end

generate(op, a1, a2, sbuf) do
	ie (op = tcode.ILDG & tcode.IOPMASK \/
		op = tcode.ILDGV & tcode.IOPMASK \/
		op = tcode.ILDL & tcode.IOPMASK \/
		op = tcode.ILDLV & tcode.IOPMASK \/
		op = tcode.ILDI & tcode.IOPMASK \/
		op = tcode.ILDIV & tcode.IOPMASK \/
		op = tcode.ILDLAB & tcode.IOPMASK \/
		op = tcode.INUM & tcode.IOPMASK \/
		op = tcode.ISELF & tcode.IOPMASK
	) do
		load(op, a1);
	end
	else ie (op = tcode.INEG & tcode.IOPMASK \/
		op = tcode.ILNOT & tcode.IOPMASK \/
		op = tcode.IBNOT & tcode.IOPMASK
	) do
		unary(op);
	end
	else ie (op = tcode.INORMB & tcode.IOPMASK \/
		op = tcode.INORM & tcode.IOPMASK \/
		op = tcode.IMUL & tcode.IOPMASK \/
		op = tcode.IDIV & tcode.IOPMASK \/
		op = tcode.IMOD & tcode.IOPMASK \/
		op = tcode.IADD & tcode.IOPMASK \/
		op = tcode.ISUB & tcode.IOPMASK \/
		op = tcode.IBAND & tcode.IOPMASK \/
		op = tcode.IBOR & tcode.IOPMASK \/
		op = tcode.IBXOR & tcode.IOPMASK \/
		op = tcode.IBSHL & tcode.IOPMASK \/
		op = tcode.IBSHR & tcode.IOPMASK \/
		op = tcode.IEQU & tcode.IOPMASK \/
		op = tcode.INEQU & tcode.IOPMASK \/
		op = tcode.ILESS & tcode.IOPMASK \/
		op = tcode.IGRTR & tcode.IOPMASK \/
		op = tcode.ILTEQ & tcode.IOPMASK \/
		op = tcode.IGTEQ & tcode.IOPMASK \/
		op = tcode.IUMUL & tcode.IOPMASK \/
		op = tcode.IUDIV & tcode.IOPMASK \/
		op = tcode.IULESS & tcode.IOPMASK \/
		op = tcode.IUGRTR & tcode.IOPMASK \/
		op = tcode.IULTEQ & tcode.IOPMASK \/
		op = tcode.IUGTEQ & tcode.IOPMASK \/
		op = tcode.IDREFB & tcode.IOPMASK \/
		op = tcode.IDEREF & tcode.IOPMASK
	) do
		binary(op);
	end
	else ie (op = tcode.ISAVG & tcode.IOPMASK \/
		op = tcode.ISAVL & tcode.IOPMASK \/
		op = tcode.ISAVI & tcode.IOPMASK \/
		op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	) do
		store(op, a1);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK \/
		op = tcode.ICALR & tcode.IOPMASK \/
		op = tcode.IICALL & tcode.IOPMASK
	) do
		pcall(op, a1);
	end
	else ie (op = tcode.IPOP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		pop();
	end
	else ie (op = tcode.ICLEAN & tcode.IOPMASK) do
		pclean(a1);
	end
	else ie (op = tcode.ISTACK & tcode.IOPMASK) do
		emitl(ADJUST_SEQ, -a1);
	end
	else ie (op = tcode.IINCG & tcode.IOPMASK) do
		emit(INC_GLOB, [(lookup(a1)),(a2)]);
	end
	else ie (op = tcode.IINCI & tcode.IOPMASK) do
		emiti2(INC_IVAR, a1, a2);
	end
	else ie (op = tcode.IINCL & tcode.IOPMASK) do
		emitl2(INC_LOC, -a1, a2);
	end
	else ie (op = tcode.IUNEXT & tcode.IOPMASK \/
		op = tcode.IDNEXT & tcode.IOPMASK
	) do
		if (Queued) spill(Iqueue, Aqueue);
		if (\Active \/ \Depth)
			internal("too few operands in GENERATE(?NEXT)");
		loop_regs(op, a1);
	end
	else ie (op = tcode.IBRF & tcode.IOPMASK \/
		op = tcode.IBRT & tcode.IOPMASK \/
		op = tcode.INBRF & tcode.IOPMASK \/
		op = tcode.INBRT & tcode.IOPMASK \/
		op = tcode.IJUMP & tcode.IOPMASK
	) do
		branch(op, a1);
	end
	else ie (op = tcode.IHDR & tcode.IOPMASK) do
		internal("HDR without CLAB");
	end
	else ie (op = tcode.IEND & tcode.IOPMASK) do
		emit(EPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END");
		Active := 0;
	end
	else ie (op = tcode.IMHDR & tcode.IOPMASK) do
		internal("MHDR without CLAB");
	end
	else ie (op = tcode.IENDM & tcode.IOPMASK) do
		emit(MEPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END");
		Active := 0;
	end
	else ie (op = tcode.IINIT & tcode.IOPMASK) do
		if (a1 \= 8) error("unsupported Tcode version");
	end
	else ie (op = tcode.IBPW & tcode.IOPMASK) do
		if (a1 \= Lbpw) error("unsupported machine word size");
	end
	else ie (op = tcode.ICLAB & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		op := rdcode();
		ie (op = tcode.IHDR & tcode.IOPMASK) do
			beginproc(a1, 0);
			Active := 0;
		end
		else ie (op = tcode.IMHDR & tcode.IOPMASK) do
			beginproc(a1, 1);
			Active := 0;
		end
		else ie (a1 = 0) do
			reject(op);
			if (Phase = 2) Out.writes("void _TXSTART_(void) {\ninit();\n");
			Active := 0;
		end
		else do
			reject(op);
			emit(DECL_CODE_LABEL, [(a1)]);
		end
	end
	else ie (op = tcode.IDLAB & tcode.IOPMASK) do
		addsym(a1);
	end
	else ie (op = tcode.IVEC & tcode.IOPMASK) do
		dgrow(a1=0-> 1: a1);
	end
	else ie (op = tcode.IDATA & tcode.IOPMASK \/
		op = tcode.IDREF & tcode.IOPMASK \/
		op = tcode.ICREF & tcode.IOPMASK
	) do
		ie (op = tcode.IDATA & tcode.IOPMASK) do
			emitd(DECL_WORD, [(Dtop, a1)]);
			dgrow(1);
		end
		else ie (op = tcode.IDREF & tcode.IOPMASK) do
			emitd(DECL_DREF, [(Dtop, lookup(a1))]);
			dgrow(1);
		end
		else do
			emitd(DECL_CREF, [(Dtop, a1)]);
			dgrow(1);
		end
	end
	else ie (op = tcode.ISTR & tcode.IOPMASK) do
		genstr(sbuf, a1);
	end
	else ie (op = tcode.IGLUE & tcode.IOPMASK) do
	end
	else ie (op = tcode.IHALT & tcode.IOPMASK) do
		emit(HALT_SEQ, [(a1)]);
	end
	else ie (op = tcode.IPUB & tcode.IOPMASK) do
		!ignore XXX
	end
	else ie (op = tcode.IEXT & tcode.IOPMASK) do
		!ignore XXX
	end
	else ie (op = tcode.IDUP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		push();
		Active := 1;
	end
	else ie (op = tcode.ISWAP & tcode.IOPMASK) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Depth < 1) internal("too few operands in GENERATE(SWAP)");
		emit(SWAP_SEQ, 0);
	end
	else ie (op = tcode.IIPROC & tcode.IOPMASK) do var n, m;
		op := rdcode();
		if (op \=tcode.ICMAP & tcode.IOPMASK) internal("no CMAP after IPROC");
		n := rdarg(L1);
		m := rdarg(L2);
		addinfproc(a1, a2, n, m, sbuf);
	end
	else ie (op = tcode.IHINT & tcode.IOPMASK \/
		op = tcode.ICMAP & tcode.IOPMASK \/
		op = tcode.ILIB & tcode.IOPMASK
	) do
		! ignore
	end
	else ie (op = tcode.ICALX & tcode.IOPMASK) do
		error("unresolved CALX (external reference)");
	end
	else ie (op = tcode.IICALX & tcode.IOPMASK) do
		error("unresolved ICALX (external interface)");
	end
	else do
		internal("unknown instruction in GENERATE()");
	end
end

xlate() do
	var	op, o2, a1, a2, i, j, ch, k;
	var	sbuf::256;

	reset();
	op := rdcode();
	if (op \= tcode.IINIT & tcode.IOPMASK) error("magic match failed");
	while (\In.eof()) do
		o2 := op & tcode.IOPMASK;
		if (o2 >= tcode.IENDOFSET) error("invalid opcode");
		if (Na) do
			a1 := rdarg(L1);
			k := a1;
			if (Na > 1) do
				a2 := rdarg(L2);
				k := a2;
			end
		end
		if (Sf) do
			for (i=0, k) sbuf::i := rdch();
			sbuf::i := 0;
		end
		generate(op, a1, a2, sbuf);
		op := rdcode();
	end
end

declprotos() do var i, b::100;
	for (i=0, Ptop)
		Out.writes(str.format(b, "void P%d(void);\n", [(Prototbl[i])]));
end

phase0() do
	Out.create(T3X.SYSOUT, Outbuf, BUFSIZE, IOSTREAM.FWRITE);
	if (Tmp.open(Tmpfile, Tmpbuf, BUFSIZE, IOSTREAM.FWRITE))
		error("failed to create work file");
	In.create(T3X.SYSIN, Inbuf, BUFSIZE, IOSTREAM.FREAD);
	Phase := 0;
	xlate();
	Tmp.close();
	In.flush();
	Dsize := Dtop;
end

phase1() do
	if (In.open(Tmpfile, Inbuf, BUFSIZE, IOSTREAM.FREAD))
		error("failed to re-open work file");
	Phase := 1;
	emitd(BANNER, 0);
	emitd(HEAD, [(Dsize + STACKSIZE)]);
	xlate();
	emitd(TAIL, 0);
	In.close();
end

phase2() do
	if (In.open(Tmpfile, Inbuf, BUFSIZE, IOSTREAM.FREAD))
		error("failed to re-open work file");
	Phase := 2;
	declinfprocs();
	declprotos();
	xlate();
	emit(TAIL, 0);
	In.close();
	Out.flush();
end

do var buf::4;
	init();
	if (t.getarg(1, buf, 4) = 2 /\ \t.memcomp(buf, "-v", 3))
		Verbose := 1;
	phase0();
	phase1();
	phase2();
	t.remove(Tmpfile);
	if (Verbose) do var b::100;
		t.write(T3X.SYSERR, str.format(b, "data=%UD%S",
			[(Dsize*Lbpw, t.newline(buf))]), str.length(b));
	end
end
