! TXNCG* -- New Tcode->Assembly translator
! Copyright (C) 2000,2001,2002,2003 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txncg(t3x, tcode, string, iostream, mdesc);

object	t[t3x], str[string];

object	md[mdesc];

const	BUFSIZE	=	1024;	! I/O buffer size
const	MAXINF =	1024;	! max. interface procedures
const	NAMEPOOLSIZE =	8192;	! interface name pool size

struct	SEGMENTS = NONE, TEXT, DATA, BSS;

struct	INFPROC =
	INFLABEL,
	INFNAME;

object	In[iostream], Out[iostream];
var	R_op, R_fl;		! Rejection queue
var	Inbuf::BUFSIZE;
var	Outbuf::BUFSIZE;
var	Tmpfile::256;
var	Segment;
var	Addr;
var	Tlabel;
var	Iqueue, Aqueue,		! 1-instruction queue (instr, arg)
	Queued,			! queue active indicator (1=active),
	Active,			! accumulator active flag (1=active),
	Depth;			! RT stack depth
var	Dataname;
var	Tprefix, Aprefix;
var	DoNL;			! emit newlines
var	Infprocs[MAXINF*INFPROC],	! interface procedure table
	Infmax;				! # of interfaces
var	Names::NAMEPOOLSIZE, Ntop;	! Name pool, top of pool
var	Scaling_limit;
var	Fl, Na, Sf, L1, L2;		! Tcode flags
var	Lbpw;				! t.bpw() cache
var	XFFFFFFFF;			! Constant 0xFFFFFFFF

const	BPW = MDESC.BPW;
const	LONGJUMP = MDESC.LONGJUMP;
const	FLAGS_REGISTER = MDESC.FLAGS_REGISTER;
const	REGISTER_ARGS = MDESC.REGISTER_ARGUMENTS;
const	RUNTIME_SCALING = MDESC.RUNTIME_SCALING;
const	MULTIPLE_DATA  = MDESC.MULTIPLE_DATA;

! variables to bind code sequences
var	BANNER,
	ARGUMENT_REGISTERS,
	FETCH_ARG_SEQ,
	HEAD,
	TAIL,
	SELECT_TEXT,
	SELECT_DATA,
	SELECT_BSS,
	DECL_WORD,
	DECL_OCTET,
	DECL_DATA_LABEL,
	DECL_DATA_ALIAS,
	DECL_VECTOR,
	DECL_REF,
	DECL_CODE_LABEL,
	DECL_TMP_LABEL,
	DECL_INFPROC,
	LOAD_GLOB,
	LOAD_LOC,
	LOAD_IVAR,
	LOAD_GLOB_VEC,
	LOAD_LOC_VEC,
	LOAD_IVAR_VEC,
	LOAD_LABEL,
	LOAD_CONST,
	LOAD_SELF,
	PUSH_SEQ,
	POP_SEQ,
	DEREF_SEQ,
	DEREF_BYTE_SEQ,
	STORE_GLOB,
	STORE_LOC,
	STORE_IVAR,
	STORE_SEQ,
	STORE_BYTE,
	NORM_SEQ,
	NORM_BYTE_SEQ,
	INC_GLOB,
	INC_LOC,
	INC_IVAR,
	NEG_SEQ,
	NOT_SEQ,
	LNOT_SEQ,
	EXT_BOOL_SEQ,
	CMP_ZERO_SEQ,
	SAVE_CMP_RESULT,
	MUL_SEQ,
	DIV_SEQ,
	MOD_SEQ,
	UMUL_SEQ,
	UDIV_SEQ,
	ADD_SEQ,
	SUB_SEQ,
	AND_SEQ,
	OR_SEQ,
	XOR_SEQ,
	SHL_SEQ,
	SHR_SEQ,
	CMP_FLAGS_SEQ,
	CMP_EQ_SEQ,
	CMP_NE_SEQ,
	CMP_LT_SEQ,
	CMP_GT_SEQ,
	CMP_LE_SEQ,
	CMP_GE_SEQ,
	CMP_ULT_SEQ,
	CMP_UGT_SEQ,
	CMP_ULE_SEQ,
	CMP_UGE_SEQ,
	SWAP_SEQ,
	BE_SEQ,
	BNE_SEQ,
	BL_SEQ,
	BG_SEQ,
	BLE_SEQ,
	BGE_SEQ,
	BUL_SEQ,
	BUG_SEQ,
	BULE_SEQ,
	BUGE_SEQ,
	JMP_SEQ,
	JMPS_SEQ,
	TRUE_SEQ,
	FALSE_SEQ,
	PROLOGUE_SEQ,
	EPILOGUE_SEQ,
	MPROLOGUE_SEQ,
	MEPILOGUE_SEQ,
	CALL_REG,
	CALL_SEQ,
	CALL_INF,
	ADJUST_SEQ,
	CLEANUP_SEQ,
	HALT_SEQ;

! variables to bind addressing modes
struct	MODES =
	MODE_GLOB,	! acc/memory
	MODE_LOC,	! acc/(local)
	MODE_IVAR,	! acc/(instance)
	MODE_CONST,	! acc/immediate
	MODE_SELF,	! acc/(self)
	MODE_STACK,	! acc/(popped value)
	MODE_LOC_R,	! acc/(local), runtime scaled
	MODE_IVAR_R;	! acc/(instance), runtime scaled

mkopcodes() do
	BANNER := md.banner();
	ARGUMENT_REGISTERS := md.argument_registers();
	FETCH_ARG_SEQ := md.fetch_arg_seq();
	HEAD := md.head();
	TAIL := md.tail();
	SELECT_TEXT := md.select_text();
	SELECT_DATA := md.select_data();
	SELECT_BSS := md.select_bss();
	DECL_WORD := md.decl_word();
	DECL_OCTET := md.decl_octet();
	DECL_DATA_LABEL := md.decl_data_label();
	DECL_DATA_ALIAS := md.decl_data_alias();
	DECL_REF := md.decl_ref();
	DECL_VECTOR := md.decl_vector();
	DECL_CODE_LABEL := md.decl_code_label();
	DECL_TMP_LABEL := md.decl_tmp_label();
	DECL_INFPROC := md.decl_infproc();
	LOAD_GLOB := md.load_glob();
	LOAD_LOC := md.load_loc();
	LOAD_IVAR := md.load_ivar();
	LOAD_GLOB_VEC := md.load_glob_vec();
	LOAD_LOC_VEC := md.load_loc_vec();
	LOAD_IVAR_VEC := md.load_ivar_vec();
	LOAD_LABEL := md.load_label();
	LOAD_CONST := md.load_const();
	LOAD_SELF := md.load_self();
	PUSH_SEQ := md.push_seq();
	POP_SEQ := md.pop_seq();
	DEREF_SEQ := md.deref_seq();
	DEREF_BYTE_SEQ := md.deref_byte_seq();
	STORE_GLOB := md.store_glob();
	STORE_LOC := md.store_loc();
	STORE_IVAR := md.store_ivar();
	STORE_SEQ := md.store_seq();
	STORE_BYTE := md.store_byte();
	NORM_SEQ := md.norm_seq();
	NORM_BYTE_SEQ := md.norm_byte_seq();
	INC_GLOB := md.inc_glob();
	INC_LOC := md.inc_loc();
	INC_IVAR := md.inc_ivar();
	NEG_SEQ := md.neg_seq();
	NOT_SEQ := md.not_seq();
	LNOT_SEQ := md.lnot_seq();
	EXT_BOOL_SEQ := md.ext_bool_seq();
	CMP_ZERO_SEQ := md.cmp_zero_seq();
	SAVE_CMP_RESULT := md.save_cmp_result();
	MUL_SEQ := md.mul_seq();
	DIV_SEQ := md.div_seq();
	MOD_SEQ := md.mod_seq();
	UMUL_SEQ := md.umul_seq();
	UDIV_SEQ := md.udiv_seq();
	ADD_SEQ := md.add_seq();
	SUB_SEQ := md.sub_seq();
	AND_SEQ := md.and_seq();
	OR_SEQ := md.or_seq();
	XOR_SEQ := md.xor_seq();
	SHL_SEQ := md.shl_seq();
	SHR_SEQ := md.shr_seq();
	CMP_FLAGS_SEQ := md.cmp_flags_seq();
	CMP_EQ_SEQ := md.cmp_eq_seq();
	CMP_NE_SEQ := md.cmp_ne_seq();
	CMP_GT_SEQ := md.cmp_gt_seq();
	CMP_LT_SEQ := md.cmp_lt_seq();
	CMP_GE_SEQ := md.cmp_ge_seq();
	CMP_LE_SEQ := md.cmp_le_seq();
	CMP_UGT_SEQ := md.cmp_ugt_seq();
	CMP_ULT_SEQ := md.cmp_ult_seq();
	CMP_UGE_SEQ := md.cmp_uge_seq();
	CMP_ULE_SEQ := md.cmp_ule_seq();
	SWAP_SEQ := md.swap_seq();
	BE_SEQ := md.be_seq();
	BNE_SEQ := md.bne_seq();
	BL_SEQ := md.bl_seq();
	BG_SEQ := md.bg_seq();
	BLE_SEQ := md.ble_seq();
	BGE_SEQ := md.bge_seq();
	BUL_SEQ := md.bul_seq();
	BUG_SEQ := md.bug_seq();
	BULE_SEQ := md.bule_seq();
	BUGE_SEQ := md.buge_seq();
	JMP_SEQ := md.jmp_seq();
	JMPS_SEQ := md.jmps_seq();
	TRUE_SEQ := md.true_seq();
	FALSE_SEQ := md.false_seq();
	PROLOGUE_SEQ := md.prologue_seq();
	EPILOGUE_SEQ := md.epilogue_seq();
	MPROLOGUE_SEQ := md.mprologue_seq();
	MEPILOGUE_SEQ := md.mepilogue_seq();
	CALL_REG := md.call_reg();
	CALL_SEQ := md.call_seq();
	CALL_INF := md.call_inf();
	ADJUST_SEQ := md.adjust_seq();
	CLEANUP_SEQ := md.cleanup_seq();
	HALT_SEQ := md.halt_seq();
end

init() do var i;
	Lbpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	R_op := %1;
	Addr := 0;
	Segment := NONE;
	Tlabel := 0;
	Active := 0;
	Queued := 0;
	Depth := 0;
	Dataname := 0;
	Aprefix := "A";
	Tprefix := "T";
	DoNL := 1;
	Infmax := 0;
	Ntop := 0;
	Scaling_limit := RUNTIME_SCALING-1;
	mkopcodes();
end

error(msg) do var buf::128;
	t.write(T3X.SYSERR, "TXNCG: ", 7);
	t.write(T3X.SYSERR, str.numtostr(buf, Addr, 10), str.length(buf));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	halt 1;
end

internal(msg) do
	t.write(T3X.SYSERR, "internal error: ", 16);
	error(msg);
end

rdch() do var ch;
	ch := In.rdch();
	Addr := Addr+1;
	return ch;
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Lbpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Lbpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdarg(k) do var n, m;
	n := rdch();
	if (k = 0) return sext8(n);
	n := n << 8 | rdch();
	if (k = 1) return sext16(n);
	m := n;
	n := n << 8 | rdch();
	n := n << 8 | rdch();
	if (Lbpw < 4 /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		error("value out of range");
	return sext32(n);
end

rdcode() do var op;
	ie (R_op >= 0) do
		op := R_op;
		Fl := R_fl;
		R_op := %1;
	end
	else do
		op := rdch();
		ie (op & tcode.IARFLAG) do
			Fl := op;
			op := rdch();
		end
		else do
			Fl := 0;
		end
	end
	Na := Fl & tcode.IARMASK >> tcode.IAMOFF;
	Sf := Fl & tcode.ISTFLAG;
	if (Na) do
		L1 := Fl & tcode.IL1MASK >> tcode.IL1OFF;
		if  (Na > 1)
			L2 := Fl & tcode.IL2MASK >> tcode.IL2OFF;
	end
	return op;
end

reject(op) do
	R_op := op;
	R_fl := Fl;
end

wrch(ch) Out.wrch(ch);

! emit a code sequence with arguments merged in
_emit(vcode, args) do var i, j, c, b::33, code, radix;
	j := 0;
	while (vcode[j]) do
		code := vcode[j]; j := j+1;
		i := 0;
		while (1) do
			c := code::i; i := i+1;
			if (\c) leave;
			ie (c = '%') do
				radix := -10;
				c := code::i; i := i+1;
				if (c = '+') do
					radix := 10;
					c := code::i; i := i+1;
				end
				ie ('1' <= c /\ c <= '9')
					Out.writes(str.numtostr(b,
						args[c-'1'], radix));
				else ie ('A' <= c /\ c <= 'Z')
					Out.writes(args[c-'A']);
				else ie (\c)
					i := i-1;
				else
					Out.wrch(c);
			end
			else do
				Out.wrch(c);
			end
		end
		if (DoNL) Out.wrch('\n');
	end
end

emit(code, args) _emit(code[0], args);

newlabel() do
	Tlabel := Tlabel +1;
	return Tlabel;
end

select(seg) do
	if (seg = Segment) return;
	Segment := seg;
	ie (Segment = TEXT) emit(SELECT_TEXT, 0);
	else ie (Segment = DATA) emit(SELECT_DATA, 0);
	else ie (Segment = BSS) emit(SELECT_BSS, 0);
	else internal("bad segment in SELECT()");
end

push() do
	if (\Active) internal("accumulator not active in PUSH()");
	emit(PUSH_SEQ, 0);
	Depth := Depth+1;
	Active := 0;
end

pop() do
	if (\Depth) do
		if (Active) do
			Active := 0;
			return;
		end
		internal("stack underflow in POP()");
	end
	emit(POP_SEQ, 0);
	Depth := Depth-1;
	Active := 1;
end

emitlr(seq, arg) do
	if (RUNTIME_SCALING)
		if (arg < -Scaling_limit \/ arg > Scaling_limit)
			return _emit(seq[1], [(-arg)]);
	emit(seq, [(-arg*BPW)]);
end

emitir(seq, arg) do
	if (RUNTIME_SCALING)
		if (arg > Scaling_limit)
			return _emit(seq[1], [(arg)]);
	emit(seq, [(arg*BPW)]);
end

emitlr2(seq, arg, a2) do
	if (RUNTIME_SCALING)
		if (arg < -Scaling_limit \/ arg > Scaling_limit)
			return _emit(seq[1], [(-arg), (a2)]);
	emit(seq, [(-arg*BPW), (a2)]);
end

emitir2(seq, arg, a2) do
	if (RUNTIME_SCALING)
		if (arg > Scaling_limit)
			return _emit(seq[1], [(arg), (a2)]);
	emit(seq, [(arg*BPW), (a2)]);
end

! load queued value into accumulator
spill(m, a) do
	select(TEXT);
	if (Active) push();
	ie (m = tcode.ILDG & tcode.IOPMASK) emit(LOAD_GLOB, [(a)]);
	else ie (m = tcode.ILDGV & tcode.IOPMASK) emit(LOAD_GLOB_VEC, [(a)]);
	else ie (m = tcode.ILDL & tcode.IOPMASK) emitlr(LOAD_LOC, a);
	else ie (m = tcode.ILDLV & tcode.IOPMASK) emitlr(LOAD_LOC_VEC, a);
	else ie (m = tcode.ILDI & tcode.IOPMASK) emitir(LOAD_IVAR, a);
	else ie (m = tcode.ILDIV & tcode.IOPMASK) emitir(LOAD_IVAR_VEC, a);
	else ie (m = tcode.ILDLAB & tcode.IOPMASK) emit(LOAD_LABEL, [(a)]);
	else ie (m = tcode.INUM & tcode.IOPMASK) emit(LOAD_CONST, [(a)]);
	else ie (m = tcode.ISELF & tcode.IOPMASK) emit(LOAD_SELF, 0);
	else internal("unknown instruction in LOAD()");
	Active := 1;
	Queued := 0;
end

! generate load instruction
load(m, a) do
	ie (\Queued) do
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
		return;
	end
	else do
		spill(Iqueue, Aqueue);
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
	end
end

unary(op) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("accumulator not active in UNARY()");
		pop();
	end
	ie (op = tcode.INEG & tcode.IOPMASK) emit(NEG_SEQ, 0);
	else ie (op = tcode.IBNOT & tcode.IOPMASK) emit(NOT_SEQ, 0);
	else ie (op = tcode.ILNOT & tcode.IOPMASK) emit(LNOT_SEQ, 0);
	else internal("unknown instruction in UNARY()");
end

emitb(seq, mode) do
	if (RUNTIME_SCALING) do
		if (	mode = MODE_LOC /\
			(Aqueue > Scaling_limit \/ Aqueue < -Scaling_limit)
		)
			return _emit(seq[MODE_LOC_R], [(-Aqueue)]);
		if (mode = MODE_IVAR /\ Aqueue > Scaling_limit)
			return _emit(seq[MODE_IVAR_R], [(Aqueue)]);
	end
	ie (mode = MODE_LOC)
		Aqueue := -Aqueue*BPW;
	else if (mode = MODE_IVAR)
		Aqueue := Aqueue*BPW;
	_emit(seq[mode], [(Aqueue)]);
end

cmp_flags(op, mode) do var op2, rel, flg, lab, t1;
	emitb(CMP_FLAGS_SEQ, mode);
	op2 := rdcode();
	ie (	op2 = tcode.IBRT & tcode.IOPMASK \/
		op2 = tcode.IBRF & tcode.IOPMASK
	) do
		pop();
		flg := op2 = tcode.IBRF & tcode.IOPMASK;
		if (LONGJUMP) flg := \flg;
		lab := rdarg(L1);
		ie (op = tcode.IEQU & tcode.IOPMASK)
			rel := flg-> BE_SEQ: BNE_SEQ;
		else ie (op = tcode.INEQU & tcode.IOPMASK)
			rel := flg-> BNE_SEQ: BE_SEQ;
		else ie (op = tcode.ILESS & tcode.IOPMASK)
			rel := flg-> BL_SEQ: BGE_SEQ;
		else ie (op = tcode.IGRTR & tcode.IOPMASK)
			rel := flg-> BG_SEQ: BLE_SEQ;
		else ie (op = tcode.ILTEQ & tcode.IOPMASK)
			rel := flg-> BLE_SEQ: BG_SEQ;
		else ie (op = tcode.IGTEQ & tcode.IOPMASK)
			rel := flg-> BGE_SEQ: BL_SEQ;
		else ie (op = tcode.IULESS & tcode.IOPMASK)
			rel := flg-> BUL_SEQ: BUGE_SEQ;
		else ie (op = tcode.IUGRTR & tcode.IOPMASK)
			rel := flg-> BUG_SEQ: BULE_SEQ;
		else ie (op = tcode.IULTEQ & tcode.IOPMASK)
			rel := flg-> BULE_SEQ: BUG_SEQ;
		else ie (op = tcode.IUGTEQ & tcode.IOPMASK)
			rel := flg-> BUGE_SEQ: BUL_SEQ;
		else internal("unknown relation in BINARY()");
		ie (LONGJUMP) do
			emit(rel, [(Aprefix), (lab)]);
		end
		else do
			newlabel();
			emit(rel, [(Tprefix), (Tlabel)]);
			emit(JMP_SEQ, [(Aprefix), (lab)]);
			emit(DECL_TMP_LABEL, [(Tlabel)]);
		end
	end
	else do
		reject(op2);
		ie (op = tcode.IEQU & tcode.IOPMASK) rel := BNE_SEQ;
		else ie (op = tcode.INEQU & tcode.IOPMASK) rel := BE_SEQ;
		else ie (op = tcode.ILESS & tcode.IOPMASK) rel := BGE_SEQ;
		else ie (op = tcode.IGRTR & tcode.IOPMASK) rel := BLE_SEQ;
		else ie (op = tcode.ILTEQ & tcode.IOPMASK) rel := BG_SEQ;
		else ie (op = tcode.IGTEQ & tcode.IOPMASK) rel := BL_SEQ;
		else ie (op = tcode.IULESS & tcode.IOPMASK) rel := BUGE_SEQ;
		else ie (op = tcode.IUGRTR & tcode.IOPMASK) rel := BULE_SEQ;
		else ie (op = tcode.IULTEQ & tcode.IOPMASK) rel := BUG_SEQ;
		else ie (op = tcode.IUGTEQ & tcode.IOPMASK) rel := BUL_SEQ;
		else internal("unknown relation in BINARY()");
		t1 := newlabel();
		newlabel();
		emit(rel, [(Tprefix), (t1)]);
		emit(TRUE_SEQ, 0);
		emit(JMPS_SEQ, [(Tprefix), (Tlabel)]);
		emit(DECL_TMP_LABEL, [(t1)]);
		emit(FALSE_SEQ, 0);
		emit(DECL_TMP_LABEL, [(Tlabel)]);
	end
end

cmp_regs(op, mode) do var op2, rel, flg, lab, t1;
	ie (op = tcode.IEQU) rel := CMP_EQ_SEQ;
	else ie (op = tcode.INEQU & tcode.IOPMASK) rel := CMP_NE_SEQ;
	else ie (op = tcode.ILESS & tcode.IOPMASK) rel := CMP_LT_SEQ;
	else ie (op = tcode.IGRTR & tcode.IOPMASK) rel := CMP_GT_SEQ;
	else ie (op = tcode.ILTEQ & tcode.IOPMASK) rel := CMP_LE_SEQ;
	else ie (op = tcode.IGTEQ & tcode.IOPMASK) rel := CMP_GE_SEQ;
	else ie (op = tcode.IULESS & tcode.IOPMASK) rel := CMP_ULT_SEQ;
	else ie (op = tcode.IUGRTR & tcode.IOPMASK) rel := CMP_UGT_SEQ;
	else ie (op = tcode.IULTEQ & tcode.IOPMASK) rel := CMP_ULE_SEQ;
	else ie (op = tcode.IUGTEQ & tcode.IOPMASK) rel := CMP_UGE_SEQ;
	else internal("unknown relation in BINARY()");
	emitb(rel, mode);
	op2 := rdcode();
	ie (	op2 = tcode.IBRT & tcode.IOPMASK \/
		op2 = tcode.IBRF & tcode.IOPMASK
	) do
		pop();
		lab := rdarg(L1);
		ie (LONGJUMP) do
			emit(op2 = tcode.IBRF & tcode.IOPMASK-> BE_SEQ: BNE_SEQ,
				[(Aprefix), (lab)]);
		end
		else do
			newlabel();
			emit(op2 = tcode.IBRF & tcode.IOPMASK-> BNE_SEQ: BE_SEQ,
				[(Tprefix), (Tlabel)]);
			emit(JMP_SEQ, [(Aprefix), (lab)]);
			emit(DECL_TMP_LABEL, [(Tlabel)]);
		end
	end
	else do
		reject(op2);
		emit(EXT_BOOL_SEQ, 0);
	end
end

wm(s) do var nl::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
end

undefd_addr() do
	wm("TXNCG: Addresses of local variables and anonymous vectors");
	wm("       are not supported in binary operations.");
	wm("       This may be considered a bug of TXNCG.");
	wm("       See the documentation for details.");
	internal("unsupported addressing mode in binary()");
end

binary(op) do var mode;
	if (\Active /\ Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in BINARY()");
		pop();
	end
	ie (Queued) do
		ie (Iqueue = tcode.ILDG & tcode.IOPMASK)
			mode := MODE_GLOB;
		else ie (Iqueue = tcode.ILDGV & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ILDL & tcode.IOPMASK)
			mode := MODE_LOC;
		else ie (Iqueue = tcode.ILDLV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDI & tcode.IOPMASK)
			mode := MODE_IVAR;
		else ie (Iqueue = tcode.ILDIV & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.ILDLAB & tcode.IOPMASK)
			undefd_addr();
		else ie (Iqueue = tcode.INUM & tcode.IOPMASK)
			mode := MODE_CONST;
		else ie (Iqueue = tcode.ISELF & tcode.IOPMASK)
			mode := MODE_SELF;
		else
			internal("bad mode in BINARY()");
	end
	else do
		if (\Depth) internal("stack underflow in BINARY()");
		mode := MODE_STACK;
		Depth := Depth-1;
	end
	Queued := 0;
	Active := 1;
	! convert Tcode-style local/ivar addresses into real addresses
	ie (op = tcode.INORMB & tcode.IOPMASK) emitb(NORM_BYTE_SEQ, mode);
	else ie (op = tcode.INORM & tcode.IOPMASK) emitb(NORM_SEQ, mode);
	else ie (op = tcode.IDEREF & tcode.IOPMASK) emitb(DEREF_SEQ, mode);
	else ie (op = tcode.IDREFB & tcode.IOPMASK) emitb(DEREF_BYTE_SEQ,mode);
	else ie (op = tcode.IMUL & tcode.IOPMASK) emitb(MUL_SEQ, mode);
	else ie (op = tcode.IDIV & tcode.IOPMASK) emitb(DIV_SEQ, mode);
	else ie (op = tcode.IMOD & tcode.IOPMASK) emitb(MOD_SEQ, mode);
	else ie (op = tcode.IUMUL & tcode.IOPMASK) emitb(UMUL_SEQ, mode);
	else ie (op = tcode.IUDIV & tcode.IOPMASK) emitb(UDIV_SEQ, mode);
	else ie (op = tcode.IADD & tcode.IOPMASK) emitb(ADD_SEQ, mode);
	else ie (op = tcode.ISUB & tcode.IOPMASK) emitb(SUB_SEQ, mode);
	else ie (op = tcode.IBAND & tcode.IOPMASK) emitb(AND_SEQ, mode);
	else ie (op = tcode.IBOR & tcode.IOPMASK) emitb(OR_SEQ, mode);
	else ie (op = tcode.IBXOR & tcode.IOPMASK) emitb(XOR_SEQ, mode);
	else ie (op = tcode.IBSHL & tcode.IOPMASK) emitb(SHL_SEQ, mode);
	else ie (op = tcode.IBSHR & tcode.IOPMASK) emitb(SHR_SEQ, mode);
	else do
		ie (FLAGS_REGISTER)
			cmp_flags(op, mode);
		else
			cmp_regs(op, mode);
	end
end

store(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in STORE()");
		pop();
	end
	ie (op = tcode.ISAVG & tcode.IOPMASK) emit(STORE_GLOB, [(arg)]);
	else ie (op = tcode.ISAVL & tcode.IOPMASK) emitlr(STORE_LOC, arg);
	else ie (op = tcode.ISAVI & tcode.IOPMASK) emitir(STORE_IVAR, arg);
	else ie (op = tcode.ISTORB & tcode.IOPMASK) emit(STORE_BYTE, 0);
	else ie (op = tcode.ISTORE & tcode.IOPMASK) emit(STORE_SEQ, 0);
	else internal("bad mode in STORE()");
	if (	op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	)
		Depth := Depth-1;
	pop();
	! Leftover values are OK when computing dynamic tables
	! if (Depth) internal("leftover values in STORE()");
end

findinfproc(id) do var y, i;
	for (i=0, Infmax, INFPROC) do
		y := @Infprocs[i];
		if (y[INFLABEL] = id) return y[INFNAME];
	end
	error("undefined interface procedure");
end

addinfproc(lab, len, name) do var inf;
	select(TEXT);
	emit(DECL_INFPROC, [(name)]);
	if (	len + Ntop + 1 >= NAMEPOOLSIZE \/
		Infmax + INFPROC >= MAXINF
	)
		error("too many interface procedures");
	inf := @Infprocs[Infmax];
	Infmax := Infmax + INFPROC;
	inf[INFLABEL] := lab;
	inf[INFNAME] := @Names::Ntop;
	Ntop := Ntop + len+1;
	t.memcopy(inf[INFNAME], name, len+1);
end

pclean(arg) do
	Depth := Depth-arg;
	if (arg) emitir(CLEANUP_SEQ, arg);
end

fixargs() do var op, narg, i, nra;
	op := rdcode();
	if (op \= tcode.ICLEAN & tcode.IOPMASK)
		internal("CLEAN required in fixargs()");
	narg := rdarg(L1);
	if (\narg) return 0;
	nra := narg > REGISTER_ARGS-> REGISTER_ARGS: narg;
	for (i=0, nra) do
		emit(FETCH_ARG_SEQ, [
			(ARGUMENT_REGISTERS[i]),
			(i*BPW)]);
	end
	pclean(nra);
	return narg;
end

pcall(op, arg) do var narg;
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.ICALR & tcode.IOPMASK) do
		if (\Active) do
			if (\Depth)
				internal("no active reg in PCALL(CALR)");
			pop();
		end
		emit(CALL_REG, 0);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK) do
		if (Active) push();
		emit(CALL_SEQ, [(arg)]);
	end
	else do	! op = tcode.IICALL & tcode.IOPMASK
		if (Active) push();
		ie (REGISTER_ARGS) do
			narg := fixargs();
			emit(CALL_INF, [(findinfproc(arg))]);
			if (narg > REGISTER_ARGS)
				pclean(narg-REGISTER_ARGS);
		end
		else do
			emit(CALL_INF, [(findinfproc(arg))]);
		end
	end
	Active := 1;
end

branch(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.IJUMP & tcode.IOPMASK) do
		emit(JMP_SEQ, [(Aprefix), (arg)]);
		if (Depth \/ Active) pop();
	end
	else do
		if (\Active) do
			if (\Depth) internal("too few operands in BRANCH()");
			pop();
		end
		emit(FLAGS_REGISTER-> CMP_ZERO_SEQ: SAVE_CMP_RESULT, 0);
		if (	op = tcode.IBRT & tcode.IOPMASK \/
			op = tcode.IBRF & tcode.IOPMASK
		)
			pop();
		ie (LONGJUMP) do
			emit(op=tcode.IBRT & tcode.IOPMASK \/
				op=tcode.INBRT & tcode.IOPMASK-> BNE_SEQ:
				BE_SEQ, [(Aprefix), (arg)]);
		end
		else do
			newlabel();
			emit(op=tcode.IBRT & tcode.IOPMASK \/
				op=tcode.INBRT & tcode.IOPMASK-> BE_SEQ:
				BNE_SEQ, [(Tprefix), (Tlabel)]);
			emit(JMP_SEQ, [(Aprefix), (arg)]);
			emit(DECL_TMP_LABEL, [(Tlabel)]);
		end
	end
end

loop_flags(op, a1) do
	_emit(CMP_FLAGS_SEQ[MODE_STACK], 0);
	Depth := Depth-1;
	ie (LONGJUMP) do
		emit(op=tcode.IUNEXT & tcode.IOPMASK-> BGE_SEQ: BLE_SEQ,
			[(Aprefix), (a1)]);
	end
	else do
		newlabel();
		emit(op=tcode.IUNEXT & tcode.IOPMASK-> BL_SEQ: BG_SEQ,
			[(Tprefix), (Tlabel)]);
		emit(JMP_SEQ, [(Aprefix), (a1)]);
		emit(DECL_TMP_LABEL, [(Tlabel)]);
	end
	pop();
end

loop_regs(op, a1) do
	ie (LONGJUMP) do
		_emit(op=tcode.IUNEXT & tcode.IOPMASK-> CMP_GE_SEQ[MODE_STACK]:
			CMP_LE_SEQ[MODE_STACK], 0);
		emit(BNE_SEQ, [(Aprefix), (a1)]);
		Depth := Depth-1;
	end
	else do
		newlabel();
		_emit(op=tcode.IUNEXT & tcode.IOPMASK-> CMP_LT_SEQ[MODE_STACK]:
			CMP_GT_SEQ[MODE_STACK], 0);
		Depth := Depth-1;
		emit(BE_SEQ, [(Tprefix), (Tlabel)]);
		emit(JMP_SEQ, [(Aprefix), (a1)]);
		emit(DECL_TMP_LABEL, [(Tlabel)]);
	end
	pop();
end

genmdstr(s, k) do var i;
	DoNL := 0;
	for (i=0, k) do
		ie (\i \/ \(i mod 16)) do
			if (i) Out.wrch('\n');
			emit(DECL_OCTET, [(s::i)]);
		end
		else do
			_emit([",%1",0], [(s::i)]);
		end
	end
	Out.wrch('\n');
	emit(DECL_OCTET, [(0)]);
	k := k+1;
	while (k & BPW-1) do
		_emit([",%1",0], [(0)]);
		k := k+1;
	end
	Out.wrch('\n');
	DoNL := 1;
end

genstr(s, k) do var i;
	if (MULTIPLE_DATA) return genmdstr(s, k);
	for (i=0, k) emit(DECL_OCTET, [(s::i)]);
	emit(DECL_OCTET, [(0)]);
	k := k+1;
	while (k & BPW-1) do
		emit(DECL_OCTET, [(0)]);
		k := k+1;
	end
end

generate(op, a1, a2, sbuf) do
	ie (	op = tcode.ILDG & tcode.IOPMASK \/
		op = tcode.ILDGV & tcode.IOPMASK \/
		op = tcode.ILDL & tcode.IOPMASK \/
		op = tcode.ILDLV & tcode.IOPMASK \/
		op = tcode.ILDI & tcode.IOPMASK \/
		op = tcode.ILDIV & tcode.IOPMASK \/
		op = tcode.ILDLAB & tcode.IOPMASK \/
		op = tcode.INUM & tcode.IOPMASK \/
		op = tcode.ISELF & tcode.IOPMASK
	) do
		load(op, a1);
	end
	else ie (op = tcode.INEG & tcode.IOPMASK \/
		op = tcode.ILNOT & tcode.IOPMASK \/
		op = tcode.IBNOT & tcode.IOPMASK
	) do
		select(TEXT);
		unary(op);
	end
	else ie (op = tcode.INORMB & tcode.IOPMASK \/
		op = tcode.INORM & tcode.IOPMASK \/
		op = tcode.IMUL & tcode.IOPMASK \/
		op = tcode.IDIV & tcode.IOPMASK \/
		op = tcode.IMOD & tcode.IOPMASK \/
		op = tcode.IADD & tcode.IOPMASK \/
		op = tcode.ISUB & tcode.IOPMASK \/
		op = tcode.IBAND & tcode.IOPMASK \/
		op = tcode.IBOR & tcode.IOPMASK \/
		op = tcode.IBXOR & tcode.IOPMASK \/
		op = tcode.IBSHL & tcode.IOPMASK \/
		op = tcode.IBSHR & tcode.IOPMASK \/
		op = tcode.IEQU & tcode.IOPMASK \/
		op = tcode.INEQU & tcode.IOPMASK \/
		op = tcode.ILESS & tcode.IOPMASK \/
		op = tcode.IGRTR & tcode.IOPMASK \/
		op = tcode.ILTEQ & tcode.IOPMASK \/
		op = tcode.IGTEQ & tcode.IOPMASK \/
		op = tcode.IUMUL & tcode.IOPMASK \/
		op = tcode.IUDIV & tcode.IOPMASK \/
		op = tcode.IULESS & tcode.IOPMASK \/
		op = tcode.IUGRTR & tcode.IOPMASK \/
		op = tcode.IULTEQ & tcode.IOPMASK \/
		op = tcode.IUGTEQ & tcode.IOPMASK \/
		op = tcode.IDREFB & tcode.IOPMASK \/
		op = tcode.IDEREF & tcode.IOPMASK
	) do
		select(TEXT);
		binary(op);
	end
	else ie (op = tcode.ISAVG & tcode.IOPMASK \/
		op = tcode.ISAVL & tcode.IOPMASK \/
		op = tcode.ISAVI & tcode.IOPMASK \/
		op = tcode.ISTORE & tcode.IOPMASK \/
		op = tcode.ISTORB & tcode.IOPMASK
	) do
		select(TEXT);
		store(op, a1);
	end
	else ie (op = tcode.ICALL & tcode.IOPMASK \/
		op = tcode.ICALR & tcode.IOPMASK \/
		op = tcode.IICALL & tcode.IOPMASK
	) do
		select(TEXT);
		pcall(op, a1);
	end
	else ie (op = tcode.IPOP & tcode.IOPMASK) do
		select(TEXT);
		if (Queued) spill(Iqueue, Aqueue);
		pop();
	end
	else ie (op = tcode.ICLEAN & tcode.IOPMASK) do
		select(TEXT);
		pclean(a1);
	end
	else ie (op = tcode.ISTACK & tcode.IOPMASK) do
		select(TEXT);
		emitlr(ADJUST_SEQ, -a1);
	end
	else ie (op = tcode.IINCG & tcode.IOPMASK) do
		select(TEXT);
		emit(INC_GLOB, [(a1),(a2)]);
	end
	else ie (op = tcode.IINCI & tcode.IOPMASK) do
		select(TEXT);
		emitir2(INC_IVAR, a1, a2);
	end
	else ie (op = tcode.IINCL & tcode.IOPMASK) do
		select(TEXT);
		emitlr2(INC_LOC, a1, a2);
	end
	else ie (op = tcode.IUNEXT & tcode.IOPMASK \/
		op = tcode.IDNEXT & tcode.IOPMASK
	) do
		select(TEXT);
		if (Queued) spill(Iqueue, Aqueue);
		if (\Active \/ \Depth)
			internal("too few operands in GENERATE(?NEXT)");
		ie (FLAGS_REGISTER)
			loop_flags(op, a1);
		else
			loop_regs(op, a1);
	end
	else ie (op = tcode.IBRF & tcode.IOPMASK \/
		op = tcode.IBRT & tcode.IOPMASK \/
		op = tcode.INBRF & tcode.IOPMASK \/
		op = tcode.INBRT & tcode.IOPMASK \/
		op = tcode.IJUMP & tcode.IOPMASK
	) do
		select(TEXT);
		branch(op, a1);
	end
	else ie (op = tcode.IHDR & tcode.IOPMASK) do
		select(TEXT);
		emit(PROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IEND & tcode.IOPMASK) do
		select(TEXT);
		emit(EPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END");
		Active := 0;
	end
	else ie (op = tcode.IMHDR & tcode.IOPMASK) do
		select(TEXT);
		emit(MPROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IENDM & tcode.IOPMASK) do
		select(TEXT);
		emit(MEPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END");
		Active := 0;
	end
	else ie (op = tcode.IINIT & tcode.IOPMASK) do
		if (a1 \= 8) error("unsupported Tcode version");
	end
	else ie (op = tcode.IBPW & tcode.IOPMASK) do
		if (a1 \= t.bpw()) error("unsupported machine word size");
	end
	else ie (op = tcode.ICLAB & tcode.IOPMASK) do
		select(TEXT);
		if (Queued) spill(Iqueue, Aqueue);
		emit(DECL_CODE_LABEL, [(a1)]);
	end
	else ie (op = tcode.IDLAB & tcode.IOPMASK) do
		if (Dataname) do
			select(DATA);
			emit(DECL_DATA_ALIAS, [(Dataname),(a1)]);
		end
		Dataname := a1;
	end
	else ie (op = tcode.IVEC & tcode.IOPMASK) do
		select(BSS);
		op := rdcode();
		ie (op = tcode.IHINT & tcode.IOPMASK) do
			a2 := rdarg(L1);
			if (a2) a1 := (a1 + (BPW/a2-1)) / (BPW/a2);
		end
		else do
			reject(op);
		end
		emit(DECL_VECTOR, [(Dataname), (a1=0-> 1: a1)]);
		Dataname := 0;
	end
	else ie (op = tcode.IDATA & tcode.IOPMASK \/
		op = tcode.IDREF & tcode.IOPMASK \/
		op = tcode.ICREF & tcode.IOPMASK
	) do
		select(DATA);
		if (Dataname) emit(DECL_DATA_LABEL, [(Dataname)]);
		ie (op = tcode.IDATA & tcode.IOPMASK) do
			emit(DECL_WORD, [(a1)]);
		end
		else do
			emit(DECL_REF, [(a1)]);
		end
		Dataname := 0;
	end
	else ie (op = tcode.ISTR & tcode.IOPMASK) do
		select(DATA);
		if (Dataname) emit(DECL_DATA_LABEL, [(Dataname)]);
		genstr(sbuf, a1);
		Dataname := 0;
	end
	else ie (op = tcode.IGLUE & tcode.IOPMASK) do
	end
	else ie (op = tcode.IHALT & tcode.IOPMASK) do
		select(TEXT);
		emit(HALT_SEQ, [(a1)]);
	end
	else ie (op = tcode.IPUB & tcode.IOPMASK) do
		!ignore XXX
	end
	else ie (op = tcode.IEXT & tcode.IOPMASK) do
		!ignore XXX
	end
	else ie (op = tcode.IDUP & tcode.IOPMASK) do
		select(TEXT);
		if (Queued) spill(Iqueue, Aqueue);
		push();
		Active := 1;
	end
	else ie (op = tcode.ISWAP & tcode.IOPMASK) do
		select(TEXT);
		if (Queued) spill(Iqueue, Aqueue);
		if (Depth < 1) internal("too few operands in GENERATE(SWAP)");
		emit(SWAP_SEQ, 0);
	end
	else ie (op = tcode.IIPROC & tcode.IOPMASK) do
		addinfproc(a1, a2, sbuf);
	end
	else ie (op = tcode.IHINT & tcode.IOPMASK \/
		op = tcode.ICMAP & tcode.IOPMASK \/
		op = tcode.ILIB & tcode.IOPMASK
	) do
		! ignore
	end
	else ie (op = tcode.ICALX & tcode.IOPMASK) do
		error("unresolved CALX (external reference)");
	end
	else ie (op = tcode.IICALX & tcode.IOPMASK) do
		error("unresolved ICALX (external interface)");
	end
	else do
		internal("unknown instruction in GENERATE()");
	end
end

xlate() do
	var	op, o2, a1, a2, i, j, ch, k;
	var	sbuf::256;

	op := rdcode();
	if (op \= tcode.IINIT & tcode.IOPMASK) error("magic match failed");
	while (\In.eof()) do
		o2 := op & tcode.IOPMASK;
		if (o2 >= tcode.IENDOFSET) error("invalid opcode");
		if (Na) do
			a1 := rdarg(L1);
			k := a1;
			if (Na > 1) do
				a2 := rdarg(L2);
				k := a2;
			end
		end
		if (Sf) do
			for (i=0, k) sbuf::i := rdch();
			sbuf::i := 0;
		end
		generate(op, a1, a2, sbuf);
		op := rdcode();
	end
end

do
	init();
	Out.create(T3X.SYSOUT, Outbuf, BUFSIZE, iostream.FWRITE);
	In.create(T3X.SYSIN, Inbuf, BUFSIZE, iostream.FREAD);
	Addr := 0;
	Active := 0;
	emit(BANNER, 0);
	select(TEXT);
	emit(HEAD, 0);
	xlate();
	emit(TAIL, 0);
	Out.flush();
end
