#!/bin/sh

TARGET=$1
CMP_OFFSET=$2

if [ ! -f txtrn.t ]; then
	echo
	echo "The 'fulldiff' script must be run in the compiler/ directory."
	echo
	exit 1
fi

grep '(build config)' Makefile >/dev/null 2>&1
if [ $? != 0 ]; then
	echo
	echo "No Build configuration found."
	echo "Please go to your port directory and run './Config build'."
	echo
	exit 1
fi

tx -v |grep 'version 8\.' >/dev/null 2>&1
if [ $? != 0 ]; then
	echo
	echo "T3X 8.x must be installed in order to run a fulldiff."
	echo
	exit 1
fi

echo
echo "Running Triple Test, this may take some time."
echo

TESTDIR=`(cd ..; pwd)`/tmp

PATH=$TESTDIR/bin:${PATH}

make clean >/dev/null 2>&1
echo "Building stage #1"
echo "----- Stage #1 -----" >Logfile
make >>Logfile 2>&1
INSTALL_PREFIX=$TESTDIR make install >/dev/null 2>&1

mv Logfile Logfile.safe
make clean >/dev/null 2>&1
mv Logfile.safe Logfile
echo "Building stage #2"
echo >>Logfile
echo "----- Stage #2 -----" >>Logfile
T3XDIR=$TESTDIR/t3x/r8 make >>Logfile 2>&1
INSTALL_PREFIX=$TESTDIR make install >/dev/null 2>&1

mv Logfile Logfile.safe
make clean >/dev/null 2>&1
mv Logfile.safe Logfile
echo "Building stage #3"
echo >>Logfile
echo "----- Stage #3 -----" >>Logfile
T3XDIR=$TESTDIR/t3x/r8 make >>Logfile 2>&1
strip ../native/txcg$TARGET \
	../native/txcgexe \
	../native/txcguc \
	../native/s86 \
	../tools/txcm \
	../compiler/txdpf \
	../tools/txdump \
	../compiler/txld \
	../compiler/txopt \
	../tools/txprint \
	../tools/txsum \
	../compiler/txtrn >/dev/null 2>&1

echo >>Logfile
echo "----- Diff -----" >>Logfile
OK="yes"
if [ "$TARGET" != "" ]; then
	cmp $TESTDIR/t3x/r8/bin/txcg$TARGET ../native/txcg$TARGET \
		>>Logfile 2>&1
	if [ $? != 0 ]; then OK="no"; fi
fi
for testpair in \
	"$TESTDIR/t3x/r8/bin/txcgexe ../native/txcgexe" \
	"$TESTDIR/t3x/r8/bin/txcguc ../native/txcguc" \
	"$TESTDIR/t3x/r8/bin/s86 ../native/s86" \
	"$TESTDIR/t3x/r8/bin/txcm ../tools/txcm" \
	"$TESTDIR/t3x/r8/bin/txdpf ../compiler/txdpf" \
	"$TESTDIR/t3x/r8/bin/txdump ../tools/txdump" \
	"$TESTDIR/t3x/r8/bin/txld ../compiler/txld" \
	"$TESTDIR/t3x/r8/bin/txopt ../compiler/txopt" \
	"$TESTDIR/t3x/r8/bin/txprint ../tools/txprint" \
	"$TESTDIR/t3x/r8/bin/txsum ../tools/txsum" \
	"$TESTDIR/t3x/r8/bin/txtrn ../compiler/txtrn" \
	"$TESTDIR/t3x/r8/classes/t3x.cls ../compiler/t3x.cls" \
	"$TESTDIR/t3x/r8/classes/char.cls ../classes/char.cls" \
	"$TESTDIR/t3x/r8/classes/iostream.cls ../classes/iostream.cls" \
	"$TESTDIR/t3x/r8/classes/memory.cls ../classes/memory.cls" \
	"$TESTDIR/t3x/r8/classes/string.cls ../classes/string.cls" \
	"$TESTDIR/t3x/r8/classes/util.cls ../classes/util.cls" \
	"$TESTDIR/t3x/r8/lib/t3x.x ../native/t3x.x" \
	"$TESTDIR/t3x/r8/lib/system.x ../native/system.x" \
	"$TESTDIR/t3x/r8/lib/ttyctl.x ../native/ttyctl.x";
do
	cmp $testpair $CMP_OFFSET >>Logfile 2>&1
	if [ $? != 0 ]; then OK="no"; fi
done

rm -r $TESTDIR

echo

if [ "$OK" = "yes" ]; then
	echo "Triple Test passed."
	echo
	echo "You may now run 'make install' (as root)"
	echo "to install the production compiler."
else
	echo "Triple Test FAILED."
	echo
	echo "See 'compiler/Logfile' for details."
fi

echo

