# install script for @{IDENT}
# Copyright (C) 2003 Nils M Holm. All rights reserved.
# See the file LICENSE for conditions of use.

# TXBASE	where to install the package
# TXDIR		don't change
# BINDIR	location of the binaries (tx, txx, etc)
# INCDIR	where the local include files go
# MANDIR	where to place the man page
# OWNER		owner (uid) of the T3X files
# GROUP		group (gid) of the T3X files

if [ "$INSTALL_PREFIX" = "" ]; then
	PREFIX=/usr/local
else
	PREFIX=$INSTALL_PREFIX
fi
TXBASE=$PREFIX/t3x
TXDIR=$TXBASE/r8
BINDIR=$PREFIX/bin
INCDIR=`eval echo @{INCDIR}`
MANDIR=`eval echo @{MANDIR}`
if [ "@{MANSECT}" != "" ]; then
	MANPATH="$MANDIR/man@{MANSECT}"
else
	MANPATH=$MANDIR
fi
TESTING=no
OWNER=bin
GROUP=bin

TXCM="../tools/txcm -b $TXDIR"

shell_fn_test() {
	return 203
}
shell_fn_test >/dev/null 2>&1
if [ $? != 203 ]; then
	echo
	echo "The installation script requires a Bourne-compatible shell"
	echo "that supports shell functions."
	echo
	echo "On your system, you may try one out of these shells:"
	for p in /bin /usr/bin /usr/local/bin ${PREFIX}/bin; do
		for s in ksh bash; do
			if [ -f $p/$s ]; then echo $p/$s; fi
		done
	done
	echo
	exit 1
fi

id | grep root >/dev/null
ID=$?
echo $INSTALL_PREFIX | grep '/tmp$' >/dev/null
if [ $? = 0 ]; then
	TESTING=yes
fi

if [ $TESTING != yes -a $ID != 0 ]; then
	echo 
	echo "The install script must be run by root."
	echo
	exit 1
fi

(cd ../classes
 echo char: >char.ndx
 echo iostream:t3x >iostream.ndx
 echo memory:t3x >memory.ndx
 echo string:t3x,char >string.ndx
 echo system:t3x >system.ndx
 echo ttyctl: >ttyctl.ndx
 echo util:t3x,string,iostream >util.ndx
)

if [ -d ../classes/xclasses ]; then
	(cd ../classes/xclasses
	 echo tcode: >tcode.ndx
	 echo rtcode: >rtcode.ndx
	 echo char:t3x >char.ndx
	 echo iostream:t3x >iostream.ndx
	 echo memory:t3x >memory.ndx
	 echo string:t3x,char >string.ndx
	 echo system:t3x >system.ndx
	 echo ttyctl:t3x >ttyctl.ndx
	 echo util:t3x,string,iostream >util.ndx
	)
fi

install() {
	if [	"x$1" != "x-o" -o \
		"x$3" != "x-g" -o \
		"x$6" != "x-m" ]
	then
		echo install: bad arguments
		exit 1
	fi
	if [ "x$5" = "x-d" ]; then
		if [ ! -d $8 ]; then
			mkdir $8
			if [ $TESTING = no ]; then
				chown $2 $8
				chgrp $4 $8
			fi
			chmod $7 $8
		fi
	elif [ "x$5" = "x-C" ]; then
		if [ -d $9 ]; then
			file=$9/`basename $8`
		else
			file=$9
		fi
		cp $8 $file
		if [ $TESTING = no ]; then
			chown $2 $file
			chgrp $4 $file
		fi
		chmod $7 $file
	else
		echo install: bad arguments
		exit 1
	fi
}

BINLINK="@{SYMLINK_CMD}"
ln -fs Makefile .symtest.$$ >/dev/null 2>&1
if [ $? != 0 ]; then
	BINLINK=binlink
fi
rm -f .symtest.$$

binlink() {
	echo "#!/bin/sh" >$2
	echo "$1 \$*" >>$2
	chmod 0755 $2
	if [ $TESTING = no ]; then
		chown bin $2
		chgrp bin $2
	fi
}


@{PREPARE_INSTALL}

if [ ! -d $PREFIX ]; then install -o $OWNER -g $GROUP -d -m 0755 $PREFIX; fi
if [ "$INSTALL_PREFIX" != "" -a ! -d $PREFIX/man ];
	then install -o $OWNER -g $GROUP -d -m 0755 $PREFIX/man
fi
if [ ! -d $INCDIR ]; then
	install -o $OWNER -g $GROUP -d -m 0755 $INCDIR
fi
if [ ! -d $BINDIR ]; then install -o $OWNER -g $GROUP -d -m 0755 $BINDIR; fi
if [ ! -d $MANDIR ]; then install -o $OWNER -g $GROUP -d -m 0755 $MANDIR; fi
if [ ! -d $MANPATH ]; then install -o $OWNER -g $GROUP -d -m 0755 $MANPATH; fi

install -o $OWNER -g $GROUP -d -m 0755 $TXBASE
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/bin
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/classes
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/lib
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/doc

install -o $OWNER -g $GROUP -C -m 0755 txdpf $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txld $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txopt $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txtrn $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txx $TXDIR/bin
$BINLINK $TXDIR/bin/txx $BINDIR/txx

touch $TXDIR/classes/CLASSNDX
(cd ../compiler; $TXCM -qi t3x)
(cd ../compiler; $TXCM -qi tcode)
(cd ../compiler; $TXCM -qi rtcode)
(cd ../classes; $TXCM -i char)
(cd ../classes; $TXCM -i iostream)
(cd ../classes; $TXCM -i memory)
(cd ../classes; $TXCM -i string)
(cd ../classes; $TXCM -i util)
(cd ../classes; $TXCM -i system)
(cd ../classes; $TXCM -i ttyctl)
if [ $TESTING = no ]; then
	chown bin $TXDIR/classes/* ; chgrp bin $TXDIR/classes/*
	chown bin $TXDIR/lib/* ; chgrp bin $TXDIR/lib/*
fi

if [ -d ../classes/xclasses ]; then
	install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/classes/2
	(cd ../classes/xclasses;
	 cat *.ndx >$TXDIR/classes/2/CLASSNDX
	 cp *.cls $TXDIR/classes/2)
	if [ $TESTING = no ]; then
		chown bin $TXDIR/classes/2/* ; chgrp bin $TXDIR/classes/2/*
	fi
fi

install -o $OWNER -g $GROUP -C -m 0755 ../native/txcg@{TARGET} $TXDIR/bin
if [ @{TARGET} = 386a -o @{TARGET} = 386c -o @{TARGET} = 386e ]; then
	rm -f $TXDIR/bin/txcg386
	@{LINK_CMD} $TXDIR/bin/txcg@{TARGET} $TXDIR/bin/txcg386
fi
rm -f $TXDIR/bin/txcgl
@{LINK_CMD} $TXDIR/bin/txcg@{TARGET} $TXDIR/bin/txcgl
install -o $OWNER -g $GROUP -C -m 0755 ../native/txcgexe $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 ../native/txcguc $TXDIR/bin
rm -f $TXDIR/bin/txcg86
@{LINK_CMD} $TXDIR/bin/txcgexe $TXDIR/bin/txcg86

install -o $OWNER -g $GROUP -C -m 0755 ../native/s86 $TXDIR/bin
$BINLINK $TXDIR/bin/s86 $BINDIR/s86
# should be done by TXCM, but sources are in wrong directory
install -o $OWNER -g $GROUP -C -m 0644 ../native/t3x.x $TXDIR/lib
install -o $OWNER -g $GROUP -C -m 0644 ../native/system.x $TXDIR/lib
install -o $OWNER -g $GROUP -C -m 0644 ../native/ttyctl.x $TXDIR/lib

install -o $OWNER -g $GROUP -C -m 0755 ../tools/tx $TXDIR/bin
$BINLINK $TXDIR/bin/tx $BINDIR/tx
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txmake $TXDIR/bin
$BINLINK $TXDIR/bin/txmake $BINDIR/txmake
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txhelp $TXDIR/bin
$BINLINK $TXDIR/bin/txhelp $BINDIR/txhelp
install -o $OWNER -g $GROUP -C -m 0644 ../tools/txprint.rules $TXDIR
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txcm $TXDIR/bin
$BINLINK $TXDIR/bin/txcm $BINDIR/txcm
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txdump $TXDIR/bin
$BINLINK $TXDIR/bin/txdump $BINDIR/txdump
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txprint $TXDIR/bin
$BINLINK $TXDIR/bin/txprint $BINDIR/txprint
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txsum $TXDIR/bin
$BINLINK $TXDIR/bin/txsum $BINDIR/txsum

install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc \
	$TXDIR/system.txrc
install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.dos \
	$TXDIR/dos.txrc
install -o $OWNER -g $GROUP -C -m 0644 ../ports/Common-unix/txrc.uc \
	$TXDIR/uc.txrc
if [ ! -d ../classes/xclasses ]; then
	echo \
	"DOS/EXE cross compilation requires to build a production compiler" \
	> $TXDIR/dos.txrc
fi
if [ -f ../ports/@{PORT}/txrc.static ]; then
	install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.static \
		$TXDIR/static.txrc
fi

install -o $OWNER -g $GROUP -C -m 0644 ../README $TXDIR/doc
install -o $OWNER -g $GROUP -C -m 0644 ../LICENSE $TXDIR/doc
install -o $OWNER -g $GROUP -C -m 0644 ../docs/T3X.html $TXDIR/doc
(cd ../docs; for docfile in *.sd; do
	install -o $OWNER -g $GROUP -C -m 0644 $docfile $TXDIR/doc
 done)
install -o $OWNER -g $GROUP -C -m 0644 ../docs/sdtoc $TXDIR/doc/.sdtoc

install -o $OWNER -g $GROUP -C -m 0644 txx.h @{INCDIR}

install -o $OWNER -g $GROUP -C -m 0644 ../docs/t3x.7 $MANPATH/t3x.7

strip $TXDIR/bin/* >/dev/null 2>&1

@{FINISH_INSTALL}

echo
echo "Don't forget to define T3XDIR = $TXDIR ."
echo
echo "You may want to install $TXDIR/system.txrc as ~/.txrc, too."
echo

