/*
 *	T3X -- Common Runtime Support Routines for T3X R8.
 *	Copyright (C) 1999,2000,2002,2003,2004 Nils M Holm.
 *	All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifdef __TURBOC__
 #include <io.h>
#endif

#include "../compiler/txx.h"

void libinit(void) {
#ifdef dos
	setmode(0, O_BINARY);
	setmode(1, O_BINARY);
#endif
}

#ifdef LIBRARY

int	Argcount;
char	**Argvec;

int main(argc, argv)
int	argc;
char	**argv;
{
	Argcount = argc;
	Argvec = argv;
	libinit();
	_TXSTART_();
	return 0;
}

#endif

XCELL t3x_bpw S0() {
	return BPW;
}

XCELL t3x_memcopy S3(char *d, char *s, XCELL n) {
	memmove(d, s, n);
	return 0;
}

XCELL t3x_memcomp S3(char *s1, char *s2, XCELL n) {
	return memcmp(s1, s2, n);
}

XCELL t3x_memfill S3(char *p, XCELL c, XCELL n) {
	memset(p, c, n);
	return 0;
}

XCELL t3x_memscan S3(char *p, XCELL c, XCELL l) {
	int	i;

	for (i=0; i<l; i++)
		if (p[i] == c) return i;
	return -1;
}

XCELL t3x_getarg S3(XCELL i, char *b, XCELL n) {
	int	k;

	if (i < 1 || i >= Argcount) {
		b[0] = 0;
		return -1;
	}
	k = strlen(Argvec[i]);
	k = k>=n? n-1: k;
	memcpy(b, Argvec[i], k);
	b[k] = 0;
	return k;
}

XCELL t3x_getenv S3(char *vn, char *b, XCELL n) {
	char	*vv;
	int	k;

	vv = getenv(vn);
	if (vv == NULL) {
		b[0] = 0;
		return -1;
	}
	k = strlen(vv);
	k = k>=n? n-1: k;
	memcpy(b, vv, k);
	b[k] = 0;
	return k;
}

XCELL t3x_open S2(char *file, XCELL mode) {
	int	fd;

	if (mode == 1) {
		fd = creat(file, 0666);
	}
	else if (mode == 0 || mode == 2 || mode == 3) {
		fd = open(file, mode==3? O_WRONLY: mode==2? O_RDWR:
			O_RDONLY);
	}
	else {
		return -1;
	}
	if (fd > 0 && mode == 3) lseek(fd, 0L, 2);
	return fd;
}

#define T_SEEK_SET	0
#define T_SEEK_FWD	1
#define T_SEEK_END	2
#define T_SEEK_BCK	3

XCELL t3x_seek S3(XCELL fd, XUCELL n, XCELL o) {
	n &= 0xffff;
	if (o == T_SEEK_SET)
		return lseek(fd, (off_t) n * 1L, SEEK_SET) == -1? -1: 0;
	else if (o == T_SEEK_END)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_END) == -1? -1: 0;
	else if (o == T_SEEK_FWD)
		return (int) lseek(fd, (off_t) n * 1L, SEEK_CUR) == -1? -1: 0;
	else if (o == T_SEEK_BCK)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_CUR) == -1? -1: 0;
	else
		return -1;
}

XCELL t3x_close S1(XCELL fd) {
	return close(fd);
}

XCELL t3x_read S3(XCELL fd, char *b, XCELL n) {
	return read(fd, b, n);
}

XCELL t3x_write S3(XCELL fd, char *b, XCELL n) {
	return write(fd, b, n);
}

UCHAR *t3x_newline S1(UCHAR *s) {
#ifdef NL_IS_LF
	s[0] = '\n';
	s[1] = 0;
#endif
#ifdef NL_IS_CRLF
	s[0] = '\r';
	s[1] = '\n';
	s[2] = 0;
#endif
#ifdef LIBRARY
	return s;
#else
	return (UCHAR *) (s - Data);
#endif
}

XCELL t3x_remove S1(char *s) {
	return unlink(s);
}

XCELL t3x_rename S2(char *so, char *sn) {
	if (close(open(sn, 0)) >= 0) return -1;
	return rename(so, sn);
}

#ifdef HOST16

XCELL t3x_cvalist S4(XCELL n, XUCELL bmap, XCELL *ilist, XCELL *olist) {
	bmap = bmap;	/*USED*/
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(int));
	else
		memcpy(olist, ilist, n*sizeof(int));
	return 0;
}

#endif /*HOST16*/

#ifndef HOST16

#ifdef LIBRARY

XCELL t3x_cvalist S4(XCELL n, XUCELL bmap, void *ilist, void **olist) {
	bmap = bmap;	/*USED*/
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(void*));
	else
		memcpy(olist, ilist, n*sizeof(void*));
	return 0;
}

#else	/* !LIBRARY */

XCELL t3x_cvalist S4(XCELL n, XUCELL bmap, XUCELL *ilist, void **olist) {
	int	i;
	ULONG	p, o;

	/* Align olist at 8-byte word boundary */
	o = (ULONG) olist;
	p = (o + 7) & ~7UL;
	olist = (void**) p;
	if (n < 0) {
		n = -n;
		for (i=0; i<n; i++) ilist[i] = (XUCELL) olist[i];
	}
	else {
		for (i=0; i<n; i++) {
			if (bmap&1)
				olist[i] = &Data[ilist[i]];
			else
				olist[i] = (void *) ilist[i];
			bmap>>=1;
		}
	}
	return (int)(p-o);
}

#endif	/* !LIBRARY */

#endif /*!HOST16*/

struct infproc	InfT3X[17] = {
	{ "t3x_bpw",		t3x_bpw		},
	{ "t3x_memcopy",	t3x_memcopy	},
	{ "t3x_memcomp",	t3x_memcomp	},
	{ "t3x_memfill",	t3x_memfill	},
	{ "t3x_memscan",	t3x_memscan	},
	{ "t3x_getarg",		t3x_getarg	},
	{ "t3x_getenv",		t3x_getenv	},
	{ "t3x_open",		t3x_open	},
	{ "t3x_read",		t3x_read	},
	{ "t3x_write",		t3x_write	},
	{ "t3x_newline",	(XCELL(*)()) t3x_newline	},
	{ "t3x_seek",		t3x_seek	},
	{ "t3x_close",		t3x_close	},
	{ "t3x_rename",		t3x_rename	},
	{ "t3x_remove",		t3x_remove	},
	{ "t3x_cvalist",	t3x_cvalist	},
	{ NULL,			NULL		}
};

#ifdef NEED_RENAME

rename(old, new)
char	*old, *new;
{
	int	r;

	if ((r = link(old, new)) < 0) return r;
	return unlink(old);
}

#endif
