#! /bin/sh

# TX.SH -- A Bourne shell clone of TX.
# Copyright (C) 2000,2001,2002,2004 Nils M Holm. All rights reserved.
# See the file LICENSE for conditions of use
#
# This is an attempt to simulate the TX compiler driver in a shell script.
# Not all options of the original version are supported.
# This program may be used on platforms not supporting the system class.

T3X_VERSION="T3X version 8.1.5"

if [ "$T3XDIR" = "" ]; then
	T3XDIR=/usr/local/t3x/r8
fi

BINDIR=$T3XDIR/bin
CLASSDIR=$T3XDIR/classes
LIBDIR=$T3XDIR/lib

TXOPT_DFLT="txopt"
TXDPF_DFLT="txdpf"
TXCG_DFLT="txcg@{TARGET}"

TXTRN="txtrn"
TXLD="txld"
TXOPT="$TXOPT_DFLT"
TXDPF="$TXDPF_DFLT"
TXCG="NA"
TXCM="txcm"

rm -f T3X.ERR

DOSFLAG=0
VERBOSE=0
PRESERVE=0
LOAD=1
BANGHDR=NA
MAKECLASS=0
OUTFILE=NA
LONGUSAGE=0
LICENSE=0

USAGE="Usage: tx [-?h] [-B dir] [-C dir] [-L dir] [-o file]"
USAGE2="[-m type] [-Ados] [-DNOceipv] file"

while true; do
	case "x$1" in
		x-\?|x-h)
			LONGUSAGE=1
			break
			;;
		x-B)
			if [ "$2" != "" ]; then
				shift
			fi
			BINDIR=$1
			;;
		x-C)
			if [ "$2" != "" ]; then
				shift
			fi
			CLASSDIR=$1
			;;
		x-U*)
			if [ "$2" != "" ]; then
				shift
			fi
			BINDIR=../compiler
			CLASSDIR=../classes
			LIBDIR=../native
			TXCG=$TXCG_DFLT
			echo \
			 "WARNING: -U ignored -- using boot defaults" 1>&2
			;;
		x-o)
			if [ "$2" != "" ]; then
				shift
			fi
			OUTFILE=$1
			;;
		x-N)
			TXOPT=NA
			TXDPF=NA
			;;
			;;
		x-O)
			TXOPT=$TXOPT_DFLT
			;;
		x-D)
			TXDPF=$TXDPF_DFLT
			;;
		x-m)
			if [ "$2" != "" ]; then
				shift
			fi
			TXCG="txcg$1"
			;;
		x-m386)
			TXCG="txcg386c"
			;;
		x-m386a)
			TXCG="txcg386a"
			;;
		x-m386c)
			TXCG="txcg386c"
			;;
		x-m386e)
			TXCG="txcg386e"
			;;
		x-Ados)
			TXCG="txcgexe"
			;;
		x-V)
			LICENSE=1
			break
			;;
		x-c)
			LOAD=0
			;;
		x-i)
			MAKECLASS=1
			;;
		x-e)
			BANGHDR="#! $BINDIR/txx"
			;;
		x-p)
			PRESERVE=1
			;;
		x-v)
			VERBOSE=1
			echo "$T3X_VERSION"
			;;
		x-vv)
			VERBOSE=2
			echo "$T3X_VERSION"
			;;
		x-*|x)
			echo $USAGE
			echo "	"$USAGE2
			exit 1
			;;
		x*)	break
			;;
	esac
	shift
done

if [ $BINDIR/$TXCG = txcgexe ]; then
	DOSFLAG=1
fi

if [ $LONGUSAGE = 1 ]; then
echo
echo $USAGE
echo "	"$USAGE2
cat <<EOT

-Ados	generate a DOS/EXE file
-B dir  search binaries in 'dir'
-C dir  search classes in 'dir'
-D      enable dead procedure filter (DPF)
-N      disable optimizer and DPF
-O      enable tcode optimization
-V      display the conditions of use
-c      compile only (don't load)
-e      create a #! header (Unix only)
-h      display this summary (also -?)
-i      create installable class files
-m type	generate code for target 'type'
-o file direct output to 'file'
-p      preserve CLASSDIR
-v      verbose operation (more v's = more output)

EOT
echo -n "Installed targets:"
for type in 386 386a 386c 386e 86 86s exe; do
	if [ -f $BINDIR/txcg$type ]; then
		echo -n " $type"
	fi
done
echo ; echo
exit 0
fi

if [ $LICENSE = 1 ]; then
echo
cat <<EOT
T3X -- A Compiler for the Minimum Procedural Language T3X
Copyright (C) 1996-2004 Nils M Holm.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
EOT
echo
exit 0
fi

FILE=$1
OFILE=`basename $FILE .t`
TMP=tx$$.tmp

if [ ! -f $FILE ]; then
	echo "TX: $FILE: no such file"
	exit 1
fi

if [ ! -f $OFILE.t ]; then
	echo "TX: $FILE: file has no .t suffix"
	exit 1
fi

if [ "$OUTFILE" != "NA" ]; then
	OFILE=$OUTFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR
fi

if [ $VERBOSE != 0 ]; then
	echo $BINDIR/$TXTRN: $FILE '->' $OFILE
fi
$BINDIR/$TXTRN <$FILE >$OFILE
if [ -f T3X.ERR -o $? != 0 ]; then
	rm -f T3X.ERR $TMP $OFILE
	echo "TX: error in compilation. stopped."
	exit 1
fi

if [ $MAKECLASS = 1 ]; then
	rm -f $OFILE.cls
	mv $OFILE $OFILE.cls
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXCM -qfx $OFILE CLASSDIR
	fi
	$BINDIR/$TXCM -qfx $OFILE CLASSDIR
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXCM -o $OFILE.ndx -n $OFILE CLASSDIR
	fi
	$BINDIR/$TXCM -o $OFILE.ndx -n $OFILE CLASSDIR
	if [ $PRESERVE = 0 ]; then
		rm -f CLASSDIR
	fi
	exit 0
fi

if [ $LOAD = 0 ]; then
	exit 0
fi

if [ -f CLASSDIR ]; then
	rm -f $TMP
	(strings CLASSDIR | grep '\*[a-zA-Z_0-9]*\*' | \
		sed -e 's/.*\*\([a-zA-Z_0-9]*\)\*/\1/' >CLASSDIR.names) \
			2>/dev/null
	for libname in `cat CLASSDIR.names`; do
		if [ $libname = t3x ]; then
			continue
		fi
		if [ ! -f $libname.cls ]; then
			libname=$CLASSDIR/$libname
		fi
		if [ ! -f $libname.cls ]; then
			rm -f $TMP $OFILE
			echo "class $libname is not available"
			exit 1
		fi
		if [ $VERBOSE = 2 ]; then
			echo "(load) $libname.cls"
		fi
		cat $libname.cls >>$TMP
	done
	rm -f CLASSDIR.names
	cat $OFILE >>$TMP
	mv -f $TMP $OFILE
fi

if [ $VERBOSE != 0 ]; then
	echo $BINDIR/$TXLD $TMP '->' $OFILE
fi
mv -f $OFILE $TMP
$BINDIR/$TXLD <$TMP >$OFILE
if [ -f T3X.ERR -o $? != 0 ]; then
	rm -f T3X.ERR $TMP $OFILE
	echo "TX: error in compilation. stopped."
	exit 1
fi

if [ "$TXOPT" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXOPT $TMP '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$BINDIR/$TXOPT <$TMP >$OFILE
	if [ -f T3X.ERR -o $? != 0 ]; then
		rm -f T3X.ERR $TMP $OFILE
		echo "TX: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$TXDPF" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXDPF $TMP '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$BINDIR/$TXDPF <$TMP >$OFILE
	if [ -f T3X.ERR -o $? != 0 ]; then
		rm -f T3X.ERR $TMP $OFILE
		echo "TX: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$DOSFLAG" = "1" ]; then
	mv -f $OFILE $TMP
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXCG $TMP '->' $OFILE.exe
	fi
	$BINDIR/$TXCG <$TMP >$OFILE.exe
elif [ "$TXCG" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $BINDIR/$TXCG $OFILE '->' $OFILE.s
	fi
	$BINDIR/$TXCG <$OFILE >$OFILE.s
	if [ $VERBOSE != 0 ]; then
		echo as -o $OFILE $OFILE.s
	fi
	as -o $OFILE.o $OFILE.s
	rm -f $OFILE.s
	if [ $VERBOSE != 0 ]; then
		echo $CC_CMD -o $OFILE $OFILE.o $LIBDIR/t3x.o
	fi
	$CC_CMD -o $OFILE $OFILE.o $LIBDIR/t3x.o
	rm -f $OFILE.o
elif [ "$BANGHDR" != "NA" ]; then
	mv -f $OFILE $TMP
	echo "$BANGHDR" >$OFILE
	cat $TMP >>$OFILE
	chmod +x $OFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR
fi
rm -f $TMP
