#!/bin/sh

# Installation Script for the Generic Bytecode Port
# Copyright (C) 2003 Nils M Holm. All rights reserved.
# See the file LICENSE for conditions of use.

# install script for the Tcode port

# TXBASE	where to install the package
# TXDIR		don't change
# BINDIR	location of the binaries (tx, txx, etc)
# INCDIR	where the local include files go
# DOCDIR	online documents
# MANDIR	base directory of the manual pages
# MANSECT	in which section to place the man page
# OWNER		owner:group of the T3X files
# GROUP		group (gid) of the T3X files

if [ "$INSTALL_PREFIX" = "" ]; then
	PREFIX=/usr/local
else
	PREFIX=$INSTALL_PREFIX
fi
TXBASE=$PREFIX/t3x
TXDIR=$TXBASE/r8
BINDIR=$PREFIX/bin
DOCDIR=$TXDIR/doc
MANDIR=$PREFIX/man
MANSECT=7
OWNER=bin
GROUP=bin
MANPATH="$MANDIR/man$MANSECT"

TXCM="../compiler/txx ../tools/txcm -b $TXDIR"

rm -f $BINDIR/tx $BINDIR/txx $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint \
	$BINDIR/txsum $BINDIR/txmake $BINDIR/txhelp $BINDIR/s86

if [ ! -d $PREFIX ]; then
	mkdir $PREFIX && chmod 0755 $PREFIX && chown $OWNER:$GROUP $PREFIX
fi
if [ ! -d $TXBASE ]; then
	mkdir $TXBASE && chmod 0755 $TXBASE && chown $OWNER:$GROUP $TXBASE
fi
if [ ! -d $TXDIR ]; then
	mkdir $TXDIR && chmod 0755 $TXDIR && chown $OWNER:$GROUP $TXDIR
fi
if [ ! -d $TXDIR/bin ]; then
	mkdir $TXDIR/bin && chmod 0755 $TXDIR/bin && \
		chown $OWNER:$GROUP $TXDIR/bin
fi
if [ ! -d $TXDIR/classes ]; then
	mkdir $TXDIR/classes && chmod 0755 $TXDIR/classes && \
		chown $OWNER:$GROUP $TXDIR/classes
fi
if [ ! -d $TXDIR/lib ]; then
	mkdir $TXDIR/lib && chmod 0755 $TXDIR/lib && \
		chown $OWNER:$GROUP $TXDIR/lib
fi
if [ ! -d $BINDIR ]; then
	mkdir $BINDIR && chmod 0755 $BINDIR && chown $OWNER:$GROUP $BINDIR
fi
if [ ! -d $DOCDIR ]; then
	mkdir $DOCDIR && chmod 0755 $DOCDIR && chown $OWNER:$GROUP $DOCDIR
fi
if [ ! -d $MANDIR ]; then
	mkdir $MANDIR && chmod 0755 $MANDIR && chown $OWNER:$GROUP $MANDIR
fi
if [ ! -d $MANPATH ]; then
	mkdir $MANPATH && chmod 0755 $MANPATH && chown $OWNER:$GROUP $MANPATH
fi

touch $TXDIR/classes/CLASSNDX

(cd ../compiler
 $TXCM -qi t3x
 $TXCM -qi tcode
 $TXCM -qi rtcode
)

(cd ../classes
 echo char: >char.ndx ; $TXCM -i  char
 echo iostream:t3x >iostream.ndx ; $TXCM -i iostream
 echo memory:t3x >memory.ndx ; $TXCM -i memory
 echo string:t3x,char >string.ndx ; $TXCM -i string
 echo system:t3x >system.ndx ; $TXCM -i system
 echo ttyctl: >ttyctl.ndx ; $TXCM -i ttyctl
 echo util:t3x,string,iostream >util.ndx ; $TXCM -i util
)

chmod 0644 $TXDIR/classes/*
chown $OWNER:$GROUP $TXDIR/classes/*

if [ -d ../classes/xclasses ]; then
	if [ ! -d $TXDIR/classes/2 ]; then
		mkdir $TXDIR/classes/2 && chown $OWNER:$GROUP $TXDIR/classes/2
	fi
	chmod 0755 $TXDIR/classes/2
	(cd ../classes/xclasses
	 echo tcode: >tcode.ndx
	 echo rtcode: >rtcode.ndx
	 echo char:t3x >char.ndx
	 echo iostream:t3x >iostream.ndx
	 echo memory:t3x >memory.ndx
	 echo string:t3x,char >string.ndx
	 echo system:t3x >system.ndx
	 echo ttyctl:t3x >ttyctl.ndx
	 echo util:t3x,string,iostream >util.ndx
	 cat *.ndx >$TXDIR/classes/2/CLASSNDX
	 cp *.cls $TXDIR/classes/2)
	chmod 0644 $TXDIR/classes/2/*
	chown $OWNER $TXDIR/classes/2/*
	chgrp $GROUP $TXDIR/classes/2/*
fi

cp ../native/t3x.x $TXDIR/lib
cp ../native/system.x $TXDIR/lib
cp ../native/ttyctl.x $TXDIR/lib

cp ../compiler/txx $TXDIR/bin
cp ../tools/tx $TXDIR/bin
cp ../tools/txcm $TXDIR/bin
cp ../tools/txdump $TXDIR/bin
cp ../tools/txprint $TXDIR/bin
cp ../tools/txsum $TXDIR/bin
if [ -f ../tools/txmake ]; then
	cp ../tools/txmake $TXDIR/bin
fi
if [ -f ../tools/txhelp ]; then
	cp ../tools/txhelp $TXDIR/bin
fi
cp ../compiler/txtrn $TXDIR/bin
cp ../compiler/txld $TXDIR/bin
cp ../compiler/txopt $TXDIR/bin
cp ../compiler/txdpf $TXDIR/bin
cp ../native/txcguc $TXDIR/bin
cp ../native/txcgexe $TXDIR/bin
cp ../native/s86 $TXDIR/bin
chown $OWNER:$GROUP $TXDIR/bin/*
chmod 0644 $TXDIR/bin/*
chmod 0755 $TXDIR/bin/txx $TXDIR/bin/tx

cp $TXDIR/bin/txx $BINDIR/txx && chown $OWNER:$GROUP $BINDIR/txx && \
	chmod 0755 $BINDIR/txx

if [ -f ../tools/tx2 ]; then
	echo "#! /bin/sh" >$BINDIR/tx
	echo "$TXDIR/bin/txx $TXDIR/bin/tx \$*" >>$BINDIR/tx
	chown $OWNER:$GROUP $BINDIR/tx
	chmod 0755 $BINDIR/tx
	chmod 0644 $TXDIR/bin/tx
else
	cp $TXDIR/bin/tx $BINDIR/tx && chown $OWNER:$GROUP $BINDIR/tx && \
		chmod 0755 $BINDIR/tx
fi

echo "#! /bin/sh" >$BINDIR/txcm
echo "$TXDIR/bin/txx $TXDIR/bin/txcm \$*" >>$BINDIR/txcm
echo "#! /bin/sh" >$BINDIR/txdump
echo "$TXDIR/bin/txx $TXDIR/bin/txdump \$*" >>$BINDIR/txdump
echo "#! /bin/sh" >$BINDIR/txprint
echo "$TXDIR/bin/txx $TXDIR/bin/txprint \$*" >>$BINDIR/txprint
echo "#! /bin/sh" >$BINDIR/txsum
echo "$TXDIR/bin/txx $TXDIR/bin/txsum \$*" >>$BINDIR/txsum
if [ -f ../tools/txmake ]; then
	echo "#! /bin/sh" >$BINDIR/txmake
	echo "$TXDIR/bin/txx $TXDIR/bin/txmake \$*" >>$BINDIR/txmake
 	chown $OWNER:$GROUP $BINDIR/txmake
 	chmod 0755 $BINDIR/txmake
fi
if [ -f ../tools/txhelp ]; then
	echo "#! /bin/sh" >$BINDIR/txhelp
	echo "$TXDIR/bin/txx $TXDIR/bin/txhelp \$*" >>$BINDIR/txhelp
 	chown $OWNER:$GROUP $BINDIR/txhelp
 	chmod 0755 $BINDIR/txhelp
fi
echo "#! /bin/sh" >$BINDIR/s86
echo "$TXDIR/bin/txx $TXDIR/bin/s86 \$*" >>$BINDIR/s86
chown $OWNER:$GROUP $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint $BINDIR/txsum \
	$BINDIR/s86
chmod 0755 $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint $BINDIR/txsum \
	$BINDIR/s86

cp ../tools/txprint.rules $TXDIR/txprint.rules && \
	chown $OWNER:$GROUP $TXDIR/txprint.rules && \
	chmod 0644 $TXDIR/txprint.rules && \

cp ../README $DOCDIR
cp ../LICENSE $DOCDIR
cp ../docs/*.sd $DOCDIR
cp ../docs/sdtoc $DOCDIR/.sdtoc
cp ../docs/*.html $DOCDIR
chown $OWNER:$GROUP $DOCDIR/*
chmod 0644 $DOCDIR/*

cp ../ports/Generic-unix/txrc $TXDIR/system.txrc
chown $OWNER:$GROUP $TXDIR/system.txrc
chmod 0644 $TXDIR/system.txrc

cp ../ports/Common-unix/txrc.dos $TXDIR/dos.txrc
chown $OWNER:$GROUP $TXDIR/dos.txrc
chmod 0644 $TXDIR/dos.txrc

if [ ! -d ../classes/xclasses ]; then
	echo \
	"DOS/EXE cross compilation requires to build a production compiler" \
	> $TXDIR/dos.txrc
fi

cp ../docs/t3x.7 $MANPATH/t3x.7
chown $OWNER:$GROUP $MANPATH/t3x.7
chmod 0644 $MANPATH/t3x.7

echo
echo "Don't forget to define T3XDIR = $TXDIR ."
echo
echo "You may want to install $TXDIR/system.txrc as ~/.txrc, too."
echo
