/*
 *	SYS -- Runtime Support Routines for T3X under DOS.
 *	Copyright (C) 1999,2000,2002,2004 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include "../compiler/txx.h"

#ifdef EXT_SYSTEM

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dir.h>
#include <time.h>
#include <process.h>

#define uchar	unsigned char
#define ushort	unsigned short
#define uint	unsigned int
#define ulong	unsigned long

#define NDIR	64
struct ffblk	Dirfd[NDIR];

#define MAXARGS	128

#define RDOK	0x01
#define WROK	0x02
#define EXOK	0x04
#define ISDIR	0x08

XCELL system_init S0() {
	int	i;

	for (i=0; i<NDIR; i++) Dirfd[i].ff_name[0] = NULL;
	return 0;
}

XCELL system_fini S0() {
	return 0;
}

XCELL system_pipe S1(fds)
int	*fds;
{
	fds[0] = fds[1];	/* fail */
	return -1;
}

XCELL system_dup S1(fd)
XCELL	fd;
{
	return dup(fd);
}

XCELL system_dup2 S2(ofd, nfd)
XCELL	ofd, nfd;
{
	return dup2(ofd, nfd);
}

XCELL system_rdchk S1(fd)
XCELL	fd;
{
	fd = fd;	/* fail */
	write(2, "SYSTEM.RDCHK: not implemented\n", 30);
	_exit(-1);
}

XCELL system_mkdir S1(s)
char	*s;
{
	return mkdir(s);
}

XCELL system_chdir S1(s)
char	*s;
{
	return chdir(s);
}

XCELL system_rmdir S1(s)
char	*s;
{
	return rmdir(s);
}

XCELL system_getdir S2(buf, len)
char	*buf;
XCELL	len;
{
	if (len < 65) return -1;
	if (getcwd(buf, len) < 0) return -1;
	return strlen(buf);
}

XCELL system_opendir S1(s)
char	*s;
{
	int	i;
	char	tmp[81];

	for (i=0; i<NDIR; i++) if (Dirfd[i].ff_name[0] == 0) break;
	if (i >= NDIR) return -1;
	strcpy(tmp, s);
	strcat(tmp, "/*.*");
	if (findfirst(tmp, &Dirfd[i], 0x3f)) {
		Dirfd[i].ff_name[0] = NULL;
		return -1;
	}
	return i;
}

XCELL system_readdir S3(dd, b, n)
XCELL	dd;
char	*b;
XCELL	n;
{
	int		k;

	if (Dirfd[dd].ff_name[0] == 0) return -1;
	k = strlen(Dirfd[dd].ff_name);
	k = k>=n? n-1: k;
	memcpy(b, Dirfd[dd].ff_name, k);
	b[k] = 0;
	if (findnext(&Dirfd[dd])) Dirfd[dd].ff_name[0] = 0;
	return k;
}

XCELL system_closedir S1(dd)
XCELL	dd;
{
	if (Dirfd[dd].ff_name[0] == 0) return -1;
	Dirfd[dd].ff_name[0] = 0;
	return 0;
}

accessmap(sp)
struct stat	*sp;
{
	int	bits = 0;

	if (sp->st_mode & S_IREAD) bits |= RDOK;
	if (sp->st_mode & S_IWRITE) bits |= WROK;
	if (sp->st_mode & S_IREAD) bits |= EXOK;
	if (sp->st_mode & S_IFDIR) bits |= ISDIR;
	return bits;
}

XCELL system__stat S2(name, b)
char	*name;
uchar	*b;
{
	struct t3x_stat {
		short	dev;
		short	ino;
		short	mode;
		short	nlink;
		short	uid;
		short	gid;
		short	rdev;
		short	size;
		short	ext64;
		short	mt_year;
		short	mt_mon_day;
		short	mt_hr_min;
		short	mt_sec_hs;
	};
	struct stat	st;
	struct tm	*tp;

	if (stat(name, &st) < 0) return -1;
	b[0] = (uchar) (st.st_dev >> 8);
	b[1] = (uchar) st.st_dev;
	b[2] = (uchar) (st.st_ino >> 8);
	b[3] = (uchar) st.st_ino;
	b[4] = (uchar) 0;
	b[5] = (uchar) (st.st_mode & S_IFDIR? ISDIR: 0) | accessmap(&st);
	b[6] = (uchar) (st.st_nlink >> 8);
	b[7] = (uchar) st.st_nlink;
	b[8] = (uchar) (st.st_uid >> 8);
	b[9] = (uchar) st.st_uid;
	b[10] = (uchar) (st.st_gid >> 8);
	b[11] = (uchar) st.st_gid;
	b[12] = (uchar) (st.st_rdev >> 8);
	b[13] = (uchar) st.st_rdev;
	b[14] = (uchar) (st.st_size >> 8);
	b[15] = (uchar) st.st_size;
	b[16] = (uchar) (st.st_size / (64L*1024) >> 8);
	b[17] = (uchar) (st.st_size / (64L*1024));
	tp = localtime(&st.st_mtime);
	b[18] = (tp->tm_year+1900) / 100;
	b[19] = tp->tm_year % 100;
	b[20] = tp->tm_mon+1;
	b[21] = tp->tm_mday;
	b[22] = tp->tm_hour;
	b[23] = tp->tm_min;
	b[24] = tp->tm_sec;
	b[25] = 0;
	return 0;
}

XCELL system_wait S1(pid)
XCELL	pid;
{
	pid = pid;	/* ignore */
	return -1;	/* fail */
}

XCELL system__spawn S3(p, av, w)
char	*p;
char	**av;
XCELL	w;
{
	int	i;
	char	*av2[MAXARGS];

	if (w != 1) return -1;
	for (i = 0; av[i]; i++)
		if (i >= MAXARGS-1) return -1;
	for (++i; i>0; i--) av2[i] = av[i-1];
	av2[0] = p;
	return spawnv(P_WAIT, p, av2);
}

XCELL system_fork S0() {
	write(2, "SYSTEM.FORK: not implemented\n", 29);
	return -1;	/* fail */
}

XCELL system_kill S2(pid, n)
XCELL	pid, n;
{
	pid = pid;	/* ignore */
	n = n;		/* ignore */
	return -1;	/* fail */
}

convtime(t, b)
time_t	t;
char	*b;
{
	int	mlen[] = { 0, 31,28,31,30,31,30,31,31,30,31,30,31 };
	int	yr, mon, day, hr, min, sec;
	long	dt;

	yr = 1970;
	while (1) {
		dt = (!(yr&3)? 366: 365) * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		yr++;
	}
	mon = 1;
	mlen[2] = !(yr&3)? 29: 28;
	while (mon<12) {
		dt = mlen[mon] * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		mon++;
	}
	day = (int) (t / 86400L + 1);
	t = (time_t) (t % 86400L);
	hr = (int) (t / 3600);
	t = (int) (t % 3600);
	min = (int) (t / 60);
	sec = (int) (t % 60);
	b[0] = yr / 100;
	b[1] = yr % 100;
	b[2] = mon;
	b[3] = day;
	b[4] = hr;
	b[5] = min;
	b[6] = sec;
	b[7] = 0;
}

XCELL system_time S1(b)
char	*b;
{
	time_t	t;

	time(&t);
	convtime(t, b);
	return 0;
}

struct infproc	InfSystem[MAXINFPROC] = {
	{ "system__spawn",	system__spawn	},
	{ "system__stat",	system__stat	},
	{ "system_chdir",	system_chdir	},
	{ "system_closedir",	system_closedir	},
	{ "system_dup",		system_dup	},
	{ "system_dup2",	system_dup2	},
	{ "system_fini",	system_fini	},
	{ "system_fork",	system_fork	},
	{ "system_getdir",	system_getdir	},
	{ "system_init",	system_init	},
	{ "system_kill",	system_kill	},
	{ "system_mkdir",	system_mkdir	},
	{ "system_opendir",	system_opendir	},
	{ "system_pipe",	system_pipe	},
	{ "system_rdchk",	system_rdchk	},
	{ "system_readdir",	system_readdir	},
	{ "system_rmdir",	system_rmdir	},
	{ "system_time",	system_time	},
	{ "system_wait",	system_wait	},
	{ NULL,			NULL		}
};

#else	/* EXT_SYSTEM */

struct infproc	InfSystem[MAXINFPROC] = { { 0L, 0L } };

#endif	/* EXT_SYSTEM */
