/*
 *	TTY -- Terminal I/O Routines for DOS.
 *	Copyright (C) 2002 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include "../compiler/txx.h"

#ifdef EXT_TTYCTL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#define K_HOME	327
#define K_WLFT	371
#define K_LEFT	331
#define K_RGHT	333
#define K_WRGT	372
#define K_END	335
#define K_BKSP	'\b'
#define K_DEL	339
#define K_KILL	127
#define K_INS	338
#define K_CR	'\r'
#define K_UP	328
#define K_DOWN	336
#define K_ESC	'\033'
#define K_PGUP	329
#define K_PGDN	337
#define K_CPGU	388
#define K_CPGD	374
#define K_F1	315
#define K_F2	316
#define K_F3	317
#define K_F4	318
#define K_F5	319
#define K_F6	320
#define K_F7	321
#define K_F8	322
#define K_F9	323
#define K_F10	324

int	tty_cols,
	tty_lines,
	tty_attr,
	tty_x,
	tty_y;

int	Ansicolor;

XCELL ttyctl_writec S1(c)
XCELL	c;
{
	putch(c);
	return c;
}

XCELL ttyctl_writes S1(s)
char	*s;
{
	cputs(s);
	return 0;
}

XCELL ttyctl_clear S0() {
	clrscr();
	return 0;
}

XCELL ttyctl_clreol S0() {
	clreol();
	return 0;
}

XCELL ttyctl_color S1(c)
XCELL	c;
{
	int	f, b;

	tty_attr = c;
	f = c & 7;
	b = (c>>4) & 7;
	if (Ansicolor) {
		textattr(c);
	}
	else {
		if (b > f)
			textattr(0x70);
		else
			textattr(0x07);
	}
	return 0;
}

XCELL ttyctl_query S0() {
	return kbhit();
}

XCELL ttyctl_readc S0() {
	int	c;

	c = getch();
	if (c == 0) return getch() + 256;
	return c;
}

XCELL ttyctl_move S2(x, y)
XCELL	x, y;
{
	tty_x = x;
	tty_y = y;
	gotoxy(x+1, y+1);
	return 0;
}

XCELL ttyctl_scroll S2(top, bot)
XCELL	top, bot;
{
	window(1, top+1, tty_cols, bot+1);
	gotoxy(1, 1);
	delline();
	window(1, 1, tty_cols, tty_lines);
	gotoxy(tty_x, tty_y);
	return 0;
}

XCELL ttyctl_rscroll S2(top, bot)
XCELL	top, bot;
{
	window(1, top+1, tty_cols, bot+1);
	gotoxy(1, 1);
	insline();
	window(1, 1, tty_cols, tty_lines);
	gotoxy(tty_x, tty_y);
	return 0;
}

XCELL ttyctl_init S0() {
	char	*t;

	t = getenv("ANSICOLOR");
	if (t != NULL) {
		Ansicolor = !strcmp(t, "no")? 0:
			!strcmp(t, "No")? 0:
			!strcmp(t, "NO")? 0: 1;
	}
	return 0;
}

XCELL ttyctl_mode S1(rawmode)
XCELL	rawmode;
{
	rawmode = rawmode;	/* ignore */
	return 0;
}

XCELL ttyctl_fini S0() {
	return 0;
}


XCELL ttyctl_lines S0() { return tty_lines; }

XCELL ttyctl_columns S0() { return tty_cols; }

XCELL ttyctl_colors S0() { return Ansicolor; }

struct infproc	InfTTYCtl[MAXINFPROC] = {
	{ "ttyctl_clear",	ttyctl_clear	},
	{ "ttyctl_clreol",	ttyctl_clreol	},
	{ "ttyctl_color",	ttyctl_color	},
	{ "ttyctl_colors",	ttyctl_colors	},
	{ "ttyctl_columns",	ttyctl_columns	},
	{ "ttyctl_fini",	ttyctl_fini	},
	{ "ttyctl_init",	ttyctl_init	},
	{ "ttyctl_lines",	ttyctl_lines	},
	{ "ttyctl_mode",	ttyctl_mode	},
	{ "ttyctl_move",	ttyctl_move	},
	{ "ttyctl_query",	ttyctl_query	},
	{ "ttyctl_readc",	ttyctl_readc	},
	{ "ttyctl_rscroll",	ttyctl_rscroll	},
	{ "ttyctl_scroll",	ttyctl_scroll	},
	{ "ttyctl_writec",	ttyctl_writec	},
	{ "ttyctl_writes",	ttyctl_writes	},
	{ NULL,			NULL		}
};

#else	/* EXT_TTYCTL */

struct infproc	InfTTYCtl[MAXINFPROC] = { { 0L, 0L } };

#endif	/* EXT_TTYCTL */
