/*
 *	SYS -- Runtime Support Routines for T3X under Unix.
 *	Copyright (C) 1999,2000,2002,2003,2004 Nils M Holm.
 *	All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include "../compiler/txx.h"

#ifdef EXT_SYSTEM

#include <u.h>
#include <libc.h>

#include <stdio.h>

struct DIR {
	Dir	*buf;
	int	here, end;
};

#define NDIR	64
struct DIR	Dirs[NDIR];

#define MAXARGS	128

#define NPROC	64
int	Procd[NPROC];

#define RDOK	0x01
#define WROK	0x02
#define EXOK	0x04
#define ISDIR	0x08

XCELL system_init S0() {
	int	i;

	for (i=0; i<NDIR; i++) Dirs[i].buf = NULL;
	for (i=0; i<NPROC; i++) Procd[i] = -1;
	return 0;
}

XCELL system_fini S0() {
	return 0;
}

int nextprocd(int pid) {
	int	i;

	for (i=1; i<NPROC; i++) if (Procd[i] == -1) {
		Procd[i] = pid;
		return i;
	}
	write(2, "SYSTEM: out of process descriptors\n", 35);
	exits("Out of ProcFds");
	return 0;
}

XCELL system_pipe S1(XCELL *fds) {
	int	a[2], r;

	r = pipe(a);
	fds[0] = a[0];
	fds[1] = a[1];
	return r;
}

XCELL system_dup S1(XCELL fd) {
	return dup(fd, -1);
}

XCELL system_dup2 S2(XCELL ofd, XCELL nfd) {
	return dup(ofd, nfd);
}

XCELL system_rdchk S1(XCELL fd) {
	USED(fd);
	write(2, "SYSTEM.RDCHK: not implemented\n", 31);
	exits("SYSTEM.RDCHK: NYI");
	return 0;
}

XCELL system_mkdir S1(char *s) {
	return close(create(s, OREAD, 0755|DMDIR));
}

XCELL system_chdir S1(char *s) {
	return chdir(s);
}

XCELL system_rmdir S1(char *s) {
	return remove(s);
}

XCELL system_getdir S2(char *buf, XCELL len) {
	if (len < 65) return -1;
	if (getwd(buf, len) == 0) return -1;
	return strlen(buf);
}

XCELL system_opendir S1(char *s) {
	int	i, fd, k;

	for (i=0; i<NDIR; i++) if (Dirs[i].buf == NULL) break;
	if (i >= NDIR) return -1;
	if ((fd = open(s, OREAD)) == 0) return -1;
	if ((k = dirreadall(fd, &Dirs[i].buf)) < 0) return -1;
	Dirs[i].here = 0;
	Dirs[i].end = k;
	close(fd);
	return i;
}

XCELL system_readdir S3(XCELL dd, char *b, XCELL n) {
	int	k;

	if (Dirs[dd].buf == NULL) return -1;
	if (Dirs[dd].here >= Dirs[dd].end) return -1;
	k = strlen(Dirs[dd].buf[Dirs[dd].here].name);
	k = k>=n? n-1: k;
	memcpy(b, Dirs[dd].buf[Dirs[dd].here].name, k);
	b[k] = 0;
	Dirs[dd].here++;
	return k;
}

XCELL system_closedir S1(XCELL dd) {
	if (Dirs[dd].buf == 0) return -1;
	free(Dirs[dd].buf);
	Dirs[dd].buf = 0;
	return 0;
}

uint accessmap(char *s) {
	int	bits = 0;

	if (access(s, 4) == 0) bits |= RDOK;
	if (access(s, 2) == 0) bits |= WROK;
	if (access(s, 1) == 0) bits |= EXOK;
	return bits;
}

XCELL system__stat S2(char *name, uchar *b) {
	struct t3x_stat {
		short	dev;
		short	ino;
		short	mode;
		short	nlink;
		short	uid;
		short	gid;
		short	rdev;
		short	size;
		short	ext64;
		short	mt_year;
		short	mt_mon_day;
		short	mt_hr_min;
		short	mt_sec_hs;
	};
	Dir		*d;
	Tm		*tp;

	if ((d = dirstat(name)) == NULL) return -1;
	b[0] = (uchar) 0;
	b[1] = (uchar) 0;
	b[2] = (uchar) 0;
	b[3] = (uchar) 0;
	b[4] = (uchar) 0;
	b[5] = (uchar) (d->mode & DMDIR? ISDIR: 0) | accessmap(name);
	b[6] = (uchar) 0;
	b[7] = (uchar) 1;
	b[8] = (uchar) 0;
	b[9] = (uchar) 0;
	b[10] = (uchar) 0;
	b[11] = (uchar) 0;
	b[12] = (uchar) 0;
	b[13] = (uchar) 0;
	b[14] = (uchar) (d->length >> 8);
	b[15] = (uchar) d->length;
	b[16] = (uchar) (d->length / (64L*1024) >> 8);
	b[17] = (uchar) (d->length / (64L*1024));
	tp = localtime(d->mtime);
	b[18] = (tp->year+1900) / 100;
	b[19] = tp->year % 100;
	b[20] = tp->mon+1;
	b[21] = tp->mday;
	b[22] = tp->hour;
	b[23] = tp->min;
	b[24] = tp->sec;
	b[25] = 0;
	free(d);
	return 0;
}

int _system_wait(void) {
	Waitmsg	*w;
	int		r;

	w = wait();
	if (w == NULL) return -1;
	r = w->msg[0]? 1: 0;
	free(w);
	return r;
}

XCELL system_wait S1(XCELL pid) {
	int	r;

	if (Procd[pid] < 0) return -1;
	r = _system_wait();
	Procd[pid] = -1;
	return r;
}

int _fork(void) {
	return rfork(RFPROC|RFNAMEG|RFENVG|RFFDG|RFREND);
}

XCELL system__spawn S3(char *p, char **av, XCELL w) {
	int	r, i;
	char	*av2[MAXARGS];

	for (i = 0; av[i]; i++)
		if (i >= MAXARGS-1) return -1;
	for (++i; i>0; i--) av2[i] = av[i-1];
	av2[0] = p;
	if ((r = _fork()) < 0) return -1;
	if (r) {
		if (w) return _system_wait();
	}
	else {
		exec(p, av2);
		exits("exec failed");
	}
	return nextprocd(r);
}

XCELL system_fork S0() {
	int	r;

	r = _fork();
	if (!r || r < 0) return r;
	return nextprocd(r);
}

int _kill(int pid, int sig) {
	char	b[100];
	int		fd, k;

	if (sig == 0) {
		sprint(b, "/proc/%d", pid);
		return close(open(b, OREAD));
	}
	sprint(b, "/proc/%d/%s", pid, sig==9? "ctl": "note");
	if ((fd = open(b, OWRITE)) < 0) return -1;
	k = write(fd, "kill", 4);
	close(fd);
	return k == 4? 0: -1;
}

XCELL system_kill S2(XCELL pid, XCELL n) {
	int	r;

	if (Procd[pid] < 0) return -1;
	r = _kill(Procd[pid], n);
	if (r < 0) return r;
	if (n) Procd[pid] = -1;
	return r;
}

void convtime(long t, char *b) {
	int		mlen[13] = { 0, 31,28,31,30,31,30,31,31,30,31,30,31 };
	int		yr, mon, day, hr, min, sec;
	long	dt;

	yr = 1970;
	while (1) {
		dt = (!(yr&3)? 366: 365) * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		yr++;
	}
	mon = 1;
	mlen[2] = !(yr&3)? 29: 28;
	while (mon<12) {
		dt = mlen[mon] * 86400L;
		if (t - dt < 0) break;
		t -= dt;
		mon++;
	}
	day = (int) (t / 86400L + 1);
	t = (int) (t % 86400L);
	hr = (int) (t / 3600);
	t = (int) (t % 3600);
	min = (int) (t / 60);
	sec = (int) (t % 60);
	b[0] = yr / 100;
	b[1] = yr % 100;
	b[2] = mon;
	b[3] = day;
	b[4] = hr;
	b[5] = min;
	b[6] = sec;
	b[7] = 0;
}

XCELL system_time S1(char *b) {
	long	t;

	time(&t);
	convtime(t, b);
	return 0;
}

#ifndef LIBRARY

struct infproc	InfSystem[20] = {
	{ "system__spawn",	system__spawn	},
	{ "system__stat",	system__stat	},
	{ "system_chdir",	system_chdir	},
	{ "system_closedir",	system_closedir	},
	{ "system_dup",		system_dup	},
	{ "system_dup2",	system_dup2	},
	{ "system_fini",	system_fini	},
	{ "system_fork",	system_fork	},
	{ "system_getdir",	system_getdir	},
	{ "system_init",	system_init	},
	{ "system_kill",	system_kill	},
	{ "system_mkdir",	system_mkdir	},
	{ "system_opendir",	system_opendir	},
	{ "system_pipe",	system_pipe	},
	{ "system_rdchk",	system_rdchk	},
	{ "system_readdir",	system_readdir	},
	{ "system_rmdir",	system_rmdir	},
	{ "system_time",	system_time	},
	{ "system_wait",	system_wait	},
	{ NULL,			NULL		}
};

#endif

#else	/* EXT_SYSTEM */

#ifndef LIBRARY
struct infproc	InfSystem[MAXINFPROC] = { { 0L, 0L } };
#endif

#endif	/* EXT_SYSTEM */
