/*
 *	T3X -- Common Runtime Support Routines for T3X R8.
 *	Plan 9 version.
 *	Copyright (C) 1999,2000,2002,2003,2004 Nils M Holm.
 *	See the file LICENSE for conditions of use.
 */

#include <u.h>
#include <libc.h>
#include <stdio.h>

#include "../compiler/txx.h"

#ifdef LIBRARY
	#define USE(X)	USED(X)
#else
	#define USE(X)
#endif

void libinit(void) {
}

#ifdef LIBRARY

int		Argcount;
char	**Argvec;

int main(int argc, char **argv) {
	Argcount = argc;
	Argvec = argv;
	libinit();
	_TXSTART_();
	return 0;
}

#endif

XCELL t3x_bpw S0() {
	USE(__self);
	return BPW;
}

XCELL t3x_memcopy S3(char *d, char *s, XCELL n) {
	USE(__self);
	memmove(d, s, n);
	return 0;
}

XCELL t3x_memcomp S3(char *s1, char *s2, XCELL n) {
	USE(__self);
	return memcmp(s1, s2, n);
}

XCELL t3x_memfill S3(char *p, XCELL c, XCELL n) {
	USE(__self);
	memset(p, c, n);
	return 0;
}

XCELL t3x_memscan S3(char *p, XCELL c, XCELL l) {
	int	i;

	USE(__self);
	for (i=0; i<l; i++)
		if (p[i] == c) return i;
	return -1;
}

XCELL t3x_getarg S3(XCELL i, char *b, XCELL n) {
	int	k;

	USE(__self);
	if (i < 1 || i >= Argcount) {
		b[0] = 0;
		return -1;
	}
	k = strlen(Argvec[i]);
	k = k>=n? n-1: k;
	memcpy(b, Argvec[i], k);
	b[k] = 0;
	return k;
}

XCELL t3x_getenv S3(char *vn, char *b, XCELL n) {
	char	*vv;
	int		k;

	USE(__self);
	vv = getenv(vn);
	if (vv == NULL) {
		b[0] = 0;
		return -1;
	}
	k = strlen(vv);
	k = k>=n? n-1: k;
	memcpy(b, vv, k);
	b[k] = 0;
	return k;
}

XCELL t3x_open S2(char *file, XCELL mode) {
	int	fd;

	USE(__self);
	if (mode == 1) {
		fd = create(file, OWRITE, 0666);
	}
	else if (mode == 0 || mode == 2 || mode == 3) {
		fd = open(file, mode==3? OWRITE: mode==2? ORDWR: OREAD);
	}
	else {
		return -1;
	}
	if (fd > 0 && mode == 3) seek(fd, 0L, 2);
	return fd;
}

#define T_SEEK_SET	0
#define T_SEEK_FWD	1
#define T_SEEK_END	2
#define T_SEEK_BCK	3

XCELL t3x_seek S3(XCELL fd, XUCELL n, XCELL o) {
	USE(__self);
	n &= 0xffff;
	if (o == T_SEEK_SET)
		return (XCELL) seek(fd, (vlong) n * 1L, 0) == -1? -1: 0;
	else if (o == T_SEEK_END)
		return (XCELL) seek(fd, (vlong) n * -1L, 2) == -1? -1: 0;
	else if (o == T_SEEK_FWD)
		return (XCELL) seek(fd, (vlong) n * 1L, 1) == -1? -1: 0;
	else if (o == T_SEEK_BCK)
		return (XCELL) seek(fd, (vlong) n * -1L, 1) == -1? -1: 0;
	else
		return -1;
}

XCELL t3x_close S1(XCELL fd) {
	USE(__self);
	return close(fd);
}

XCELL t3x_read S3(XCELL fd, char *b, XCELL n) {
	USE(__self);
	return read(fd, b, n);
}

XCELL t3x_write S3(XCELL fd, char *b, XCELL n) {
	USE(__self);
	return write(fd, b, n);
}

UCHAR *t3x_newline S1(UCHAR *s) {
	USE(__self);
#ifdef NL_IS_LF
	s[0] = '\n';
	s[1] = 0;
#endif
#ifdef NL_IS_CRLF
	s[0] = '\r';
	s[1] = '\n';
	s[2] = 0;
#endif
#ifdef LIBRARY
	return s;
#else
	return (UCHAR *) (s - Data);
#endif
}

XCELL t3x_remove S1(char *s) {
	USE(__self);
	return remove(s);
}

XCELL t3x_rename S2(char *so, char *sn) {
	Dir		*d;
	int		r;
	char	*p;

	USE(__self);
	p = strrchr(sn, '/');
	if (p != NULL) {
		p++;
		while (*so && sn != p && *so == *sn) {
			so++;
			sn++;
		}
	}
	if (strrchr(sn, '/')) return -1;
	d = dirstat(sn);
	r = d != NULL;
	free(d);
	if (r) return -1;
	if ((d = dirstat(so)) == NULL) return -1;
	d->name = sn;
	r = dirwstat(so, d);
	free(d);
	if (r < 0) return -1;
	d = dirstat(sn);
	r = d == NULL? -1: 0;
	free(d);
	return r;
}

#ifdef LIBRARY

XCELL t3x_cvalist S4(XCELL n, XUCELL bmap, void *ilist, void **olist) {
	USE(__self);
	USE(bmap);
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(void*));
	else
		memcpy(olist, ilist, n*sizeof(void*));
	return 0;
}

#else	/* !LIBRARY */

XCELL t3x_cvalist S4(XCELL n, XUCELL bmap, XUCELL *ilist, void **olist) {
	int	i;
	ULONG	p, o;

	USE(__self);
	/* Align olist at 8-byte word boundary */
	o = (ULONG) olist;
	p = (o + 7) & ~7UL;
	olist = (void**) p;
	if (n < 0) {
		n = -n;
		for (i=0; i<n; i++) ilist[i] = (XUCELL) olist[i];
	}
	else {
		for (i=0; i<n; i++) {
			if (bmap&1)
				olist[i] = &Data[ilist[i]];
			else
				olist[i] = (void *) ilist[i];
			bmap>>=1;
		}
	}
	return (int)(p-o);
}

#endif	/* !LIBRARY */

struct infproc	InfT3X[17] = {
	{ "t3x_bpw",		t3x_bpw		},
	{ "t3x_memcopy",	t3x_memcopy	},
	{ "t3x_memcomp",	t3x_memcomp	},
	{ "t3x_memfill",	t3x_memfill	},
	{ "t3x_memscan",	t3x_memscan	},
	{ "t3x_getarg",		t3x_getarg	},
	{ "t3x_getenv",		t3x_getenv	},
	{ "t3x_open",		t3x_open	},
	{ "t3x_read",		t3x_read	},
	{ "t3x_write",		t3x_write	},
	{ "t3x_newline",	(XCELL(*)()) t3x_newline	},
	{ "t3x_seek",		t3x_seek	},
	{ "t3x_close",		t3x_close	},
	{ "t3x_rename",		t3x_rename	},
	{ "t3x_remove",		t3x_remove	},
	{ "t3x_cvalist",	t3x_cvalist	},
	{ NULL,			NULL		}
};
