! MKTEST2 - Meta test program generator
! Copyright (C) 2004 Nils M Holm
!
! This program generates a program (GENTEST2) which in turn
! generates the TEST2 program. TEST2 tests the expression part
! of a T3X compiler.
!
! MKTEST2 constructs the expressions to be tested.
! GENTEST2 evaluates the expressions, giving a presumably
! correct result. It emits both the result and the expression
! leading to the result.
! TEST2 prints both the result computed by GENTEST2 and the
! result computed by itself.
! Given that TEST2 was generated by a correctly working compiler,
! equal results of TEST2 indicate that the compiler that was
! used to compile TEST2 works correctly, too.
!
! *** BOTH MKTEST2 AND GENTEST2 MUST BE COMPILED BY A COMPILER
! *** THAT IS WORKING CORRECTLY. OTHERWISE, THE RESULT OF TEST2
! *** IS MEANINGLESS.

module mktest2(string, util);

object	str[string], u[util];

var	ID;
var	Factors1, Factors2, Unary_ops, Binary_ops, Prec_grps;

var buf::100;
esc(s) do var i, j;
	i := 0;
	j := 0;
	while (s::i) do
		if (s::i = '\\') do
			buf::j := '\\';
			j := j+1;
		end
		buf::j := s::i;
		i := i+1;
		j := j+1;
	end
	buf::j := 0;
	return buf;
end

head() do
	u.printf("module gentest2(util);\n", 0);
	u.printf("object u[util];\n", 0);
	u.printf("var V, B;\n", 0);
	u.printf("P() return 23;\n", 0);
	u.printf("init() do\n", 0);
	u.printf("V := [ 0, 97, 0 ];\n", 0);
	u.printf("B := packed[ 0, 101, 0 ];\n", 0);
	u.printf("end\n", 0);
end

main() do var i, j;
	u.printf("do\n", 0);
	u.printf("init();\n", 0);
	u.printf(
	"u.printf(\"! This file was automatically generated.\\n\", 0);\n", 0);
	u.printf(
	"u.printf(\"! DO NOT RE-GENERATE THIS FILE.\\n\", 0);\n", 0);
	u.printf(
	"u.printf(\"! See mktest2.t and mkfile for details.\\n\", 0);\n", 0);
	u.printf("u.printf(\"module test2(util);\\n\", 0);\n", 0);
	u.printf("u.printf(\"object u[util];\\n\", 0);\n", 0);
	u.printf("u.printf(\"var V, B;\\n\", 0);\n", 0);
	u.printf("u.printf(\"P() return 23;\\n\", 0);\n", 0);
	u.printf("u.printf(\"init() do\\n\", 0);\n", 0);
	u.printf("u.printf(\"V := [ 0, 97, 0 ];\\n\", 0);\n", 0);
	u.printf("u.printf(\"B := packed[ 0, 101, 0 ];\\n\", 0);\n", 0);
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("unary1();\n", 0);
	u.printf("unary2();\n", 0);
	u.printf("binary1();\n", 0);
	u.printf("binary2();\n", 0);
	for (i=0, 7)
		for (j=0, 7)
			u.printf("precedence%d%d();\n", [(i,j)]);
	u.printf("shortcircuit1();\n", 0);
	u.printf("shortcircuit2();\n", 0);
	u.printf("u.printf(\"do\\n\", 0);\n", 0);
	u.printf("u.printf(\"init();\\n\", 0);\n", 0);
	u.printf("u.printf(\"unary1();\\n\", 0);\n", 0);
	u.printf("u.printf(\"unary2();\\n\", 0);\n", 0);
	u.printf("u.printf(\"binary1();\\n\", 0);\n", 0);
	u.printf("u.printf(\"binary2();\\n\", 0);\n", 0);
	u.printf("u.printf(\"shortcircuit1();\\n\", 0);\n", 0);
	u.printf("u.printf(\"shortcircuit2();\\n\", 0);\n", 0);
	for (i=0, 7)
		for (j=0, 7)
			u.printf("u.printf(\"precedence%d%d();\\n\", 0);\n",
				[(i,j)]);
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("end\n", 0);
end

unary(n, facs) do var i, j;
	i := 0;
	u.printf("unary%d() do\n", [(n)]);
	u.printf("u.printf(\"unary%d() do\\n\", 0);\n", [(n)]);
	while (Unary_ops[i]) do
		j := 0;
		while (facs[j]) do
			u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%s%s)]);\\n\q, [(%s%s)]);\n",
				[(ID, esc(Unary_ops[i]), facs[j],
				 Unary_ops[i], facs[j])]);
			j := j+1;
			ID := ID+1;
		end
		i := i+1;
	end
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("end\n", 0);
end

binary(n, facs) do var i, j;
	i := 0;
	u.printf("binary%d() do\n", [(n)]);
	u.printf("u.printf(\"binary%d() do\\n\", 0);\n", [(n)]);
	while (Binary_ops[i]) do
		j := 0;
		while (facs[j] /\ facs[j+1]) do
			if (\str.comp(facs[j+1], "0") /\
				(\str.comp(Binary_ops[i], "/") \/
				 \str.comp(Binary_ops[i], "./") \/
				 \str.comp(Binary_ops[i], "mod"))
			) do
				j := j+1;
				loop;
			end
			u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%s %s %s)]);\\n\q, [(%s %s %s)]);\n",
				[(ID,
				 facs[j], esc(Binary_ops[i]), facs[j+1],
				 facs[j], Binary_ops[i], facs[j+1])
				]);
			j := j+1;
			ID := ID+1;
		end
		i := i+1;
	end
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("end\n", 0);
end

precedence(n,m) do var i, j, k, escb::100;
	u.printf("precedence%d%d() do\n", [(n,m)]);
	u.printf("u.printf(\"precedence%d%d() do\\n\", 0);\n", [(n,m)]);
	i := 0;
	while (Prec_grps[n][i]) do
		j := 0;
		while (Prec_grps[m][j]) do
			str.copy(escb, esc(Prec_grps[n][i]));
			u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(17 %s 19 %s 23)]);\\n\q, [(17 %s 19 %s 23)]);\n",
				[(ID,
				 escb, esc(Prec_grps[m][j]),
				 Prec_grps[n][i], Prec_grps[m][j])
				]);
			j := j+1;
			ID := ID+1;
		end
		i := i+1;
	end
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("end\n", 0);
end

shortcircuit(n,pt) do var i;
	u.printf("shortcircuit%d() do\n", [(n)]);
	u.printf("u.printf(\"shortcircuit%d() do\\n\", 0);\n", [(n)]);
	i := 0;
	while (pt[i]) do
		u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%c %s %c %s %c)]);\\n\q, [(%c %s %c %s %c)]);\n",
			[(ID,
			 pt[i]::0, "\\\\/", pt[i]::1, "\\\\/", pt[i]::2,
			 pt[i]::0, "\\/", pt[i]::1, "\\/", pt[i]::2,
			]);
		u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%c %s %c %s %c)]);\\n\q, [(%c %s %c %s %c)]);\n",
			[(ID,
			 pt[i]::0, "\\\\/", pt[i]::1, "/\\\\", pt[i]::2,
			 pt[i]::0, "\\/", pt[i]::1, "/\\", pt[i]::2,
			]);
		u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%c %s %c %s %c)]);\\n\q, [(%c %s %c %s %c)]);\n",
			[(ID,
			 pt[i]::0, "/\\\\", pt[i]::1, "\\\\/", pt[i]::2,
			 pt[i]::0, "/\\", pt[i]::1, "\\/", pt[i]::2,
			]);
		u.printf("u.printf(\qu.printf(\\\qEX%3:0d: %%d = %%%%d\\\\n\\\q, [(%c %s %c %s %c)]);\\n\q, [(%c %s %c %s %c)]);\n",
			[(ID,
			 pt[i]::0, "/\\\\", pt[i]::1, "/\\\\", pt[i]::2,
			 pt[i]::0, "/\\", pt[i]::1, "/\\", pt[i]::2,
			]);
		i := i+1;
	end
	u.printf("u.printf(\"end\\n\", 0);\n", 0);
	u.printf("end\n", 0);
end

init() do
	Factors1 := [
		"53", "53", "0", "1", "89",
	0 ];
	Factors2 := [
		"P()", "V[1]", "B::1",
	0 ];
	Unary_ops := [ "~", "\\", "-", 0 ];
	Binary_ops  := [
		"*", "/", ".*", "./", "mod", "+", "-", "&", "|", "^",
		"<", ">", "<=", ">=", ".<", ".>", ".<=", ".>=", "=",
		"\\=",
	0 ];
	Prec_grps := [
		[ "*", "/", ".*", "./", "mod", 0 ],
		[ "+", "-", 0 ],
		[ "&", "|", "^", 0 ],
		[ "<", ">", "<=", ">=", ".<", ".>", ".<=", ".>=", 0 ],
		[ "=", "\\=",0 ],
		[ "/\\", 0 ],
		[ "\\/", 0 ],
	0, ];
	ID := 0;
end

do var i, j;
	init();
	head();
	unary(1, Factors1);
	unary(2, Factors2);
	binary(1, Factors1);
	binary(2, Factors2);
	for (i=0, 7)
		for (j=0,7)
			precedence(i,j);
	shortcircuit(1, [ "012", "021", "102", "120", "201", "210", 0 ]);
	shortcircuit(2, [ "001", "010", "100", "011", "101", "110", 0 ]);
	main();
end

