! TX -- T3X Compiler Driver
! Copyright (C) 1999-2004 Nils M Holm. All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: TX [options] file ...

#classpath "../classes";

module tx(t3x, system, string, iostream);

object	t[t3x], sys[system], str[string];

struct	XSYMREC =	! From TXTRN
	XHHVAL, XHLVAL, XLHVAL, XLLVAL,
	XHHSIZE, XHLSIZE, XLHSIZE, XLLSIZE,
	XHFLGS, XLFLGS;

const	MAXARGLEN	= 128;	! max. command line argument length
const	STRPOOLSIZE	= 4096;	! space reserved for strings
const	MAXNAMES	= 128;	! max. strings
const	MAXCLASSES	= 128;	! max. number of classes to link
const	BUFLEN		= 512;	! stream I/O buffer size
const	LIBMAX		= 256,	! space for library names
	LIBPATHMAX	= 1024;	! space for library paths

! options ...
var	O_bindir,	! location of the compiler binaries
	O_classdir,	! location of class files
	O_libdir,	! location of RT libraries
	O_dpfilter,	! enable dead procedure filter
	O_optimize,	! enable optimizer
	O_userconfig,	! file to laod user config from
	O_verbose,	! verbosity level (0...4)
	O_compile_only,	! disable link phase
	O_make_glue,	! disable local backend (emit glue)
	O_make_class,	! create installable class files
	O_keep_tmpfiles, ! do not remove tmp files
	O_local_class,	! create a local (not installable) class file
	O_output_file,	! output file name (w/o suffix, if any)
	O_tmpin,	! temporary input file
	O_tmpout,	! temporary output file
	O_tmppfx,	! temporary file prefix
	O_tmpdir,	! where to create tmp files (directory path)
	O_keep_dirs,	! keep CLASSDIR when finished
	O_bang_header,	! create #! header (Unix only)
	O_target,	! name of target (0=Tcode)
	O_as_cmd,	! command to assemble CG output
	O_ld_cmd,	! command to link AS output
	O_ldctrl,	! loader control file
	O_exesfx,	! suffix of compiler executables (eg .exe on DOS)
	O_outsfx,	! suffix of executable output files (eg .exe on DOS)
	O_libsfx,	! suffix of extension libraries
	O_gluesfx,	! glue file suffix
	O_native,	! running in native mode
	O_libs,		! libraries to link against
	O_force_bpw;	! force Tcode output with a given word size

var	Basedir;	! compiler base dir ($T3XDIR)

var	Strpool::STRPOOLSIZE, St;	! string pool
var	Names[MAXNAMES], Nt;		! name pool (module names)
var	Keeplist::MAXNAMES;		! files to keep after translation
var	Classes[MAXCLASSES], Ct;	! classes to link
var	Libraries::LIBPATHMAX, Lt;	! library paths, top ptr
object	Ndx[iostream];			! CLASSNDX stream

init() do
	sys.init();
	St := 0;
	Nt := 0;
	Ct := 0;
	Lt := 0;
end

wr(s) t.write(1, s, str.length(s));

ws(s) do var nl::3;
	t.write(1, s, str.length(s));
	wr(t.newline(nl));
end

wm(s) t.write(2, s, str.length(s));

wf(tp, a) do var buf::256;
	return ws(str.format(buf, tp, a));
end

remove(path, force) do
	if (O_keep_tmpfiles /\ \force) do
		if (O_verbose > 1) wf("(keep) %s", [(path)]);
		return;
	end
	if (O_verbose > 1) wf("(delete) %s", [(path)]);
	return t.remove(path);
end

rmdirfiles()
	if (\O_keep_dirs) do
		remove("CLASSDIR", 0);
		remove("LIBDIR", 0);
	end

removeall() do var buf::MAXARGLEN+4;
	rmdirfiles();
	if (O_tmpin) remove(O_tmpin, 0);
	if (O_tmpout) remove(O_tmpout, 0);
	if (O_tmppfx) do
		remove(str.format(buf, "%s%s", [(O_tmppfx, O_gluesfx)]), 0);
		remove(str.format(buf, "%s%s", [(O_tmppfx, O_libsfx)]), 0);
	end
end

error(m, n) do var nl::4;
	wm("TX: ");
	wm(m);
	if (n) do
		wm(": ");
		wm(n);
	end
	wm(t.newline(nl));
	removeall();
	halt 1;
end

savestr(s) do var k;
	k := str.length(s)+1;
	if (k+St >= STRPOOLSIZE) error("out of string space", 0);
	t.memcopy(@Strpool::St, s, k);
	St := St+k;
	return @Strpool::(St-k);
end

defaults() do var tmp::MAXARGLEN+1, tmp2::MAXARGLEN+10;
	ie (t.getenv("T3XDIR", tmp, MAXARGLEN) < 0 /\
		t.getenv("t3xdir", tmp, MAXARGLEN) < 0
	) do
		O_bindir := ".";
		O_classdir := ".";
		O_libdir := ".";
		Basedir := ".";
	end
	else do
		O_bindir := savestr(str.format(tmp2, "%S/bin", [(tmp)]));
		O_classdir := savestr(str.format(tmp2, "%S/classes",
			[(tmp)]));
		O_libdir := savestr(str.format(tmp2, "%S/lib", [(tmp)]));
		Basedir := savestr(tmp);
	end
	O_dpfilter := 1;
	O_optimize := 1;
	O_userconfig := "~/.txrc";
	O_verbose := 0;
	O_compile_only := 0;
	O_make_glue := 0;
	O_make_class := 0;
	O_keep_tmpfiles := 0;
	O_local_class := 0;
	O_output_file := 0;
	O_tmpin := 0;
	O_tmpout := 0;
	O_tmpdir := 0;
	O_keep_dirs := 0;
	O_bang_header := 0;
	O_target := 0;
	O_as_cmd := "/usr/bin/as -o %S %S";
	O_ld_cmd := "/usr/bin/cc -o %S %S";
	O_libs := "-ltermcap";
	O_ldctrl := "";
	O_exesfx := 0;
	O_outsfx := 0;
	O_libsfx := ".o";
	O_gluesfx := ".s";
	O_native := 1;
	O_force_bpw := -1;
end

savename(s, keep) do
	if (Nt >= MAXNAMES) error("too many file names", 0);
	Names[Nt] := savestr(s);
	Keeplist::Nt := keep;
	Nt := Nt+1;
end

saveclass(s) do var i;
	for (i=0, Ct)
		if (\str.comp(Classes[i], s))
			return;
	if (Ct >= MAXCLASSES) error("too many classes", 0);
	Classes[Ct] := savestr(s);
	Ct := Ct+1;
end

license() do
ws(
"T3X -- A Compiler for the Minimum Procedural Language T3X");
ws(
"Copyright (C) 1996-2004 Nils M Holm.  All rights reserved.");
ws("");
ws(
"Redistribution and use in source and binary forms, with or without");
ws(
"modification, are permitted provided that the following conditions");
ws(
"are met:");
ws(
"1. Redistributions of source code must retain the above copyright");
ws(
"   notice, this list of conditions and the following disclaimer.");
ws(
"2. Redistributions in binary form must reproduce the above copyright");
ws(
"   notice, this list of conditions and the following disclaimer in the");
ws(
"   documentation and/or other materials provided with the distribution.");
ws("");
ws(
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
ws(
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
ws(
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
ws(
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
ws(
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
ws(
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
ws(
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
ws(
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
ws(
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
ws(
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
ws(
"SUCH DAMAGE.");
ws("");
halt 0;
end

usage(hlt) do
	wr("Usage: tx [-V] [-h?] [-B dir] [-C dir]");
	ws(" [-L dir] [-A arch]");
	wr("\t [-U file] [-m target] [-o file] [-DGNOceiklpv]");
	ws(" file ...");
	if (hlt) halt 1;
end

listtargets() do var first, i, targets, tp::MAXARGLEN*2;
	targets := [
		"386", "386a", "386c", "386e", "86",
		"86s", "alpha", "uc", "exe", "l",
		0
	];
	ws("");
	wr("Installed targets: ");
	first := 1;
	i := 0; while (targets[i]) do
		str.format(tp, "%S/txcg%S%S", [(O_bindir),
			(targets[i]), (O_exesfx-> O_exesfx: "")]);
		if (t.close(t.open(tp, T3X.OREAD)) = 0) do
			if (\first) wr(", ");
			first := 0;
			wr(targets[i]);
		end
		i := i+1;
	end
	ws("");
	ws("");
end

longusage() do
	ws("");
	usage(0);
	ws("");
	ws("-A name   use architecture setup 'name'");
	ws("-A dos    generate DOS EXE file");
	ws("-A static generate static executable (where available)");
	ws("-B dir    search binaries in 'dir'");
	ws("-C dir    search classes in 'dir'");
	ws("-D        enable dead procedure filter (DPF)");
	ws("-G        generate glue (assembly language)");
	ws("-L dir    search libraries in 'dir'");
	ws("-N        disable optimizer and DPF");
	ws("-O        enable tcode optimization");
	ws("-U file   load user config from 'file'");
	ws("-V        display the conditions of use");
	ws("-c        compile only (don't load)");
	ws("-e        create #! header (Unix only)");
	ws("-h        display this summary (also -?)");
	ws("-i        create installable class files");
	ws("-k        keep intermediate files");
	ws("-l        create local (non-installable) class files");
	ws("-m type   generate native code for target 'type' (-m? = list)");
	ws("-ml       generate native code for local target");
	ws("-o file   direct output to 'file'");
	ws("-p        preserve CLASSDIR and LIBDIR");
	ws("-v        verbose operation (more v's = more output)");
	ws("");
	halt 0;
end

flag(s, i, p) do var c;
	c := s::(i+1);
	if (c = '0') do
		p[0] := 0;
		return i+1;
	end
	if (c = '1') do
		p[0] := 1;
		return i+1;
	end
	p[0] := 1;
	return i;
end

yesno(x) return x-> "YES": "NO";

optarg(s) return s-> s: "(none)";

printargs() do
	wf("verbosity level = %D", [(O_verbose)]);
	wf("running in native mode = %S", [(yesno(O_native))]);
	wf("userconfig = %S", [(optarg(O_userconfig))]);
	wf("bindir = %S", [(O_bindir)]);
	wf("classdir = %S", [(O_classdir)]);
	wf("libdir = %S", [(O_libdir)]);
	wf("exe suffix = %S", [(optarg(O_exesfx))]);
	wf("output suffix = %S", [(optarg(O_outsfx))]);
	wf("lib suffix = %S", [(optarg(O_libsfx))]);
	wf("glue suffix = %S", [(optarg(O_gluesfx))]);
	wf("output file = %S", [(O_output_file-> O_output_file:
		"(automatic)")]);
	wf("tmp prefix = %S", [(O_tmppfx)]);
	wf("dead procedure filter enabled = %S",
		[(yesno(O_dpfilter))]);
	wf("optimization enabled = %S", [(yesno(O_optimize))]);
	wf("target = %S", [(O_target-> O_target: "(Tcode)")]);
	wf("compile only = %S", [(yesno(O_compile_only))]);
	wf("make class = %S", [(yesno(O_make_class))]);
	wf("make local class = %S", [(yesno(O_local_class))]);
	wf("make glue = %S", [(yesno(O_make_glue))]);
	wf("keep class/lib directories = %S", [(yesno(O_keep_dirs))]);
	wf("keep tmpfiles = %S", [(yesno(O_keep_tmpfiles))]);
	wf("assembler command = %S", [(optarg(O_as_cmd))]);
	wf("loader command = %S", [(optarg(O_ld_cmd))]);
	if (O_verbose > 3) wf("loader control file = %S",
		[(optarg(O_ldctrl))]);
	wf("libraries = %S", [(optarg(O_libs))]);
end

source_p(s) do var i;
	i := str.rscan(s, '.');
	if (i < 0) return 0;
	return str.comp(@s::i, ".t") = 0;
end

exists(name) return t.close(t.open(name, T3X.OREAD)) = 0;

copy(in, out, outname) do
	const	CPBUFLEN = 1024;
	var	cpbuf::CPBUFLEN;
	var	k;

	k := t.read(in, cpbuf, CPBUFLEN);
	while (k > 0) do
		if (t.write(out, cpbuf, k) \= k)
			error("file write error", outname);
		k := t.read(in, cpbuf, CPBUFLEN);
	end
end

rename(old, new, cp) do var buf::MAXARGLEN*2+5, in, out;
	if (O_verbose > 1) wf("(rename) %s -> %s", [(old), (new)]);
	t.remove(new);
	if (t.rename(old, new) < 0) do
		ie (cp) do
			if (O_verbose > 1) wf(
				"(copying instead) %s -> %s", [(old), (new)]);
			in := t.open(old, T3X.OREAD);
			if (in < 0) error("could not open", old);
			out := t.open(new, T3X.OWRITE);
			if (out < 0) error("could not create", old);
			copy(in, out, new);
			t.close(in);
			t.close(out);
			t.remove(old);
		end
		else do
			str.format(buf, "%S -> %S", [(old), (new)]);
			error("failed to rename", buf);
		end
	end
end

! run command with arguments
run(cmd, ext) do
	const	NARG = 18;		! max. args+1
	var	args[NARG];		! argument list
	var	txxpath::MAXARGLEN+5;	! path to txx
	var	p, i, q;

	i := 0;
	p := 0;
	if (\(O_native \/ ext)) do
		str.format(txxpath, "%S/txx", [(O_bindir)]);
		args[p] := txxpath;
		p := 1;
	end
	while (1) do
		while (cmd::i = '\s' \/ cmd::i = '\t')
			i := i+1;
		if (\cmd::i) leave;
		if (p >= NARG-1) error("too many aruments", cmd);
		args[p] := @cmd::i;
		p := p+1;
		while (cmd::i /\ cmd::i \= '\s' /\ cmd::i \= '\t')
			i := i+1;
		if (cmd::i) do
			cmd::i := 0;
			i := i+1;
		end
	end
	args[p] := 0;
	if (O_verbose) do var nl::3;
		for (i=0, p) do
			if (i) wr("\s");
			wr(args[i]);
		end
		wr(t.newline(nl));
	end
	if (sys.spawn(args[0], @args[1], system.SPAWN_WAIT))
		error("command returned failure", args[0]);
end

run_with_redirection(prog, opts, infile, outfile, append, ext) do
	const	NARG = 17;		! max. args+1
	var	old0, old1, in, out;
	var	trnpath::MAXARGLEN+6;
	var	auxpath::MAXARGLEN+5;
	var	rc, k, i;
	var	xopts::17;

	! Paths beginning with '!' are absolute. Just strip the '!'
	ie (prog::0 = '!') do
		str.copy(trnpath, @prog::1);
	end
	! Paths beginning with '/' are also absolute, but an exe
	! suffix may be appended
	else ie (prog::0 = '/') do
		ie (O_exesfx)
			str.format(trnpath, "%S%S",
				[(prog), (O_exesfx)]);
		else
			str.copy(trnpath, prog);
	end
	! else assume the program is in O_bindir
	else do
		str.format(trnpath, "%S/%S%S",
			[(O_bindir), (prog), (O_exesfx-> O_exesfx: "")]);
	end
	if (\(O_native \/ ext)) do
		i := 0;
		while (opts[i]) do
			if (i >= NARG-2) error("too many aruments", prog);
			xopts[i+1] := opts[i];
			i := i+1;
		end
		xopts[i+1] := 0;
		str.copy(auxpath, trnpath);
		xopts[0] := auxpath;
		str.format(trnpath, "%S/txx", [(O_bindir)]);
		opts := xopts;
	end
	if (O_verbose) do
		wr(trnpath);
		i := 0;
		while (opts[i]) do
			wr("\s");
			wr(opts[i]);
			i := i+1;
		end
		wr(": ");
		wf("%S -> %S", [(infile),
			(outfile-> outfile: "(null)")]);
	end
	in := t.open(infile, T3X.OREAD);
	if (in < 0) error("cannot open", infile);
	ie (append) do
		out := t.open(outfile, T3X.ORDWR);
		if (out < 0 \/ t.seek(out, 0, T3X.SEEK_END) < 0) do
			t.close(in);
			! 'out' is closed when TX exits
			error("cannot append", outfile);
		end
	end
	else if (outfile) do
		out := t.open(outfile, T3X.OWRITE);
		if (out < 0) do
			t.close(in);
			error("cannot create", outfile);
		end
	end
	old0 := sys.dup(0);
	if (outfile) old1 := sys.dup(1);
	if (old0 < 0 \/ sys.dup2(in, 0) < 0)
		error("redirection failed (this should not happen)", 0);
	if (outfile /\ (old1 < 0 \/ sys.dup2(out, 1) < 0))
		error("redirection failed (this should not happen)", 0);
	rc := sys.spawn(trnpath, opts, system.SPAWN_WAIT);
	t.close(in);
	if (outfile) t.close(out);
	if (sys.dup2(old0, 0) < 0 \/ t.close(old0) < 0)
		error("redirection cleanup failed (oops!)", 0);
	if (outfile /\ (sys.dup2(old1, 1) < 0 \/ t.close(old1) < 0))
		error("redirection cleanup failed (oops!)", 0);
	if (rc = 255) error("program not found", trnpath);
	if (rc) do
		k := str.length(infile);
		if (k > 4 /\ \str.comp(@infile::(k-4), ".tmp"))
			remove(infile, 0);
		if (outfile) remove(outfile, 0);
		error("error in compilation -- terminating", 0);
	end
end

! copy basename of a to b
basename(a, b) do var i;
	str.copy(b, a);
	i := str.rscan(b, '.');
	if (i = %1) error("file name has no suffix", a);
	b::i := 0;
	return b;
end

forcebpw(s) do var in, out, buf::20;
	if (O_verbose > 1)
		wf("(force bpw=%d) %s -> %s", [(O_force_bpw), (s), (O_tmpin)]);
	in := t.open(s, T3X.OREAD);
	if (in < 0) error("cannot open file", s);
	out := t.open(O_tmpin, T3X.OWRITE);
	if (in < 0) error("cannot create file", s);
	t.write(out, str.format(buf, "#BPW %d; ", [(O_force_bpw)]),
		str.length(buf));
	copy(in, out, O_tmpin);
	t.close(in);
	t.close(out);
	return O_tmpin;
end

compile(s) do
	var	buf::MAXARGLEN+11, tname::MAXARGLEN*2+6;
	var	i;

	basename(s, buf);
	i := str.rscan(buf, '/');
	str.format(tname, "%S/%S.tcode", [(O_tmpdir), (@buf::(i+1))]);
	if (O_force_bpw \= -1 /\ O_force_bpw \= t.bpw())
		s := forcebpw(s);
	run_with_redirection("txtrn", [0], s, tname, 0, 0);
	savename(tname, 0);
end

copyclasses(outfile) do
	var	modname::MAXARGLEN*2+1, i;
	var	mf;

	for (i=0, Ct) do
		! The T3X core class has no Tcode text file
		str.format(modname, "./%S.tcode", [(Classes[i])]);
		mf := t.open(modname, T3X.OREAD);
		if (mf < 0) do
			! first attempt to find the module in the CWD
			str.format(modname, "./%S.cls", [(Classes[i])]);
			mf := t.open(modname, T3X.OREAD);
			if (mf < 0) do
				! and then in O_classdir
				str.format(modname, "%S/%S.cls",
					[(O_classdir), (Classes[i])]);
				mf := t.open(modname, T3X.OREAD);
				if (mf < 0)
					error("cannot open class",
						modname);
			end
		end
		if (O_verbose > 1) wf("(load) %S", [(modname)]);
		copy(mf, outfile, modname);
		t.close(mf);
	end
end

! resolve class dependencies defined in CLASSNDX
resolvcls(clss) do
	var	line::MAXARGLEN+1, name::MAXARGLEN+1;
	var	k, j;

	Ndx.reset();
	Ndx.move(0, IOSTREAM.SEEK_SET);
	k := Ndx.reads(line, MAXARGLEN);
	while (\Ndx.eof()) do
		line::(k-1) := 0;
		if (line::(k-2) = '\r') line::(k-2) := 0;
		j := str.length(clss);
		if (\t.memcomp(clss, line, j) /\ line::j = ':') do
			while (	str.parse(@line::j, "%[:,]%S,",
				[(name)]) = 1
			) do
				j := j+str.length(name)+1;
				resolvcls(name);
				saveclass(name);
			end
			leave;
		end
		k := Ndx.reads(line, MAXARGLEN);
	end
end

! resolve class dependencies defined in CLASSNDX
resolve(clss) do
	var	nbuf::BUFLEN;
	var	path::MAXARGLEN*2;

	str.format(path, "%S/CLASSNDX", [(O_classdir)]);
	if (Ndx.open(path, nbuf, BUFLEN, iostream.FREAD) = %1)
		return;
	resolvcls(clss);
	Ndx.close();
end

! load all required classes into outfile
loadclasses(outfile) do
	object	cdir[iostream];
	var	cdbuf::BUFLEN;
	var	magic::4;
	var	corrdir;
	var	nb::4, n;
	var	c;
	var	modname::MAXARGLEN*2+1, i, j, k, p;

	corrdir := "corrupt CLASSDIR";
	! extract the required modules from CLASSDIR
	if (cdir.open("CLASSDIR", cdbuf, BUFLEN, iostream.FREAD) = %1)
		return;
	if (cdir.read(magic, 4) \= 4) error(corrdir, 0);
	if (t.memcomp(magic, "CDR", 3)) error(corrdir, 0);
	if (cdir.read(nb, 2) \= 2) error(corrdir, 0);
	n := nb::0 << 8 | nb::1;
	! go to the name pool of CLASSDIR (skip over the symbol table)
	if (cdir.move(n.*XSYMREC+6, iostream.SEEK_SET) < 0)
		error(corrdir, 0);
	c := cdir.rdch();
	if (cdir.eof()) error(corrdir, 0);
	while (\cdir.eof()) do
		! module names begin with a '*'
		if (c = '*') do
			c := cdir.rdch();
			i := 0;
			! the next '*' is the end of the module name
			while (c /\ c \= '*' /\ \cdir.eof()) do
				if (i >= MAXARGLEN) do
					modname::i := 0;
					error("module name too long",
						modname);
				end
				modname::i := c; i := i+1;
				c := cdir.rdch();
			end
			modname::i := 0;
			! check if the module is already in the load list
			for (j=0, Nt) do
				p := Names[j];
				p := @p::(str.rscan(p, '/')+1);
				if (	\t.memcomp(p, modname, i) /\
					p::i = '.'
				)
					leave;
			end
			if (j<Nt) do	! if so, skip it
				c := cdir.rdch();
				loop;
			end
			resolve(modname);
			if (j >= Nt) saveclass(modname);
		end
		c := cdir.rdch();
	end
	cdir.close();
	copyclasses(outfile);
end

! load all classes and user supplied modules into a file
! and then run TXLD on that file
loadall() do
	var	i;
	var	outfile;
	var	tmpname::MAXARGLEN+5, tmpfile;
	var	infile;
	var	ofile;

	outfile := t.open(O_tmpin, T3X.OWRITE);
	if (outfile < 0)
		error("cannot create loader input file", O_tmpin);
	loadclasses(outfile);
	for (i=0, Nt) do
		infile := t.open(Names[i], T3X.OREAD);
		if (infile < 0) error("cannot reopen", Names[i]);
		if (O_verbose > 1) wf("(load) %S", [(Names[i])]);
		copy(infile, outfile, O_tmpin);
		t.close(infile);
		if (\Keeplist::i) remove(Names[i], 0);
	end
	t.close(outfile);
	run_with_redirection("txld", [0], O_tmpin, O_tmpout, 0, 0);
	remove(O_tmpin, 0);
end

optimize(prog, opts) do
	rename(O_tmpout, O_tmpin, 0);
	run_with_redirection(prog, opts, O_tmpin, O_tmpout, 0, 0);
	remove(O_tmpin, 0);
end

rdlibdir() do var buf::LIBMAX+1, ld, k, i, j, kp, ks, n, msg;
	msg := "too many libraries";
	ld := t.open("LIBDIR", T3X.OREAD);
	if (ld = -1) return;
	n := t.read(ld, buf, LIBMAX+1);
	t.close(ld);
	if (n > LIBMAX) error(msg, 0);
	kp := str.length(O_libdir);
	ks := O_libsfx-> str.length(O_libsfx): 0;
	for (i=0, n) if (buf::i = '\n' \/ buf::i = '\r') buf::i := 0;
	i := 0;
	while (i < n) do
		k := str.length(@buf::i);
		j := Lt;
		if (Lt + kp + ks + k + 2 >= LIBPATHMAX) error(msg, 0);
		t.memcopy(@Libraries::Lt, @buf::i, k);
		i := i + k + 1;
		Lt := Lt + k + 1;
		if (ks) do
			t.memcopy(@Libraries::(Lt-1), O_libsfx, ks);
			Lt := Lt+ks;
		end
		if (\exists(@Libraries::j)) do
			Lt := Lt - k - 1 - ks;
			i := i - k - 1;
			t.memcopy(@Libraries::Lt, O_libdir, kp);
			Libraries::(Lt+kp) := '/';
			t.memcopy(@Libraries::(Lt+kp+1), @buf::i, k);
			i := i + k+1;
			Lt := Lt + kp + k + 2;
			if (ks) do
				t.memcopy(@Libraries::(Lt-1), O_libsfx, ks);
				Lt := Lt+ks;
			end
		end
		while (buf::i = 0 /\ i < n) i := i+1;
		if (O_verbose > 1) wf("(interface) %S", [(@Libraries::j)]);
		Libraries::(Lt-1) := '\s';
	end
	Lt := Lt-1;
	Libraries::Lt := 0;
end

tcode_to_native(m) do
	var	cgprog::MAXARGLEN+10,	! code generator path
		s_path::MAXARGLEN+10,	! assembly language file name
		o_path::MAXARGLEN+10,	! object file name
		tmp::MAXARGLEN+LIBPATHMAX+1,
		command::MAXARGLEN*2+1;
	var	k, ctlfd;
	var	cgargs[16], cgc;

	! The local target (-ml) is magic, since on a Tcode machine,
	! it means not to run any generator at all.
	if (O_native = 0 /\ \str.comp(m, "l")) do
		str.format(tmp, "%S%S", [(O_output_file),
			(O_outsfx-> O_outsfx: "")]);
		remove(tmp, 1);
		rename(O_tmpout, tmp, 1);
		return;
	end
	cgargs[0] := 0;
	cgc := 0;
	if (O_verbose) do
		cgargs[0] := "-v";
		cgargs[1] := 0;
		cgc := cgc+1;
	end
	str.format(cgprog, "%S/txcg%S%S",
		[(O_bindir), (m), (O_exesfx-> O_exesfx: "")]);
	if (t.close(t.open(cgprog, 0)) < 0)
		error("unsupported machine type", m);
	str.format(cgprog, "txcg%S", [(m)]);
	if (\O_as_cmd /\ \O_ld_cmd) do
		! There is neither an assembler nor a linker command,
		! so assume that the generator produres executable
		! output.
		! If an output suffix is given, append it to the name
		! of the output file.
		str.format(tmp, "%S%S", [(O_output_file),
			(O_outsfx-> O_outsfx: "")]);
		run_with_redirection(cgprog, cgargs, O_tmpout, tmp, 0, 0);
		remove(O_tmpout, 0);
		return;
	end
	str.format(s_path, "%S%S", [(O_tmppfx, O_gluesfx)]);
	run_with_redirection(cgprog, cgargs, O_tmpout, s_path, 0, 0);
	remove(O_tmpout, 0);
	if (O_make_glue) do
		! just making a glue file, stop here
		str.format(o_path, "%S%S", [(O_output_file, O_gluesfx)]);
		remove(o_path, 1);
		rename(s_path, o_path, 1);
		return;
	end
	str.format(o_path, "%S%S", [(O_tmppfx, O_libsfx)]);
	! if there is an assembler, assemble the CG output
	ie (O_as_cmd) do
		ie (str.find(O_as_cmd, "%S") \= -1) do
			str.format(command, O_as_cmd, [(o_path), (s_path)]);
			run(command, 1);
		end
		else do
			run_with_redirection(O_as_cmd, [0], s_path, o_path,
				0, 1);
		end
		remove(s_path, 0);
	end
	else do
		remove(o_path, 1);
		rename(s_path, o_path, 0);
	end
	! if there is a linker, link the output of the assembler or CG
	ie (O_ld_cmd) do
		! read LIBDIR
		rdlibdir();
		! build library paths
		t.memcopy(tmp, Libraries, Lt+1);
		if (O_libs) do
			tmp::Lt := '\s';
			str.format(@tmp::(Lt+1), O_libs, [(O_libdir)]);
		end
		! if the linker command contains %S, assume that
		! is is controlled using command line arguments
		ie (str.find(O_ld_cmd, "%S") \= -1) do
			str.format(command, O_ld_cmd,
				[(O_output_file), (o_path)]);
			k := str.length(command);
			command::k := '\s';
			str.copy(@command::(k+1), tmp);
			run(command, 1);
		end
		! else generate a linker control file and connect
		! the linker input to that file
		else do
			str.format(command, O_ldctrl,
				[(O_output_file), (o_path)]);
			k := str.length(command);
			str.copy(@command::k, tmp);
			ctlfd := t.open("__ldctrl", T3X.OWRITE);
			if (ctlfd < 0)
				error("failed to create __ldctrl file", 0);
			k := str.length(command);
			command::k := '\n'; k := k+1;
			if (t.write(ctlfd, command, k) \= k)
				error("failed to write __ldctrl file", 0);
			t.close(ctlfd);
			run_with_redirection(O_ld_cmd, [0], "__ldctrl",
				0, 0, 1);
			remove("__ldctrl", 0);
		end
		remove(o_path, 0);
	end
	else do
		! no linker? just rename the CG/assembler output file
		str.format(tmp, "%S%S", [(O_output_file),
			(O_outsfx-> O_outsfx: "")]);
		remove(tmp, 1);
		rename(o_path, tmp, 1);
	end
end

make_class() do
	var	cmd::MAXARGLEN+10,
		cls_path::MAXARGLEN+1,
		cdir_path::MAXARGLEN+1;

	! append a .cls suffix to the Tcode text file
	str.format(cls_path, "%S.cls", [(O_output_file)]);
	str.format(cdir_path, "%S.dir", [(O_output_file)]);
	remove(cls_path, 1);
	rename(Names[Nt-1], cls_path, 1);
	ie (O_local_class) do
		! rename common class directory
		remove(cdir_path, 1);
		rename("CLASSDIR", cdir_path, 1);
	end
	else do
		! extract the class directory
		str.format(cmd, "%S/txcm%S -qfx %S CLASSDIR", [(O_bindir),
				(O_exesfx-> O_exesfx: ""), (O_output_file)]);
		run(cmd, 0);
		! generate a CLASSNDX entry
		str.format(cmd, "%S/txcm%S -o %S.ndx -n %S CLASSDIR",
			[(O_bindir), (O_exesfx-> O_exesfx: ""), (O_output_file),
			(O_output_file)]);
		run(cmd, 0);
	end
end

make_bang_header(infile, outfile) do
	var	in, out;
	var	xhdr::MAXARGLEN+10;

	if (O_verbose > 1)
		wf("(make #!-file) %s -> %s", [(infile), (outfile)]);
	in := t.open(infile, T3X.OREAD);
	if (in < 0) error("cannot reopen", infile);
	out := t.open(outfile, T3X.OWRITE);
	if (out < 0) error("cannot create", outfile);
	str.format(xhdr, "#! %S/txx\n", [(O_bindir)]);
	t.write(out, xhdr, str.length(xhdr));
	copy(in, out, outfile);
	remove(infile, 0);
	t.close(in);
	t.close(out);
end

! get the remaining lines of a multi-line option
! and append them to the  specified buffer
gettherest(buf, in, lim) do var k, j, i, add::MAXARGLEN+1;
	k := str.length(buf);
	while (1) do
		if (k /\ buf::(k-1) = '\n') k := k-1;
		if (k /\ buf::(k-1) = '\r') k := k-1;
		if (k /\ buf::(k-1) \= '\\') leave;
		k := k-1;
		j := send(in, iostream, reads(add, MAXARGLEN));
		if (j /\ add::(j-1) = '\n') j := j-1;
		if (j /\ add::(j-1) = '\r') j := j-1;
		add::j := 0;
		for (i=0, j) if (add::i \= '\s' /\ add::i \= '\t') leave;
		if (j+k-i+1 >= lim)
			error("config: string too long", 0);
		str.copy(@buf::k, @add::i);
		k := k+j-i;
	end
	return buf;
end

saveopt(s) do
	if (\str.comp(s, "(none)")) return 0;
	return savestr(s);
end

configure(path, reqd) do
	object	conf[iostream];
	var	buffer::BUFLEN;
	var	line::MAXARGLEN+1;
	var	tmp::MAXARGLEN*3+1;
	var	k, dummy;

	ie (path) do
		!
	end
	else ie (O_userconfig) do
		ie (O_userconfig::0 = '~') do
			if (	t.getenv("HOME", tmp, MAXARGLEN) < 0 /\
				t.getenv("home", tmp, MAXARGLEN) < 0
			)
				error("cannot determine $HOME", 0);
			k := str.length(tmp);
			str.copy(@tmp::k, @O_userconfig::1);
			path := tmp;
		end
		else do
			path := O_userconfig;
		end
	end
	else do
		return;
	end
	if (O_verbose > 1) do
		wf("(configure) %S", [(path)]);
	end
	if (conf.open(path, buffer, BUFLEN, iostream.FREAD) = %1) do
		if (reqd) error("failed to open config file", path);
		if (\path) O_userconfig := 0;
		return;
	end
	k := conf.reads(line, MAXARGLEN);
	while (\conf.eof()) do
		line::(k-1) := 0;
		if (k > 1 /\ line::(k-2) = '\r') line::(k-2) := 0;
		if (\line::0 \/ line::0 = '#') do
			k := conf.reads(line, MAXARGLEN);
			loop;
		end
		ie (	line::0 = 'a' /\
			str.parse(line, "as_cmd%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_as_cmd := saveopt(tmp);
		end
		else ie (line::0 = 'b' /\
			str.parse(line, "bindir%W=%W%S", [(@tmp)]) = 1
		)
			O_bindir := saveopt(tmp);
		else ie (line::0 = 'c' /\
			str.parse(line, "classdir%W=%W%S", [(@tmp)]) = 1
		)
			O_classdir := saveopt(tmp);
		else ie (line::0 = 'd' /\
			str.parse(line, "dpfilter%W=%W%D",
			[(@O_dpfilter)]) = 1
		)
			;
		else ie (line::0 = 'e' /\
			str.parse(line, "exesfx%W=%W%S", [(@tmp)]) = 1
		)
			O_exesfx := saveopt(tmp);
		else ie (line::0 = 'f' /\
			str.parse(line, "force_bpw%W=%W%D",
			[(@O_force_bpw)]) = 1
		)
			;
		else ie (line::0 = 'g' /\
			str.parse(line, "gluesfx%W=%W%S", [(@tmp)]) = 1
		)
			O_gluesfx := saveopt(tmp);
		else ie (line::0 = 'l' /\
			str.parse(line, "libdir%W=%W%S", [(@tmp)]) = 1
		)
			O_libdir := saveopt(tmp);
		else ie (line::0 = 'l' /\
			str.parse(line, "ld_cmd%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_ld_cmd := saveopt(tmp);
		end
		else ie (line::0 = 'l' /\
			str.parse(line, "ldctrl%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_ldctrl := saveopt(tmp);
			if (O_ldctrl) str.xlate(O_ldctrl, ';', '\n');
		end
		else ie (line::0 = 'l' /\
			str.parse(line, "libs%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_libs := saveopt(tmp);
		end
		else ie (line::0 = 'l' /\
			str.parse(line, "libsfx%W=%W%S", [(@tmp)]) = 1
		)
			O_libsfx := saveopt(tmp);
		else ie (line::0 = 'n' /\
			str.parse(line, "native%W=%W%D", [(@O_native)]) = 1
		)
			;
		else ie (line::0 = 'o' /\
			str.parse(line, "optimize%W=%W%D",
			[(@O_optimize)]) = 1
		)
			;
		else ie (line::0 = 'o' /\
			str.parse(line, "outsfx%W=%W%S", [(@tmp)]) = 1
		)
			O_outsfx := saveopt(tmp);
		else ie (line::0 = 't' /\
			str.parse(line, "target%W=%W%S", [(@tmp)]) = 1
		)
			O_target := saveopt(tmp);
		else
			error("unknown option in user config", line);
		k := conf.reads(line, MAXARGLEN);
	end
	conf.close();
end

loadconf(id, reqd) do var buf::MAXARGLEN*2+1;
	str.format(buf, "%S/%S.txrc", [(Basedir), (id)]);
	if (reqd) do
		if (t.close(t.open(buf, T3X.OREAD)) < 0)
			error("unsupported architecture", id);
	end
	configure(buf, reqd);
end

mktmpfiles() do
	var	buf::MAXARGLEN*2+17;
	var	tb::8;
	var	fd;

	ie (O_tmpdir)
		;
	else ie (t.getenv("TMP", buf, MAXARGLEN+1) > 0)
		O_tmpdir := savestr(buf);
	else ie (t.getenv("tmp", buf, MAXARGLEN+1) > 0)
		O_tmpdir := savestr(buf);
	else
		O_tmpdir := ".";
	sys.time(tb);
	str.format(buf, "%S/tx%2:0D%2:0D%2:0D.in",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmpin := savestr(buf);
	str.format(buf, "%S/tx%2:0D%2:0D%2:0D.out",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmpout := savestr(buf);
	str.format(buf, "%S/tx%2:0D%2:0D%2:0D",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmppfx := savestr(buf);
	fd := t.open(O_tmpin, T3X.OWRITE);
	ie (fd < 0) do
		error("cannot create tmp file", O_tmpin);
	end
	else do
		t.close(fd);
		t.remove(O_tmpin);
	end
end

! extract file argument from command line
! + accept -o file as well as -ofile
! + update the pointers
file(pi, s, pj, pk, p) do var i, j, k;
	i := pi[0];
	j := pj[0];
	k := pk[0];
	ie (s::(j+1)) do
		j := j+1;
	end
	else do
		i := i+1;
		k := t.getarg(i, s, MAXARGLEN);
		if (k < 0) usage(1);
		j := 0;
	end
	p[0] := savestr(@s::j);
	pk[0] := k;
	pj[0] := str.length(s)-1;
	pi[0] := i;
end

do
	var	abuf::MAXARGLEN*2+1, k, i, j;
	var	altarc, dflname;

	init();
	defaults();
	i := 1;
	loadconf("system", 0);
	configure(0, 0);
	while (1) do
		k := t.getarg(i, abuf, MAXARGLEN);
		if (k < 0 \/ abuf::0 \= '-') leave;
		for (j=1, k) do
		ie (abuf::j = 'A') do
			file(@i, abuf, @j, @k, @altarc);
			loadconf(altarc, 1);
		end
		else ie (abuf::j = 'B') file(@i, abuf, @j, @k, @O_bindir);
		else ie (abuf::j = 'C') file(@i, abuf, @j, @k, @O_classdir);
		else ie (abuf::j = 'D') O_dpfilter := 1;
		else ie (abuf::j = 'G') O_make_glue := 1;
		else ie (abuf::j = 'L') file(@i, abuf, @j, @k, @O_libdir);
		else ie (abuf::j = 'N') do
				O_optimize := 0;
				O_dpfilter := 0;
		end
		else ie (abuf::j = 'O') O_optimize := 1;
		else ie (abuf::j = 'U') do
			file(@i, abuf, @j, @k, @O_userconfig);
			configure(0, 1);
		end
		else ie (abuf::j = 'V') license();
		else ie (abuf::j = 'c') O_compile_only := 1;
		else ie (abuf::j = 'e') O_bang_header := 1;
		else ie (abuf::j = 'h' \/ abuf::j = '?') longusage();
		else ie (abuf::j = 'i') O_make_class := 1;
		else ie (abuf::j = 'k') O_keep_tmpfiles := 1;
		else ie (abuf::j = 'l') do
			O_make_class := 1;
			O_local_class := 1;
		end
		else ie (abuf::j = 'm') file(@i, abuf, @j, @k, @O_target);
		else ie (abuf::j = 'o') file(@i, abuf, @j, @k, @O_output_file);
		else ie (abuf::j = 'p') O_keep_dirs := 1;
		else ie (abuf::j = 'v') O_verbose := O_verbose+1;
		else usage(1);
		end
		i := i+1;
	end
	if (O_target /\ \t.memcomp(O_target, "?", 2)) do
		listtargets();
		halt 0;
	end
	mktmpfiles();
	if (O_verbose) ws("T3X version 8.1.5");
	if (O_verbose > 2) printargs();
	if (k < 0) usage(1);
	rmdirfiles();
	! compile each file specified on the command line to Tcode
	while (1) do
		k := t.getarg(i, abuf, MAXARGLEN);
		if (k < 0) leave;
		ie (source_p(abuf)) do
			compile(abuf);
		end
		else do
			if (O_verbose)
				wf("%S: no .t suffix (not compiling)",
					[(abuf)]);
			savename(abuf, 1);
		end
		i := i+1;
	end
	! use the last name on the command line, if -o is not used
	if (\O_output_file) do
		basename(Names[Nt-1], abuf);
		i := str.rscan(abuf, '/');
		O_output_file := savestr(@abuf::(i+1));
		dflname := 1;
	end
	ie (O_compile_only \/ O_make_class) do
		if (Nt > 1) error("multiple input -- ignoring -o", 0);
		ie (O_compile_only) do
			if (O_output_file /\ \dflname) do
				remove(O_output_file, 1);
				rename(Names[Nt-1], O_output_file, 1);
			end
		end
		else do ! if(O_make_class)
			make_class();
		end
	end
	else do
		loadall();
		if (O_optimize) optimize("txopt", [0]);
		if (O_dpfilter)
			optimize("txdpf", [0]);
		ie (O_target)
			tcode_to_native(O_target);
		else do
			remove(O_output_file, 1);
			ie (O_bang_header)
				make_bang_header(O_tmpout, O_output_file);
			else
				rename(O_tmpout, O_output_file, 1);
		end
	end
	removeall();
	sys.fini();
end
