! TXCM -- T3X Class Manager
! Copyright (C) 2000,2002,2003,2004 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcm(t3x, string, iostream);

object	t[t3x], str[string];

struct	MODI =
	M_NONE,
	M_LIST,		! list class dir
	M_SPLIT,	! split class dir (extract all)
	M_EXTR,		! extract class dir
	M_INDEX,	! create index file
	M_INST,		! install class
	M_UNINST,	! uninstall class
	M_RESOLV;	! resolve class

const	BUFLEN =	512;
const	STRLEN =	128;
const	MAXARGLEN =	STRLEN;
const	MAXMEMBERS =	1024;	! max. # of members in class definition
const	MAXNAMEPOOL =	8192;	! space reserved for symbol names
const	CACHESPACE =	4096;	! symbol table (class dir cache) size
const	MAXNAMES =	512;	! max. # of symbols to keep in core

var	B4F, B8F,		! constant 0xFFFF and 0xFFFFFFFF
	B16F;			! and 0xFFFFFFFFFFFFFFFF
var	B15, B31;		! constant 0x8000 and 0x80000000


!--- From TXTRN ---!
struct	SYMREC =	! Symbol table record
	SNAME,
	SVAL,
	SSIZE,
	SFLAGS;	
struct XSYMREC =	! byte-order-independent SYMREC
	XHHVAL,	XHLVAL, XLHVAL, XLLVAL,
	XHHSIZE, XHLSIZE, XLHSIZE, XLLSIZE,
	XHFLGS,	XLFLGS;
const	CNST=	1,	! Symbol Flags
	GLOBL=	2,
	PROC=	4,
	PROTO=	8,
	IPROC=	32,
	PUBLC=	64,
	EXTRN=	128,
	CLSS=	256,
	OBJCT=	512,
	IVAR=	1024,
	REDIR=	2048,
	LINKD=	4096;
!------------------!

struct MEMBER = MNAME, MSIZE, MVAL, MFLAGS;

var	Members[MAXMEMBERS*MEMBER], Mt;	! members of new class dir
var	Names::MAXNAMEPOOL+1, Nt;	! name pool
var	Symbols[CACHESPACE];		! symbol table
var	Line, Arg;			! input line and arg #
object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Classname::STRLEN+1;		! class to manipulate
var	Outfile::STRLEN+1;		! where the output goes (if not SYSOUT)
var	Basedir::STRLEN+1;		! $T3XDIR
var	Quiet, Force;			! verbosity flags
var	Itable;				! installable class suffices
var	Xbpw;				! BPW of CLASSDIR
var	Nameptrs[MAXNAMES];		! pointers to symbol names
object	Cndx[iostream];			! CLASSNDX file
var	Clist::BUFLEN+1, Ct;		! Class list for resolve

init() do
	Mt := 0;
	Nt := 0;
	Line := 0;
	Classname::0 := 0;
	Outfile::0 := 0;
	Basedir::0 := 0;
	Quiet := 0;
	Force := 0;
	Itable := [
		! suffix  directory   required
		[ ".cls", "/classes",	1 ],	! tcode class
		[ ".dir", "/classes",	1 ],	! class directory
		[ ".o",   "/lib",	0 ],	! Unix object
		[ ".8",   "/lib",	0 ],	! Plan9 386 object
		[ ".x",   "/lib",	0 ],	! DOS extension
	];
	In.create(0, Inbuf, BUFLEN, iostream.FREAD);
	Out.create(1, Outbuf, BUFLEN, iostream.FWRITE);
end

error(m, n) do var b::10;
	t.write(2, "TXCM: ", 6);
	t.write(2, m, str.length(m));
	if (n) do
		t.write(2, ": ", 2);
		t.write(2, n, str.length(n));
	end
	t.write(2, t.newline(b), str.length(b));
	halt 1;
end

fini() do
	if (Out.flush()) error("file write error", 0);
end

writep(s) Out.writes(s);

exists(file) do var fd;
	fd := t.open(file, T3X.OREAD);
	if (fd < 0) return 0;
	t.close(fd);
	return 1;
end

var	ntoa_buf::290;
ntoa(n, w) do var k;
	t.memfill(ntoa_buf, '\s', 290);
	str.numtostr(@ntoa_buf::256, n, -10);
	k := str.length(@ntoa_buf::256);
	return k > w -> @ntoa_buf::256: @ntoa_buf::(256-w+k);
end


showflag(f, m, nm) do var lm;
	if (f & m) do
		! is there any bit with a lower significance than F?
		lm := ~(~m - (m-1))-m;
		if (lm & f) Out.writes("|");
		Out.write(nm, str.length(nm));
	end
end

! sign-extend value
extend(hh, hl, lh, ll) do var lbpw, x;
	lbpw := t.bpw();
	if (	lbpw < 4 /\
		\(hh = 0x00 /\ hl = 0x00 \/
		  hh = 0xFF /\ hl = 0xFF)
	)
		error("CLASSDIR has long values", 0);
	x := (hh << 24) | (hl << 16) | (lh << 8) | ll;
	if (Xbpw < lbpw) do
		if (Xbpw = 2 /\ x & B15) x := B8F & ~B4F | x;
		if (lbpw > 4 /\ x & B31) x := B16F & ~B8F | x;
	end
	return x;
end

dump(nm, raw_y) do var f, y[SYMREC], b::10;
	y[SVAL] := extend(raw_y::XHHVAL, raw_y::XHLVAL,
			raw_y::XLHVAL, raw_y::XLLVAL);
	y[SSIZE] := extend(raw_y::XHHSIZE, raw_y::XHLSIZE,
			raw_y::XLHSIZE, raw_y::XLLSIZE);
	y[SFLAGS] := raw_y::XHFLGS << 8 | raw_y::XLFLGS;
	Out.writes(ntoa(y[SSIZE], 6));
	Out.writes(ntoa(y[SVAL], 6));
	Out.write("\s", 1);
	Out.writes(nm);
	Out.write("\t", 1);
	f := y[SFLAGS];
	showflag(f, CNST, "CNST");
	showflag(f, GLOBL, "GLOBL");
	showflag(f, PROC, "PROC");
	showflag(f, PROTO, "PROTO");
	showflag(f, IPROC, "IPROC");
	showflag(f, PUBLC, "PUBLC");
	showflag(f, EXTRN, "EXTRN");
	showflag(f, CLSS, "CLSS");
	showflag(f, OBJCT, "OBJCT");
	showflag(f, IVAR, "IVAR");
	showflag(f, REDIR, "REDIR");
	showflag(f, LINKD, "LINKD");
	Out.write(t.newline(b), str.length(b));
end

readcldir() do var hd::4, nb::2, n;
	if (In.read(hd, 4) \= 4 \/ t.memcomp(hd, "CDR", 3))
		error("magic match failed", 0);
	Xbpw := hd::3;
	if (In.read(nb, 2) \= 2) error("short file", 0);
	n := nb::0 << 8 | nb::1;
	if (In.read(Symbols, n*XSYMREC) \= n*XSYMREC)
 		error("short file", 0);
	if (In.read(Names, MAXNAMEPOOL+1) = MAXNAMEPOOL+1)
		error("file too long", 0);
	return n;
end

listcldir() do var i, n, p;
	n := readcldir();
	p := 0;
	for (i=0, n*XSYMREC, XSYMREC) do
		dump(@Names::p, @Symbols::i);
		p := p+str.length(@Names::p)+1;
	end
end

writedir(nlist, tbl, n) do
	var	k, i, dirname::256, sfx, s, b::10;
	object	dir[iostream];
	var	dirbuf::BUFLEN;
	var	fwerr;
	var	nb::2;

	fwerr := "file write error";
	str.copy(dirname, nlist[0]);
	k := str.length(dirname);
	sfx := ".dir";
	for (i=0, 5) dirname::(k+i) := sfx::i;
	if (\Quiet) do
		Out.writes(dirname);
		Out.writes(": ");
	end
	if (\Force /\ exists(dirname)) do
		if (\Quiet) do
			Out.writes("exists -- not replaced");
			Out.writes(t.newline(b));
		end
		return 0;
	end
	if (dir.open(dirname, dirbuf, BUFLEN, iostream.FWRITE) = %1)
		error("file creation failed", dirname);
	if (	dir.write("CDR", 3) \= 3 \/
		dir.wrch(Xbpw) < 0
	)
		error(fwerr, 0);
	nb::0 := n >> 8;
	nb::1 := n & 255;
	if (dir.write(nb, 2) \= 2) error(fwerr, 0);
	if (dir.write(tbl, n*XSYMREC) \= n*XSYMREC) error(fwerr, 0);
	for (i=0, n) do
		s := nlist[i];
		k := str.length(s);
		k := k+1;
		if (dir.write(s, k) \= k) error(fwerr, 0);
	end
	if (dir.close()) error(fwerr, 0);
	if (\Quiet) do
		Out.writes("extracted"); Out.writes(t.newline(b));
	end
end

! Extract class directory of class 'cname' from input file.
! If cname=0, extract all class dirs.
splitdir(n, cname) do
	var	i, j, f, cn, x;
	var	fnd;

	fnd := 0;
	for (x=0, n) do
		i := x*XSYMREC;
		f := Symbols::(i+XHFLGS) << 8 | Symbols::(i+XLFLGS);
		if (f & CLSS) do
			j := i;
			cn := Nameptrs[x];
			while (1) do
				if (	\(Symbols::(j+XHFLGS) << 8 |
					  Symbols::(j+XLFLGS))
				)
					leave;
				if (j >= n*XSYMREC)
					error("missing end of class", 0);
				j := j+XSYMREC;
			end
			if (\cname \/ \str.comp(cname, cn)) do
				writedir(@Nameptrs[x], @Symbols::i,
					(j+XSYMREC-i)/XSYMREC);
				fnd := 1;
			end
		end
	end
	if (\fnd /\ \Quiet) do
		ie (cname)
			error("directory not found", cname);
		else
			error("no directories found", 0);
	end
end

splitcldir(cname) do var i, n, p;
	n := readcldir();
	p := 0;
	! Fix name pool references.
	for (i=0, n) do
		Nameptrs[i] := @Names::p;
		p := p+str.length(@Names::p)+1;
	end
	splitdir(n, cname);
end

! make CLASSNDX entry for module 'bname'
mkindex(bname) do
	var	n, i, f, cn, p, k;
	var	first;

	n := readcldir();
	p := 0;
	first := 1;
	Out.writes(bname);
	Out.wrch(':');
	for (i=0, n*XSYMREC, XSYMREC) do
		f := Symbols::(i+XHFLGS) << 8 | Symbols::(i+XLFLGS);
		k := str.length(@Names::p);
		if (\f) do
			Names::(p+k-1) := 0;
			if (str.comp(@Names::(p+1), bname)) do
				if (\first) Out.wrch(',');
				Out.writes(@Names::(p+1));
				first := 0;
			end
		end
		p := p + k+1;
	end
	Out.wrch('\n');
end

! copy file with error checking
copy(srcname, destname) do
	var	src, dest;
	var	copybuf::BUFLEN;
	var	k, w;

	src := t.open(srcname, T3X.OREAD);
	if (src < 0) error("cannot open file", srcname);
	dest := t.open(destname, T3X.OWRITE);
	if (dest < 0) error("cannot create file", destname);
	while (1) do
		k := t.read(src, copybuf, BUFLEN);
		if (k < 0) error("file read error", srcname);
		if (k = 0) leave;
		w := t.write(dest, copybuf, k);
		if (w \= k) error("file write error", destname);
		if (k < BUFLEN) leave;
	end
	t.close(src);
	t.close(dest);
end

getbasedir() do
	if (Basedir::0) return;
	if (t.getenv("T3XDIR", basedir, STRLEN+1) < 0 /\
		t.getenv("t3xdir", basedir, STRLEN+1) < 0
	)
		error("cannot find T3X base dir. Use -b or set $T3XDIR", 0);
end

! insert an entry into CLASSNDX
insertndx(cls, ndx) do
	var	ndxpath::STRLEN+30, tmp::STRLEN+31, ndxent::STRLEN+1;
	var	srcbuf::BUFLEN, destbuf::BUFLEN;
	object	src[iostream], dest[iostream];
	var	k, l, replaced, fd;

	str.format(ndxpath, "%S/classes/CLASSNDX", [(Basedir)]);
	if (\exists(ndxpath)) error("cannot open class index", 0);
	str.copy(tmp, ndxpath);
	k := str.length(ndxpath);
	tmp::(k-1) := '-';
	tmp::k := 0;
	copy(ndxpath, tmp);	! make backup copy (CLASSND-)
	! copy in new index entry
	fd := t.open(ndx, T3X.OREAD);
	if (fd < 0) error("cannot open index file", ndx);
	k := t.read(fd, ndxent, STRLEN+1);
	if (k > STRLEN) error("index entry too long", ndx);
	ndxent::k := 0;
	t.close(fd);
	if (src.open(tmp, srcbuf, BUFLEN, iostream.FREAD) = %1)
		error("cannot reopen index copy", 0);
	if (dest.open(ndxpath, destbuf, BUFLEN, iostream.FWRITE) = %1) do
		t.rename(tmp, ndxpath);
		error("cannot create new index file", 0);
	end
	replaced := 0;
	l := str.length(cls);
	while (1) do
		k := src.reads(tmp, STRLEN);
		if (src.eof()) leave;
		if (\t.memcomp(tmp, cls, l) /\ tmp::l = ':') do
			str.copy(tmp, ndxent);
			replaced := 1;
		end
		dest.writes(tmp);
	end
	src.close();
	! if no entry was replaced, append the new one
	if (\replaced) dest.writes(ndxent);
	if (dest.close()) error("error updating index", 0);
end

wo(s) do var b::10;
	Out.writes(s);
	Out.writes(t.newline(b));
end

install(clsname) do
	var	file::STRLEN+5, path::STRLEN*2+20;
	var	i;

	getbasedir();
	! install the files
	for (i=0, 4) do
		str.format(file, "%S%S", [(clsname), (Itable[i][0])]);
		ie (exists(file)) do
			str.format(path, "%S%S/%S",
				[(Basedir), (Itable[i][1]), (file)]);
			copy(file, path);
		end
		else do
			if (Itable[i][2])
				error("cannot open required file", file);
		end
	end
	! update CLASSNDX
	str.format(file, "%S.ndx", [(clsname)]);
	ie (exists(file)) do
		insertndx(clsname, file);
	end
	else do
		if (\Quiet) do
			wo("*** No index file has been found for the class");
			wo(str.format(file, "*** '%S'", [(clsname)]));
			wo("*** If this class depends on other classes, it");
			wo("*** will not link correctly.");
		end
	end
end

deletendx(cls) do
	var	ndxpath::STRLEN+30, tmp::STRLEN+31;
	var	srcbuf::BUFLEN, destbuf::BUFLEN;
	object	src[iostream], dest[iostream];
	var	k, l;

	str.format(ndxpath, "%S/classes/CLASSNDX", [(Basedir)]);
	if (\exists(ndxpath)) error("cannot open class index", 0);
	str.copy(tmp, ndxpath);
	k := str.length(ndxpath);
	tmp::(k-1) := '-';
	tmp::k := 0;
	copy(ndxpath, tmp);	! create backup copy (CLASSND-)
	if (src.open(tmp, srcbuf, BUFLEN, iostream.FREAD) = %1)
		error("cannot reopen index copy", 0);
	if (dest.open(ndxpath, destbuf, BUFLEN, iostream.FWRITE) = %1) do
		t.rename(tmp, ndxpath);
		error("cannot create new index file", 0);
	end
	l := str.length(cls);
	while (1) do
		k := src.reads(tmp, STRLEN);
		if (src.eof()) leave;
		if (t.memcomp(tmp, cls, l) \/ tmp::l \= ':')
			dest.writes(tmp);
	end
	src.close();
	if (dest.close()) error("error updating index", 0);
end

uninstall(clsname) do
	var	path::STRLEN*2+20;
	var	i;

	getbasedir();
	for (i=0, 3) do
		str.format(path, "%S%S/%S%S",
			[(Basedir), (Itable[i][1]), (clsname),
			 (Itable[i][0])]);
		ie (exists(path))
			if (t.remove(path) < 0)
				error("cannot remove file", path);
		else
			if (Itable[i][2]) error("file not found", path);
	end
	deletendx(clsname);
end

addcls(name) do var b::4, i, k;
	i := 0;
	while (i < Ct) do
		if (\str.comp(@Clist::i, name)) return;
		i := i + str.length(@Clist::i)+1;
	end
	k := str.length(name)+1;
	if (Ct + k >= BUFLEN) error("resolve: class list overflow", 0);
	t.memcopy(@Clist::Ct, name, k);
	Ct := Ct+k;
	Out.writes(name);
	Out.writes(t.newline(b));
end

resolvcls(cls) do
	var	lbuf::MAXARGLEN+1, name::MAXARGLEN+1;
	var	k, j;

	Cndx.reset();
	Cndx.move(0, IOSTREAM.SEEK_SET);
	k := Cndx.reads(lbuf, MAXARGLEN);
	while (\Cndx.eof()) do
		lbuf::(k-1) := 0;
		if (lbuf::(k-2) = '\r') lbuf::(k-2) := 0;
		j := str.length(cls);
		if (\t.memcomp(cls, lbuf, j) /\ lbuf::j = ':') do
			while (	str.parse(@lbuf::j, "%[:,]%S,",
				[(name)]) = 1
			) do
				j := j+str.length(name)+1;
				resolvcls(name);
				addcls(name);
			end
			leave;
		end
		k := Cndx.reads(lbuf, MAXARGLEN);
	end
end

! resolve class dependencies defined in CLASSNDX
resolve(ap) do
	var	nbuf::BUFLEN;
	var	path::MAXARGLEN*2;
	var	k;

	getbasedir();
	str.format(path, "%S/classes/CLASSNDX", [(Basedir)]);
	if (Cndx.open(path, nbuf, BUFLEN, iostream.FREAD) = %1) do
		error("no class index", path);
		return;
	end
	Ct := 0;
	addcls("t3x");
	while (1) do
		k := t.getarg(ap, Classname, STRLEN+1);
		if (k < 1) leave;
		resolvcls(Classname);
		addcls(Classname);
		ap := ap+1;
	end
	Cndx.close();
end

ws(fd, s) do var b::10;
	t.write(fd, s, str.length(s));
	t.write(fd, t.newline(b), str.length(b));
end

usage(fd, q) do
	ws(fd, "Usage:  txcm -l [-o file] [file]");
	ws(fd, "        txcm -n baseclass [-o file] [file]");
	ws(fd, "        txcm {-x class | -s} [-qf] [file]");
	ws(fd, "        txcm {-i |-u} class [-q] [-b dir]");
	ws(fd, "        txcm [-o file] -r class [...]");
	ws(fd, "        txcm [-?h]");
	if (q) halt 1;
end

longusage() do
	ws(1, "");
	usage(1, 0);
	ws(1, "");
	ws(1, "-l      list class directory");
	ws(1, "-n base create class index entry for 'base'");
	ws(1, "-s      extract all directories (split)");
	ws(1, "-x clss extract directory of 'clss'");
	ws(1, "-i clss install class");
	ws(1, "-u clss uninstall class");
	ws(1, "-r clss resolve class (print required classes)");
	ws(1, "-b dir  define base directory for -i and -u");
	ws(1, "-o file store output in 'file'");
	ws(1, "-f      overwrite existing directories");
	ws(1, "-q      quiet mode (don't report errors)");
	ws(1, "");
	halt 0;
end

argarg(abuf, pap, pi, plen) do var i;
	pi[0] := pi[0]+1;
	if (\abuf::(pi[0])) do
		pap[0] := pap[0]+1;
		plen[0] := t.getarg(pap[0], abuf, STRLEN+1);
		if (plen[0] < 0) usage(2, 1);
		pi[0] := 0;
	end
	i := pi[0];
	pi[0] := plen[0]-1;
	return @abuf::i;
end

do
	var	abuf::STRLEN+1, ap, i, k, c;
	var	mode;

	init();
	mode := M_NONE;
	ap := 1;
	while (1) do
		k := t.getarg(ap, abuf, STRLEN+1);
		if (k < 0 \/ abuf::0 \= '-') leave;
		for (i=1, k) do
			c := abuf::i;
			ie (c = '?' \/ c = 'h') longusage();
			else ie (c = 'b') do
				str.copy(Basedir, argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'f') Force := 1;
			else ie (c = 'i') do
				mode := M_INST;
				str.copy(Classname, argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'l') mode := M_LIST;
			else ie (c = 'n') do
				mode := M_INDEX;
				str.copy(Classname, argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'o') do
				str.copy(Outfile, argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'q') Quiet := 1;
			else ie (c = 'r') do
				mode := M_RESOLV;
				argarg(abuf, @ap, @i, @k);
				leave;
			end
			else ie (c = 's') mode := M_SPLIT;
			else ie (c = 'u') do
				mode := M_UNINST;
				str.copy(Classname, argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'x') do
				mode := M_EXTR;
				str.copy(Classname, argarg(abuf, @ap, @i, @k));
			end
			else usage(2, 1);
		end
		if (mode = M_RESOLV) leave;
		ap := ap+1;
	end
	if (mode = M_NONE) usage(2, 1);
	if (mode \= M_RESOLV /\ k > 0) do
		if (mode = M_INST \/ mode = M_UNINST) usage(2, 1);
		if (t.getarg(ap+1, abuf, STRLEN) > 0) usage(2, 1);
		t.getarg(ap, abuf, STRLEN+1);
		if (In.open(abuf, Inbuf, BUFLEN, iostream.FREAD) = %1)
			error("cannot open", abuf);
	end
	if (Outfile::0) do
		! extract and split mode have no redirectable output
		if (mode = M_EXTR \/ mode = M_SPLIT) usage(2, 1);
		if (Out.open(Outfile, Outbuf, BUFLEN, iostream.FWRITE) = %1)
			error("cannot create", Outfile);
	end
	ie (mode = M_LIST) listcldir();
	else ie (mode = M_INDEX) mkindex(Classname);
	else ie (mode = M_SPLIT) splitcldir(0);
	else ie (mode = M_EXTR) splitcldir(Classname);
	else ie (mode = M_INST) install(Classname);
	else ie (mode = M_UNINST) uninstall(Classname);
	else ie (mode = M_RESOLV) resolve(ap);
	else usage(2, 1);
	fini();
end
