! TXDUMP -- Examine Tcode programs
! Copyright (C) 2000,2001,2002 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txdump(t3x, tcode, rtcode, iostream, string, util);

object	t[t3x], str[string], rt[rtcode], u[util];

const	BUFSIZE = 1024;

const	M_INFO=0,	! info mode (print size, etc)
	M_UX=1,		! UX (disassembler) mode
	M_BANG=2;	! print #! paths

var	Addr;			! input address
var	Decls, Ndcl;		! declaration tables, # of decl codes
object	In[iostream], Out[iostream];
var	Inbuf::BUFSIZE, Outbuf::BUFSIZE;
var	Nlstr::4;		! newline
var	O_mode,			! mode
	O_hex, O_one,		! hex output flag, single-line output flag
	O_force;		! force dump
var	Name;			! module name
var	Header::256;		! #! header
var	Bpw;			! t.bpw() cache
var	XFFFFFFFF;			! Constant 0xFFFFFFFF

tc_decls() return [		! declaration opcodes
	tcode.IBPW,
	tcode.ICLAB,
	tcode.ICMAP,
	tcode.ICREF,
	tcode.IDATA,
	tcode.IDLAB,
	tcode.IDREF,
	tcode.IEXT,
	tcode.IGSYM,
	tcode.IINIT,
	tcode.IIPROC,
	tcode.IISYM,
	tcode.ILIB,
	tcode.ILINE,
	tcode.ILSYM,
	tcode.IPUB,
	tcode.ISTR,
	tcode.IVEC,
0 ];

init() do
	Bpw := t.bpw();
	XFFFFFFFF := 0xFFFF;
	XFFFFFFFF := XFFFFFFFF << 16 | XFFFFFFFF;
	Addr := 0;
	t.newline(Nlstr);
	Decls := tc_decls();
end

ws(s) Out.write(s, str.length(s));

wf(f, a) do var b::1024;
	ws(str.format(b, f, a));
	Out.flush();
end

wm(s) t.write(2, s, str.length(s));

wl(s) do
	t.write(2, s, str.length(s));
	t.write(2, Nlstr, str.length(Nlstr));
end

fail(m, n) do var b::256;
	wm(str.format(b, "TXDUMP: %D: ", [(Addr)]));
	wm(str.format(b, m, [(n)]));
	wm(Nlstr);
	halt 1;
end

rdc() do
	Addr := Addr+1;
	return In.rdch();
end

sext8(n) return n .> 0x7F-> -1 & ~0x7F | n: n;

sext16(n) return Bpw <= 2-> n: n .> 0x7FFF-> -1 & ~0x7FFF | n: n;

sext32(n) return Bpw <= 4-> n: n>>31-> XFFFFFFFF << 32 | n: n;

rdop(k) do var n, m;
	n := rdc();
	if (k = 0) return sext8(n);
	n := n << 8 | rdc();
	if (k = 1) return sext16(n);
	if (k = 2) fail("bad argument length", 0);
	m := n;
	n := n << 8 | rdc();
	n := n << 8 | rdc();
	if (Bpw < 4 /\ m /\ (m \= 0xFFFF \/ n & 0x8000 = 0))
		fail("value out of range", 0);
	return sext32(n);
end

! print visual string (with non-printable characters escaped)
vstr(s, k) do var b::1024, i, j, c;
	i := 0;
	j := 0;
	for (i=0, k) do
		c := s::i;
		ie (c < '\s' \/ c > '~') do
			str.format(@b::j, "\\%2:0X", [(c)]);
			j := j+3;
		end
		else ie (c = '\\') do
			b::j := '\\';
			b::(j+1) := '\\';
			j := j+2;
		end
		else do
			b::j := c;
			j := j+1;
		end
	end
	b::j := 0;
	wf("  \q%S\q", [(b)]);
end

getversion() do var op, a1, hdlen;
	op := rdc();
	Header::0 := 0;
	hdlen := 0;
	if (op = '#') do
		Header::0 := op;
		for (hdlen=1, 256) do
			a1 := rdc();
			Header::hdlen := a1;
			if (a1 = '\n') leave;
		end
		Header::hdlen := 0;
		hdlen := hdlen+1;
		op := rdc();
	end
	op := rdc();
	if (op \= tcode.IINIT & tcode.IOPMASK /\ \O_force)
		fail("not a Tcode program (magic match failed)", 0);
	a1 := rdop(0);
	if (a1 \= 8 /\ \O_force) fail("unsupported Tcode version: %D", a1);
	! move to beginning of Tcode (skip #! header, if any)
	In.move(hdlen, iostream.SEEK_SET);
	for (Ndcl=0, 256)
		if (\Decls[Ndcl])
			leave;
	return a1;
end

disasm() do
	var	op, op2, a1;
	var	sbuf::128;
	var	na, l1, l2, sf;
	var	n, i;

	getversion();
	Addr := 0;
	while (1) do
		op := rdc();
		if (In.eof()) leave;
		ie (op & tcode.IARFLAG) do
			na := op & tcode.IARMASK >> tcode.IAMOFF;
			l1 := op & tcode.IL1MASK >> tcode.IL1OFF;
			l2 := op & tcode.IL2MASK >> tcode.IL2OFF;
			sf := op & tcode.ISTFLAG;
			op := rdc();
		end
		else do
			na := 0;
			sf := 0;
		end
		op2 := op;
		if (op2 .>= tcode.IENDOFSET)
			fail("invalid instruction (opcode=0x%2:0UX)", op);
		wf(O_hex-> "%4:0UX  %5LS": "%5UD  %5LS",
			[(na-> Addr-2: Addr-1), (rt.lookup(op2))]);
		if (na) do
			a1 := rdop(l1);
			wf(O_hex-> "  %8:0UX": "  %11D", [(a1)]);
			if (na > 1) do
				a1 := rdop(l2);
				wf(O_hex-> "  %8:0UX": "  %11D", [(a1)]);
			end
			if (sf) do
				if (a1 > 255)
					fail("suspicious string size", a1);
				for (i=0, a1) sbuf::i := rdc();
				vstr(sbuf, a1);
			end
		end
		ws(Nlstr);
	end
end

info() do
	var	op, op2, a1, a2;
	var	version, v;
	var	sbuf::128;
	var	na, l1, l2, sf;
	var	n, i;
	var	dsize, fsize, fp;
	var	isize, dclsize;
	var	xbpw, sbpw::10;

	v := getversion();
	Addr := 0;
	dsize := 0;	! size of data declared by program
	fsize := 0;	! max stack frame size used by program
	fp := 0;	! used for frame pointer tracking
	dclsize := 0;	! size of declaration part of program
	xbpw := 0;
	while (1) do
		isize := 1;	! instruction size
		op := rdc();
		if (In.eof()) leave;
		ie (op & tcode.IARFLAG) do
			na := op & tcode.IARMASK >> tcode.IAMOFF;
			l1 := op & tcode.IL1MASK >> tcode.IL1OFF;
			l2 := op & tcode.IL2MASK >> tcode.IL2OFF;
			sf := op & tcode.ISTFLAG;
			op := rdc();
			isize := 2;
		end
		else do
			na := 0;
			sf := 0;
		end
		op2 := op;
		if (op2 .>= tcode.IENDOFSET)
			fail("invalid instruction (opcode=0x%2:0UX)", op);
		if (na) do
			isize := isize + l1 + 1;
			a1 := rdop(l1);
			if (na > 1) do
				isize := isize + l2 + 1;
				a1 := rdop(l2);
			end
		end
		if (sf) do
			for (i=0, a1) sbuf::i := rdc();
			isize := isize+a1;
		end
		ie (	op = tcode.IDATA & tcode.IOPMASK \/
			op = tcode.ICREF & tcode.IOPMASK \/
			op = tcode.IDREF & tcode.IOPMASK
		)
			dsize := dsize + 1;
		else ie (op = tcode.IVEC & tcode.IOPMASK)
			dsize := dsize + (a1-> a1: 1);
		else ie (op = tcode.ISTR & tcode.IOPMASK)
			dsize := dsize + a1 + 1;
		else ie (op = tcode.ISTACK & tcode.IOPMASK)
			fp := fp + a1;
		else ie (op = tcode.IHDR & tcode.IOPMASK \/
			op = tcode.IMHDR & tcode.IOPMASK
		)
			fp := 0;
		else if (op = tcode.IBPW & tcode.IOPMASK)
			xbpw := xbpw /\ xbpw \= a1 -> -1 : a1;
		if (fp > fsize) fsize := fp;
		for (i=0, Ndcl)
			if (Decls[i] & tcode.IOPMASK = op) do
				dclsize := dclsize + isize;
				leave;
			end
	end
	Addr := Addr-1;
	if (xbpw /\ xbpw \= -1) do
		fsize := fsize .* xbpw;
		dsize := dsize .* xbpw;
	end
	ie (xbpw = 0)
		str.copy(sbpw, "portable");
	else ie (xbpw = -1)
		str.copy(sbpw, "invalid");
	else
		str.format(sbpw, "bpw=%d", [(xbpw)]);
	ie (O_one) do
		 wf(O_hex-> "v=%D %8S i=%4:0X t=%4:0X d=%4:0X s=%4:0X %S%S":
			"v=%D %8S i=%5:0UD t=%5:0UD d=%5:0UD s=%5:0UD %S%S",
			[(v), (sbpw), (Addr), (Addr-dclsize), (dsize),
			 (fsize), (Name), (Nlstr)]);
	end
	else do
		ie (O_hex) do
		 wf("Tcode version:    %4D (%S)%S", [(v), (sbpw), (Nlstr)]);
		 if (Header::0)
		  wf("script header:        %S%S", [(Header), (Nlstr)]);
		 wf("load image size:  %4UX%S", [(Addr), (Nlstr)]);
		 wf("text length:      %4UX%S", [(Addr-dclsize), (Nlstr)]);
		 wf("data length:      %4UX%S", [(dsize), (Nlstr)]);
		 wf("min stack size:   %4UX + 100%S", [(fsize), (Nlstr)]);
		end
		else do
		 wf("Tcode version:    %5D (%S)%S", [(v), (sbpw), (Nlstr)]);
		 if (Header::0)
		  wf("script header:        %S%S", [(Header), (Nlstr)]);
		 wf("load image size:  %5UD%S", [(Addr), (Nlstr)]);
		 wf("text length:      %5UD%S", [(Addr-dclsize), (Nlstr)]);
		 wf("data length:      %5UD%S", [(dsize), (Nlstr)]);
		 wf("min stack size:   %5UD + 256%S", [(fsize), (Nlstr)]);
		end
	end
end

tm_path() do var i;
	getversion();
	ie (O_one) do
		if (\Header::0) return;
		i := str.scan(Header, '/');
		if (i<0) i := 0;
		ws(@Header::i);
	end
	else do
		wf("%S:\t%S", [(Name), (Header::0-> Header: "(none)")]);
	end
	ws(Nlstr);
end

process() do
	ie (O_mode = M_UX) disasm();
	else ie (O_mode = M_INFO) info();
	else ie (O_mode = M_BANG) tm_path();
	else ;
end

usage() do
	wm(Nlstr);
	wl("Usage: txdump [-imux1] [file ...]");
	wm(Nlstr);
	wl("-f - force dump");
	wl("-i - summarize Tcode sizes (default)");
	wl("-m - print the #!header of the module");
	wl("-u - unassemble module");
	wl("-x - use hex numbers (default = decimal)");
	wl("-1 - (with -i) print one-line summaries");
	wl("     (with -m) print only the interpreter path");
	wm(Nlstr);
	halt 1;
end

defaults() do
	O_mode := M_INFO;
	O_hex := 0;
	O_one := 0;
	O_force := 0;
end

do
	var	arg::128;
	var	i, j, c, k;
	var	mode;

	init();
	defaults();
	i := 1; while (1) do
		k := t.getarg(i, arg, 127);
		if (k < 0 \/ arg::0 \= '-') leave;
		for (j=1, k) do
			c := arg::j;
			ie (c = 'f') O_force := 1;
			else ie (c = 'i') O_mode := M_INFO;
			else ie (c = 'm') O_mode := M_BANG;
			else ie (c = 'u') O_mode := M_UX;
			else ie (c = 'x') O_hex := 1;
			else ie (c = '1') O_one := 1;
			else usage();
		end
		i := i+1;
	end
	Out.create(1, Outbuf, BUFSIZE, iostream.FWRITE);
	ie (k < 0) do
		Name := "(stdin)";
		In.create(0, Inbuf, BUFSIZE, iostream.FREAD);
		Out.create(1, Outbuf, BUFSIZE, iostream.FWRITE);
		process();
	end
	else while (k >= 0) do
		Name := arg;
		ie (In.open(arg, Inbuf, BUFSIZE, iostream.FREAD) = %1) do
			fail("%S: cannot open", arg);
		end
		else do
			process();
			In.close();
		end
		i := i+1;
		k := t.getarg(i, arg, 127);
	end
	Out.flush();
end
