! TXHELP -- Browse Structured Documents (SDs)
! Copyright (C) 2000,2002,2003 Nils M Holm.  All rights reserved.
! Based on SDV, Copyright (C) 1999 Nils M Holm.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txhelp(t3x, iostream, ttyctl, string, system, util);

object	t[t3x], str[string], tty[ttyctl], sys[system], u[util];

const   SEGLEN  =       24576;	! max. size of text segment in memory
const	BUFLEN	=	1024;
const	SBUFLEN	=	256;	! small buffer size
const	NLINES	=	100;	! max. lines per topic
const	NCOLS	=	80;	! max. columns per topic
const	MAXCOLS	=	NCOLS;	! max. table columns
const   MAXNDX  =       64;	! max. number of level-2 index entries
const   MAXNDX2 =       256;	! max. number of top-level entries
const	MAXPOOL=	4096;	! size of name pool (for index entries)
const	MAXFILES=	NLINES+%4;

! special return codes of browser
const	NEXTTOPIC = 256, PREVTOPIC = 257, LIST_KEYWORDS = 258;

! browser state
struct	MODES =
	M_TOP,	! at top level
	M_INDEX,! at sub index level
	M_VIEW,	! viewing topic
	M_KEY;	! listing keywords

! index entry
struct LOCATION =
	L_hseg,		! segment, high word
	L_lseg,		! segment, low word
	L_offset,	! offset in segment
	L_name;		! ptr to topic/section name

object	Stream[iostream];		! file to browse
var	Streambuf::BUFLEN;
var	Filelist[MAXFILES], Nfiles;	! list of browsable files
var	Titles[MAXFILES];		! list of document titles
var	Ptoc[MAXFILES], Pt;		! priority TOC, top ptr
var	Segment::SEGLEN;		! in-core segment of curr. file
var     Names::MAXPOOL, Nt,		! string pool, top ptr
	Nbase;				!  top of locked area (file names)
var	Buffer::NCOLS*NLINES, Size;	! rendering buffer, lines used
var	Index[MAXNDX2*LOCATION], Nkw;	! subindex, num. of entries
var	Topindex[MAXNDX*LOCATION], Ntkw; ! Top index, num. of entries
var	Title::NCOLS;			! file title
var	Section, Topic;			! curr. section + topic
var	Lines, Cols;			! number of lines/columns on screen
var	Lineset;			! line drawing characters
var	Lowseg, Highseg;		! curr. segment (lo word, hi word)
var	Message;			! curr error message or 0
var	C_text, C_high;			! text color, standout color
var	Keyword::128,			! keyword to search
	Keyndx[NLINES];			! index containing matched topics
var	Startdir::SBUFLEN;		! WD at launch time
var	Directory::128;			! where to search browsable files

fini(clr) do
	tty.color(ttyctl.F_GREY|ttyctl.B_BLACK);
	if (clr) tty.clear();
	tty.fini();
	sys.chdir(Startdir);
	sys.fini();
end

fatal(m, n) do
	fini(1);
	t.write(T3X.SYSERR, "\r\nTXHELP ERROR: ", 16);
	t.write(T3X.SYSERR, m, str.length(m));
	if (n) do
		t.write(T3X.SYSERR, ": " , 2);
		t.write(T3X.SYSERR, n, str.length(n));
	end
	t.write(T3X.SYSERR, "\r\n\n", 3);
	halt 1;
end

init() do var k;
	sys.init();
	tty.init();
	if (tty.columns() < 64 \/ tty.lines() < 18)
		fatal("Screen size too small (less than 64x18).", 0);
	tty.mode(1);
	Message := 0;
	Section := 0;
	Topic := 0;
	Lineset := "|---++++#.";
	Keyword::0 := 0;
	k := t.getenv("T3XDIR", Directory, 123);
	if (k < 1) k := t.getenv("t3xdir", Directory, 123);
	if (k > 0) str.copy(@Directory::k, "/doc");
	Message := 0;
	C_text := ttyctl.F_GREY | ttyctl.B_BLUE;
	C_high := ttyctl.F_BLUE | ttyctl.B_GREY;
	Lines := tty.lines();
	Cols := tty.columns();
	Nfiles := 0;
	Nt := 0;
	Pt := 0;
	Nbase := 0;
	sys.getdir(Startdir, SBUFLEN);
end

msg(m, action) do
	if (action = 1) Message := m;
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	tty.move((Cols/2-1)-(str.length(m))/2, Lines-1);
	tty.writes(m);
end

savename(s) do var k, q;
	k := str.length(s);
	if (Nt+k+1 >= MAXPOOL) fatal("Name pool overflow", s);
	q := @Names::Nt;
	str.copy(q, s);
	Nt := Nt+k+1;
	return q;
end

trim(s, k) do var len;
	len := str.length(s);
	if (len > k) do
		s := @s::(len-k);
		t.memcopy(s, "...", 3);
	end
	return s;
end

gauge(s, len, n, lim) do var i, j;
	n := (n=lim)-> 10: n * 10 / lim;
	for (j=str.length(s), len) s::j := '\s';
	for (i=0, 10) s::(i+j) := Lineset::(i<n->8: 9);
	s::(i+j) := 0;
end

showtitle(pos) do
	var	buf::256, s;
	var	k, q;

	if (Message) return msg(Message, 1);
	str.copy(buf, Title);
	k := str.length(Title);
	if (Section) do
		str.copy(@buf::k, " : ");
		k := k + 3;
		q := Section;
		while (	q::0 = '\s' \/ q::0 = ')' \/
			'0' <= q::0 /\ q::0 <= '9'
		)
			q := @q::1;
		str.copy(@buf::k, q);
		k := k + str.length(q);
		if (Topic) do
			str.copy(@buf::k, " : ");
			k := k + 3;
			q := Topic;
			while (	q::0 = '\s' \/ q::0 = '.' \/
				'0' <= q::0 /\ q::0 <= '9'
			)
				q := @q::1;
			str.copy(@buf::k, q);
			k := k + str.length(q);
		end
	end
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	k := Cols-1-12;
	s := trim(buf, k);
	if (pos >= 0) gauge(s, k+2, pos, Size);
	tty.writes(s);
end

position(hi, lo) do
	Stream.move(0, iostream.SEEK_SET);
	while (hi) do
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		hi := hi-1;
	end
	Stream.move(lo, iostream.SEEK_FWD);
end

setlocation(high, low) do
	const	LINELEN = 82;
	var	line::LINELEN;
	var	k, len;

	if (Highseg = high /\ Lowseg = low) return 0;
	msg("Loading...", 0);
	Stream.reset();
	position(high, low);
	Highseg := high;
	Lowseg := low;
	len := 0;
	Stream.reads(line, LINELEN);
	while (\Stream.eof()) do
		k := str.length(line);
		if (len + k >= SEGLEN-2) leave;
		line::(k-1) := 0;
		if (k > 1 /\ line::(k-2) = '\r') do
			line::(k-2) := 0;	! DOS CR
			k := k-1;
		end
		t.memcopy(@Segment::len, line, k);
		len := len+k;
		Stream.reads(line, LINELEN);
	end
	Segment::len := 0;
	Segment::(len+1) := 0;	! Segment delimiter
	showtitle(-1);
end

loadfile(name) do
	const	LINELEN = 82;
	var	line::LINELEN, p, q;
	var	totop, k, top, lastndx;
	var	nsep;
	var	o;

	if (Stream.open(name, StreamBuf, BUFLEN, iostream.FREAD) = %1)
		fatal("Cannot open file", name);
	msg("Building index...", 0);
	Nkw := 0;
	Ntkw := 0;
	Nt := Nbase;
	top := 0;	! top (size) of segment
	nsep := 0;	! number of subsequent line separators
	Title::0 := 0;
	Lowseg := 0;
	Highseg := 0;
	lastndx := 0;
	totop := 0;	! 1=add to top index, 0=add to subindex
	Stream.reads(line, LINELEN);
	while (\Stream.eof()) do
		k := str.length(line);
		line::(k-1) := 0;
		if (k > 1 /\ line::(k-2) = '\r') do
			line::(k-2) := 0;
			k := k-1;
		end
		if (top + k >= SEGLEN-2) do
			o := Lowseg;
			Lowseg := Lowseg + lastndx;
			if (o .>= Lowseg) Highseg := Highseg+1;
			top := top - lastndx;
			o := totop-> @Topindex[(Ntkw-1)*LOCATION]:
				@Index[(Nkw-1)*LOCATION];
			o[L_hseg] := Highseg;
			o[L_lseg] := Lowseg;
			o[L_offset] := o[L_offset] - lastndx;
		end
		t.memcopy(@Segment::top, line, k+1);
		ie (k < 2) do	! empty line
			nsep := nsep+1;
		end
		else do
			! the first non-empty line contains the title
			ie (\Title::0) do
				p := @Segment::top;
				while (p::0 = '\t') p := @p::1;
				str.copy(Title, p);
			end
			! beginnig of new section or topic
			else if (nsep > 1) do
				p := @Segment::top;
				while (p::0 = '\t') p := @p::1;
				ie (p::1 = ')' \/ p::2 = ')') do
					q := Ntkw * LOCATION;
					if (Ntkw >= MAXNDX) fatal(
						"Top index overflow", 0);
					Topindex[q+L_hseg] := Highseg;
					Topindex[q+L_lseg] := Lowseg;
					Topindex[q+L_offset] := top;
					Topindex[q+L_name] := savename(p);
					Ntkw := Ntkw+1;
					lastndx := top;
					totop := 1;
				end
				else do
					q := Nkw * LOCATION;
					if (Nkw >= MAXNDX2) fatal(
						"Index overflow", 0);
					Index[q+L_hseg] := Highseg;
					Index[q+L_lseg] := Lowseg;
					Index[q+L_offset] := top;
					Index[q+L_name] := savename(p);
					Nkw := Nkw+1;
					lastndx := top;
					totop := 0;
				end
			end
			nsep := 0;
		end
		top := top+k;
		Stream.reads(line, LINELEN);
	end
	Segment::top := 0;
	Segment::(top+1) := 0;	! segment delimiter
	showtitle(-1);
	setlocation(0, 0);
end

clear() t.memfill(Buffer, '\s', NLINES*NCOLS);

render_subindex(major) do var p, i, j, k, base;
	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	j := 0;
	base := -1;
	for (i=0, Nkw) do
		p := @Buffer::((j+2)*NCOLS);
		for (k=0, 8) p::k := '\s';
		if (str.strtonum(Index[i*LOCATION+L_name], 10, 0) = major) do
			if (base = -1) base := i;
			t.memcopy(@p::8, Index[i*LOCATION+L_name],
				str.length(Index[i*LOCATION+L_name])+1);
			j := j+1;
		end
	end
	Size := j+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
	return base;
end

render_topindex() do var p, i, j;
	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	for (i=0, Ntkw) do
		p := @Buffer::((i+2)*NCOLS);
		for (j=0, 8) p::j := '\s';
		t.memcopy(@p::8, Topindex[i*LOCATION+L_name],
			str.length(Topindex[i*LOCATION+L_name])+1);
	end
	Size := Ntkw+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
end

getk() do var k;
	k := tty.readc();
	Message := 0;
	return k;
end

puts(y, off, s) do var b::NCOLS;
	tty.move(0, y);
	tty.clreol();
	if (str.length(s) >= off) do
		ie (Cols < NCOLS) do
			t.memcopy(b, @s::off, NCOLS);
			b::Cols := 0;
			tty.writes(b);
		end
		else do
			tty.writes(@s::off);
		end
	end
end

findright() do var max, i, j, s;
	max := 0;
	for (i=0, Size) do
		s := @Buffer::(i*NCOLS);
		j := str.length(s);
		if (j > max) max := j;
	end
	return max < Cols-> Cols: max;
end

wh(y, s) do
	tty.move(4, y);
	tty.writes(s);
end

help() do
	tty.color(C_text);
	tty.clear();
	msg("Help screen -- press [ENTER] to return", 1);
	tty.move(0, 0);
	tty.clreol();
	tty.writes(
	"    TXHelp 1.3 Copyright (C) 1999,2000,2002,2003 Nils M Holm");
	tty.color(C_text);
	wh(3,
	"[j][k]      Move the cursor up/down         [Up][Down]");
	wh(4,
	"[^F][^B]    Move to the next/prev. page     [Next][Prev]");
	wh(5,
	"[^A][^E]    Move to beginning/end of page   [Home][End]");
	wh(6,
	"[l]         Select current item             [Right][CR]");
	wh(7,
	"[h]         Return to parent / exit on top  [Left][ESC]");
	wh(8,
	"[+][-]      Move to next/prev. section      [ and ]");
	wh(9,
	"[Q]         Quit immediately                [Z]");
	wh(10,
	"[ ]         Move screen left/right");
	wh(11,
	"[^L]        Screen refresh");
	wh(12,
	"[K]         Keyword search (single file)    [/]");
	wh(14,
	"For conditions of use, see the file LICENSE which is part");
	wh(15,
	"of the T3X distribution. ('tx -V' prints a copy.)");
	getk();
end

prompt(ps, buf) do
	var	dfl::11, k;
	var	cp, x;

	k := str.length(buf);
	ie (k > 7) do
		t.memcopy(dfl, buf, 7);
		str.copy(@dfl::7, "...");
	end
	else do
		str.copy(dfl, buf);
	end
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	tty.writes(ps);
	tty.writes(" [");
	tty.writes(dfl);
	tty.writes("]:");
	cp := 0;
	x := str.length(ps)+str.length(dfl)+5;
	while (1) do
		tty.move(x+cp, Lines-1);
		k := tty.readc();
		ie (k = '\b' /\ cp) do
			cp := cp-1;
			tty.move(x+cp, Lines-1);
			tty.writec('\s');
		end
		else ie (k = '\r' \/ k = '\n') do
			if (\dfl::0) buf::cp := 0;
			return;
		end
		else ie (k = ttyctl.K_ESC \/ k = 'G'-'@') do
			buf::0 := 0;
			return -1;
		end
		else ie (k = 'U'-'@') do
			cp := 0;
			tty.move(0, Lines-1);
			tty.clreol();
			tty.writes(ps);
			tty.writes(" [");
			tty.writes(dfl);
			tty.writes("]:");
		end
		else if (k >= '\s' /\ k <= '~' /\ cp+x < Cols-1) do
			tty.writec(k);
			buf::cp := k;
			cp := cp+1;
		end
		dfl::0 := 0;
	end
end

! browse a file
! base = line to place selection bar in
! slct = enable selection bar
! page = enable page keys (+ and -)
browse(base, slct, page) do
	var	curr, top, redraw, otop;
	var	k, i;
	var	offset;
	var	new;

	curr := base;
	top := base-Lines+2;
	if (top < 0) top := 0;
	tty.color(C_text);
	tty.clear();
	offset := 0;
	redraw := 1;
	while (1) do
		showtitle(top+Lines-1);
		if (redraw) do
			tty.color(C_text);
			for (i=0, Lines-1) do
				ie (i+top < Size)
					puts(i, offset,
						@Buffer::((top+i)*NCOLS));
				else
					puts(i, 0, "");
			end
			redraw := 0;
		end
		ie (slct) do
			tty.color(C_high);
			puts(curr-top, offset, @Buffer::(curr*NCOLS));
			tty.move(8, curr-top);
			k := getk();
			tty.color(C_text);
			puts(curr-top, offset, @Buffer::(curr*NCOLS));
		end
		else do
			tty.move(0, Lines-1);
			k := getk();
			! no selection bar to move:
			ie (k = ttyctl.K_DOWN \/ k = 'j') do
				! go to bottom so that DOWN will scroll
				curr := top+Lines-2;
				if (curr >= Size) curr := Size-1;
			end
			else if (k = ttyctl.K_UP \/ k = 'k') do
				! go to top so that UP will scroll
				curr := top;
			end
		end
		ie (	k = ttyctl.K_ESC \/ k = ttyctl.K_LEFT \/
			k = 'h' \/ k = 'q'
		) do
			return %1;
		end
		else ie (k = 'Q' \/ k = 'Z') do
			fini(1);
			halt 0;
		end
		else ie (k = ttyctl.K_DOWN \/ k = 'j') do
			if (curr >= Size-1) do
				loop;
			end
			curr := curr+1;
			if (curr > top+Lines-2) do
				top := top+1;
				tty.scroll(0, Lines-1);
				if (\slct) do
					tty.color(C_text);
					puts(curr-top, offset,
						@Buffer::(curr*NCOLS));
				end
			end
		end
		else ie (k = ttyctl.K_UP \/ k = 'k') do
			if (curr < 1) do
				loop;
			end
			curr := curr-1;
			if (curr < top) do
				top := top-1;
				tty.rscroll(0, Lines-1);
				if (\slct) do
					tty.color(C_text);
					puts(curr-top, offset,
						@Buffer::(curr*NCOLS));
				end
			end
		end
		else ie (k = ttyctl.K_NEXT \/ k = 'f' \/ k = 'F'-'@') do
			otop := top;
			curr := curr + Lines-2;
			top := top + Lines-2;
			if (curr >= Size) do
				curr := Size-1;
				if (curr < 0) curr := 0;
			end
			if (top > Size-Lines+2) do
				top := Size-Lines+1;
				if (top < 0) top := 0;
			end
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_PREV \/ k = 'b' \/ k = 'B'-'@') do
			otop := top;
			curr := curr - Lines+2;
			top := top - Lines+2;
			if (curr < 0) curr := 0;
			if (top < 0) top := 0;
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_END \/ k = 'E'-'@' \/ k = 'G') do
			otop := top;
			curr := Size-1;
			if (curr < 0) curr := 0;
			top := Size-Lines+1;
			if (top < 0) top := 0;
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_HOME \/ k = 'A'-'@' \/ k = '1') do
			otop := top;
			top := 0;
			curr := 0;
			redraw := otop \= 0;
		end	
		else ie (k = ttyctl.K_CR \/ k = ttyctl.K_RIGHT \/ k = 'l') do
			if (\slct \/ Buffer::(curr*NCOLS) = 0) loop;
			return curr-2;
		end
		else ie (k = '[' \/ k = ']' \/ k = '+' \/ k = '-') do
			if (\page) loop;
			return k = '[' \/ k = '-'-> PREVTOPIC: NEXTTOPIC;
		end
		else ie (k = '\s') do
			if (Cols < NCOLS) do
				offset := offset=0-> findright()-Cols: 0;
				redraw := 1;
			end
		end
		else ie (k = 'K' \/ k = '/') do
			if (Nfiles > 1) do
				msg("Multiple files -- search disabled", 1);
				loop;
			end
			if (prompt("Search", Keyword) < 0) loop;
			return LIST_KEYWORDS;
		end
		else ie (k = 'L'-'@') do
			redraw := 1;
		end
		else ie (k = ttyctl.K_F1 \/ k = '?') do
			help();
			redraw := 1;
		end
		else do
			msg("Press '?' for help", 1);
		end
	end
end

findleft(src) do var l, i, j;
	l := 256;
	i := 0;
	while (src::i \/ src::(i+1)) do
		for (j=0, l)
			if (src::(i+j) \= '\t')
				leave;
		if (src::(i+j) /\ j < l) l := j;
		while (src::i) i := i+1;
		i := i+1;
	end
	return l;
end

istable(src) do var i;
	i := 0;
	while (src::i = '\t' \/ src::i = '\s') i := i+1;
	if (src::i = '|' /\ (src::(i+1) = '\s' \/ \src::(i+1))) return 0;
	while (src::i) do
		if (src::i = '\t') return 1;
		i := i+1;
	end
	return 0;
end

! insert N times CHR at POS in LINE
insert(line, pos, chr, n) do var i;
	for (i=NCOLS-1, pos+n-1, -1) line::i := line::(i-n);
	for (i=0, n) line::(i+pos) := chr;
end

insertvbar(line, pos) do var i;
	for (i=NCOLS-1, pos+2, -1) line::i := line::(i-3);
	line::(pos+0) := '\s';
	line::(pos+1) := Lineset::0;
	line::(pos+2) := '\s';
end

inserthline(where, len) do var i, p;
	for (i=NLINES-1, where-1, -1) do
		p := @Buffer::(i*NCOLS);
		t.memcopy(p, p-NCOLS, NCOLS);
	end
	for (i=0, len) p::i := Lineset::1;
	p::i := 0;
	return p;
end

implode(line, pos) do var n, i;
	n := 0;
	while (line::(pos+n) = '\t') n := n+1;
	for (i=pos, NCOLS-n) line::i := line::(i+n);
end

format_table(from, pto) do
	var	i, to, s;
	var	z[MAXCOLS], j, k, b, nk, nb;
	var	ul, ll;

	to := pto[0];			! last column of table
	for (j=0, MAXCOLS) z[j] := 0;	! column sizes (characters)
	nk := 0;			! table width (characters)
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		j := 0;
		k := 0;
		while (k < MAXCOLS /\ s::j) do
			while (s::j = '\t') j := j+1;
			b := j;
			while (s::j \= '\t' /\ s::j) j := j+1;
			if (j-b > z[k]) z[k] := j-b;
			k := k+1;
		end
		if (k > nk) nk := k;
	end
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		j := 0;
		k := 0;
		while (s::j = '\t') j := j+1;
		while (k < MAXCOLS /\ s::j) do
			b := j;
			while (s::j \= '\t' /\ s::j) j := j+1;
			! "." = empty column
			if (j-b = 1 /\ s::b = '.') s::b := '\s';
			if (\s::j) leave;
			implode(s, j);
			nb := z[k]-j+b;	! number of padding blanks
			insert(s, j, '\s', nb);
			j := j+nb;
			insertvbar(s, j);
			j := j + 3;
			k := k+1;
		end
		nb := z[k]-j+b;
		insert(s, j, '\s', nb);
	end
	k := 0;
	for (j=0, nk) k := k+z[j];	! column sizes
	k := k + (nk-1)*3;		! separator sizes
	ul := inserthline(from, k);	! top line
	ll := inserthline(to+1, k);	! bottom line
	j := 0;
	for (i=0, nk-1) do		! insert separators into horiz. lines
		j := j+z[i]+1;
		ul::j := Lineset::2;
		ll::j := Lineset::3;
		j := j+2;
	end
	to := to+2;
	s := @Buffer::(from*NCOLS);
	k := str.length(s);
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		insert(s, k, '\s', 2);	! righthand side vertical line
		s::(k+1) := Lineset::0;
		insert(s, 0, '\s', 2);	! lefthand side vertical line
		s::0 := Lineset::0;
	end
	! draw the corners
	s := @Buffer::(from*NCOLS);
	s::0 := Lineset::4;
	s::1 := Lineset::1;
	s::(k+3) := Lineset::5;
	s::(k+2) := Lineset::1;
	s := @Buffer::((to-1)*NCOLS);
	s::0 := Lineset::6;
	s::1 := Lineset::1;
	s::(k+3) := Lineset::7;
	s::(k+2) := Lineset::1;
	pto[0] := to;
end

expand(s, k) do var i, j, buf::NCOLS*2;
	if (t.memscan(s, '\t', k) = -1) return;
	j := 0;
	for (i=0, k) do
		if (j >= NCOLS-1) leave;
		ie (s::i = '\t') do
			while (1) do
				buf::j := '\s';
				j := j+1;
				if (\(j mod 8)) leave;
			end
		end
		else do
			buf::j := s::i;
			j := j+1;
		end
	end
	j := j>NCOLS-1-> NCOLS-1: j;
	buf::j := 0;
	t.memcopy(s, buf, j+1);
end

format_topic(isrc) do
	var	left, line, i, k, dest;
	var	intbl;
	var	toolong;
	var	src;

	clear();
	msg("Formatting...", 0);
	setlocation(isrc[L_hseg], isrc[L_lseg]);
	src := @Segment::(isrc[L_offset]);
	left := findleft(src);
	Buffer::0 := 0;
	dest := 1;
	intbl := -1;
	i := 0;
	toolong := 0;
	while (src::i \/ src::(i+1)) do
		if (dest >= NLINES) do
			toolong := 1;
			leave;
		end
		ie (src::i = 0) do
			Buffer::(dest*NCOLS) := 0;
			line := "";
		end
		else do
			line := @src::(i+left);
			k := str.length(line);
			if (line::0 = '|') do
				ie (line::1 = '\s') line := @line::2;
				else if (\line::1) line := @line::1;
			end
			t.memcopy(@Buffer::(dest*NCOLS), line, k+1);
			i := i + k+left;
		end
		ie (istable(line)) do
			if (intbl = -1) intbl := dest;
		end
		else do
			if (intbl \= -1) format_table(intbl, @dest);
			intbl := -1;
		end
		if (intbl = -1) expand(@Buffer::(dest*NCOLS), k);
		dest := dest+1;
		i := i+1;
	end
	if (intbl \= -1) do
		if (dest >= NLINES-2) do
			toolong := 1;
			dest := dest-2;
		end
		format_table(intbl, @dest);
	end
	Buffer::(dest*NCOLS) := 0;
	Size := dest+1;
	ie (toolong)
		msg("Topic too long!", 1);
	else
		showtitle(-1);
end

upcase(a, b) do var i;
	i := 0;
	while (a::i) do
		b::i := 'a' <= a::i /\ a::i <= 'z'-> a::i-'a'+'A': a::i;
		i := i+1;
	end
	b::i := 0;
end

render_keylist() do var p, i, j, top, kbuf::128, tbuf::128;
	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	upcase(Keyword, kbuf);
	j := 0;
	top := 0;
	for (i=0, Nkw) do
		upcase(Index[i*LOCATION+L_name], tbuf);
		if (str.find(tbuf, kbuf) \= -1) do
			ie (j >= NLINES-4) do
				msg("Too many matches", 1);
				leave;
			end
			else do
				t.memfill(p, '\s', 8);
				str.copy(@p::8, tbuf);
				p := p+NCOLS;
				Keyndx[j] := i;
				j := j+1;
			end
		end
	end
	Size := j+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
end

viewfile(name) do var k, mode, sect_off, topic_off, sect_len, base;
	loadfile(name);
	mode := Keyword::0-> M_KEY: M_TOP;
	Section := 0;
	Topic := 0;
	sect_off := 0;	! selected section
	topic_off := 0;	! selected topic
	while (1) do
		ie (mode = M_TOP) do
			render_topindex();
			k := browse(sect_off+2, 1, 0);
		end
		else ie (mode = M_INDEX) do
			base := render_subindex(str.strtonum(Section, 10, 0));
			ie (Size >= 5) do
				! the section has subsections
				k := browse(topic_off+2, 1, 0);
				sect_len := Size-4;
			end
			else do
				! no subsections
				format_topic(@Topindex[sect_off*LOCATION]);
				k := browse(0, 0, 1);
				sect_len := 0;
			end
		end
		else ie (mode = M_VIEW) do
			format_topic(@Index[(base+topic_off)*LOCATION]);
			k := browse(0, 0, 1);
		end
		else do	! mode = M_KEY
			if (\Keyword::0) do
				mode := M_TOP;
				loop;
			end
			render_keylist();
			if (Size < 5) do
				msg("Keyword not found", 1);
				mode := M_TOP;
				loop;
			end
			! if only one match found, browse it
			k := Size=5-> 0: browse(2, 1, 0);
			ie (k < 0) do
				mode := M_TOP;
			end
			else do
				! select topic from keyword search results
				k := Keyndx[k];
				sect_off := str.strtonum(
					Index[k*LOCATION+L_name], 10, 0)-1;
				base := render_subindex(sect_off+1);
				topic_off := k-base;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
				Section := Topindex[sect_off*LOCATION+L_name];
				mode := M_VIEW;
			end
			loop;	! don't try anything else
		end
		! evaluate the return code of browse()
		ie (k = -1) do
			mode := mode-1;
			if (mode < M_VIEW) Topic := 0;
			if (mode < M_INDEX) Section := 0;
			if (mode < M_TOP) leave;
		end
		else ie (k = PREVTOPIC) do
			ie (topic_off > 0) do
				topic_off := topic_off - 1;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
			end
			else do
				msg("Beginning of section", 1);
			end
		end
		else ie (k = NEXTTOPIC) do
			ie (topic_off < sect_len-1) do
				topic_off := topic_off + 1;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
			end
			else do
				msg("End of section", 1);
			end
		end
		else ie (k = LIST_KEYWORDS) do
			mode := M_KEY;
		end
		else do	! selection made
			ie (mode = M_TOP) do
				Section := Topindex[k*LOCATION+L_name];
				sect_off := k;
				topic_off := 0;
			end
			else do
				Topic := Index[(base+k)*LOCATION+L_name];
				topic_off := k;
			end
			mode := mode+1;
		end
	end
	Stream.close();
end

! collect filenames from command line
getfileargs(ap) do var abuf::128, k;
	while (1) do
		k := t.getarg(ap, abuf, 128);
		if (k < 0) leave;
		if (Nfiles >= MAXFILES) fatal("Too many files", 0);
		Filelist[Nfiles] := savename(abuf);
		Nfiles := Nfiles+1;
		ap := ap+1;
	end
end

! search the current directory for *.sd files
scan_sd_files() do var dd, k, buf::128;
	dd := sys.opendir(".");
	while (1) do
		k := sys.readdir(dd, buf, 128);
		if (k > 3 /\ (\str.comp(@buf::(k-3), ".sd") \/
			\str.comp(@buf::(k-3), ".SD"))
		) do
			if (Nfiles >= MAXFILES)
				fatal("Too many files", 0);
			Filelist[Nfiles] := savename(buf);
			Nfiles := Nfiles+1;
		end
		if (k < 0) leave;
	end
	sys.closedir(dd);
end

! extract the title from an SD file
getname(fname, name) do var buf::256, fd, i, j;
	fd := t.open(fname, T3X.OREAD);
	if (fd < 0) fatal("Cannot open file", fname);
	t.read(fd, buf, 256);
	t.close(fd);
	i := 0;
	while (	(buf::i = '\s' \/ buf::i = '\t' \/ buf::i = '\n' \/
		buf::i = '\r') /\ i < 256
	)
		i := i+1;
	j := i;
	while (buf::i \= '\n' /\ i < 256) i := i+1;
	buf::i := 0;
	if (i-j >= NCOLS-16) buf::(j+NCOLS-17) := 0;
	str.copy(name, @buf::j);
end

render_filelist() do
	var	p, i;
	var	name::NCOLS;

	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	for (i=0, Nfiles) do
		getname(Filelist[i], name);
		t.memfill(p, '\s', 8);
		str.copy(@p::8, name);
		p := p+NCOLS;
	end
	Size := Nfiles+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
	str.copy(Title, "Select file");
end

getptoc() do var fd, k, i, j, p;
	fd := t.open(".sdtoc", T3X.OREAD);
	if (fd < 0) do
		fd := t.open("sdtoc", T3X.OREAD);
		if (fd < 0) return;
	end
	k := MAXPOOL-Nt-1;
	p := @Names::Nt;
	k := t.read(fd, p, k);
	Names::(Nt+k) := 0;
	upcase(@Names::Nt, @Names::Nt);
	j := 0;
	for (i=0, k) do
		if (p::i = '\n' \/ p::i = '\r') do
			if (Pt >= MAXFILES) leave;
			Ptoc[Pt] := @p::j;
			Pt := Pt+1;
			p::i := 0;
			j := i+1;
			if (p::j = '\n') j := j+1;
		end
	end
	if (Pt < MAXFILES /\ j < i) do
		Ptoc[Pt] := @p::j;
		Pt := Pt+1;
	end
	t.close(fd);
end

stricomp(a, b) do var s::SBUFLEN;
	if (str.length(a) >= SBUFLEN) a::(SBUFLEN-1) := 0;
	upcase(a, s);
	return str.comp(s, b);
end

applyptoc() do var newfl[MAXFILES], done::MAXFILES, i, j, n;
	n := 0;
	t.memfill(done, 0, MAXFILES);
	for (j=0, Pt) do
		for (i=0, Nfiles) do
			if (\stricomp(Filelist[i], Ptoc[j])) do
				newfl[n] := Filelist[i];
				n := n+1;
				done::i := 1;
			end
		end
	end
	for (i=0, Nfiles) do
		if (\done::i) do
			newfl[n] := Filelist[i];
			n := n+1;
		end
	end
	t.memcopy(Filelist, newfl, Nfiles*t.bpw());
end

sortfiles() do var i, j, x, buf::256;
	for (i=0, Nfiles) do
		getname(Filelist[i], buf);
		buf::64 := 0;
		Titles[i] := savename(buf);
	end
	getptoc();
	for (j=1, Nfiles) do
		for (i=0, Nfiles-j) do
			if (str.comp(Titles[i], Titles[i+1]) > 0) do
				x := Filelist[i];
				Filelist[i] := Filelist[i+1];
				Filelist[i+1] := x;
				x := Titles[i];
				Titles[i] := Titles[i+1];
				Titles[i+1] := x;
			end
		end
	end
	applyptoc();
	Nt := Nbase;
end

makefilelist(fileargs) do
	if (\Nfiles) getfileargs(fileargs);
	if (\Nfiles) scan_sd_files();
	Nbase := Nt;
	sortfiles();
end

writes(s) do var b::3;
	if (s) t.write(T3X.SYSOUT, s, str.length(s));
	t.write(T3X.SYSOUT, t.newline(b), str.length(b));
end

usage() do
	fini(0);
	writes(0);
	writes("Usage: txhelp [-l] [-m] [-s] [-c color] [-k keyword]");
	writes("              [-s section] [file ...]");
	writes(0);
	writes("-l      use nice-looking IBM line drawing characters");
	writes("-m      monochrome (actually b&w) mode");
	writes("-p      assume HP1000 palmtop (64x18) screen");
	writes("-c XX   use color code XX (where X is a hex digit)");
	writes("-d dir  search files in the given directory");
	writes("-k name search the given keyword (in single file)");
	writes("-s sect select a specific section: T=T3X language,");
	writes("        R=runtime classes, C=compiler, U=utilities,");
	writes("        F=file formats, 8=8086 assembler");
	writes("file... file(s) to display");
	writes(0);
end

! extract two digit hex number (byte) from string
xtoi(s) do var c0, c1;
	c0 := s::0;
	c1 := s::1;
	c0 := '0'<=c0/\c0<='9'-> c0-'0': 'a'<=c0/\c0<='f'-> c0-'a'+10:
		'A'<=c0/\c0<='F'-> c0-'A'+10: 0;
	if (\c1) return c0;
	c1 := '0'<=c1/\c1<='9'-> c1-'0': 'a'<=c1/\c1<='f'-> c1-'a'+10:
		'A'<=c1/\c1<='F'-> c1-'A'+10: 0;
	return c0<<4|c1;
end

! shortcuts for frequently used T3X online docs
selectsect(id) do
	Nfiles := 1;
	ie (id::0='c' \/ id::0='C') Filelist[0] := "compiler.sd";
	else ie (id::0='r' \/ id::0='R') Filelist[0] := "classes.sd";
	else ie (id::0='t' \/ id::0='T') Filelist[0] := "t3xref.sd";
	else ie (id::0='u' \/ id::0='U') Filelist[0] := "utilities.sd";
	else ie (id::0='f' \/ id::0='F') Filelist[0] := "formats.sd";
	else ie (id::0='8') Filelist[0] := "s86.sd";
	else Nfiles := 0;
end

argarg(abuf, buf, pap, pi, plen) do
	ie (abuf::(pi[0]+1)) do
		str.copy(buf, @abuf::(pi[0]+1));
	end
	else do
		pap[0] := pap[0]+1;
		plen[0] := t.getarg(pap[0], abuf, 128);
		if (plen[0] < 1) do
			usage();
			halt;
		end
		str.copy(buf, abuf);
	end
	pi[0] := plen[0]-1;
end

config() do var	abuf::128, aabuf::128, ap, i, len;
	ap := 1;
	while (1) do
		len := t.getarg(ap, abuf, 128);
		if (len < 1 \/ abuf::0 \= '-') leave;
		i := 1;
		while (abuf::i) do
			ie (abuf::i = 'c') do
				argarg(abuf, aabuf, @ap, @i, @len);
				C_text := xtoi(aabuf);
				C_high := (C_text>>4) | (C_text<<4&0xff)
					& 0xF7;
			end
			else ie (abuf::i = 'd') do
				argarg(abuf, Directory, @ap, @i, @len);
			end
			else ie (abuf::i = 'k') do
				argarg(abuf, Keyword, @ap, @i, @len);
			end
			else ie (abuf::i = 'l') do
				Lineset := packed[ 179, 196, 194, 193, 218,
				191, 192, 217, 177, 176 ];
			end
			else ie (abuf::i = 'm') do
				C_text := ttyctl.F_GREY | ttyctl.B_BLACK;
				C_high := ttyctl.F_BLACK| ttyctl.B_GREY;
			end
			else ie (abuf::i = 'p') do
				Lines := 18;
				Cols := 64;
			end
			else ie (abuf::i = 's') do
				argarg(abuf, aabuf, @ap, @i, @len);
				selectsect(aabuf);
			end
			else do
				usage();
				halt 1;
			end
			i := i+1;
		end
		ap := ap+1;
	end
	return ap;
end

do var k;
	init();
	k := config();
	sys.chdir(Directory);
	makefilelist(k);
	if (Nfiles < 1) fatal("No files found", 0);
	tty.clear();
	ie (Nfiles > 1) do
		if (Keyword::0)
			msg("Multiple files -- keyword ignored", 1);
		Keyword::0 := 0;
		k := 0;
		while (1) do
			render_filelist();
			k := browse(k+2, 1, 0);
			if (k < 0) leave;
			viewfile(Filelist[k]);
		end
	end
	else do
		viewfile(Filelist[0]);
	end
	fini(1);
end
