! TXMAKE -- T3X Make Surrogate
! Copyright (C) 2001,2002,2003 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.
!
! SYNTAX: TXMAKE [-f makefile] [-cnq] [-d[mtx]] [target]

#classpath "../classes";

module txmake(t3x, string, iostream, memory, system);

const	POOLSIZE	=	16384;
const	IOBUFLEN	=	1024;
const	TEXTLEN		=	512;
const	PATHLEN		=	256;

object	t[t3x], str[string], mem[memory], sys[system];

struct	MACRO = M_name, M_value, M_next;
struct	RULE = R_target, R_mtime, R_depend, R_action, R_line, R_next;
struct	LIST = L_value, L_next;

var	Pool::POOLSIZE;
object	Mk[iostream];
var	Mkbuf::IOBUFLEN;
var	Lno;
var	Macros;
var	Rules;
var	Target, Usrtgt;
var	Binpath;
var	DOSmode;
var	Comspec;
var	Level;
var	O_noexec,
	O_quiet,
	O_dumptarg,
	O_dumpmac,
	O_debug,
	O_mkfile,
	O_nocd,
	O_stupid;

fini(x) do
	sys.fini();
	ie (x)
		halt 1;
	else
		halt 0;
end

error(m, n) do var buf::TEXTLEN;
	t.write(T3X.SYSERR, "TXMAKE: ", 8);
	t.write(T3X.SYSERR, str.format(buf, "%D: ", [(Lno)]),
		str.length(buf));
	t.write(T3X.SYSERR, str.format(buf, m, n), str.length(buf));
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	fini(1);
end

writes(s) t.write(T3X.SYSOUT, s, str.length(s));

writeln(s) do var nl::3;
	writes(s);
	writes(t.newline(nl));
end

alloc(n) do var p;
	p := mem.alloc(n);
	if (p = 0) error("out of memory", 0);
	return p;
end

readline(lbuf) do var k, m, p, buf::TEXTLEN;
	p := 0;
	while (1) do
		k := Mk.reads(buf, TEXTLEN);
		Lno := Lno+1;
		if (k < 1 \/ Mk.eof()) return -1;
		buf::(k-1) := 0;
		k := k-1;
		if (k > 0 /\ buf::(k-1) = '\r') do
			buf::(k-1) := 0;
			k := k-1;
		end
		if (k = 0) Target := 0;
		m := str.scan(buf, '#');
		if (m >= 0) do
			buf::m := 0;
			k := m;
		end
		if (p+k+1 >= TEXTLEN) error("line too long", 0);
		str.copy(@lbuf::p, buf);
		p := p+k;
		if (buf::(k-1) \= '\\') leave;
		lbuf::(p-1) := 0;
		p := p-1;
	end
end

strsave(s) do var p;
	p := alloc(str.length(s)+1);
	str.copy(p, s);
	return p;
end

split(src, dlm, a1, a2) do var p, q;
	p := str.scan(src, dlm);
	if (p < 0) error("internal: missing delimiter in split()", 0);
	q := p-1;
	while (q /\ (src::q = '\s' \/ src::q = '\t')) q := q-1;
	if (\q /\ (src::q = '\s' \/ src::q = '\t'))
		error("missing name before '%C'", [(dlm)]);
	src::(q+1) := 0;
	a1[0] := src;
	q := p+1;
	while (src::q /\ (src::q = '\s' \/ src::q = '\t')) q := q+1;
	a2[0] := @src::q;
end

append(lst, link, data) do
	while (lst[0]) lst := @lst[0][link];
	lst[0] := data;
	data[link] := 0;
end

unquote(s) do var k;
	if (s::0 = '"') str.copy(s, @s::1);
	k := str.length(s);
	if (s::(k-1) = '"') s::(k-1) := 0;
end

scatter(line) do var p, q, lst, c, n, quoted;
	p := 0;
	q := 0;
	lst := 0;
	quoted := 0;
	while (1) do
		c := line::p;
		if (((c = '\s' \/ c = '\t') /\ \quoted) \/ \c) do
			line::p := 0;
			n := alloc(LIST * t.bpw());
			n[L_value] := strsave(@line::q);
			if (n[L_value]::0 = '"') unquote(n[L_value]);
			if (p \= q) append(@lst, L_next, n);
			line::p := c;
			while (line::p = '\s' \/ line::p = '\t') p := p+1;
			if (\line::p) leave;
			q := p;
			p := p-1;
		end
		if (c = '"') quoted := \quoted;
		p := p+1;
	end
	return lst;
end

getpath() do var path::TEXTLEN, dlm, p, q, n, c;
	if (t.getenv("COMSPEC", path, TEXTLEN) >= 0) do
		Comspec := strsave(path);
		DOSmode := 1;
	end
	if (t.getenv("PATH", path, TEXTLEN) < 0) return;
	dlm := str.scan(path, '\\') < 0-> ':': ';';
	p := 0;
	q := 0;
	if (DOSmode) do
		n := alloc(LIST * t.bpw());
		n[L_value] := ".";
		append(@Binpath, L_next, n);
	end
	while (1) do
		if (path::p = dlm \/ \path::p) do
			c := path::p;
			n := alloc(LIST * t.bpw());
			path::p := 0;
			n[L_value] := strsave(@path::q);
			append(@Binpath, L_next, n);
			if (\c) leave;
			q := p+1;
		end
		p := p+1;
	end
end

findmacro(n) do var m;
	m := Macros;
	while (m) do
		if (\str.comp(n, m[M_name])) return m;
		m := m[M_next];
	end
end

findtarget(n) do var r;
	r := Rules;
	while (r) do
		if (\str.comp(n, r[R_target])) return r;
		r := r[R_next];
	end
end

addmacro(line) do var m, n, v;
	split(line, '=', @n, @v);
	m := findmacro(n);
	ie (m) do
		m[M_value] := strsave(v);
	end
	else do
		m := alloc(MACRO * t.bpw());
		m[M_name] := strsave(n);
		m[M_value] := strsave(v);
		m[M_next] := Macros;
		Macros := m;
	end
end

getfiledate(name, p) do var sb[system.STATBUF], tm::8;
	ie (sys.stat(name, sb) >= 0) do
		t.memcopy(tm, @sb[system.ST_MTIME], 8);
	end
	else do
		t.memfill(tm, 0x00, 8);
	end
	if (\p) p := alloc(8);
	t.memcopy(p, tm, 8);
	return p;
end

addtarget(line) do var r, tg, dl;
	r := alloc(RULE * t.bpw());
	split(line, ':', @tg, @dl);
	r[R_target] := strsave(tg);
	r[R_mtime] := getfiledate(tg, 0);
	r[R_depend] := scatter(dl);
	r[R_line] := Lno;
	r[R_action] := 0;
	Target := r;
	append(@Rules, R_next, r);
end

addaction(line) do var p, n;
	if (Target = 0) error("missing target for action", 0);
	p := 0;
	while (line::p = '\t') p := p+1;
	n := alloc(LIST * t.bpw());
	n[L_value] := strsave(@line::p);
	append(@Target[R_action], L_next, n);
end

addmacval(buf, name, ppos, lim) do var v, k;
	v := findmacro(name);
	if (\v) error("undefined macro: %S", [(name)]);
	v := v[M_value];
	k := str.length(v);
	if (ppos[0]+k >= lim) error("line buffer overflow", 0);
	str.copy(@buf::ppos[0], v);
	ppos[0] := ppos[0] + k;
end

expand(s) do var c, p, buf::TEXTLEN, i, j, n::2;
	i := 0;
	j := 0;
	n::1 := 0;
	while (s::i) do
		if (j >= TEXTLEN-1) error("line buffer overflow", 0);
		c := s::i;
		i := i+1;
		if (c \= '$') do
			buf::j := c;
			j := j+1;
			loop;
		end
		c := s::i;
		i := i+1;
		ie (c = '(') do
			p := i;
			while (s::i /\ s::i \= ')') i := i+1;
			if (\s::i) error("missing ')'", 0);
			s::i := 0;
			i := i+1;
			addmacval(buf, @s::p, @j, TEXTLEN);
		end
		else ie ('A' <= c /\ c <= 'Z' \/ 'a' <= c /\ c <= 'z') do
			n::0 := c;
			addmacval(buf, n, @j, TEXTLEN);
			i := i+1;
		end
		else do
			buf::j := c;
			j := j+1;
		end
	end
	buf::j := 0;
	str.copy(s, buf);
end

readmkfile() do
	var	line::TEXTLEN, k;
	var	p, q;
	var	mknames, i, r;

	mknames := [
		"txmkfile",
		"makefile",
		"Makefile",
	0 ];
	ie (O_mkfile) do
		if (Mk.open(O_mkfile, Mkbuf, IOBUFLEN, iostream.FREAD) < 0)
			error("no such file: %S", [(O_mkfile)]);
	end
	else do
		i := 0; while (mknames[i]) do
			r := Mk.open(mknames[i], Mkbuf, IOBUFLEN,
				iostream.FREAD);
			if (r >= 0) leave;
			i := i+1;
		end
		if (r < 0) error("no makefile", 0);
	end
	while (1) do
		k := readline(line);
		if (k < 0) leave;
		if (line::0 = '#' \/ \line::0) loop;
		expand(line);
		ie (line::0 = '\t') do
			addaction(line);
		end
		else do
			p := str.scan(line, ':');
			q := str.scan(line, '=');
			ie (q .< p) addmacro(line);
			else ie (p .< q) addtarget(line);
			else if (p < 0 /\ q < 0) error("syntax error", 0);
		end
	end
end

checkdepends(r, tgt) do var d;
	d := r[R_depend];
	while (d) do
		if (\str.comp(d[L_value][R_target], tgt))
			error("cyclic dependency: %S", [(tgt)]);
		checkdepends(d[L_value], tgt);
		d := d[L_next];
	end
end

makedepends() do var r, d, p;
	r := Rules;
	while (r) do
		d := r[R_depend];
		while (d) do
			p := findtarget(d[L_value]);
			if (\p) do
				p := alloc(RULE * t.bpw());
				p[R_target] := d[L_value];
				p[R_mtime] := getfiledate(d[L_value], 0);
				p[R_depend] := 0;
				p[R_action] := 0;
				p[R_line] := 0;
				append(@Rules, R_next, p);
			end
			d[L_value] := p;
			d := d[L_next];
		end
		r := r[R_next];
	end
	r := Rules;
	while (r) do
		checkdepends(r, r[R_target]);
		r := r[R_next];
	end
end

dumpmacros() do var m, nl::3;
	m := Macros;
	while (m) do
		writes(m[M_name]);
		writes("\t=\t");
		writes(m[M_value]);
		writes(t.newline(nl));
		m := m[M_next];
	end
end

dumpmtime(tm, buf) do
	ie (tm::0 = 0x00)
		str.copy(buf, "NONEXISTANT");
	else
		str.format(buf, "%2:0D%2:0D-%2:0D-%2:0D %2:0D:%2:0D.%2:0D",
			[(tm::0), (tm::1), (tm::2), (tm::3), (tm::4),
			(tm::5), (tm::6)]);
	return buf;
end

dumptargets() do var r, b::80, nl::3, p;
	r := Rules;
	while (r) do
		writes("# date(");
		writes(r[R_target]);
		writes(") = ");
		writes(dumpmtime(r[R_mtime], b));
		writes(t.newline(b));
		writes(r[R_target]-> r[R_target]: "(anonymous)");
		writes(":\t");
		p := r[R_depend];
		while (p) do
			if (p \= r[R_depend]) writes("\s");
			writes(p[L_value][R_target]);
			p := p[L_next];
		end
		writes(t.newline(nl));
		p := r[R_action];
		if (p) while (p) do
			writes("\t");
			writes(p[L_value]);
			p := p[L_next];
			writes(t.newline(nl));
		end
		r := r[R_next];
	end
end

debug(dlevel, msg, args) do var i, buf::TEXTLEN;
	if (dlevel > O_debug) return;
	for (i=0, Level) writes("+");
	writes("\s");
	writes(str.format(buf, msg, args));
	writes(t.newline(buf));
end

try_suffices(buf) do var k, sfx, i, sb[system.STATBUF];
	sfx := [ ".com", ".exe", ".bat", 0 ];
	if (sys.stat(buf, sb) >= 0) return buf;
	k := str.length(buf);
	i := 0;
	while (sfx[i]) do
		str.copy(@buf::k, sfx[i]);
		debug(4, "locating program: %S", [(buf)]);
		if (sys.stat(buf, sb) >= 0) return buf;
		i := i+1;
	end
	buf::k := 0;
end

findprog(name, buf) do var bp, sb[system.STATBUF];
	ie (str.scan(name, '/') \= -1) do
		str.copy(buf, name);
		return buf;
	end
	else if (str.scan(name, '\\') \= -1) do
		str.copy(buf, name);
		return try_suffices(buf);
	end
	bp := Binpath;
	while (bp) do
		str.format(buf, "%S/%S", [(bp[L_value]), (name)]);
		debug(4, "locating program: %S", [(buf)]);
		ie (DOSmode)
			if (try_suffices(buf) \= 0) return buf;
		else
			if (sys.stat(buf, sb) >= 0) return buf;
		bp := bp[L_next];
	end
	! Plan 9 executables live in /bin
	str.format(buf, "/bin/%S", [(name)]);
	debug(4, "locating program: %S", [(buf)]);
	if (sys.stat(buf, sb) >= 0) return buf;
end

runcmd(cmd) do
	var	i;
	var	infile, outfile;
	var	infd, outfd, inold, outold;
	var	doappend;
	var	args[20], ac, alist;
	var	path::PATHLEN;
	var	s, r, k, m;
	var	ignore, subsh;

	infile := 0;
	outfile := 0;
	ignore := 0;
	subsh := 0;
	if (cmd::0 = '-') do
		cmd := @cmd::1;
		ignore := 1;
	end
	ie (cmd::0 = '(') do
		k := str.length(cmd)-1;
		while (k /\ (cmd::k = '\s' \/ cmd::k = '\t')) k := k-1;
		if (cmd::k \= ')') error("unbalanced parentheses", 0);
		m := findmacro("SHELL");
		if (\m) error("$SHELL is undefined", 0);
		alist := scatter(m[M_value]);
		m := alloc(LIST * t.bpw());
		m[L_value] := cmd;
		append(@alist, L_next, m);
		subsh := 1;
	end
	else do
		alist := scatter(cmd);
	end
	if (\alist) return;
	if (	\O_nocd /\
		\str.comp(alist[L_value], "cd") /\
		alist[L_next] /\ \alist[L_next][L_next]
	) do
		if (sys.chdir(alist[L_next][L_value]) < 0)
			error("cd: no such directory: %S",
				[(alist[L_next][L_value])]);
		return;
	end
	ac := 0;
	doappend := 0;
	while (alist) do
		s := alist[L_value];
		if (\subsh /\ (s::0 = '>' \/ s::0 = '<')) do
			if (s::0 = '>' /\ s::1 = '>') do
				s := @s::1;
				doappend := 1;
			end
			ie (s::1) do
				ie (s::0 = '<')
					infile := @s::1;
				else
					outfile := @s::1;
			end
			else do
				alist := alist[L_next];
				if (\alist)
					error("missing name after '<' or '>'",
						0);
				ie (s::0 = '<')
					infile := alist[L_value];
				else
					outfile := alist[L_value];
			end
			alist := alist[L_next];
			if (\alist) leave;
			loop;
		end
		if (ac >= 17) error("too many arguments: %S", [(cmd)]);
		args[ac] := alist[L_value];
		ac := ac+1;
		alist := alist[L_next];
	end
	if (findprog(args[0], path) = 0) do
		ie (O_stupid) do
			if (Comspec = 0) error("$COMSPEC is undefined", 0);
			for (i=ac+1, 1, -1) args[i] := args[i-2];
			args[1] := "/c";
			ac := ac+2;
			str.copy(path, Comspec);
		end
		else do
			error("program not found: %S", [(args[0])]);
		end
	end
	args[ac] := 0;
	if (ac >= 17) error("too many arguments: %S", [(cmd)]);
	if (infile) do
		infd := t.open(infile, T3X.OREAD);
		if (infd < 0) error("file not found: %S", [(infile)]);
		inold := sys.dup(T3X.SYSIN);
		sys.dup2(infd, T3X.SYSIN);
	end
	if (outfile) do
		outfd := t.open(outfile, doappend-> T3X.OAPPND: T3X.OWRITE);
		if (outfd < 0)
			error(doappend->
				"cannot append to file: %S":
				"cannot create file: %S",
					[(outfile)]);
		outold := sys.dup(T3X.SYSOUT);
		if (outold < 0) error("redirection failed", 0);
		sys.dup2(outfd, T3X.SYSOUT);
	end
	r := sys.spawn(path, @args[1], system.SPAWN_WAIT);
	if (infile) do
		if (sys.dup2(inold, T3X.SYSIN) < 0)
			error("redirection cleanup failed", 0);
		t.close(infd);
		t.close(inold);
	end
	if (outfile) do
		if (sys.dup2(outold, T3X.SYSOUT) < 0)
			error("redirection cleanup failed", 0);
		t.close(outfd);
		t.close(outold);
	end
	if (r \= 0 /\ \ignore)
		error("command returned failure: %S", [(args[0])]);
end

newer(a, b) do
	if (a[R_mtime]::0 = 0 \/ b[R_mtime]::0 = 0) return -1;
	if (t.memcomp(a[R_mtime], b[R_mtime], 7) > 0) return -1;
end

make(tgt, lvl) do var d, a, s, doit, b1::20, b2::20;
	Level := Level+1;
	doit := 0;
	d := tgt[R_depend];
	debug(1, "entering %S", [(tgt[R_target])]);
	while (d) do
		ie (newer(d[L_value], tgt)) do
			dumpmtime(d[L_value][R_mtime], b1);
			dumpmtime(tgt[R_mtime], b2);
			debug(2, "%S (%S) is newer than %S (%S)",
				[(d[L_value][R_target]), (b1),
				(tgt[R_target]), (b2)]);
			make(d[L_value], lvl+1);
			doit := 1;
		end
		else do
			dumpmtime(d[L_value][R_mtime], b1);
			dumpmtime(tgt[R_mtime], b2);
			debug(3, "%S (%S) is older than %S (%S)",
				[(d[L_value][R_target]), (b1),
				(tgt[R_target]), (b2)]);
		end
		d := d[L_next];
	end
	if (tgt[R_mtime]::0 = 0x00 \/ doit) do
		debug(1, "rebuilding %S", [(tgt[R_target])]);
		a := tgt[R_action];
		Lno := tgt[R_line];
		while (a) do
			s := a[L_value];
			Lno := Lno+1;
			if (\O_quiet /\ s::0 \= '@') do
				writes("\t"); writeln(s);
			end
			if (\O_noexec) runcmd(s::0 = '@'-> @s::1: s);
			a := a[L_next];
		end
	end
	getfiledate(tgt[R_target], tgt[R_mtime]);
	debug(1, "leaving %S", [(tgt[R_target])]);
	Level := Level-1;
end

usage() do
	writeln("");
	writeln(
	"Usage: txmake [-cnqS] [-d[mt][x[1-9]]] [-f makefile] [target]");
	writeln("");
	writeln("-c      do not use builtin 'cd' command");
	writeln("-n      do not execute receipes");
	writeln("-q      do not echo commands");
	writeln("-d[mtx] dump: -dm = macros, -dt = targets -dx = trace");
	writeln("        -dx[1-9] = set trace level");
	writeln("-f file read rules from 'file'");
	writeln("-S      pass unknown commands to $COMSPEC (DOS only)");
	writeln("target  rebuild 'target' (default: first in file)");
	writeln("");
	fini(1);
end

getopts() do var abuf::TEXTLEN, p, i, k;
	p := 1;
	while (1) do
		k := t.getarg(p, abuf, TEXTLEN);
		if (k < 0 \/ abuf::0 \= '-') leave;
		for (i=1, k) do
			ie (abuf::i = 'c') O_nocd := 1;
			else ie (abuf::i = 'n') O_noexec := 1;
			else ie (abuf::i = 'q') O_quiet := 1;
			else ie (abuf::i = 'S') O_stupid := 1;
			else ie (abuf::i = 'd') do
				ie (abuf::(i+1) = 0) do
					O_dumptarg := 1;
					O_dumpmac := 1;
					O_debug := 1;
				end
				else ie (abuf::(i+1) = 'x') do
					if (\O_debug) O_debug := 1;
					i := i+1;
				end
				else ie (abuf::(i+1) = 't') do
					O_dumptarg := 1;
					i := i+1;
				end
				else ie (abuf::(i+1) = 'm') do
					O_dumpmac := 1;
					i := i+1;
				end
				else usage();
			end
			else ie (abuf::i = 'f') do
				ie (abuf::(i+1)) do
					O_mkfile := strsave(@abuf::(i+1));
				end
				else do
					p := p+1;
					k := t.getarg(p, abuf, TEXTLEN);
					if (k < 0) usage();
					O_mkfile := strsave(abuf);
				end
				i := str.length(abuf)-1;
			end
			else ie ('1' <= abuf::i /\ abuf::i <= '9') do
				if (O_debug) O_debug := abuf::i - '0';
			end
			else usage();
		end
		p := p+1;
	end
	if (k < 0) return;
	Usrtgt := strsave(abuf);
	if (t.getarg(p+1, abuf, TEXTLEN) >= 0) usage();
end

init() do
	mem.init(Pool, POOLSIZE);
	sys.init();
	Lno := 0;
	Macros := 0;
	Rules := 0;
	Target := 0;
	Usrtgt := 0;
	Binpath := 0;
	DOSmode := 0;
	Comspec := 0;
	Level := 1;
	O_noexec := 0;
	O_quiet := 0;
	O_dumptarg := 0;
	O_dumpmac := 0;
	O_debug := 0;
	O_mkfile := 0;
	O_nocd := 0;
	O_stupid := 0;
end

initmacs() do
	addmacro("MAKE = txmake");
	addmacro(DOSmode-> "SHELL = command.com /c": "SHELL = /bin/sh -c");
end

do
	init();
	getopts();
	getpath();
	initmacs();
	readmkfile();
	makedepends();
	if (O_dumpmac) dumpmacros();
	if (O_dumptarg) dumptargets();
	if (Rules = 0) error("no targets", 0);
	if (Usrtgt) do
		Target := findtarget(Usrtgt);
		if (\Target) error("don't know how to make '%S'",
				[(Usrtgt)]);
	end
	make(Usrtgt-> Target: Rules, 0);
	fini(0);
end
