! TXSUM -- Print/Check/Update Checksums
! Copyright (C) 2001,2002 Nils M Holm.  All rights reserved.
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txsum(t3x, string, iostream);

object	t[t3x], str[string];

const	BUFLEN		= 1024;
const	BLOCKSIZE	= 512;
const	MAXPATHL	= 256;

struct	MODES = M_MAKE, M_CHECK, M_UPDATE, M_DIGEST;

object	In[iostream], Out[iostream];
var	Ibuf::BUFLEN, Obuf::BUFLEN;
var	Mode;

init() do
	In.create(T3X.SYSIN, Ibuf, BUFLEN, iostream.FREAD|iostream.FKILLCR);
	Out.create(T3X.SYSOUT, Obuf, BUFLEN, iostream.FWRITE);
	Mode := M_CHECK;
end

fini() do
	Out.flush();
end

msg(m, n) do var b::3;
	t.write(T3X.SYSERR, m, str.length(m));
	if (n) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, n, str.length(n));
	end
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
	return -1;
end

mksum(file, sum, sizes) do
	var	fd, buf::BLOCKSIZE;
	var	k, i, c, o, j, nb;

	fd := file-> t.open(file, T3X.OREAD): T3X.SYSIN;
	if (fd < 0) return msg("TXSUM: could not open", file);
	for (i=0, 8) sum[i] := 0;
	nb := 0;
	while (1) do
		k := t.read(fd, buf, BLOCKSIZE);
		if (k < 0) do
			msg("TXSUM: file read error", file);
			t.close(fd);
			return -1;
		end
		if (k < BLOCKSIZE) t.memfill(@buf::k, 0, BLOCKSIZE-k);
		for (i=0, BLOCKSIZE) do
			c := sum[0] & 1;
			for (j=0, 7)
				sum[j] := sum[j] >> 1 |
					(sum[j+1] & 1 -> 0x80 : 0);
			sum[7] := sum[7] >> 1 | (c -> 0x80 : 0);
			sum[0] := sum[0] + buf::i;
			for (j=1, 8) do
				if (sum[j-1] > 0xFF) do
					sum[j] := sum[j] + 1;
					sum[j-1] := sum[j-1] & 0xFF;
				end
			end
			sum[7] := sum[7] & 0xFF;
		end
		nb := nb+1;
		if (k < BLOCKSIZE) leave;
	end
	if (file) t.close(fd);
	sizes[0] := nb;
	sizes[1] := k;
end

check_sums() do
	var	line::MAXPATHL*2, name::MAXPATHL;
	var	sum[8], osum[8], sizes[2], osizes[2];
	var	k;

	while (1) do
		k := In.reads(line, MAXPATHL);
		if (k < 1 \/ In.eof()) leave;
		line::(k-1) := 0;
		k := str.parse(line, "%2X%2X%2X%2X%2X%2X%2X%2X %4X %4X %S",
			[(@osum[7]), (@osum[6]), (@osum[5]), (@osum[4]),
			(@osum[3]), (@osum[2]), (@osum[1]), (@osum[0]),
			(@osizes[0]), (@osizes[1]), (name)]);
		if (k \= 11) do
			msg("TXSUM: corrupt checksum entry", 0);
			loop;
		end
		if (mksum(name, sum, sizes) < 0) loop;
		if (sizes[0] \= osizes[0] \/ sizes[1] \= osizes[1]) do
			msg("TXSUM: file size mismatch", name);
			loop;
		end
		if (t.memcomp(sum, osum, 8*t.bpw()))
			msg("TXSUM: invalid checksum", name);
	end
end

make_sums() do var file::MAXPATHL, buf::MAXPATHL*2, k, sum[8], sizes[2];
	while (1) do
		k := In.reads(file, MAXPATHL);
		if (k < 1 \/ In.eof()) leave;
		file::(k-1) := 0;
		mksum(file, sum, sizes);
		Out.writes(str.format(buf,
		"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X %4:0UX %4:0UX %S\n",
			[(sum[7]),(sum[6]),(sum[5]),(sum[4]),
			 (sum[3]),(sum[2]),(sum[1]),(sum[0]),
			 (sizes[0]), (sizes[1]), (file)]));
	end
end

update_sums() do
	var	line::MAXPATHL*2, name::MAXPATHL;
	var	sum[8], osum[8], sizes[2], osizes[2];
	var	k;

	while (1) do
		k := In.reads(line, MAXPATHL);
		if (k < 1 \/ In.eof()) leave;
		line::(k-1) := 0;
		k := str.parse(line, "%2X%2X%2X%2X%2X%2X%2X%2X %4X %4X %S",
			[(@osum[7]), (@osum[6]), (@osum[5]), (@osum[4]),
			(@osum[3]), (@osum[2]), (@osum[1]), (@osum[0]),
			(@osizes[0]), (@osizes[1]), (name)]);
		if (k \= 11) do
			msg("TXSUM: corrupt checksum entry", 0);
			loop;
		end
		if (mksum(name, sum, sizes) < 0) loop;
		ie (sizes[0] \= osizes[0] \/ sizes[1] \= osizes[1])
			msg("TXSUM: updating", name);
		else if (t.memcomp(sum, osum, 8*t.bpw()))
			msg("TXSUM: updating", name);
		Out.writes(str.format(line,
		"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X %4:0UX %4:0UX %S\n",
			[(sum[7]),(sum[6]),(sum[5]),(sum[4]),
			 (sum[3]),(sum[2]),(sum[1]),(sum[0]),
			 (sizes[0]), (sizes[1]), (name)]));
	end
end

make_digest() do var buf::MAXPATHL*2, sum[8], sizes[2];
	mksum(0, sum, sizes);
	Out.writes(str.format(buf,
	"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X\n",
		[(sum[7]),(sum[6]),(sum[5]),(sum[4]),
		 (sum[3]),(sum[2]),(sum[1]),(sum[0])]));
end

usage() do
	msg("Usage: txsum [-cdmu]", 0);
	msg("-c - check checksums for files listed in sysin", 0);
	msg("-d - create digest for data in sysin", 0);
	msg("-m - make checksums for files listed in sysin", 0);
	msg("-u - update checksums of files listed in sysin", 0);
	halt 1;
end

do var abuf::128, p, i;
	init();
	p := 1;
	while (1) do
		if (t.getarg(p, abuf, 128) < 1) leave;
		if (abuf::0 \= '-') leave;
		i := 1;
		while (abuf::i) do
			ie (abuf::i = 'c') Mode := M_CHECK;
			else ie (abuf::i = 'd') Mode := M_DIGEST;
			else ie (abuf::i = 'm') Mode := M_MAKE;
			else ie (abuf::i = 'u') Mode := M_UPDATE;
			else usage();
			i := i+1;
		end
		p := p+1;
	end
	ie (Mode = M_CHECK) check_sums();
	else ie (Mode = M_MAKE) make_sums();
	else ie (Mode = M_UPDATE) update_sums();
	else if (Mode = M_DIGEST) make_digest();
	fini();
end
